// C++ program file boxcount.cpp

/*
  By Nathan C. Hearn
     December 11, 2007

  Program for counting boxes of scalar data containing values above or crossing
  a threshold.
*/


#ifdef USE_MPI
#include <mpi.h>
#endif  // USE_MPI

#include <stdlib.h>  // random()
#include <cstdlib>
#include <iostream>
#include <string>
#include <vector>
#include <list>
#include <sstream>
#include <fstream>
#include "quickflash_file_datafile.hpp"
#include "quickflash_file_dataset.hpp"
#include "quickflash_file_meshinfo.hpp"
#include "quickflash_utils_text.hpp"
#include "quickflash_utils_file.hpp"
#include "quickflash_utils_commandline.hpp"
#include "quickflash_counters.hpp"
#include "quickflash_cache_indexcache.hpp"
#include "boxdata.hpp"
#include "decomp_utils.hpp"
#include "box_utils.hpp"


#ifdef USE_MPI
const int mpiTag_Transition_Count = 100;
const int mpiTag_Box_Bounds = 200;
const int mpiTag_Proc_Time = 300;
const int mpiTag_File_Time = 400;
#endif  // USE_MPI


const unsigned int default_search_grid_divisions = 0;
const char base_output_file_suffix[] = "boxcount.out";

const unsigned int buffer_size = 200;
const unsigned int cache_size = 4000;



double get_random()
  {
  // NEEDS CLEAN UP!

  // Returns a random number in the range [0, 1)

  const long rand_num_range = (1 << 31);  // Max is (2^31 - 1)

  const double oneOver_rand_num_range 
    = 1.0 / static_cast<double>(rand_num_range);

  // Just use the standard C random number generator for now

  const double rand_num = static_cast<double>(random());

  return rand_num * oneOver_rand_num_range;
  }


int find_volume_limits_value(const QuickFlash::File::DataFile & df, 
			     const std::string & dataset_name, 
			     const double region_limit_min_value,
			     const double region_limit_max_value, 
			     std::vector<double> & region_min_coords, 
			     std::vector<double> & region_max_coords)
  {
  // Here, we ignore any periodicity of the data set -- IS THIS A PROBLEM???

  int ret_val = -1;  // Error state

  // Get the simulation information

  const QuickFlash::File::MeshInfo & meshinfo = df.get_mesh_info();

  const unsigned int dims = meshinfo.get_dims();

  // Open the data set

  const QuickFlash::File::Dataset & dset = df.get_dataset(dataset_name);

  // Run through the blocks

  bool interior_found = false;

  QuickFlash::Block::BlockData<double> block_data;

  std::vector<double> cell_center;

  QuickFlash::Tree::ConstTreeIter node_iter = meshinfo.get_tree_iter();

  while (node_iter.valid_node())
    {
    const QuickFlash::Block::BlockInfo & block_info 
      = node_iter.get_block_info();

    bool skip_branch = false;

    if (interior_found)
      {
      const std::vector<double> & block_min_coords 
	= block_info.get_min_coords();

      const std::vector<double> & block_max_coords 
	= block_info.get_max_coords();

      if (sub_region_enclosed(region_min_coords, region_max_coords, 
			      block_min_coords, block_max_coords))
	skip_branch = true;
      }

    if (node_iter.is_leaf() && (!skip_branch))
      {
      const unsigned int block_index = node_iter.get_block_index();

      dset.get_block_data(block_index, block_data);

      const unsigned int num_cells = block_data.get_num_cells();

      for (unsigned int cell_index = 0; cell_index < num_cells; cell_index++)
	{
	const double value = block_data[cell_index];

	if ((!(value < region_limit_min_value)) 
	    && (!(value > region_limit_max_value)))
	  {
	  block_info.get_cell_center(cell_index, cell_center);

	  if (interior_found)
	    {
	    for (unsigned int i = 0; i < dims; i++)
	      {
	      const double cell_center_coord = cell_center[i];

	      if (cell_center_coord < region_min_coords[i])
		region_min_coords[i] = cell_center_coord;
	      else if (cell_center_coord > region_max_coords[i])
		region_max_coords[i] = cell_center_coord;
	      }
	    }
	  else
	    {
	    region_min_coords = cell_center;
	    region_max_coords = cell_center;

	    interior_found = true;
	    }
	  }
	}
      }

    if (skip_branch)
      node_iter.next_branch();
    else
      node_iter.next();
    }

  if (interior_found)
    ret_val = 0;  // OK

  return ret_val;
  }


#ifdef USE_MPI
int find_volume_limits_value(const QuickFlash::File::DataFile & df, 
			     const std::string & dataset_name, 
			     const double region_limit_min_value,
			     const double region_limit_max_value, 
			     std::vector<double> & region_min_coords, 
			     std::vector<double> & region_max_coords,
			     const MPI_Comm mpi_comm)
  {
  int ret_val = -1;  // Error state

  const int root_rank = 0;

  int mpi_rank = 0;
  int mpi_size = 0;

  MPI_Comm_rank(mpi_comm, &mpi_rank);
  MPI_Comm_size(mpi_comm, &mpi_size);

  const bool is_root = (mpi_rank == root_rank) ? true : false;

  if (is_root)
    {
    ret_val = find_volume_limits_value(df, dataset_name, 
				       region_limit_min_value,
				       region_limit_max_value,
				       region_min_coords, region_max_coords);

    if (mpi_size > 0)
      {
      MPI_Bcast(&ret_val, 1, MPI_INT, 0, mpi_comm);

      if (ret_val >= 0)
	{
	unsigned int dims = region_min_coords.size();

	MPI_Bcast(&dims, 1, MPI_UNSIGNED, 0, mpi_comm);

	if (dims > 0)
	  {
	  double * coord_vect = new double[dims];

	  for (unsigned int i = 0; i < dims; i++)
	    coord_vect[i] = region_min_coords[i];

	  MPI_Bcast(coord_vect, dims, MPI_DOUBLE, 0, mpi_comm);

	  for (unsigned int i = 0; i < dims; i++)
	    coord_vect[i] = region_max_coords[i];

	  MPI_Bcast(coord_vect, dims, MPI_DOUBLE, 0, mpi_comm);

	  delete[] coord_vect;
	  coord_vect = 0;
	  }
	}
      }
    }
  else
    {
    MPI_Bcast(&ret_val, 1, MPI_INT, 0, mpi_comm);

    if (ret_val >= 0)
      {
      unsigned int dims = 0;

      MPI_Bcast(&dims, 1, MPI_UNSIGNED, 0, mpi_comm);

      if (dims > 0)
	{
	double * coord_vect = new double[dims];

	MPI_Bcast(coord_vect, dims, MPI_DOUBLE, 0, mpi_comm);

	region_min_coords.resize(dims);

	for (unsigned int i = 0; i < dims; i++)
	  region_min_coords[i] = coord_vect[i];

	MPI_Bcast(coord_vect, dims, MPI_DOUBLE, 0, mpi_comm);

        region_max_coords.resize(dims);

	for (unsigned int i = 0; i < dims; i++)
	  region_max_coords[i] = coord_vect[i];

	delete[] coord_vect;
	coord_vect = 0;
	}
      }
    }

  return ret_val;
  }
#endif  // USE_MPI


int find_volume_limits_min_refine(const QuickFlash::File::DataFile & df, 
				  const unsigned int min_refinement_level,
				  std::vector<double> & region_min_coords, 
				  std::vector<double> & region_max_coords)
  {
  // Here we ignore any periodicity of the volume -- IS THIS A PROBLEM???

  int ret_val = -1;  // Error state

  // NOTE: Refinement level is an index with origin one, as in Flash native

  // Get the simulation information

  const QuickFlash::File::MeshInfo & meshinfo = df.get_mesh_info();

  const unsigned int dims = meshinfo.get_dims();

  // Run through the blocks

  bool interior_found = false;

  QuickFlash::Tree::ConstTreeIter node_iter = meshinfo.get_tree_iter();

  while (node_iter.valid_node())
    {
    const QuickFlash::Tree::NodeData & node_data = node_iter.get_node_data();

    bool skip_branch = false;

    const QuickFlash::Block::BlockInfo & block_info 
      = node_iter.get_block_info();

    const std::vector<double> & block_min_coords 
      = block_info.get_min_coords();

    const std::vector<double> & block_max_coords 
      = block_info.get_max_coords();

    if (interior_found)
      {
      if (sub_region_enclosed(region_min_coords, region_max_coords, 
			      block_min_coords, block_max_coords))
	skip_branch = true;
      }

    if (!skip_branch)
      {
      const unsigned int block_index = node_data.get_block_index();
    
      const unsigned int refine_level = meshinfo.get_refine_level(block_index);

      if (refine_level >= min_refinement_level)
	{
	if (interior_found)
	  {
	  for (unsigned int i = 0; i < dims; i++)
	    {
	    const double axis_block_min = block_min_coords[i];
	    const double axis_block_max = block_max_coords[i];

	    if (axis_block_min < region_min_coords[i])
	      region_min_coords[i] = axis_block_min;

	    if (axis_block_max > region_max_coords[i])
	      region_max_coords[i] = axis_block_max;
	    }
	  }
	else
	  {
	  region_min_coords = block_min_coords;
	  region_max_coords = block_max_coords;

	  interior_found = true;
	  }

	skip_branch = true;
	}
      }

    if (skip_branch)
      node_iter.next_branch();
    else
      node_iter.next();
    }

  if (interior_found)
    ret_val = 0;  // OK

  return ret_val;
  }


#ifdef USE_MPI
int find_volume_limits_min_refine(const QuickFlash::File::DataFile & df, 
				  const unsigned int min_refinement_level,
				  std::vector<double> & region_min_coords, 
				  std::vector<double> & region_max_coords,
				  const MPI_Comm mpi_comm)
  {
  int ret_val = -1;  // Error state

  const int root_rank = 0;

  int mpi_rank = 0;
  int mpi_size = 0;

  MPI_Comm_rank(mpi_comm, &mpi_rank);
  MPI_Comm_size(mpi_comm, &mpi_size);

  const bool is_root = (mpi_rank == root_rank) ? true : false;

  if (is_root)
    {
    ret_val = find_volume_limits_min_refine(df, min_refinement_level, 
					    region_min_coords, 
					    region_max_coords);

    if (mpi_size > 0)
      {
      MPI_Bcast(&ret_val, 1, MPI_INT, 0, mpi_comm);

      if (ret_val >= 0)
	{
	unsigned int dims = region_min_coords.size();

	MPI_Bcast(&dims, 1, MPI_UNSIGNED, 0, mpi_comm);

	if (dims > 0)
	  {
	  double * coord_vect = new double[dims];

	  for (unsigned int i = 0; i < dims; i++)
	    coord_vect[i] = region_min_coords[i];

	  MPI_Bcast(coord_vect, dims, MPI_DOUBLE, 0, mpi_comm);

	  for (unsigned int i = 0; i < dims; i++)
	    coord_vect[i] = region_max_coords[i];

	  MPI_Bcast(coord_vect, dims, MPI_DOUBLE, 0, mpi_comm);

	  delete[] coord_vect;
	  coord_vect = 0;
	  }
	}
      }
    }
  else
    {
    MPI_Bcast(&ret_val, 1, MPI_INT, 0, mpi_comm);

    if (ret_val >= 0)
      {
      unsigned int dims = 0;

      MPI_Bcast(&dims, 1, MPI_UNSIGNED, 0, mpi_comm);

      if (dims > 0)
	{
	double * coord_vect = new double[dims];

	MPI_Bcast(coord_vect, dims, MPI_DOUBLE, 0, mpi_comm);

	region_min_coords.resize(dims);

	for (unsigned int i = 0; i < dims; i++)
	  region_min_coords[i] = coord_vect[i];

	MPI_Bcast(coord_vect, dims, MPI_DOUBLE, 0, mpi_comm);

        region_max_coords.resize(dims);

	for (unsigned int i = 0; i < dims; i++)
	  region_max_coords[i] = coord_vect[i];

	delete[] coord_vect;
	coord_vect = 0;
	}
      }
    }

  return ret_val;
  }
#endif  // USE_MPI


void fill_boxes(const QuickFlash::File::DataFile & df, 
		const std::string & dataset_name, BoxGrid & boxgrid,
		double & file_read_time, 
		const std::vector<bool> & periodic_axis=std::vector<bool>(0))
  {
  using std::cerr;
  using std::endl;

  if (boxgrid.get_num_boxes() > 0)
    {
    // Get the simulation information

    const QuickFlash::File::MeshInfo & meshinfo = df.get_mesh_info();

    const unsigned int dims = meshinfo.get_dims();

    const std::vector<double> & global_min_bounds 
      = meshinfo.get_volume_minbounds();

    const std::vector<double> & global_max_bounds 
      = meshinfo.get_volume_maxbounds();

    const unsigned int num_blocks = meshinfo.get_num_blocks();

    // Get boundary information

    bool periodic_boundary = false;

    if (periodic_axis.size() > 0)
      {
      if (periodic_axis.size() != dims)
	throw QuickFlash::Except("Incompatible periodic axis flags vector",
			       __FILE__, __LINE__);

      for (unsigned int i = 0; i < dims; i++)
	if (periodic_axis[i])
	  periodic_boundary = true;
      }

    // Open the data set

    const QuickFlash::File::Dataset & dset = df.get_dataset(dataset_name);
    
    const double start_read_time = dset.get_file_read_time();

    QuickFlash::Cache::IndexCache< QuickFlash::Block::BlockData<double> >
      block_cache(num_blocks, cache_size);

    unsigned int cache_hits = 0;
    unsigned int cache_misses = 0;

    // Run through the boxes

    const unsigned int start_box_index = boxgrid.get_start_box_index();
    const unsigned int end_box_index = boxgrid.get_end_box_index();

    QuickFlash::Block::BlockData<double> file_block_data;

    std::vector<double> box_min_coords;
    std::vector<double> box_max_coords;

    std::vector<double> cell_min_coords;
    std::vector<double> cell_max_coords;

    QuickFlash::VectorCounter counter;

    for (unsigned int box_index = start_box_index; box_index < end_box_index;
	 box_index++)
      {
      boxgrid.get_box_boundary(box_index, box_min_coords, box_max_coords);

      QuickFlash::Tree::ConstTreeIter node_iter = meshinfo.get_tree_iter();

      while (node_iter.valid_node())
	{
	const QuickFlash::Tree::NodeData & node_data = node_iter.get_node_data();

	bool skip_branch = false;

	const QuickFlash::Block::BlockInfo & block_info 
	  = node_iter.get_block_info();

	const std::vector<double> & block_min_coords 
	  = block_info.get_min_coords();

	const std::vector<double> & block_max_coords 
	  = block_info.get_max_coords();

	if (periodic_boundary)
	  {
	  if (regions_exclusive(periodic_axis, box_min_coords, box_max_coords, 
				block_min_coords, block_max_coords,
				global_min_bounds, global_max_bounds))
	    skip_branch = true;
	  }
	else 
	  {
	  if (regions_exclusive(box_min_coords, box_max_coords, 
				block_min_coords, block_max_coords))
	    skip_branch = true;
	  }

	if ((!skip_branch) && node_iter.is_leaf())
	  {
	  const unsigned int block_index = node_data.get_block_index();

	  if (!(block_cache.data_present(block_index)))
	    {
	    dset.get_block_data(block_index, file_block_data);

	    block_cache.add_data(block_index, file_block_data);

	    cache_misses++;
	    }
	  else
	    cache_hits++;

	  const QuickFlash::Block::BlockData<double> & block_data 
	    = block_cache[block_index];

	  // Run through the cells

	  counter.reset(block_data.get_block_dims());

	  while (counter.in_bounds())
	    {
	    const std::vector<unsigned int> & cell_index = counter.get_state();

	    block_info.get_cell_bounds(cell_index, cell_min_coords, 
				       cell_max_coords);

	    bool use_cell = true;  // Use the cell by default

	    if (periodic_boundary)
	      {
	      if (regions_exclusive(periodic_axis,
				    box_min_coords, box_max_coords,
				    cell_min_coords, cell_max_coords,
				    global_min_bounds, global_max_bounds))
		use_cell = false;
	      }
	    else
	      {
	      if (regions_exclusive(box_min_coords, box_max_coords,
				    cell_min_coords, cell_max_coords))
		use_cell = false;
	      }

	    if (use_cell)
	      {
	      const double cell_data = block_data[cell_index];

	      boxgrid[box_index].add_value(cell_data);
	      }

	    counter.increment();
	    }
	  }

	if (skip_branch)
	  node_iter.next_branch();
	else
	  node_iter.next();
	}
      }

    file_read_time = dset.get_file_read_time() - start_read_time;

    cerr << "    " << "** Cache hits " << cache_hits << " misses " 
	 << cache_misses << " cleans " << block_cache.get_clean_counter() 
	 << endl;
    }
  else
    file_read_time = 0.0;
  }


#ifdef USE_MPI
int get_boxes(const QuickFlash::File::DataFile & df, 
	      const std::string & dataset_name, 
	      const double threshold_value,
	      const std::vector<double> & grid_min_coords,
	      const std::vector<double> & box_width,
	      const std::vector<unsigned int> & grid_dims,
	      unsigned int & active_boxes, unsigned int & boxes_above,
	      unsigned int & boxes_below, unsigned int & boxes_transition,
	      const bool keep_box_list, std::list<BoxBoundary> & box_list, 
	      std::vector<double> & grid_min_bounds, 
	      std::vector<double> & grid_max_bounds, const MPI_Comm mpi_comm,
	      const std::vector<bool> & periodic_axis=std::vector<bool>(0))
  {
  using std::cerr;
  using std::endl;

  const int root_rank = 0;

  int mpi_rank = 0;
  int mpi_size = 0;

  MPI_Comm_rank(mpi_comm, &mpi_rank);
  MPI_Comm_size(mpi_comm, &mpi_size);

  const bool is_root = (mpi_rank == root_rank) ? true : false;

  const unsigned int dims = grid_dims.size();

  unsigned int num_boxes = 1;

  for (unsigned int i = 0; i < dims; i++)
    num_boxes *= grid_dims[i];

  // Determine the decomposition

  std::vector<unsigned int> subset_start_index_list;
  std::vector<unsigned int> subset_index_count_list;

  simple_linear_decomp(num_boxes, static_cast<unsigned int>(mpi_size),
		       subset_start_index_list, subset_index_count_list);

  const unsigned int local_start_index = subset_start_index_list[mpi_rank];
  const unsigned int local_index_count = subset_index_count_list[mpi_rank];

  // Report if this process is participating

  if (mpi_size > 1)
    {
    unsigned int participant = (local_index_count > 0) ? 1 : 0;

    unsigned int participant_count = 0;

    MPI_Reduce(&participant, &participant_count, 1, MPI_UNSIGNED, MPI_SUM,
	       root_rank, mpi_comm);

    if (is_root)
      cerr << "    " << "** Active processes: " << participant_count << endl;
    }

  // Build and fill the block

  const double start_proc_time = MPI_Wtime();

  box_list.clear();

  BoxGrid boxgrid;

  double file_read_time = 0.0;

  const QuickFlash::File::MeshInfo & meshinfo = df.get_mesh_info();

  const std::vector<double> & volume_min = meshinfo.get_volume_minbounds();
  const std::vector<double> & volume_max = meshinfo.get_volume_maxbounds();

  if (local_index_count > 0)
    {
    boxgrid.reset(grid_min_coords, box_width, grid_dims, 
		  periodic_axis, volume_min, volume_max,
		  true, local_start_index, local_index_count);

    // Collect the data

    fill_boxes(df, dataset_name, boxgrid, file_read_time, periodic_axis);

    active_boxes = boxgrid.count_boxes_threshold(threshold_value, boxes_above, 
						 boxes_below, 
						 boxes_transition);

    boxgrid.get_transition_boxes(threshold_value, box_list);
    }
  else
    {
    active_boxes = 0;

    boxes_above = 0;
    boxes_below = 0;
    boxes_transition = 0;
    }

  double proc_time = MPI_Wtime() - start_proc_time;

  // Return the data to the root process

  if (!is_root)
    {
    // Send data to root

    MPI_Reduce(&active_boxes, 0, 1, MPI_UNSIGNED, MPI_SUM, root_rank, 
	       mpi_comm);

    MPI_Reduce(&boxes_above, 0, 1, MPI_UNSIGNED, MPI_SUM, root_rank, mpi_comm);
    MPI_Reduce(&boxes_below, 0, 1, MPI_UNSIGNED, MPI_SUM, root_rank, mpi_comm);
    MPI_Reduce(&boxes_transition, 0, 1, MPI_UNSIGNED, MPI_SUM, root_rank, 
	       mpi_comm);

    MPI_Send(&boxes_transition, 1, MPI_UNSIGNED, root_rank, 
	     mpiTag_Transition_Count, mpi_comm);

    if ((boxes_transition > 0) && keep_box_list)
      {
      const unsigned int coord_vect_len = boxes_transition * dims * 2;

      double * coord_vect_elems = new double[coord_vect_len];

      // Get the boundaries

      unsigned int elem_ptr = 0;

      std::list<BoxBoundary>::const_iterator box_iter = box_list.begin();

      while (box_iter != box_list.end())
	{
	const BoxBoundary & box_bounds = *box_iter;

	const std::vector<double> & min_coords = box_bounds.get_min_bounds();
	const std::vector<double> & max_coords = box_bounds.get_max_bounds();

	for (unsigned int i = 0; i < dims; i++)
	  coord_vect_elems[elem_ptr++] = min_coords[i];

	for (unsigned int i = 0; i < dims; i++)
	  coord_vect_elems[elem_ptr++] = max_coords[i];

	++box_iter;
	}

      if (elem_ptr != coord_vect_len)
	{
	cerr << endl << "Error " << mpi_rank 
	     << " : Proper box count not reached" << endl << endl;
	MPI_Abort(mpi_comm, -2);
	}

      MPI_Send(coord_vect_elems, coord_vect_len, MPI_DOUBLE, root_rank, 
	       mpiTag_Box_Bounds, mpi_comm);

      // Clean up

      delete[] coord_vect_elems;
      coord_vect_elems = 0;
      }

    if (local_index_count > 0)
      {
      MPI_Send(&proc_time, 1, MPI_DOUBLE, root_rank, mpiTag_Proc_Time, 
	       mpi_comm);
      MPI_Send(&file_read_time, 1, MPI_DOUBLE, root_rank, mpiTag_File_Time,
	       mpi_comm);
      }
    }
  else
    {
    bool found_proc_time = false;
    bool found_file_time = false;

    double min_proc_time = 0.0;
    double max_proc_time = 0.0;

    double min_file_time = 0.0;
    double max_file_time = 0.0;

    if (local_index_count > 0)
      {
      min_proc_time = proc_time;
      max_proc_time = proc_time;

      found_proc_time = true;

      min_file_time = file_read_time;
      max_file_time = file_read_time;

      found_file_time = true;
      }

    if (mpi_size > 1)
      {
      // Collect the data

      unsigned int global_active_boxes = 0;

      unsigned int global_boxes_above = 0;
      unsigned int global_boxes_below = 0;
      unsigned int global_boxes_transition = 0;

      MPI_Reduce(&active_boxes, &global_active_boxes, 1, MPI_UNSIGNED, 
		 MPI_SUM, 0, mpi_comm);

      MPI_Reduce(&boxes_above, &global_boxes_above, 1, MPI_UNSIGNED, MPI_SUM, 
		 0, mpi_comm);
      MPI_Reduce(&boxes_below, &global_boxes_below, 1, MPI_UNSIGNED, MPI_SUM, 
		 0, mpi_comm);
      MPI_Reduce(&boxes_transition, &global_boxes_transition, 1, MPI_UNSIGNED,
		 MPI_SUM, 0, mpi_comm);

      // Replace the local values with the global ones

      active_boxes = global_active_boxes;

      boxes_above = global_boxes_above;
      boxes_below = global_boxes_below;
      boxes_transition = global_boxes_transition;

      // Get the transition box boundaries

      std::vector<double> min_box_coords(dims);
      std::vector<double> max_box_coords(dims);
    
      BoxBoundary box_bounds;

      for (int src_rank = 1; src_rank < mpi_size; src_rank++)
	{
	MPI_Status mpi_status;

	unsigned int remote_transition_boxes = 0;

	MPI_Recv(&remote_transition_boxes, 1, MPI_UNSIGNED, src_rank, 
		 mpiTag_Transition_Count, mpi_comm, &mpi_status);

	if ((remote_transition_boxes > 0) && keep_box_list)
	  {
	  const unsigned int coord_vect_len 
	    = remote_transition_boxes * dims * 2;
	  
	  double * coord_vect_elems = new double[coord_vect_len];

	  MPI_Recv(coord_vect_elems, coord_vect_len, MPI_DOUBLE, src_rank,
		   mpiTag_Box_Bounds, mpi_comm, &mpi_status);

	  unsigned int elem_ptr = 0;

	  for (unsigned int remote_box_index = 0; 
	       remote_box_index < remote_transition_boxes; remote_box_index++)
	    {
	    for (unsigned int i = 0; i < dims; i++)
	      min_box_coords[i] = coord_vect_elems[elem_ptr++];

	    for (unsigned int i = 0; i < dims; i++)
	      max_box_coords[i] = coord_vect_elems[elem_ptr++];

	    box_bounds.reset(min_box_coords, max_box_coords);

	    box_list.push_back(box_bounds);
	    }
	  
	  // Clean up

	  delete[] coord_vect_elems;
	  coord_vect_elems = 0;
	  }

	if (subset_index_count_list[src_rank] > 0)
	  {
	  MPI_Status mpi_status;
	  
	  double src_proc_time = 0.0;

	  MPI_Recv(&src_proc_time, 1, MPI_DOUBLE, src_rank, mpiTag_Proc_Time,
		   mpi_comm, &mpi_status);

	  if (found_proc_time)
	    {
	    if (src_proc_time < min_proc_time)
	      min_proc_time = src_proc_time;
	    else if (src_proc_time > max_proc_time)
	      max_proc_time = src_proc_time;
	    }
	  else
	    {
	    min_proc_time = src_proc_time;
	    max_proc_time = src_proc_time;

	    found_proc_time = true;
	    }

	  double src_file_time = 0.0;

	  MPI_Recv(&src_file_time, 1, MPI_DOUBLE, src_rank, mpiTag_File_Time,
		   mpi_comm, &mpi_status);

	  if (found_file_time)
	    {
	    if (src_file_time < min_file_time)
	      min_file_time = src_file_time;
	    else if (src_file_time > max_file_time)
	      max_file_time = src_file_time;
	    }
	  else
	    {
	    min_file_time = src_file_time;
	    max_file_time = src_file_time;

	    found_file_time = true;
	    }
	  }
	}
      }

    if (found_proc_time)
      cerr << "    " << "Width Proc Time Min/Max [ " << min_proc_time 
	   << " | " << max_proc_time << " ]" << endl;

    if (found_file_time)
      cerr << "    " << "Width Time Min/Max [ " << min_file_time 
	   << " | " << max_file_time << " ]" << endl;
    }

  grid_min_bounds = boxgrid.get_global_min_bounds();
  grid_max_bounds = boxgrid.get_global_max_bounds();

  if (!keep_box_list)
    box_list.clear();

  return 0;
  }
#else
int get_boxes(const QuickFlash::File::DataFile & df, 
	      const std::string & dataset_name, 
	      const double threshold_value,
	      const std::vector<double> & grid_min_coords,
	      const std::vector<double> & box_width,
	      const std::vector<unsigned int> & grid_dims,
	      unsigned int & active_boxes, unsigned int & boxes_above,
	      unsigned int & boxes_below, unsigned int & boxes_transition,
	      const bool keep_box_list, std::list<BoxBoundary> & box_list, 
	      std::vector<double> & grid_min_bounds, 
	      std::vector<double> & grid_max_bounds,
	      const std::vector<bool> & periodic_axis=std::vector<bool>(0))
  {
  box_list.clear();

  const QuickFlash::File::MeshInfo & meshinfo = df.get_mesh_info();

  const std::vector<double> & volume_min = meshinfo.get_volume_minbounds();
  const std::vector<double> & volume_max = meshinfo.get_volume_maxbounds();
  
  BoxGrid boxgrid(grid_min_coords, box_width, grid_dims, periodic_axis,
		  volume_min, volume_max);

  // Collect the data

  double file_read_time = 0.0;

  fill_boxes(df, dataset_name, boxgrid, file_read_time, periodic_axis);

  active_boxes = boxgrid.count_boxes_threshold(threshold_value, boxes_above, 
					       boxes_below, boxes_transition);

  boxgrid.get_transition_boxes(threshold_value, box_list);

  grid_min_bounds = boxgrid.get_global_min_bounds();
  grid_max_bounds = boxgrid.get_global_max_bounds();

  if (!keep_box_list)
    box_list.clear();

  return 0;
  }
#endif  // USE_MPI


int main(int argc, char * argv[])
  {
  using std::cout;
  using std::cerr;
  using std::endl;

#ifdef USE_MPI
  const MPI_Comm mpi_comm = MPI_COMM_WORLD;

  MPI_Init(&argc, &argv);

  const int root_rank = 0;

  int mpi_rank = 0;
  int mpi_size = 0;

  MPI_Comm_rank(mpi_comm, &mpi_rank);
  MPI_Comm_size(mpi_comm, &mpi_size);

  const bool is_root = (mpi_rank == root_rank) ? true : false;

  const double start_time = MPI_Wtime();

  if (is_root)
    cerr << endl << "MPI Size: " << mpi_size << endl;
#else
  const bool is_root = true;
#endif  // USE_MPI

  std::vector<std::string> argument_list;
  std::list<QuickFlash::Utils::Option> option_list;

  QuickFlash::Utils::parse_command_line(argc, argv, argument_list, option_list);

  const unsigned int num_args = argument_list.size();

  const unsigned int min_args = 4;

  if (num_args < min_args)
    {
    if (is_root)
      {
      const std::string full_exec = argv[0];

      std::string exec_path;
      std::string exec_name;

      QuickFlash::Utils::split_name_path_file(full_exec, exec_path, exec_name);

      cerr << endl << "Usage: " << exec_name << " dataset_name"
	   << endl << "\t" << "length_scale1[,length_scale2,...] "
	   << "threshold_value"
	   << endl << "\t" << "[ --transition-box-list ]"
	   << endl << "\t" << "[ --limit-volume-value=MIN_VALUE,MAX_VALUE ]"
	   << endl << "\t" << "[ --limit-volume-refine=MIN_REFINEMENT_LEVEL ]"
	   << endl << "\t" << "[ --output-file-prefix=PREFIX ]"
	   << endl << "\t" << "[ --buffer-box-fract=BOX_WIDTH_FRACTION ]"
	   << endl << "\t" << "[ --jitter-average=NUM_CYCLES ]"
	   << endl << "\t" << "[ --periodic-axes[=axis1,axis2,...] ]"
	   << endl << "\t" << "[ --positive-start-offset ]"
	   << endl << "\t" << "datafile1 [ datafile2 ... ]"
	   << endl << endl;
      }

#ifdef USE_MPI
    MPI_Abort(mpi_comm, 1);
#else
    return 1;
#endif  // USE_MPI
    }

  // Read the arguments

  unsigned int argptr = 0;

  const std::string dataset_name = argument_list[argptr++];

  const std::string length_scale_user_str = argument_list[argptr++];

  const std::string threshold_value_str = argument_list[argptr++];
  const double threshold_value = std::atof(threshold_value_str.c_str());

  std::list<std::string> datafile_name_list;

  while (argptr < num_args)
    datafile_name_list.push_back(argument_list[argptr++]);

  // Parse the options

  bool print_transition_boxes = false;

  bool limit_volume_by_value = false;
  double region_limit_min_value = 0.0;
  double region_limit_max_value = 0.0;

  bool limit_volume_by_refinement = false;
  unsigned int min_refinement_level = 0;

  bool use_output_file_prefix = false;
  std::string output_file_prefix;

  bool use_volume_buffer = false;
  double volume_buffer_box_width_fraction = 0.0;

  bool use_jitter_average = false;
  unsigned int num_jitter_cycles = 1;  // Used to control main loop

  bool periodic_boundary = false;

  std::vector<unsigned int> user_periodic_axes;

  bool use_positive_start_offset = false;

  std::list<QuickFlash::Utils::Option>::const_iterator option_iter
    = option_list.begin();

  while (option_iter != option_list.end())
    {
    const QuickFlash::Utils::Option & current_option = *option_iter;

    const std::string & keyword = current_option.get_key();

    if (keyword == "transition-box-list")
      {
      print_transition_boxes = true;
      }
    else if (keyword == "limit-volume-value")
      {
      if (!current_option.value_present())
	{
	if (is_root)
	  cerr << endl << "Error: Minimum and maximum values must be "
	       << "specified for volume limiting by value" << endl << endl;

#ifdef USE_MPI
	MPI_Abort(mpi_comm, -1);
#else
	return -1;
#endif  // USE_MPI
	}

      std::vector<double> volume_limit_values;

      current_option.get_value(volume_limit_values);

      if (volume_limit_values.size() != 2)
	{
	if (is_root)
	  cerr << endl << "Error: Minimum and maximum values must be "
	       << "specified for volume limiting by value" << endl << endl;

#ifdef USE_MPI
	MPI_Abort(mpi_comm, -1);
#else
	return -1;
#endif  // USE_MPI
	}

      region_limit_min_value = volume_limit_values[0];
      region_limit_max_value = volume_limit_values[1];

      if (!(region_limit_min_value < region_limit_max_value))
	{
	if (is_root)
	  cerr << endl << "Error: Min limit value must be less than max" 
	       << endl << endl;

#ifdef USE_MPI
	MPI_Abort(mpi_comm, -1);
#else
	return -1;
#endif  // USE_MPI
	}

      limit_volume_by_value = true;
      }
    else if (keyword == "limit-volume-refine")
      {
      if (!current_option.value_present())
	{
	if (is_root)
	  cerr << endl << "Error: Minimum refinement level must be  "
	       << "specified for volume limiting by refinement" << endl 
	       << endl;

#ifdef USE_MPI
	MPI_Abort(mpi_comm, -1);
#else
	return -1;
#endif  // USE_MPI
	}

      current_option.get_value(min_refinement_level);

      limit_volume_by_refinement = true;
      }
    else if (keyword == "output-file-prefix")
      {
      if (!current_option.value_present())
	{
	if (is_root)
	  cerr << endl << "Error: Output file prefix option requires file name"
	       << endl << endl;

#ifdef USE_MPI
	MPI_Abort(mpi_comm, -1);
#else
	return -1;
#endif  // USE_MPI
	}

      use_output_file_prefix = true;

      output_file_prefix = current_option.get_value();
      }
    else if (keyword == "buffer-box-fract")
      {
      if (!current_option.value_present())
	{
	if (is_root)
	  cerr << endl << "Error: Buffer option requires a width fraction"
	       << endl << endl;

#ifdef USE_MPI
	MPI_Abort(mpi_comm, -1);
#else
	return -1;
#endif  // USE_MPI
	}

      current_option.get_value(volume_buffer_box_width_fraction);

      if (!(volume_buffer_box_width_fraction > 0.0))
	{
	if (is_root)
	  cerr << endl << "Error: Buffer box width fraction must be positive"
	       << endl << endl;

#ifdef USE_MPI
	MPI_Abort(mpi_comm, -1);
#else
	return -1;
#endif  // USE_MPI
	}

      use_volume_buffer = true;
      }
    else if (keyword == "jitter-average")
      {
      if (!current_option.value_present())
	{
	if (is_root)
	  cerr << endl << "Error: Jitter average option requires a cycle count"
	       << endl << endl;

#ifdef USE_MPI
	MPI_Abort(mpi_comm, -1);
#else
	return -1;
#endif  // USE_MPI
	}

      current_option.get_value(num_jitter_cycles);

      if (num_jitter_cycles < 1)
	{
	if (is_root)
	  cerr << endl << "Error: At least one jitter cycle required"
	       << endl << endl;

#ifdef USE_MPI
	MPI_Abort(mpi_comm, -1);
#else
	return -1;
#endif  // USE_MPI
	}

      use_jitter_average = true;
      }
    else if (keyword == "periodic-axes")
      {
      periodic_boundary = true;

      if (current_option.value_present())
	current_option.get_value(user_periodic_axes);
      }
    else if (keyword == "positive-start-offset")
      {
      use_positive_start_offset = true;
      }
    else
      {
      if (is_root)
	cerr << endl << "Error: Unrecognized option " << keyword << endl 
	     << endl;

#ifdef USE_MPI
	MPI_Abort(mpi_comm, -1);
#else
	return -1;
#endif  // USE_MPI
      }

    ++option_iter;
    }


  // Get the length scales and check for errors

  std::vector<std::string> length_scale_str_vect;

  QuickFlash::Utils::read_vector_string(length_scale_user_str, 
				      length_scale_str_vect);

  std::vector<double> length_scale_vect;

  QuickFlash::Utils::read_vector_string(length_scale_user_str, 
				      length_scale_vect);

  const unsigned int num_length_scales = length_scale_vect.size();

  for (unsigned int scale_index = 0; scale_index < num_length_scales;
       scale_index++)
    {
    const double length_scale = length_scale_vect[scale_index];

    if (!(length_scale > 0.0))
      {
      if (is_root)
	cerr << endl << "Error: Box width [ " << length_scale 
	     << " ] must be positive" << endl << endl;

#ifdef USE_MPI
      MPI_Abort(mpi_comm, -1);
#else
      return -1;
#endif  // USE_MPI
      }
    }


  // Perform some additional error checking

  if (limit_volume_by_value && limit_volume_by_refinement)
    {
    if (is_root)
      cerr << endl << "Error: Volume limit by refinement and value not "
	   << "currently supported" << endl << endl;

#ifdef USE_MPI
      MPI_Abort(mpi_comm, -1);
#else
      return -1;
#endif  // USE_MPI
    }

  if (use_jitter_average && print_transition_boxes && (num_jitter_cycles > 1))
    {
    if (is_root)
      cerr << endl 
	   << "Error: Jitter average is incompatible with transition box "
	   << "output" << endl << endl;

#ifdef USE_MPI
      MPI_Abort(mpi_comm, -1);
#else
      return -1;
#endif  // USE_MPI
    }

  if (use_jitter_average && use_volume_buffer)
    {
    if (is_root)
      cerr << endl 
	   << "Error: Jitter average is incompatible with volume buffer "
	   << endl << endl;

#ifdef USE_MPI
      MPI_Abort(mpi_comm, -1);
#else
      return -1;
#endif  // USE_MPI
    }


  // Refine the output file prefix (add underscore if non-path component)

  if (use_output_file_prefix)
    {
    std::string prefix_path;
    std::string prefix_name;

    QuickFlash::Utils::split_name_path_file(output_file_prefix, prefix_path, 
					  prefix_name);

    if (prefix_name.size() > 0)
      output_file_prefix += "_";
    }


  std::string comment_marker = "";

  if (print_transition_boxes)
    comment_marker = "# ";


  if (is_root)
    cerr << "Threshold value [ " << threshold_value << " ]" << endl;

  // Run through the files

  std::list<std::string>::const_iterator filename_iter 
    = datafile_name_list.begin();

  while (filename_iter != datafile_name_list.end())
    {
    const std::string & datafile_name = *filename_iter;

#ifdef USE_MPI
    const double file_start_time = MPI_Wtime();
#endif  // USE_MPI

    if (is_root)
      cerr << "Processing file [ " << datafile_name << " ]" << endl;

    // Open the file

    const QuickFlash::File::DataFile df(datafile_name, buffer_size, true);

    const QuickFlash::File::MeshInfo & meshinfo = df.get_mesh_info();

    const double sim_time = meshinfo.get_sim_time();

    const unsigned int dims = meshinfo.get_dims();

    const std::vector<double> & volume_min = meshinfo.get_volume_minbounds();
    const std::vector<double> & volume_max = meshinfo.get_volume_maxbounds();

    std::vector<double> volume_width(dims, 0.0);

    for (unsigned int i = 0; i < dims; i++)
      volume_width[i] = volume_max[i] - volume_min[i];

    // Determine the effective (finest AMR) resolution along each axis

    const unsigned int max_refine_level = meshinfo.get_max_leaf_refine_level();

    const unsigned int refine_level_res_factor = 1 << (max_refine_level - 1);

    const std::vector<unsigned int> & base_block_dims 
      = meshinfo.get_base_block_dims();  // Top level blocks

    const std::vector<unsigned int> & block_dims = meshinfo.get_block_dims();

    std::vector<unsigned int> effective_cell_count(dims, 0);
    std::vector<double> smallest_cell_width(dims, 0.0);

    for (unsigned int i = 0; i < dims; i++)
      {
      const unsigned int effective_axis_cells 
	= refine_level_res_factor * base_block_dims[i] * block_dims[i];

      effective_cell_count[i] = effective_axis_cells;

      smallest_cell_width[i] 
	= volume_width[i] / static_cast<double>(effective_axis_cells);
      }

    // Determine the indexes of the periodic axes

    std::vector<bool> periodic_axes;

    if (periodic_boundary)
      {
      const unsigned int num_user_axes = user_periodic_axes.size();

      periodic_axes.resize(dims, false);

      if (num_user_axes == 0)
	{
	for (unsigned int i = 0; i < dims; i++)
	  periodic_axes[i] = true;  // Set all to periodic
	}
      else
	{
	for (unsigned int index = 0; index < num_user_axes; index++)
	  {
	  const unsigned int axis_index = user_periodic_axes[index];

	  if (axis_index >= dims)
	    {
	    if (is_root)
	      cerr << endl 
		   << "Error: User periodic axis [ " << axis_index 
		   << " ] out of range for file " << datafile_name 
		   << endl << endl;

#ifdef USE_MPI
	    MPI_Abort(mpi_comm, -1);
#else
	    return -1;
#endif  // USE_MPI
	    }

	  periodic_axes[axis_index] = true;
	  }
	}
      }

    // Prep the dataset

    const QuickFlash::File::Dataset & dset = df.get_dataset(dataset_name);

    dset.set_report_stats(is_root);

    // Determine the volume to be analyzed

    std::vector<double> region_min;
    std::vector<double> region_max;

    if (limit_volume_by_value)
      {
#ifdef USE_MPI
      if (find_volume_limits_value(df, dataset_name, region_limit_min_value,
				   region_limit_max_value, region_min, 
				   region_max, mpi_comm) < 0)
	{
	if (is_root)
	  cerr << endl << "Error: Unable to find volume limits by value" 
	       << endl << endl;

	MPI_Abort(mpi_comm, -2);
	}
#else
      if (find_volume_limits_value(df, dataset_name, region_limit_min_value,
				   region_limit_max_value, region_min, 
				   region_max) < 0)
	{
	cerr << endl << "Error: Unable to find volume limits by value" 
	     << endl << endl;

	return -2;
	}
#endif // USE_MPI
      }
    else if (limit_volume_by_refinement)
      {
#ifdef USE_MPI
      if (find_volume_limits_min_refine(df, min_refinement_level, region_min,
					region_max, mpi_comm) < 0)
	{
	if (is_root)
	  cerr << endl 
	       << "Error: Unable to find volume limits by refinement" 
	       << endl << endl;

	MPI_Abort(mpi_comm, -2);
	}
#else
      if (find_volume_limits_min_refine(df, min_refinement_level, region_min,
					region_max) < 0)
	{
	cerr << endl << "Error: Unable to find volume limits by refinement" 
	     << endl << endl;

	return -2;
	}
#endif // USE_MPI
      }
    else
      {
      region_min = volume_min;
      region_max = volume_max;
      }

#ifdef USE_MPI
    const double file_find_region_time = MPI_Wtime();

    if (is_root)
      {
      const double delta_time = file_find_region_time - file_start_time;

      cerr << "Found region in time [ " << delta_time << " s ]" << endl;
      }
#else
    cerr << "Found region" << endl;
#endif  // USE_MPI

    // Run through the box widths

    for (unsigned int scale_index = 0; scale_index < num_length_scales;
	 scale_index++)
      {
      const double length_scale = length_scale_vect[scale_index];

      const std::string & length_scale_str 
	= length_scale_str_vect[scale_index];

      // Build the output file suffix

      std::stringstream file_suffix_stream;

      file_suffix_stream << "_" << dataset_name << "_w" << length_scale_str 
			 << "_t" << threshold_value_str << "_" 
			 << base_output_file_suffix;

      const std::string output_file_suffix = file_suffix_stream.str();

#ifdef USE_MPI
      const double file_width_start_time = MPI_Wtime();
#endif  // USE_MPI

      if (is_root)
	cerr << "  " << "Processing box width [ " << length_scale << " ]" 
	     << endl;

      // Build the output filename

      std::string output_file_name;

      if (use_output_file_prefix)
	output_file_name = output_file_prefix;

      output_file_name += datafile_name + output_file_suffix;

      // Determine the size of the buffer region (the starting position)

      if (use_volume_buffer)
	{
	const double buffer_size 
	  = length_scale * volume_buffer_box_width_fraction;

	const double start_offset 
	  = buffer_size * (use_positive_start_offset ? 1.0 : -1.0);

	for (unsigned int i = 0; i < dims; i++)
	  region_min[i] += start_offset;

	// NOTE: Because of the way the grid is currently sized, the maximum
	// boundary does not need to be extended.
	}

      // Collect the data

      unsigned int accum_active_boxes = 0;

      unsigned int accum_boxes_above = 0;
      unsigned int accum_boxes_below = 0;
      unsigned int accum_boxes_transition = 0;

      std::list<BoxBoundary> box_list;

      std::vector< std::vector<double> > 
	grid_min_bounds_list(num_jitter_cycles);

      std::vector< std::vector<double> > 
	grid_max_bounds_list(num_jitter_cycles);

      std::vector<unsigned int> active_boxes_list(num_jitter_cycles, 0);

      std::vector<unsigned int> active_boxes_above_list(num_jitter_cycles, 0);

      std::vector<unsigned int> active_boxes_below_list(num_jitter_cycles, 0);

      std::vector<unsigned int> 
	active_boxes_transition_list(num_jitter_cycles, 0);

      for (unsigned int cycle_index = 0; cycle_index < num_jitter_cycles; 
	   cycle_index++)
	{
	if (is_root)
	  cerr << "    " << "Starting cycle " << cycle_index << endl;

	box_list.clear();  // IS THIS THE BEST PLACE FOR THIS???

	const std::vector<double> box_width(dims, length_scale);

	// Set up the BoxGrid object

	std::vector<double> cycle_region_min = region_min;

	if (use_jitter_average)
	  {
	  const double offset_sign = use_positive_start_offset ? 1.0 : -1.0;

	  for (unsigned int i = 0; i < dims; i++)
	    cycle_region_min[i] += get_random() * box_width[i] * offset_sign;
	  }

	if (is_root)
	  cerr << "      " << "Region min coords [ " << cycle_region_min
	       << " ]" << endl;

	std::vector<unsigned int> grid_dims(dims);

	for (unsigned int i = 0; i < dims; i++)
	  {
	  const double axis_box_width = box_width[i];

	  // const double axis_delta = region_max[i] - cycle_region_min[i];

	  const double axis_delta = volume_width[i];

	  unsigned int axis_cells 
	    = static_cast<unsigned int>(axis_delta / axis_box_width);

	  if (axis_cells < 1)
	    axis_cells = 1;

	  if ((static_cast<double>(axis_cells) * axis_box_width) < axis_delta)
	    axis_cells++;

	  grid_dims[i] = axis_cells;
	  }

	unsigned int cycle_active_boxes = 0;

	unsigned int cycle_boxes_above = 0;
	unsigned int cycle_boxes_below = 0;
	unsigned int cycle_boxes_transition = 0;

#ifdef USE_MPI
	get_boxes(df, dataset_name, threshold_value, cycle_region_min, 
		  box_width, grid_dims, cycle_active_boxes, cycle_boxes_above, 
		  cycle_boxes_below, cycle_boxes_transition, 
		  print_transition_boxes, box_list, 
		  grid_min_bounds_list[cycle_index], 
		  grid_max_bounds_list[cycle_index], mpi_comm, 
		  periodic_axes);
#else
	get_boxes(df, dataset_name, threshold_value, cycle_region_min, 
		  box_width, grid_dims, cycle_active_boxes, cycle_boxes_above, 
		  cycle_boxes_below, cycle_boxes_transition, 
		  print_transition_boxes, box_list, 
		  grid_min_bounds_list[cycle_index],
		  grid_max_bounds_list[cycle_index], periodic_axes);
#endif  // USE_MPI

	if (is_root)
	  cerr << "      " << "Grid bounds min [ " 
	       << grid_min_bounds_list[cycle_index] << " ] max [ " 
	       << grid_max_bounds_list[cycle_index] << " ]" << endl;

	accum_active_boxes += cycle_active_boxes;
	
	accum_boxes_above += cycle_boxes_above;
	accum_boxes_below += cycle_boxes_below;
	accum_boxes_transition += cycle_boxes_transition;

	active_boxes_list[cycle_index] = cycle_active_boxes;
	active_boxes_above_list[cycle_index] = cycle_boxes_above;
	active_boxes_below_list[cycle_index] = cycle_boxes_below;
	active_boxes_transition_list[cycle_index] = cycle_boxes_transition;
	}

      // Compute averages

      const unsigned int active_boxes = accum_active_boxes / num_jitter_cycles;

      const unsigned int boxes_above = accum_boxes_above / num_jitter_cycles;
      const unsigned int boxes_below = accum_boxes_below / num_jitter_cycles;
    
      const unsigned int boxes_transition 
	= accum_boxes_transition / num_jitter_cycles;

      // Report the results

      std::vector<unsigned int> periodic_axes_indexes;

      if (periodic_boundary)
	{
	for (unsigned int i = 0; i < dims; i++)
	  if (periodic_axes[i])
	    periodic_axes_indexes.push_back(i);
	}

      if (is_root)
	{
	std::ofstream f(output_file_name.c_str());

	f << comment_marker << "File " << datafile_name << endl;
	f << comment_marker << "  Dimensions " << dims << endl;
	f << comment_marker << "  SimTime " << sim_time << endl;
	f << comment_marker << "  VolumeMinBounds " << volume_min << endl;
	f << comment_marker << "  VolumeMaxBounds " << volume_max << endl;
	f << comment_marker << "  Dataset " << dataset_name << endl;

	f << comment_marker << "  EffectiveMeshCells " << effective_cell_count
	  << endl;

	f << comment_marker << "  SmallestCellWidth " << smallest_cell_width
	  << endl;

	f << endl;

	f << comment_marker << "  LengthScale " << length_scale << endl;
	f << comment_marker << "  Threshold " << threshold_value << endl;

	f << endl;

	f << comment_marker << "  ActiveBoxes " << active_boxes << endl;
	f << comment_marker << "    AboveBoxes " << boxes_above << endl;
	f << comment_marker << "    BelowBoxes " << boxes_below << endl;
	f << comment_marker << "    TransitionBoxes " << boxes_transition 
	  << endl;

	f << endl;

	if (periodic_boundary)
	  {
	  f << comment_marker << "PeriodicAxes " << periodic_axes_indexes 
	    << endl;

	  f << endl;
	  }

	f << comment_marker << "BoxcountCycles " << num_jitter_cycles << endl;

	for (unsigned int cycle_index = 0; cycle_index < num_jitter_cycles;
	       cycle_index++)
	  {
	  f << comment_marker << "  Cycle " << cycle_index 
	    << " GridMinBounds " << grid_min_bounds_list[cycle_index] 
	    << " GridMaxBounds " << grid_max_bounds_list[cycle_index] 
	    << " ActiveBoxes " << active_boxes_list[cycle_index]
	    << " AboveBoxes " << active_boxes_above_list[cycle_index]
	    << " BelowBoxes " << active_boxes_below_list[cycle_index]
	    << " TransitionBoxes " 
	    << active_boxes_transition_list[cycle_index] << endl;
	  }

	f << endl;

	if (print_transition_boxes)
	  {
	  // Reprint the information above without comment markers

	  f << "DataFile " << datafile_name << endl;
	  f << "  Dimensions " << dims << endl;
	  f << "  SimTime " << sim_time << endl;
	  f << "  Dataset " << dataset_name << endl;

	  f << "  EffectiveMeshCells " << effective_cell_count << endl;

	  f << "  SmallestCellWidth " << smallest_cell_width << endl;

	  f << endl;
    
	  f << "Threshold " << threshold_value << endl;
	  f << "LengthScale " << length_scale << endl;

	  f << endl;

	  f << "ActiveBoxes " << active_boxes << endl;
	  f << "  AboveBoxes " << boxes_above << endl;
	  f << "  BelowBoxes " << boxes_below << endl;
	  f << "  TransitionBoxes " << boxes_transition << endl;

	  f << endl;

	  if (periodic_boundary)
	    {
	    f << "PeriodicAxes " << periodic_axes_indexes << endl;

	    f << endl;
	    }

	  f << "BoxcountCycles " << num_jitter_cycles << endl;

	for (unsigned int cycle_index = 0; cycle_index < num_jitter_cycles;
	       cycle_index++)
	  {
	  f << "  Cycle " << cycle_index 
	    << " GridMinBounds " << grid_min_bounds_list[cycle_index] 
	    << " GridMaxBounds " << grid_max_bounds_list[cycle_index] 
	    << " ActiveBoxes " << active_boxes_list[cycle_index]
	    << " AboveBoxes " << active_boxes_above_list[cycle_index]
	    << " BelowBoxes " << active_boxes_below_list[cycle_index]
	    << " TransitionBoxes " 
	    << active_boxes_transition_list[cycle_index] << endl;
	  }

	  f << endl;

	  // Write the data for the individual boxes

	  f << "BoxData" << endl;

	  f << endl;

	  std::list<BoxBoundary>::const_iterator box_iter = box_list.begin();

	  while (box_iter != box_list.end())
	    {
	    const BoxBoundary & box_bounds = *box_iter;

	    f << "TransitionBox " << box_bounds.get_min_bounds() << " " 
	      << box_bounds.get_max_bounds() << endl;

	    ++box_iter;
	    }

	  f << endl;
	  }
	}
#ifdef USE_MPI
      const double file_width_finish_time = MPI_Wtime();

      if (is_root)
	{
	const double delta_file_width_time 
	  = file_width_finish_time - file_width_start_time;

	cerr << "\t" << "\t" << "Found boxes in time [ " 
	     << delta_file_width_time << " s ]" << endl;
	}
#endif  // USE_MPI
      }

#ifdef USE_MPI
    const double file_finish_time = MPI_Wtime();

    if (is_root)
      {
      const double delta_box_time = file_finish_time - file_find_region_time;

      cerr << "Found boxes in time [ " << delta_box_time << " s ]" << endl;

      const double delta_file_time = file_finish_time - file_start_time;

      cerr << "Processed file in time [ " << delta_file_time << " s ]" << endl;
      }
#endif  // USE_MPI

    ++filename_iter;
    }

#ifdef USE_MPI
  const double end_time = MPI_Wtime();

  if (is_root)
    {
    const double delta_time = end_time - start_time;

    cerr << "Finished files in time [ " << delta_time << " s ]" << endl
	 << endl;
    }

  MPI_Finalize();
#endif  // USE_MPI

  return 0;
  }
