# boxcount_sorter.py

# By Nathan C. Hearn
#    June 10, 2008
#
# Sorter for BoxcountInfo objects


import math
from rounding import round_value, value_close

class TimestepBoxes :

    def __init__(self, time, width_precision=None) :

        self.__width_precision = None

        if (width_precision is not None) :
            if (width_precision > 0.0) :
                self.__width_precision = width_precision

        self.__time = time

        self.__boxcount_info_list = dict()


    def add_boxcount_info(self, boxcount_info) :

        box_width = boxcount_info.get_box_width()

        if (self.__width_precision is not None) :

            ## Round width to nearest integer multiple of width_precision

            round_width = round_value(box_width, self.__width_precision)

            ## Search for a width close to this one

            width_list = self.get_box_width_list()

            num_widths = len(width_list)

            found_width = False

            width_index = int(0)

            while ((not found_width) and (width_index < num_widths)) :

                test_width = width_list[width_index]

                found_width = value_close(round_width, test_width,
                                          self.__width_precision)

                if (not found_width) :
                    width_index += 1

            if (not found_width) :
                
                self.__boxcount_info_list[round_width] = list()

                box_width = round_width
                
            else :
                
                box_width = width_list[width_index]

        else :
            
            if (not self.__boxcount_info_list.has_key(box_width)) :
                self.__boxcount_info_list[box_width] = list()
            
        self.__boxcount_info_list[box_width].append(boxcount_info)


    def get_time(self) :
        return self.__time

    def get_box_width_list(self) :
        return self.__boxcount_info_list.keys()

    def width_present(self, box_width) :

        ret_val = False

        if (self.__boxcount_info_list.has_key(box_width)) :
            ret_val = True

        return ret_val

    def get_info_list(self, box_width) :
        return self.__boxcount_info_list[box_width]


class BoxcountSorter :

    def __init__(self, time_precision=None, width_precision=None,
                 speed_precision=None) :

        self.__time_precision = None

        if (time_precision is not None) :
            if (time_precision > 0.0) :
                self.__time_precision = time_precision

        self.__width_precision = width_precision

        self.__speed_precision = None

        if (speed_precision is not None) :
            if (speed_precision > 0.0) :
                self.__speed_precision = speed_precision

        self.__timestep_boxcounts = dict()

        self.__resolution_speed_list = dict()
        self.__speed_resolution_list = dict()


    def add_boxcount_info(self, boxcount_info) :

        time = boxcount_info.get_sim_time()

        if (self.__time_precision is not None) :

            ## Round time to nearest integer multiple of time_precision

            round_time = round_value(time, self.__time_precision)

            ## Search for a time close to this one

            time_list = self.get_time_list()

            num_times = len(time_list)

            found_time = False

            time_index = int(0)

            while ((not found_time) and (time_index < num_times)) :

                test_time = time_list[time_index]

                if (math.fabs(round_time - test_time)
                    < self.__time_precision) :
                    found_time = True
                else :
                    time_index += 1

            if (not found_time) :
                
                self.__timestep_boxcounts[round_time] \
                    = TimestepBoxes(round_time, self.__width_precision)

                time = round_time
                
            else :
                
                time = time_list[time_index]

        else :
            
            if (not self.__timestep_boxcounts.has_key(time)) :
                self.__timestep_boxcounts[time] \
                    = TimestepBoxes(time, self.__width_precision)

        self.__timestep_boxcounts[time].add_boxcount_info(boxcount_info)


        resolution = boxcount_info.get_resolution()
        flame_speed = boxcount_info.get_flame_speed()

        if (self.__speed_precision is not None) :
            flame_speed = round_value(flame_speed, self.__speed_precision)


        if (not self.__resolution_speed_list.has_key(resolution)) :
            self.__resolution_speed_list[resolution] = list()

        speed_list = self.__resolution_speed_list[resolution]

        if (not (flame_speed in speed_list)) :
            speed_list.append(flame_speed)


        if (not self.__speed_resolution_list.has_key(flame_speed)) :
            self.__speed_resolution_list[flame_speed] = list()

        resolution_list = self.__speed_resolution_list[flame_speed]

        if (not (resolution in resolution_list)):
            resolution_list.append(flame_speed)


    def get_time_list(self) :
        return self.__timestep_boxcounts.keys()
    

    def time_present(self, time) :

        ret_val = False

        if (self.__boxcount_timesteps.has_key(time)) :
            ret_val = True

        return ret_val


    def get_timestep_boxes(self, time) :
        return self.__timestep_boxcounts[time]


    def get_flame_speeds(self) :
        return self.__speed_resolution_list.keys()

    def get_resolution_list(self, flame_speed) :
        return list(self.__speed_resolution_list[flame_speed])


    def get_resolutions(self) :
        return self.__resolution_speed_list.keys()
        
    def get_flame_speed_list(self, resolution) :
        return list(self.__resolution_speed_list[resolution])

        
    def get_resolution_speed_boxcount_list(self, resolution, flame_speed) :

        boxcount_list = list()

        time_list = self.get_time_list()
        time_list.sort()

        for time in time_list :

            timestep_boxes = self.__timestep_boxcounts[time]

            width_list = timestep_boxes.get_box_width_list()
            width_list.sort()

            for width in width_list :

                width_boxes = timestep_boxes.get_info_list(width)

                for boxcount_info in width_boxes :

                    include_info = False

                    if (boxcount_info.get_resolution() == resolution) :

                        info_flame_speed = boxcount_info.get_flame_speed()

                        if (self.__speed_precision is not None) :

                            include_info \
                                = value_close(flame_speed, info_flame_speed,
                                              self.__speed_precision)

                        else :

                            if (flame_speed == info_flame_speed) :
                                include_info = True
                            
                    if (include_info) :
                        boxcount_list.append(boxcount_info)

        return boxcount_list

