#!/usr/bin/env python

# boxcount_viewer.py

# By Nathan C. Hearn
#    February 8, 2008
#
# Viewer for boxcount data produced with the -transition-box-list flag


default_slicer = "slicer"
default_drawimage = "drawimage.py"
default_box_opacity = float(0.5)
default_page_width = int(800)

dims = int(2)

box_opacity = float(0.5)


from random import random


def get_random_value(min_value=0.0, max_value=1.0) :

    val_range = max_value - min_value

    rand_val = (random() * val_range) + min_value

    return rand_val


def get_random_color(min_value=0.0, max_value=1.0) :

    red = get_random_value(min_value, max_value)
    green = get_random_value(min_value, max_value)
    blue = get_random_value(min_value, max_value)

    return red, green, blue


def main() :

    from sys import exit, stdout, stderr
    from os import system
    from optparse import OptionParser
    import SciCompositor as scicomp
    from read_box_data import BoxData

    ## Parse the command-line input

    usage = str("%prog [ options ] slice_pixel_width {boxcount_output_files}")
    version = str("%prog 0.1")

    argparser = OptionParser(usage, version=version)

    skip_slicer_str = str("skip-slicer")
    argparser.add_option("--" + skip_slicer_str, action="store_true",
                         dest="skip_slicer", help="Do not generate slice")
    argparser.set_defaults(skip_slicer=False)

    slicer_str = str("slicer")
    argparser.add_option("--" + slicer_str, type="string",
                         dest="slicer", help="slicer executable")
    argparser.set_defaults(slicer=default_slicer)

    use_drawimage_str = str("use-drawimage")
    argparser.add_option("--" + use_drawimage_str, action="store_true",
                         dest="use_drawimage",
                         help="Use separate drawimage.py for slice")
    argparser.set_defaults(use_drawimage=False)

    drawimage_str = str("drawimage")
    argparser.add_option("--" + drawimage_str, type="string",
                         dest="drawimage", help="drawimage.py executable")
    argparser.set_defaults(drawimage=default_drawimage)

    output_prefix_str = str("output-prefix")
    argparser.add_option("--" + output_prefix_str, type="string",
                         dest="output_prefix", help="Output filename prefix")
    argparser.set_defaults(output_prefix=None)

    box_opacity_str = str("box-opacity")
    argparser.add_option("--" + box_opacity_str, type="float",
                         dest="box_opacity", help="Box object opacity")
    argparser.set_defaults(box_opacity=default_box_opacity)

    page_width_str = str("page-width")
    argparser.add_option("--" + page_width_str, type="int",
                         dest="page_width",
                         help="SVG page width in screen pixels")
    argparser.set_defaults(page_width=default_page_width)


    opts, args = argparser.parse_args()

    skip_slicer = opts.skip_slicer
    slicer_exec = opts.slicer

    use_drawimage = opts.use_drawimage
    drawimage_exec = opts.drawimage

    master_output_prefix = opts.output_prefix  ## May be None

    box_opacity = opts.box_opacity

    if ((box_opacity < 0.0) or (box_opacity > 1.0)) :
        stderr.write("\nError: Opacities must fall in range [0.0, 1.0]\n\n")
        exit(-1)

    page_width = opts.page_width

    if (page_width < 1) :
        stderr.write("\nError: Page width must be a positive integer\n\n")
        exit(-1)
        
    
    argc = len(args)

    min_args = int(2)

    if (argc < min_args) :
        argparser.print_help(file=stderr)
        exit(1)

    argptr = int(0)

    slice_pixel_width = float(args[argptr])
    argptr += 1

    if (not (slice_pixel_width > 0.0)) :
        stderr.write("\nError: Pixel width must be positive\n\n")
        exit(-1)

    oneOver_slice_pixel_width = 1.0 / slice_pixel_width

    boxcount_file_list = list()

    while (argptr < argc) :
        boxcount_file_list.append(str(args[argptr]))
        argptr += 1

    num_boxcount_files = len(boxcount_file_list)


    ## Read the data

    boxcount_data_list = list()

    for filename in boxcount_file_list :

        boxcount_data = BoxData(filename)

        if (boxcount_data.get_dims() != 2) :
            raise RuntimeError("Only 2D boxcount data supported at this time")
        
        boxcount_data_list.append(boxcount_data)


    ## Determine global spatial limits on the sampled region and basic info

    prim_boxcount_data = boxcount_data_list[0]

    datafile_name = prim_boxcount_data.get_data_file_name()  ## SAME FOR ALL???
    dataset_name = prim_boxcount_data.get_dataset_name()  ## SAME FOR ALL???

    threshold_value = prim_boxcount_data.get_threshold()  ## SAME FOR ALL???

    min_bounds = boxcount_data_list[0].get_min_bounds()
    max_bounds = boxcount_data_list[0].get_max_bounds()

    for index in range(1, num_boxcount_files) :

        boxcount_data = boxcount_data_list[index]

        current_min_bounds = boxcount_data.get_min_bounds()
        current_max_bounds = boxcount_data.get_max_bounds()

        for i in range(dims) :

            current_min = current_min_bounds[i]
            current_max = current_max_bounds[i]

            if (current_min < min_bounds[i]) :
                min_bounds[i] = current_min

            if (current_max > max_bounds[i]) :
                max_bounds[i] = current_max

    slice_center = list()

    for i in range(dims) :
        axis_center = 0.5 * (min_bounds[i] + max_bounds[i])
        slice_center.append(axis_center)

    slice_width = list()

    for i in range(dims) :
        axis_width = max_bounds[i] - min_bounds[i]
        slice_width.append(axis_width)

    slice_pixels = list()

    for i in range(dims) :

        axis_pixels = int((slice_width[i] * oneOver_slice_pixel_width) + 0.25)

        if (axis_pixels < 1) :
            axis_pixels = 1

        slice_pixels.append(axis_pixels)

    slice_aspect_height_width = slice_width[1] / slice_width[0]

    ## *Guess* the name of the slice image file and create the SVG file name

    output_file_prefix = ""
    
    if (master_output_prefix is not None) :
        output_file_prefix += master_output_prefix + "_"

    output_file_prefix += datafile_name + "_" + dataset_name

    slice_suffix = ""

    if (use_drawimage) :
        slice_suffix += ".csv"

    slice_suffix += ".png"

    slice_output_file = output_file_prefix + slice_suffix
    svg_output_file = output_file_prefix + "_boxes.svg"


    ## Compute the slice

    if (not skip_slicer) :
        
        slice_cmd = slicer_exec + " " + dataset_name + " " \
                    + str(slice_center[0]) + "," + str(slice_center[1]) + " " \
                    + str(slice_width[0]) + "," + str(slice_width[1]) + " " \
                    + str(slice_pixels[0]) + "," + str(slice_pixels[1]) + " " \
                    + "--subsamples=10,10"

        if (master_output_prefix is not None) :
            slice_cmd += " --output-prefix=" + master_output_prefix

        if (not use_drawimage) :
            slice_cmd += " --image-type=png --image-only"

        slice_cmd += " " + datafile_name

        system(slice_cmd)

        if (use_drawimage) :

            slice_csv_name = output_file_prefix + ".csv"

            drawimage_cmd = drawimage_exec + " " + slice_csv_name

            system(drawimage_cmd)


    ## Construct the SVG file

    canvas_width = 1.2
    canvas_height = canvas_width * (slice_aspect_height_width)

    sliceport_start_x = 0.1
    sliceport_end_x = 1.1

    sliceport_start_y = sliceport_start_x * slice_aspect_height_width
    sliceport_end_y = sliceport_end_x * slice_aspect_height_width

    sliceport_width = sliceport_end_x - sliceport_start_x
    sliceport_height = sliceport_end_y - sliceport_start_y
    
    canvas = scicomp.Canvas(0.0, canvas_width, 0.0, canvas_height)


    ## Build the slicer viewport

    sliceport = canvas.get_new_viewport("sliceport",
                                        sliceport_start_x, sliceport_end_x,
                                        sliceport_start_y, sliceport_end_y,
                                        0.0, sliceport_width,
                                        0.0, sliceport_height)

    sliceport.set_physical_coord_map(min_bounds[0], max_bounds[0],
                                     min_bounds[1], max_bounds[1])

    ## Draw the image

    slice_img = scicomp.ImageObject((dataset_name + "_slice"),
                                    slice_output_file,
                                    min_bounds[0], max_bounds[0],
                                    min_bounds[1], max_bounds[1],
                                    use_physical_coords=True)

    sliceport.add_child_object(slice_img)

    ## Draw the boxes

    for boxcount_data in boxcount_data_list :

        length_scale = boxcount_data.get_length_scale()

        viewport_name = "boxes_" + str(length_scale)

        boxport = sliceport.get_sub_viewport(viewport_name,
                                             0.0, sliceport_width,
                                             0.0, sliceport_height)

        boxport.set_physical_coord_map(min_bounds[0], max_bounds[0],
                                       min_bounds[1], max_bounds[1])

        num_boxes = boxcount_data.get_transition_boxes()

        box_red, box_green, box_blue = get_random_color(0.2, 0.9)

        for box_index in range(num_boxes) :

            box_min = boxcount_data.get_transition_box_min_bounds(box_index)
            box_max = boxcount_data.get_transition_box_max_bounds(box_index)

            boxname = viewport_name + "_" + str(box_index)

            boxobj = scicomp.BoxObject(boxname, box_min[0], box_max[0],
                                       box_min[1], box_max[1],
                                       use_physical_coords=True)

            boxobj.set_fill_color(red=box_red, green=box_green, blue=box_blue,
                                  opacity=box_opacity)
            boxobj.set_line_opacity(0.0)
            boxobj.set_line_width(0.0)

            boxport.add_child_object(boxobj)

    ## Add the axes

    slice_axes = scicomp.AxesObject((dataset_name + "_axes"),
                                    min_bounds[0], max_bounds[0],
                                    min_bounds[1], max_bounds[1],
                                    use_physical_coords=True)

    slice_axes.set_frame_color(red=0.0, green=1.0, blue=0.5)
    slice_axes.set_frame_line_width(0.001)

    slice_axes.set_x_major_ticks_regular(max_ticks=11)
    slice_axes.set_y_major_ticks_regular(max_ticks=11)

    slice_axes.set_major_ticks_line_width(0.001)

    slice_axes.set_axis_label_font_size(0.02)

    sliceport.add_child_object(slice_axes)


    ## Write the SVG file
                                             
    svg_code = canvas.get_svg_code(page_width)

    svg_file = file(svg_output_file, "w")

    for svg_line in svg_code :

        svg_file.write(svg_line + "\n")

    svg_file.write("\n")


if (__name__ == "__main__") :
    main()

    

    
