// C++ program file boxdata.cpp

/*
  By Nathan C. Hearn
     December 11, 2007

  Box data collection classes.
*/


#include "boxdata.hpp"
#include <vector>
#include <list>
#include "quickflash_except.hpp"
#include "quickflash_geometry.hpp"
#include "quickflash_counters.hpp"
#include "box_utils.hpp"


// Class BoxGrid

void BoxGrid::reset(const std::vector<double> & global_min_coords, 
		    const std::vector<double> & box_width,
		    const std::vector<unsigned int> & global_grid_dims,
		    const bool use_grid_subset,
		    const unsigned int subset_start_box_index,
		    const unsigned int subset_num_boxes)
  {
  reset();

  const unsigned int dims = global_min_coords.size();

  if (box_width.size() != dims)
    throw QuickFlash::Except("Incompatible box width vector size", __FILE__,
			   __LINE__);

  if (global_grid_dims.size() != dims)
    throw QuickFlash::Except("Incompatible grid dimensions vector size", 
			   __FILE__, __LINE__);

  if (dims < 1)
    throw QuickFlash::Except("At least one dimension required for box grid",
			   __FILE__, __LINE__);

  std::vector<double> global_max_coords = global_min_coords;

  for (unsigned int i = 0; i < dims; i++)
    global_max_coords[i] 
      += static_cast<double>(global_grid_dims[i]) * box_width[i];

  grid_info.reset(global_min_coords, global_max_coords, global_grid_dims,
		  QuickFlash::Geometry::Cartesian);

  const unsigned int total_boxes = grid_info.get_num_cells();

  if (use_grid_subset)
    {
    if ((subset_start_box_index + subset_num_boxes) > total_boxes)
      throw QuickFlash::Except("Subset indexes out of range", __FILE__, 
			     __LINE__);

    subset_start_index = subset_start_box_index;
    subset_size = subset_num_boxes;
    }
  else
    {
    subset_start_index = 0;
    subset_size = total_boxes;
    }

  subset_end_index = subset_start_index + subset_size;  // One past last box

  grid_data.resize(subset_size);

  // Clear the data

  for (unsigned int index = 0; index < subset_size; index++)
    grid_data[index].reset();
  }


void BoxGrid::reset(const std::vector<double> & global_min_coords, 
		    const std::vector<double> & box_width,
		    const std::vector<unsigned int> & global_grid_dims,
		    const std::vector<bool> & periodic_axis_flags,
		    const std::vector<double> & volume_min_coords,
		    const std::vector<double> & volume_max_coords,
		    const bool use_grid_subset,
		    const unsigned int subset_start_box_index,
		    const unsigned int subset_num_boxes)
  {
  reset();

  const unsigned int dims = global_min_coords.size();

  if (box_width.size() != dims)
    throw QuickFlash::Except("Incompatible box width vector size", __FILE__,
			   __LINE__);

  if (global_grid_dims.size() != dims)
    throw QuickFlash::Except("Incompatible grid dimensions vector size", 
			   __FILE__, __LINE__);

  if (periodic_axis_flags.size() > 0)
    {
    if (periodic_axis_flags.size() != dims)
      throw QuickFlash::Except("Incompatible periodic axis flags vector size",
			     __FILE__, __LINE__);
    else
      {
      if ((volume_min_coords.size() != dims) 
	  || (volume_max_coords.size() != dims))
	throw QuickFlash::Except("Incompatible volume coordinates size",
			       __FILE__, __LINE__);

      for (unsigned int i = 0; i < dims; i++)
	if (periodic_axis_flags[i])
	  periodic_boundary = true;

      if (periodic_boundary)
	{
	periodic_axis = periodic_axis_flags;

	volume_min_bounds = volume_min_coords;
	volume_max_bounds = volume_max_coords;

	for (unsigned int i = 0; i < dims; i++)
	  if (!(volume_max_bounds[i] > volume_min_bounds[i]))
	    throw QuickFlash::Except("Non-positive volume width", __FILE__, 
				   __LINE__);
	}
      }
    }

  if (dims < 1)
    throw QuickFlash::Except("At least one dimension required for box grid",
			   __FILE__, __LINE__);

  std::vector<double> global_max_coords = global_min_coords;

  for (unsigned int i = 0; i < dims; i++)
    {
    const double axis_box_width = box_width[i];

    if (!(axis_box_width > 0.0))
      throw QuickFlash::Except("Non-positive box width", __FILE__, __LINE__);

    global_max_coords[i] 
      += static_cast<double>(global_grid_dims[i]) * box_width[i];
    }

  grid_info.reset(global_min_coords, global_max_coords, global_grid_dims,
		  QuickFlash::Geometry::Cartesian);

  const unsigned int total_boxes = grid_info.get_num_cells();

  if (use_grid_subset)
    {
    if ((subset_start_box_index + subset_num_boxes) > total_boxes)
      throw QuickFlash::Except("Subset indexes out of range", __FILE__, 
			     __LINE__);

    subset_start_index = subset_start_box_index;
    subset_size = subset_num_boxes;
    }
  else
    {
    subset_start_index = 0;
    subset_size = total_boxes;
    }

  subset_end_index = subset_start_index + subset_size;  // One past last box

  grid_data.resize(subset_size);

  // Clear the data

  for (unsigned int index = 0; index < subset_size; index++)
    grid_data[index].reset();
  }


void BoxGrid::add_cell_data(const std::vector<double> & cell_min_coords,
			    const std::vector<double> & cell_max_coords,
			    const double value)
  {
  const unsigned int dims = grid_info.get_space_dims();

  if (cell_min_coords.size() != dims)
    throw QuickFlash::Except("Incompatible min coords vector size", __FILE__,
			   __LINE__);

  if (cell_max_coords.size() != dims)
    throw QuickFlash::Except("Incompatible max coords vector size", __FILE__,
			   __LINE__);

  // Run through all boxes

  std::vector<double> box_min_coords;
  std::vector<double> box_max_coords;

  for (unsigned int global_box_index = subset_start_index; 
       global_box_index < subset_end_index; global_box_index++)
    {
    grid_info.get_cell_bounds(global_box_index, box_min_coords, 
			      box_max_coords);

    bool overlap = false;

    if (periodic_boundary)
      overlap = regions_overlap(periodic_axis, 
				cell_min_coords, cell_max_coords, 
				box_min_coords, box_max_coords,
				volume_min_bounds, volume_max_bounds);
    else
      overlap = regions_overlap(cell_min_coords, cell_max_coords, 
				box_min_coords, box_max_coords);

    if (overlap)
      {
      const unsigned int local_box_index 
	= global_box_index - subset_start_index;

      grid_data[local_box_index].add_value(value);
      }
    }
  }


bool BoxGrid::overlaps_local_boxes(const std::vector<double> 
				     & region_min_coords,
				   const std::vector<double> 
				     & region_max_coords) const
  {
  bool ret_val = false;

  unsigned int box_index = subset_start_index;

  std::vector<double> box_min_coords;
  std::vector<double> box_max_coords;

  while ((!ret_val) && (box_index < subset_end_index))
    {
    grid_info.get_cell_bounds(box_index, box_min_coords, box_max_coords);

    if (periodic_boundary)
      ret_val = regions_overlap(periodic_axis,
				region_min_coords, region_max_coords,
				box_min_coords, box_max_coords,
				volume_min_bounds, volume_max_bounds);
    else
      ret_val = regions_overlap(region_min_coords, region_max_coords,
				box_min_coords, box_max_coords);

    box_index++;
    }

  return ret_val;
  }


unsigned int BoxGrid::count_boxes_threshold(const double threshold_value,
					    unsigned int & boxes_above,
					    unsigned int & boxes_below,
					    unsigned int & boxes_transition)
  {
  unsigned int count_above = 0;
  unsigned int count_below = 0;
  unsigned int count_trans = 0;

  unsigned int total_boxes = 0;

  for (unsigned int index = 0; index < subset_size; index++)
    {
    const BoxData & boxdata = grid_data[index];

    if (boxdata.value_set())
      {
      const bool above = boxdata.above_threshold(threshold_value);      
      const bool below = boxdata.below_threshold(threshold_value);

      if (above)
	count_above++;

      if (below)
	count_below++;

      if (above && below)
	count_trans++;

      total_boxes++;
      }
    }

  boxes_above = count_above;
  boxes_below = count_below;
  boxes_transition = count_trans;

  return total_boxes;
  }


unsigned int BoxGrid::get_transition_boxes(const double threshold_value,
					   std::list<BoxBoundary> & box_list)
  const
  {
  // SHOULD WE CLEAR BOX_LIST???

  unsigned int count_trans = 0;

  std::vector<double> min_coords;
  std::vector<double> max_coords;

  BoxBoundary work_box;

  for (unsigned int index = 0; index < subset_size; index++)
    {
    const BoxData & boxdata = grid_data[index];

    if (boxdata.value_set())
      {
      const bool above = boxdata.above_threshold(threshold_value);      
      const bool below = boxdata.below_threshold(threshold_value);

      if (above && below)
	{
	count_trans++;

	// Store the box

	const unsigned int global_index = index + subset_start_index;

	grid_info.get_cell_bounds(global_index, min_coords, max_coords);

	work_box.reset(min_coords, max_coords);

	box_list.push_back(work_box);
	}
      }
    }

  return count_trans;
  }


unsigned int BoxGrid::get_below_boxes(const double threshold_value,
				      std::list<BoxBoundary> & box_list)
  const
  {
  // SHOULD WE CLEAR BOX_LIST???

  unsigned int count_below = 0;

  std::vector<double> min_coords;
  std::vector<double> max_coords;

  BoxBoundary work_box;

  for (unsigned int index = 0; index < subset_size; index++)
    {
    const BoxData & boxdata = grid_data[index];

    if (boxdata.value_set())
      {
      const bool below = boxdata.below_threshold(threshold_value);

      if (below)
	{
	count_below++;

	// Store the box

	const unsigned int global_index = index + subset_start_index;

	grid_info.get_cell_bounds(global_index, min_coords, max_coords);

	work_box.reset(min_coords, max_coords);

	box_list.push_back(work_box);
	}
      }
    }

  return count_below;
  }


unsigned int BoxGrid::get_above_boxes(const double threshold_value,
				      std::list<BoxBoundary> & box_list) const
  {
  // SHOULD WE CLEAR BOX_LIST???

  unsigned int count_above = 0;

  std::vector<double> min_coords;
  std::vector<double> max_coords;

  BoxBoundary work_box;

  for (unsigned int index = 0; index < subset_size; index++)
    {
    const BoxData & boxdata = grid_data[index];

    if (boxdata.value_set())
      {
      const bool above = boxdata.above_threshold(threshold_value);

      if (above)
	{
	count_above++;

	// Store the box

	const unsigned int global_index = index + subset_start_index;

	grid_info.get_cell_bounds(global_index, min_coords, max_coords);

	work_box.reset(min_coords, max_coords);

	box_list.push_back(work_box);
	}
      }
    }

  return count_above;
  }
