// C++ header file boxdata.hpp

/*
  By Nathan C. Hearn
     December 11, 2007

  Box data collection classes.
*/


#ifndef BOXDATA_HPP
#define BOXDATA_HPP


#include <vector>
#include <list>
#include "quickflash_block_blockinfo.hpp"
#include "quickflash_block_blockdata.hpp"


class BoxBoundary
  {
  public :

    BoxBoundary() : min_coords(), max_coords() { }

    BoxBoundary(const std::vector<double> & min_bounds,
		const std::vector<double> & max_bounds) :
      min_coords(), max_coords()
      { reset(min_bounds, max_bounds); }

    BoxBoundary(const BoxBoundary & source) :
      min_coords(), max_coords()
      { reset(source); }

    ~BoxBoundary() { }

    BoxBoundary & operator=(const BoxBoundary & source)
      {
      reset(source);
      return *this;
      }

    void reset()
      {
      min_coords.clear();
      max_coords.clear();
      }

    void reset(const std::vector<double> & min_bounds,
	       const std::vector<double> & max_bounds)
      {
      min_coords = min_bounds;
      max_coords = max_bounds;
      }

    void reset(const BoxBoundary & source)
      {
      if (&source != this)
	{
	min_coords = source.min_coords;
	max_coords = source.max_coords;
	}
      }

    const std::vector<double> & get_min_bounds() const { return min_coords; }
    const std::vector<double> & get_max_bounds() const { return max_coords; }

  private :
    
    std::vector<double> min_coords;
    std::vector<double> max_coords;
  };


class BoxData
  {
  public :

    BoxData() : 
      initialized(false), min_value(0.0), max_value(0.0)
      { }

    BoxData(const BoxData & source) : 
      initialized(false), min_value(0.0), max_value(0.0)
      { reset(source); }

    ~BoxData() { }

    BoxData & operator=(const BoxData & source)
      {
      reset(source);
      return *this;
      }

    void reset()
      {
      initialized = false;

      min_value = 0.0;
      max_value = 0.0;
      }

    void reset(const BoxData & source)
      {
      if (&source != this)
	{
	initialized = source.initialized;

	min_value = source.min_value;
	max_value = source.max_value;
	}
      }

    void add_value(const double value)
      {
      if (initialized)
	{
	if (value < min_value)
	  min_value = value;
	else if (value > max_value)
	  max_value = value;
	}
      else
	{
	min_value = value;
	max_value = value;

	initialized = true;
	}
      }

    bool value_set() const { return initialized; }

    double get_min_value() const { return min_value; }
    double get_max_value() const { return max_value; }

    bool above_threshold(const double threshold_value) const
      { return (max_value > threshold_value); }

    bool below_threshold(const double threshold_value) const
      { return (min_value < threshold_value); }

    bool transits_threshold(const double threshold_value) const
      { 
      return (above_threshold(threshold_value) 
	      && below_threshold(threshold_value));
      }


  private :

    bool initialized;

    double min_value;
    double max_value;

  };


class BoxGrid
  {
  public :

    BoxGrid() : 
      grid_info(), grid_data(), periodic_boundary(false), periodic_axis(),
      volume_min_bounds(), volume_max_bounds(),
      subset_start_index(0), subset_size(0),
      subset_end_index(0), subset_min_coords(), subset_max_coords()
      { }

    BoxGrid(const std::vector<double> & global_min_coords, 
	    const std::vector<double> & box_width,
	    const std::vector<unsigned int> & global_grid_dims,
	    const bool use_grid_subset=false,
	    const unsigned int subset_start_box_index=0,
	    const unsigned int subset_num_boxes=0) :
      grid_info(), grid_data(), periodic_boundary(false), periodic_axis(),
      volume_min_bounds(), volume_max_bounds(),
      subset_start_index(0), subset_size(0),
      subset_end_index(0), subset_min_coords(), subset_max_coords()
      { 
      reset(global_min_coords, box_width, global_grid_dims, use_grid_subset, 
	    subset_start_box_index, subset_num_boxes); 
      }

    BoxGrid(const std::vector<double> & global_min_coords, 
	    const std::vector<double> & box_width,
	    const std::vector<unsigned int> & global_grid_dims,
	    const std::vector<bool> & periodic_axis_flags=std::vector<bool>(0),
	    const std::vector<double> 
	      & volume_min_coords=std::vector<double>(0),
	    const std::vector<double> 
	      & volume_max_coords=std::vector<double>(0),
	    const bool use_grid_subset=false,
	    const unsigned int subset_start_box_index=0,
	    const unsigned int subset_num_boxes=0) :
      grid_info(), grid_data(), periodic_boundary(false), periodic_axis(), 
      volume_min_bounds(), volume_max_bounds(),
      subset_start_index(0), subset_size(0),
      subset_end_index(0), subset_min_coords(), subset_max_coords()
      { 
      reset(global_min_coords, box_width, global_grid_dims, 
	    periodic_axis_flags, volume_min_coords, volume_max_coords, 
	    use_grid_subset, subset_start_box_index, subset_num_boxes); 
      }

    BoxGrid(const BoxGrid & source) :
      grid_info(), grid_data(), periodic_boundary(false), periodic_axis(),
      volume_min_bounds(), volume_max_bounds(),
      subset_start_index(0), subset_size(0),
      subset_end_index(0), subset_min_coords(), subset_max_coords()
      { reset(source); }

    ~BoxGrid() { }

    BoxGrid & operator=(const BoxGrid & source)
      {
      reset(source);
      return *this;
      }

    BoxData & operator[](const unsigned int global_box_index)
      { return grid_data[global_box_index - subset_start_index]; }

    const BoxData & operator[](const unsigned int global_box_index) const
      { return grid_data[global_box_index - subset_start_index]; }

    BoxData & operator[](const std::vector<unsigned int> & global_box_index)
      {
      const unsigned int global_index 
	= grid_info.get_scalar_index(global_box_index);

      return grid_data[global_index - subset_start_index];
      }

    const BoxData & operator[](const std::vector<unsigned int> 
			       & global_box_index) const
      {
      const unsigned int global_index 
	= grid_info.get_scalar_index(global_box_index);

      return grid_data[global_index - subset_start_index];
      }

    void get_box_boundary(const unsigned int global_box_index,
			  std::vector<double> & min_bounds,
			  std::vector<double> & max_bounds) const
    { grid_info.get_cell_bounds(global_box_index, min_bounds, max_bounds); }

    void get_box_boundary(const std::vector<unsigned int> & global_box_index,
			  std::vector<double> & min_bounds,
			  std::vector<double> & max_bounds) const
    { grid_info.get_cell_bounds(global_box_index, min_bounds, max_bounds); }

    unsigned int get_global_num_boxes() const 
      { return grid_info.get_num_cells(); }

    unsigned int get_start_box_index() const { return subset_start_index; }
    unsigned int get_end_box_index() const { return subset_end_index; }

    unsigned int get_num_boxes() const { return subset_size; }

    const std::vector<unsigned int> & get_global_grid_dims() const 
      { return grid_info.get_block_dims(); }

    const std::vector<double> & get_global_min_bounds() const
      { return grid_info.get_min_coords(); }

    const std::vector<double> & get_global_max_bounds() const
      { return grid_info.get_max_coords(); }

    const std::vector<double> & get_local_min_bounds() const
      { return subset_min_coords; }

    const std::vector<double> & get_local_max_bounds() const
      { return subset_max_coords; }

    void reset()
      {
      grid_info.reset();
      grid_data.clear();

      periodic_boundary = false;

      periodic_axis.clear();

      volume_min_bounds.clear();
      volume_max_bounds.clear();

      subset_start_index = 0;
      subset_size = 0;

      subset_end_index = 0;

      subset_min_coords.clear();
      subset_max_coords.clear();
      }

    void reset(const std::vector<double> & global_min_coords, 
	       const std::vector<double> & box_width,
	       const std::vector<unsigned int> & global_grid_dims,
               const bool use_grid_subset=false,
               const unsigned int subset_start_box_index=0,
               const unsigned int subset_num_boxes=0);

    void reset(const std::vector<double> & global_min_coords, 
	       const std::vector<double> & box_width,
	       const std::vector<unsigned int> & global_grid_dims,
	       const std::vector<bool> 
	         & periodic_axis_flags=std::vector<bool>(0),
	       const std::vector<double> 
	         & volume_min_coords=std::vector<double>(0),
	       const std::vector<double> 
	         & volume_max_coords=std::vector<double>(0),
               const bool use_grid_subset=false,
               const unsigned int subset_start_box_index=0,
               const unsigned int subset_num_boxes=0);

    void reset(const BoxGrid & source)
      {
      if (&source != this)
	{
	grid_info = source.grid_info;
	grid_data = source.grid_data;

	periodic_boundary = source.periodic_boundary;

	periodic_axis = source.periodic_axis;

	volume_min_bounds = source.volume_min_bounds;
	volume_max_bounds = source.volume_max_bounds;

	subset_start_index = source.subset_start_index;
	subset_size = source.subset_size;

	subset_end_index = source.subset_end_index;

	subset_min_coords = source.subset_min_coords;
	subset_max_coords = source.subset_max_coords;
	}
      }

    void add_cell_data(const std::vector<double> & cell_min_coords,
		       const std::vector<double> & cell_max_coords,
		       const double value);

    bool overlaps_local_boxes(const std::vector<double> & region_min_coords,
			      const std::vector<double> & region_max_coords) 
      const;

    unsigned int count_boxes_threshold(const double threshold_value,
				       unsigned int & boxes_above,
				       unsigned int & boxes_below,
				       unsigned int & boxes_transition);

    unsigned int get_transition_boxes(const double threshold_value,
				      std::list<BoxBoundary> & box_list) const;

    unsigned int get_below_boxes(const double threshold_value,
				 std::list<BoxBoundary> & box_list) const;

    unsigned int get_above_boxes(const double threshold_value,
				 std::list<BoxBoundary> & box_list) const;

  private :

    QuickFlash::Block::BlockInfo grid_info;
    std::vector<BoxData> grid_data;

    bool periodic_boundary;

    std::vector<bool> periodic_axis;

    std::vector<double> volume_min_bounds;
    std::vector<double> volume_max_bounds;

    unsigned int subset_start_index;
    unsigned int subset_size;

    unsigned int subset_end_index;  // One past the last index in the subset

    std::vector<double> subset_min_coords;
    std::vector<double> subset_max_coords;
  };


#endif  // BOXDATA_HPP
