# generate_plot.py

# By Nathan C. Hearn
#    October 16, 2007


import matplotlib
import pylab
import numpy


class Markers :
    
    def __init__(self, start_offset=0) :
        
        self.__marker_list = [ "s", "o", "^", ">", "v",
                               "<", "d", "p", "h", "x", "+" ]

        self.__marker_none = "None"
        
        self.__num_markers = len(self.__marker_list)
        
        self.__current_value = start_offset % self.__num_markers

        if (self.__current_value < 0) :
            self.__current_value = 0
            

    def next(self) :
        self.__current_value += 1
        
        if (not (self.__current_value < self.__num_markers)) :
            self.__current_value = 0

        return self.__marker_list[self.__current_value]
    

    def current(self) :
        return self.__marker_list[self.__current_value]
    

    def get_marker(self, marker_index=None) :

        ret_val = None

        if (marker_index is None) :

            ret_val = self.current()

            self.next()

        elif (marker_index < 0) :

            ret_val = self.__marker_none

        else :

            limited_index = marker_index % self.__num_markers

            ret_val = self.__marker_list[limited_index]

        return ret_val


class Colors :
    
    def __init__(self, start_offset=0) :
        
        self.__color_list = [ "b", "g", "r", "c", "m", "y", "k", "0.5" ]
        
        self.__num_colors = len(self.__color_list)
        
        self.__current_value = start_offset % self.__num_colors

        if (self.__current_value < 0) :
            self.__current_value = 0
            

    def next(self) :
        self.__current_value += 1

        if (not (self.__current_value < self.__num_colors)) :
            self.__current_value = 0

        return self.__color_list[self.__current_value]
    

    def current(self) :
        return self.__color_list[self.__current_value]
    

    def get_color(self, color_index=None) :

        ret_val = None

        if (color_index is None) :

            ret_val = self.current()

            self.next()

        else :

            limited_index = color_index % self.__num_colors

            ret_val = self.__color_list[limited_index]

        return ret_val


class Plot :

    def __init__(self, plot_title=None, x_axis_title=None, y_axis_title=None) :

        self.__plot_title = plot_title

        self.__page_height = float(8.0)
        self.__aspect_ratio_width_height = float(1.25)

        self.__x_axis_title = x_axis_title
        self.__y_axis_title = y_axis_title

        self.__x_ticks_values = None
        self.__x_ticks_labels = None

        self.__y_ticks_values = None
        self.__y_ticks_labels = None

        self.__x_data = list()
        self.__y_data = list()

        self.__legend_titles = list()

        self.__show_legend = True

        self.__plot_colors = list()
        self.__line_types = list()
        self.__symbols = list()
        self.__symbol_fill_color = list()

        self.__min_x = None
        self.__max_x = None

        self.__min_y = None
        self.__max_y = None

        self.__logscale_x = False
        self.__logscale_y = False

        self.__marker_size = float(0.01)

        self.__marker_gen = Markers()
        self.__color_gen = Colors()
        self.__marker_color_gen = Colors()

        self.__legend_prop \
            = matplotlib.font_manager.FontProperties(size="x-small")


    def set_limits(self, min_x=None, max_x=None, min_y=None, max_y=None) :

        self.__min_x = min_x
        self.__max_x = max_x

        self.__min_y = min_y
        self.__max_y = max_y

        if ((min_x is not None) and (max_x is not None)) :
            if (not (min_x < max_x)) :
                raise RuntimeError("x-axis bounds mismatch")

        if ((min_y is not None) and (max_y is not None)) :
            if (not (min_y < max_y)) :
                raise RuntimeError("y-axis bounds mismatch")

    def set_page_height(self, height) :

        if (not (height > 0.0)) :
            raise RuntimeError("Height must be positive")

        self.__page_height = height
        

    def set_aspect_ratio_width_height(self, ratio) :

        if (not (ratio > 0.0)) :
            raise RuntimeError("Aspect ratio must be positive")

        self.__aspect_ratio_width_height = ratio


    def show_legend(self, state=True) :
        self.__show_legend = state


    def set_x_ticks(self, value_list, label_list=None) :
        self.__x_ticks_values = value_list
        self.__x_ticks_labels = label_list


    def set_y_ticks(self, value_list, label_list=None) :
        self.__y_ticks_values = value_list
        self.__y_ticks_labels = label_list


    def set_logscale_x(self, state=True) :
        self.__logscale_x = state


    def set_logscale_y(self, state=True) :
        self.__logscale_y = state


    def set_marker_size(self, marker_size) :
        self.__marker_size = marker_size


    def add_plot(self, x_data, y_data, legend_text=None, color_index=None,
                 line_type="-", marker_index=-1, use_prev_color=False,
                 fill_color_index=None, use_prev_fill_color=False) :

        x_len = len(x_data)

        if (len(y_data) != x_len) :
            raise RuntimeError("Incompatible x-data and y-data arrays")

        self.__x_data.append(numpy.array(x_data))
        self.__y_data.append(numpy.array(y_data))

        if (legend_text is None) :
            self.__legend_titles.append("")
        else :
            self.__legend_titles.append(str(legend_text))

        if (not use_prev_color) :
            self.__plot_colors.append(self.__color_gen.get_color(color_index))
        else :
            self.__plot_colors.append(self.__color_gen.current())
            
        self.__line_types.append(line_type)
        
        self.__symbols.append(self.__marker_gen.get_marker(marker_index))

        marker_fill_color = "None"

        if (use_prev_fill_color) :

            marker_fill_color = self.__marker_color_gen.current()

        elif (fill_color_index is not None) :

            marker_fill_color \
                = self.__marker_color_gen.get_color(fill_color_index)

        self.__symbol_fill_color.append(marker_fill_color)


    def show_plot(self, filename=None) :

        if (filename is not None) :
            matplotlib.use("SVG")
        else :
            matplotlib.use("GTKAgg")

        self.__set_plot_bounds()

        page_width = self.__page_height * self.__aspect_ratio_width_height

        fig = pylab.figure(figsize=(page_width, self.__page_height))

        fig.clf()
        fig.hold(True)

        plot = fig.add_subplot(1,1,1)

        if (self.__logscale_x and self.__logscale_y) :
            plot.loglog()
        elif (self.__logscale_x) :
            plot.semilogx()
        elif (self.__logscale_y) :
            plot.semilogy()

        if (self.__plot_title is not None) :
            plot.set_title(self.__plot_title)

        if (self.__x_axis_title is not None) :
            plot.set_xlabel(self.__x_axis_title)

        if (self.__y_axis_title is not None) :
            plot.set_ylabel(self.__y_axis_title)

        num_plots = len(self.__x_data)

        for i in range(num_plots) :

            plot.plot(self.__x_data[i], self.__y_data[i],
                      c=self.__plot_colors[i], linestyle=self.__line_types[i],
                      marker=self.__symbols[i],
                      markerfacecolor=self.__symbol_fill_color[i],
                      markeredgecolor=self.__plot_colors[i],
                      label=self.__legend_titles[i])

        if (self.__x_ticks_values is not None) :
            plot.set_xticks(self.__x_ticks_values)

            if (self.__x_ticks_labels is not None) :
                plot.set_xticklabels(self.__x_ticks_labels)

        if (self.__y_ticks_values is not None) :
            plot.set_xticks(self.__y_ticks_values, self.__y_ticks_labels)

            if (self.__y_ticks_labels is not None) :
                plot.set_yticklabels(self.__y_ticks_labels)

        plot.set_xlim(self.__min_x, self.__max_x)
        plot.set_ylim(self.__min_y, self.__max_y)

        plot.grid()

        if (self.__show_legend) :
            plot.legend(loc="best", prop=self.__legend_prop)

        if (filename is not None) :
            matplotlib.use("SVG")
            pylab.savefig(filename, orientation="portrait")
        else :
            matplotlib.use("GTKAgg")
            pylab.show()


    def __set_plot_bounds(self) :

        num_plots = len(self.__x_data)

        if (num_plots < 1) :
            raise RuntimeError("No plot data present")

        local_min_x = min(self.__x_data[0])
        local_max_x = max(self.__x_data[0])

        local_min_y = min(self.__y_data[0])
        local_max_y = max(self.__y_data[0])

        for i in range(1, num_plots) :

            x_data = self.__x_data[i]
            y_data = self.__y_data[i]

            min_x = min(x_data)
            max_x = max(x_data)

            min_y = min(y_data)
            max_y = max(y_data)

            if (min_x < local_min_x) :
                local_min_x = min_x

            if (max_x > local_max_x) :
                local_max_x = max_x

            if (min_y < local_min_y) :
                local_min_y = min_y

            if (max_y > local_max_y) :
                local_max_y = max_y

        if (self.__min_x is None) :
            self.__min_x = local_min_x

        if (self.__max_x is None) :
            self.__max_x = local_max_x

        if (self.__min_y is None) :
            self.__min_y = local_min_y

        if (self.__max_y is None) :
            self.__max_y = local_max_y

        if (not (self.__min_x < self.__max_x)) :
            raise RuntimeError("Minimum x value must be less than maximum")

        if (not (self.__min_y < self.__max_y)) :
            raise RuntimeError("Minimum y value must be less than maximum")





