#!/usr/bin/env python

# plot_boxcounts.py

# By Nathan C. Hearn
#    June 10, 2008
#
# Plotter for boxcount information


try :
    import psyco
except ImportError :
    pass


default_plot_file_type = str("png")

default_time_precision = float(0.01)
default_width_precision = float(10.0)
default_speed_precision = float(0.1)

output_data_suffix = str("txt")


def main() :

    from sys import argv, exit, stderr
    from os import path
    import math

    from optparse import OptionParser

    from boxcount_info import read_boxcount_file
    from boxcount_sorter import BoxcountSorter

    from rounding import value_close

    from generate_plot import Plot

    ## Set up the command-line parser

    usage = str("%prog [options] output_prefix boxcount_file1 "
                + "[ boxcount_file2 ... ]")
    version = str("%prog 0.1")

    argparser = OptionParser(usage, version=version)

    plot_file_type_str = str("plot-file-type")
    argparser.add_option("--" + plot_file_type_str, type="string",
                         dest="plot_file_type",
                         help="Plot file type (extension)")
    argparser.set_defaults(plot_file_type=default_plot_file_type)

    output_data_str = str("output-data")
    argparser.add_option("--" + output_data_str, action="store_true",
                         dest="output_data", help="Save processed data")
    argparser.set_defaults(output_data=False)


    ## Process the command-line input

    opts, args = argparser.parse_args()

    argc = len(args)

    min_args = int(2)

    if (argc < min_args) :
        stderr.write("\n")
        argparser.print_help(file=stderr)
        stderr.write("\n")
        exit(1)


    argptr = int(0)

    output_prefix = str(args[argptr])
    argptr += 1

    boxcount_file_list = list()

    while (argptr < argc) :
        boxcount_file_list.append(str(args[argptr]))
        argptr += 1


    plot_file_type = opts.plot_file_type
    
    output_data = opts.output_data
    
    time_precision = default_time_precision
    width_precision = default_width_precision
    speed_precision = default_speed_precision


    ## Additional option processing

    if (path.split(output_prefix)[1] != "") :
        output_prefix += "_"


    ## Read the data

    sorter = BoxcountSorter(time_precision, width_precision, speed_precision)

    for boxcount_file in boxcount_file_list :
        boxcount_info = read_boxcount_file(boxcount_file)
        sorter.add_boxcount_info(boxcount_info)

    flame_speed_list = sorter.get_flame_speeds()
    flame_speed_list.sort()
    
    resolution_list = sorter.get_resolutions()
    resolution_list.sort()

    num_flame_speeds = len(flame_speed_list)

    ## Create plots by time

    time_list = sorter.get_time_list()
    time_list.sort()

    for time in time_list :

        timestep_boxcounts = sorter.get_timestep_boxes(time)

        ## We will be progressing towards smaller box widths

        box_width_list = timestep_boxcounts.get_box_width_list()
        box_width_list.sort()
        box_width_list.reverse()

        ## Collect data

        width_lists = dict()
        count_lists = dict()

        for speed in flame_speed_list :
            for res in resolution_list :
                width_lists[(res, speed)] = list()
                count_lists[(res, speed)] = list()

        for box_width in box_width_list :

            boxcount_list = timestep_boxcounts.get_info_list(box_width)

            for box_info in boxcount_list :

                count = box_info.get_transition_boxes()

                print "-> ", box_width, count

                res = box_info.get_resolution()
                speed = box_info.get_flame_speed()

                ## Get the appropriate speed key

                found_key = False

                key_index = 0

                while ((not found_key) and (key_index < num_flame_speeds)) :

                    key_speed = flame_speed_list[key_index]

                    if (value_close(speed, key_speed, speed_precision)) :
                        found_key = True
                    else :
                        key_index += 1

                if (not found_key) :
                    stderr.write("\nError: Unable to find speed key\n\n")
                    exit(-2)

                speed = flame_speed_list[key_index]

                width_lists[(res, speed)].append(box_width)
                count_lists[(res, speed)].append(float(count))

        ## Build the plot

        time_str = "%(tm)3.2f" % { "tm" : time }

        count_plot_title = "Transition Box Counts vs. Box Widths - Time = " \
                           + time_str
        
        slope_plot_title = "Fractal Dimension Plot - Time = " + time_str

        count_plot = Plot(count_plot_title, "Box Width (m)", "Box Count")

        count_plot.set_logscale_x()
        count_plot.set_logscale_y()

        slope_plot = Plot(slope_plot_title, "Mid-Width (m)",
                          "d[ log10(box_count) ] / d[ log10(1 / box_width) ]")

        slope_plot.set_logscale_x()

        ref_widths = list()
        ref_counts = list()

        ref_mid_widths = list()
        ref_slopes = list()

        res_speed_keys = width_lists.keys()
        res_speed_keys.sort()

        marker_index = int(0)

        for (res, speed) in res_speed_keys :

            widths = width_lists[(res, speed)]

            num_points = len(widths)

            if (num_points > 1) :

                counts = count_lists[(res, speed)]

                log_widths = list()
                log_oneOver_widths = list()
                log_counts = list()

                for point_index in range(num_points) :

                    log10_width = math.log10(widths[point_index])

                    log_widths.append(log10_width)
                    log_oneOver_widths.append(-1.0 * log10_width)
                    log_counts.append(math.log10(counts[point_index]))

                mid_widths = list()
                slopes = list()
                
                for point_index in range(num_points - 1) :

                    p1_log_width = log_widths[point_index]
                    p2_log_width = log_widths[point_index + 1]

                    p1_log_oneOver_width = log_oneOver_widths[point_index]
                    p2_log_oneOver_width = log_oneOver_widths[point_index + 1]

                    p1_log_count = log_counts[point_index]
                    p2_log_count = log_counts[point_index + 1]

                    delta_log_oneOver_width \
                        = p2_log_oneOver_width - p1_log_oneOver_width
                    
                    delta_log_count = p2_log_count - p1_log_count

                    log_mid_width = 0.5 * (p1_log_width + p2_log_width)

                    geom_mid_width = math.pow(10.0, log_mid_width)

                    mid_widths.append(geom_mid_width)
                    slopes.append(delta_log_count / delta_log_oneOver_width)

                legend_text = str(res) + (" s=%(sp)3.1f" % { "sp" : speed }) \
                              + " km / s"

                count_plot.add_plot(widths, counts, legend_text,
                                    marker_index=marker_index,
                                    fill_color_index=marker_index)
        
                slope_plot.add_plot(mid_widths, slopes, legend_text,
                                    marker_index=marker_index,
                                    fill_color_index=marker_index)

                marker_index += 1

                if (num_points > len(ref_widths)) :
                    ref_widths = widths
                    ref_counts = counts

                if (num_points > (len(ref_mid_widths) + 1)) :
                    ref_mid_widths = mid_widths
                    ref_slopes = slopes


        ## Set labels for x-axis

        ref_widths_labels = list()

        for width in ref_widths :
            ref_widths_labels.append("%(num)5.0f" % { "num" : (width * 0.01) })

        ref_mid_widths_labels = list()

        for width in ref_mid_widths :
            ref_mid_widths_labels.append("%(num)5.0f"
                                         % { "num" : (width * 0.01) })

        count_plot.set_x_ticks(ref_widths, ref_widths_labels)

        slope_plot.set_x_ticks(ref_mid_widths, ref_mid_widths_labels)

        ## Save the plots
                
        count_plot_file_name = output_prefix + "time_count_t" + time_str \
                               + "." + plot_file_type

        slope_plot_file_name = output_prefix + "time_slope_t" + time_str \
                               + "." + plot_file_type

        count_plot.show_plot(count_plot_file_name)
        slope_plot.show_plot(slope_plot_file_name)


    ## Create plots by flame speed

    speed_list = sorter.get_flame_speeds()
    speed_list.sort()
    
    resolution_list = sorter.get_resolutions()
    resolution_list.sort()

    for flame_speed in speed_list :

        ## Collect data

        width_lists = dict()
        count_lists = dict()

        mid_width_lists = dict()
        slope_lists = dict()

        for time in time_list :
            for res in resolution_list :

                width_lists[(res, time)] = list()
                count_lists[(res, time)] = list()

                mid_width_lists[(res, time)] = list()
                slope_lists[(res, time)] = list()

        ref_widths = list()
        ref_counts = list()

        ref_mid_widths = list()
        ref_slopes = list()

        for resolution in resolution_list :

            info_list = sorter.get_resolution_speed_boxcount_list(resolution,
                                                                  flame_speed)

            for info_obj in info_list :

                info_time = info_obj.get_sim_time()

                key_time = 0.0

                for time in time_list :
                    if (value_close(info_time, time, time_precision)) :
                        key_time = time

                width = info_obj.get_box_width()
                count = info_obj.get_transition_boxes()

                width_lists[(resolution, key_time)].append(width)
                count_lists[(resolution, key_time)].append(count)


        ## Build the plot

        speed_str = "%(sp)3.2f" % { "sp" : flame_speed }

        count_plot_title = "Transition Box Counts vs. Box Widths - " \
                           + "Flame Speed = " + speed_str
        
        slope_plot_title = "Fractal Dimension Plot - Flame Speed = " \
                           + speed_str

        count_plot = Plot(count_plot_title, "Box Width (m)",
                          "Box Count")

        count_plot.set_logscale_x()
        count_plot.set_logscale_y()

        slope_plot = Plot(slope_plot_title, "Mid-Width (m)",
                          "d[ log10(box_count) ] / d[ log10(1 / box_width) ]")

        slope_plot.set_logscale_x()

        res_time_keys = width_lists.keys()
        res_time_keys.sort()

        marker_index = int(0)

        for (res, time) in res_time_keys :

            widths = width_lists[(res, time)]

            num_points = len(widths)

            if (num_points > 1) :

                counts = count_lists[(res, time)]

                log_widths = list()
                log_oneOver_widths = list()
                log_counts = list()

                for point_index in range(num_points) :

                    log10_width = math.log10(widths[point_index])

                    log_widths.append(log10_width)
                    log_oneOver_widths.append(-1.0 * log10_width)
                    log_counts.append(math.log10(counts[point_index]))

                mid_widths = mid_width_lists[(res, time)]
                slopes = slope_lists[(res, time)]
                
                for point_index in range(num_points - 1) :

                    p1_log_width = log_widths[point_index]
                    p2_log_width = log_widths[point_index + 1]

                    p1_log_oneOver_width = log_oneOver_widths[point_index]
                    p2_log_oneOver_width = log_oneOver_widths[point_index + 1]

                    p1_log_count = log_counts[point_index]
                    p2_log_count = log_counts[point_index + 1]

                    delta_log_oneOver_width \
                        = p2_log_oneOver_width - p1_log_oneOver_width
                    
                    delta_log_count = p2_log_count - p1_log_count

                    log_mid_width = 0.5 * (p1_log_width + p2_log_width)

                    geom_mid_width = math.pow(10.0, log_mid_width)

                    mid_widths.append(geom_mid_width)
                    slopes.append(delta_log_count / delta_log_oneOver_width)

                legend_text = str(res) + (" t=%(tm)3.2f" % { "tm" : time }) \
                              + " sec"
                
                count_plot.add_plot(widths, counts, legend_text,
                                    marker_index=marker_index,
                                    fill_color_index=marker_index)
        
                slope_plot.add_plot(mid_widths, slopes, legend_text,
                                    marker_index=marker_index,
                                    fill_color_index=marker_index)

                marker_index += 1

                if (num_points > len(ref_widths)) :
                    ref_widths = widths
                    ref_counts = counts

                if (num_points > (len(ref_mid_widths) + 1)) :
                    ref_mid_widths = mid_widths
                    ref_slopes = slopes

        ## Set labels for x-axis

        ref_widths_labels = list()

        for width in ref_widths :
            ref_widths_labels.append("%(num)5.0f" % { "num" : (width * 0.01) })

        ref_mid_widths_labels = list()

        for width in ref_mid_widths :
            ref_mid_widths_labels.append("%(num)5.0f"
                                         % { "num" : (width * 0.01) })

        count_plot.set_x_ticks(ref_widths, ref_widths_labels)

        slope_plot.set_x_ticks(ref_mid_widths, ref_mid_widths_labels)

        ## Save the plots
                
        count_plot_file_name = output_prefix + "speed_count_s" + speed_str \
                               + "." + plot_file_type

        slope_plot_file_name = output_prefix + "speed_slope_s" + speed_str \
                               + "." + plot_file_type

        count_plot.show_plot(count_plot_file_name)
        slope_plot.show_plot(slope_plot_file_name)

        if (output_data) :

            count_data_file_name = output_prefix + "speed_count_s" \
                                   + speed_str + "." + output_data_suffix
            
            slope_data_file_name = output_prefix + "speed_slope_s" \
                                   + speed_str + "." + output_data_suffix
            
            count_f = open(count_data_file_name, "w")
            slope_f = open(slope_data_file_name, "w")

            count_f.write("Time\tResolution\tBox_Width\tBox_Count\n")
            slope_f.write("Time\tResolution\tBox_Mid_Width\tEffective_Dimension\n")

            for time in time_list :
                for res in resolution_list :

                    widths = width_lists[(res, time)]
                    counts = count_lists[(res, time)]

                    num_counts = len(counts)

                    for index in range(num_counts) :
                        count_f.write(str(time) + "\t" + str(res) + "\t"
                                      + str(widths[index]) + "\t" 
                                      + str(counts[index]) + "\n")

                    mid_widths = mid_width_lists[(res, time)]
                    slopes = slope_lists[(res, time)]

                    num_slopes = len(slopes)

                    for index in range(num_slopes) :
                        slope_f.write(str(time) + "\t" + str(res) + "\t"
                                      + str(mid_widths[index]) + "\t" 
                                      + str(slopes[index]) + "\n")
    

    ## Provide some information

    stderr.write("Times: " + str(sorter.get_time_list()) + "\n")
    stderr.write("Speeds: " + str(sorter.get_flame_speeds()) + "\n")
    stderr.write("Resolutions: " + str(sorter.get_resolutions()) + "\n")

    stderr.write("\n\n")


if (__name__ == "__main__") :
    main()


    
