# read_box_data.py

# By Nathan C. Hearn
#    February 7, 2008
#
# Python-based reader for output from boxcount (with -trasition-box-list
# option)


datafile_keyword = "DataFile"
dataset_keyword = "Dataset"

threshold_keyword = "Threshold"
lengthscale_keyword = "LengthScale"

activeboxes_keyword = "ActiveBoxes"
aboveboxes_keyword = "AboveBoxes"
belowboxes_keyword = "BelowBoxes"
transitionboxes_keyword = "TransitionBoxes"

minbounds_keyword = "MinBounds"
maxbounds_keyword = "MaxBounds"

transitionbox_keyword = "TransitionBox"

improper_entries = "Improper number of entries for "


def get_vector(vector_string) :

    ret_val = list()

    vector_words = vector_string.split(",")

    num_elems = len(vector_words)

    for i in range(num_elems) :
        ret_val.append(float(vector_words[i]))

    return ret_val


def read_next_line(file_obj) :

    ret_val = None  ## Return this if end of file reached

    finished = False

    while (not finished) :

        raw_line = file_obj.readline()

        if (raw_line == "") :
            
            finished = True  ## At end of file

        else :

            comment_free_line = raw_line.split("#")[0].strip()

            if (len(comment_free_line) > 0) :
                ret_val = comment_free_line
                finished = True

    return ret_val


class BoxData :

    def __init__(self, filename=None) :

        self.__data_file = None
        self.__dataset = None

        self.__threshold = None
        self.__length_scale = None

        self.__active_boxes = None
        self.__above_boxes = None
        self.__below_boxes = None
        self.__transition_boxes = None

        self.__dims = None

        self.__min_bounds = None
        self.__max_bounds = None

        self.__transition_box_min_coords = list()
        self.__transition_box_max_coords = list()

        if (filename is not None) :
            self.reset(filename)


    def reset(self, filename=None) :

        self.__data_file = None
        self.__dataset = None

        self.__threshold = None
        self.__length_scale = None

        self.__active_boxes = None
        self.__above_boxes = None
        self.__below_boxes = None
        self.__transition_boxes = None

        self.__dims = None

        self.__min_bounds = None
        self.__max_bounds = None

        self.__transition_box_min_coords = list()
        self.__transition_box_max_coords = list()

        if (filename is not None) :
            self.__read_file(filename)
            self.__check_data()


    def get_data_file_name(self) :
        return self.__data_file

    def get_dataset_name(self) :
        return self.__dataset

    def get_threshold(self) :
        return self.__threshold

    def get_length_scale(self) :
        return self.__length_scale

    def get_active_boxes(self) :
        return self.__active_boxes

    def get_above_boxes(self) :
        return self.__above_boxes

    def get_below_boxes(self) :
        return self.__below_boxes

    def get_transition_boxes(self) :
        return self.__transition_boxes

    def get_dims(self) :
        return self.__dims

    def get_min_bounds(self) :
        return list(self.__min_bounds)

    def get_max_bounds(self) :
        return list(self.__max_bounds)

    def get_transition_box_min_bounds(self, index) :
        return list(self.__transition_box_min_coords[index])

    def get_transition_box_max_bounds(self, index) :
        return list(self.__transition_box_max_coords[index])


    def __read_file(self, filename) :

        ## To be called from self.reset

        f = file(filename, "r")

        finished = False

        while (not finished) :

            next_line = read_next_line(f)

            if (next_line is None) :

                finished = True

            else :

                line_words = next_line.split()

                num_words = len(line_words)

                if (num_words > 0) :

                    keyword = line_words[0]

                    if (keyword == datafile_keyword) :

                        if (num_words == 2) :

                            self.__data_file = str(line_words[1])

                        else :

                            raise RuntimeError(improper_entries
                                               + datafile_keyword)

                    elif (keyword == dataset_keyword) :

                        if (num_words == 2) :

                            self.__dataset = str(line_words[1])

                        else :

                            raise RuntimeError(improper_entries
                                               + dataset_keyword)

                    elif (keyword == threshold_keyword) :

                        if (num_words == 2) :

                            self.__threshold = float(line_words[1])

                        else :

                            raise RuntimeError(improper_entries
                                               + threshold_keyword)
            
                    elif (keyword == lengthscale_keyword) :

                        if (num_words == 2) :

                            self.__length_scale = float(line_words[1])

                        else :

                            raise RuntimeError(improper_entries
                                               + lengthscale_keyword)
            
                    elif (keyword == activeboxes_keyword) :

                        if (num_words == 2) :

                            self.__active_boxes = int(line_words[1])

                        else :

                            raise RuntimeError(improper_entries
                                               + activeboxes_keyword)
            
                    elif (keyword == aboveboxes_keyword) :

                        if (num_words == 2) :

                            self.__above_boxes = int(line_words[1])

                        else :

                            raise RuntimeError(improper_entries
                                               + aboveboxes_keyword)
            
                    elif (keyword == belowboxes_keyword) :

                        if (num_words == 2) :

                            self.__below_boxes = int(line_words[1])

                        else :

                            raise RuntimeError(improper_entries
                                               + belowboxes_keyword)
            
                    elif (keyword == transitionboxes_keyword) :

                        if (num_words == 2) :

                            self.__transition_boxes = int(line_words[1])

                        else :

                            raise RuntimeError(improper_entries
                                               + transitionboxes_keyword)

                    elif (keyword == minbounds_keyword) :

                        if (num_words == 2) :

                            self.__min_bounds = get_vector(line_words[1])

                        else :

                            raise RuntimeError(improper_entries
                                               + minbounds_keyword)
            
                    elif (keyword == maxbounds_keyword) :

                        if (num_words == 2) :

                            self.__max_bounds = get_vector(line_words[1])

                        else :

                            raise RuntimeError(improper_entries
                                               + maxbounds_keyword)
            
                    elif (keyword == transitionbox_keyword) :

                        if (num_words == 3) :

                            min_coords = get_vector(line_words[1])
                            max_coords = get_vector(line_words[2])

                            self.__transition_box_min_coords.append(min_coords)
                            self.__transition_box_max_coords.append(max_coords)

    
    def __check_data(self) :

        from sys import stderr

        ## Make sure all data entries have been read

        ## SHOULD WE CHECK VECTOR DIMENSIONS???  MAYBE SET DIMS???

        error_state = False

        if (self.__data_file is None) :

            stderr.write("Keyword [ " + datafile_keyword + " ] missing\n")
            error_state = True

        if (self.__dataset is None) :
            
            stderr.write("Keyword [ " + dataset_keyword + " ] missing\n")
            error_state = True
            
        if (self.__threshold is None) :
            
            stderr.write("Keyword [ " + threshold_keyword + " ] missing\n")
            error_state = True
            
        if (self.__length_scale is None) :
            
            stderr.write("Keyword [ " + lengthscale_keyword + " ] missing\n")
            error_state = True
            
        if (self.__min_bounds is None) :
            
            stderr.write("Keyword [ " + minbounds_keyword + " ] missing\n")
            error_state = True

        else :

            if (self.__dims is None) :
                
                self.__dims = len(self.__min_bounds)
                
            elif (len(self.__min_bounds) != self.__dims) :
                
                stderr.write("Incompatible dimensions for keyword [ "
                             + minbounds_keyword + " ]\n")
                error_state = True
            
        if (self.__max_bounds is None) :
            
            stderr.write("Keyword [ " + maxbounds_keyword + " ] missing\n")
            error_state = True

        else :

            if (self.__dims is None) :
                
                self.__dims = len(self.__max_bounds)
                
            elif (len(self.__max_bounds) != self.__dims) :
                
                stderr.write("Incompatible dimensions for keyword [ "
                             + maxbounds_keyword + " ]\n")
                error_state = True
            
        if (self.__active_boxes is None) :
            
            stderr.write("Keyword [ " + activeboxes_keyword + " ] missing\n")
            error_state = True
            
        if (self.__above_boxes is None) :
            
            stderr.write("Keyword [ " + aboveboxes_keyword + " ] missing\n")
            error_state = True
            
        if (self.__below_boxes is None) :
            
            stderr.write("Keyword [ " + belowboxes_keyword + " ] missing\n")
            error_state = True
            
        if (self.__transition_boxes is None) :
            
            stderr.write("Keyword [ " + transitionboxes_keyword
                         + " ] missing\n")
            error_state = True

        elif ((len(self.__transition_box_min_coords)
               != self.__transition_boxes)
              or (len(self.__transition_box_max_coords)
                  != self.__transition_boxes)) :

            stderr.write("Transition boxes read does not match reported size"
                         + "\n")
            error_state = True

        else :

            ## CHECK DIMENSIONS

            for index in range(self.__transition_boxes) :

                if ((len(self.__transition_box_min_coords[index])
                     != self.__dims)
                    or (len(self.__transition_box_max_coords[index])
                        != self.__dims)) :

                    stderr.write("Incompatible dimensions for transition box "
                                 + str(index) + "\n")
                    error_state = True

        ## Throw an exception on error
            
        if (error_state) :
            raise RuntimeError("Items missing in data file")


