// C++ program file treedemo.cpp

/*
  By Nathan C. Hearn
     December 11, 2006

  Demonstration of tree-related functions in the QuickFlash library, including 
  tree node iterators for Flash data files.
*/


#include <hdf5.h>
#include <string>
#include <vector>
#include <iostream>
#include <stdlib.h>
#include "quickflash_file_datafile.hpp"
#include "quickflash_block_blockinfo.hpp"
#include "quickflash_block_blockdata.hpp"
#include "quickflash_file_meshinfo.hpp"
#include "quickflash_tree_block.hpp"
#include "quickflash_tree_nodedata.hpp"


bool bxor(const bool b1, const bool b2)
  {
  bool value = false;

  if ((b1 && (!b2)) || ((!b1) && (b2)))
    value = true;

  return value;
  }


int main(int argc, char * argv[])
  {
  using std::cout;
  using std::cerr;
  using std::endl;

  if (argc < 6)
    {
    cerr << "Usage: " << argv[0] << " filename dataset buffer_size " 
	 << "cache_size <position>" << endl
	 << endl;
    cerr << "Gets the data for the specified position" 
	 << endl << endl;
    return 1;
    }

  int argPtr = 1;

  const std::string file_name = argv[argPtr++];
  const std::string dataset_name = argv[argPtr++];

  const unsigned int buffer_size = atoi(argv[argPtr++]);

  const unsigned int cache_size = atoi(argv[argPtr++]);

  std::vector<double> pos;

  while (argPtr < argc)
    pos.push_back(atof(argv[argPtr++]));

  // Open the file

  QuickFlash::File::DataFile dfile(file_name, buffer_size, true);

  // Get simulation info

  const QuickFlash::File::MeshInfo & meshinfo = dfile.get_mesh_info();

  const unsigned int num_blocks = meshinfo.get_num_blocks();
  const unsigned int dims = meshinfo.get_dims();

  cerr << "Dims [ " << dims << " ]" << endl;
  cerr << "Num blocks [ " << num_blocks << " ]" << endl;

  unsigned int num_leaves = 0;

  for (unsigned int block_index = 0; block_index < num_blocks; block_index++)
    if (meshinfo.is_leaf(block_index))
      num_leaves++;

  cerr << "Num leaves [ " << num_leaves << " ]" << endl;

  // Get information

  const std::vector<unsigned int> & block_dims = meshinfo.get_block_dims();

  cerr << "Block dimensions [ ";

  for (unsigned int i = 0; i < dims; i++)
    cerr << block_dims[i] << " ";

  cerr << "]" << endl;

  const std::vector<unsigned int> & base_block_dims 
    = meshinfo.get_base_block_dims();

  cerr << "Base block dimensions [ ";

  for (unsigned int i = 0; i < dims; i++)
    cerr << base_block_dims[i] << " ";

  cerr << "]" << endl;

  // Count the blocks

  QuickFlash::Tree::ConstTreeIter node_iter = meshinfo.get_tree_iter();

  unsigned int forest_block_count = 0;
  unsigned int forest_leaf_count = 0;

  while (node_iter.valid_node())
    {
    forest_block_count++;

    const QuickFlash::Tree::NodeData & node_data = node_iter.get_node_data();

    const unsigned int block_index = node_data.get_block_index();

    // Check leaves

    bool node_leaf = false;
    bool block_leaf = false;

    if (node_iter.is_leaf())
      {
      forest_leaf_count++;
      node_leaf = true;
      }

    if (meshinfo.is_leaf(block_index))
      block_leaf = true;

    if (bxor(node_leaf, block_leaf))
      cerr << "ERROR: Leaf mismatch at block " << block_index << endl;

    node_iter.next();
    }

  cerr << "Num tree blocks " << forest_block_count << endl;
  cerr << "Num tree leaves " << forest_leaf_count << endl;

  // Check the leaves

  node_iter.first();

  // Open a dataset

  const QuickFlash::File::Dataset & dset 
    = dfile.get_dataset(dataset_name);

  dset.set_cache_size(cache_size);
  dset.set_report_stats();

  cerr << "Dataset loaded" << endl;

  const unsigned int block_index = meshinfo.get_block_index(pos);

  cerr << "Block index [ " << block_index << " ]" << endl;

  const QuickFlash::Block::BlockInfo & block_info 
    = meshinfo.get_block_info(block_index);

  QuickFlash::Block::BlockData<double> data;

  if (block_info.in_block(pos))
    cerr << "Position in block" << endl;
  else
    cerr << "ERROR: Incorrect block found" << endl;

  const std::vector<double> & block_min_coords = block_info.get_min_coords();
  const std::vector<double> & block_max_coords = block_info.get_max_coords();

  cerr << "Block min [ ";

  for (unsigned int i = 0; i < dims; i++)
    cerr << block_min_coords[i] << " ";

  cerr << "] max [ ";

  for (unsigned int i = 0; i < dims; i++)
    cerr << block_max_coords[i] << " ";

  cerr << "]" << endl;

  dset.get_block_data(block_index, data);

  const unsigned int cell_index = block_info.get_nearest_cell(pos);

  cerr << "Cell index [ " << cell_index << " ]" << endl;

  cerr << "Data [ " << data[cell_index] << " ]" << endl;

  return 0;
  }
