// C++ program file binrt.cpp

/*
  By Nathan C. Hearn
     September 21, 2006

  Program for estimating mixing layer boundary positions for rt_mmode_nch2
  Flash simulation data.


  Single-Mode Boundaries

    ref_bounds_poly_lower_a =  7.37757762e-03
    ref_bounds_poly_lower_b =  6.13329193e+06
    ref_bounds_poly_lower_c = -8.75648952e+13
    ref_bounds_poly_lower_d =  7.43771014e+20

    ref_bounds_poly_upper_a =  1.70299864e-02
    ref_bounds_poly_upper_b =  6.71604177e+06
    ref_bounds_poly_upper_c = -4.03671469e+13
    ref_bounds_poly_upper_d = -1.93651537e+20

  Double-Mode Boundaries

    ref_bounds_poly_lower_a =  7.59735880e-03
    ref_bounds_poly_lower_b =  5.95425811e+06
    ref_bounds_poly_lower_c = -7.57874769e+13
    ref_bounds_poly_lower_d =  5.34742737e+20

    ref_bounds_poly_upper_a =  1.75142951e-02
    ref_bounds_poly_upper_b =  6.64159562e+06
    ref_bounds_poly_upper_c = -4.31708314e+13
    ref_bounds_poly_upper_d = -7.74470832e+19
*/


#include <iostream>
#include <vector>
#include <string>
#include <list>
#include "moving_boundary.hpp"
#include "cyl_domain.hpp"
#include "part_map.hpp"
#include "quickflash_file_datafile.hpp"
#include "quickflash_file_meshinfo.hpp"
#include "stat_bins.hpp"
#include "quickflash_file_dataset.hpp"
#include "quickflash_block_blockinfo.hpp"
#include "quickflash_block_blockdata.hpp"



// const std::string poly_name = "poly";
// const std::string chbr_name = "chbr";
const std::string dens_name = "dens";

const double domain_rad = 0.03;

const double bin_width = 0.0002;
const unsigned int buffer_size = 200;


// Single-Mode tinitial = 0.0

// const double single_lower_a =  7.37757762e-03;
// const double single_lower_b =  6.13329193e+06;
// const double single_lower_c = -8.75648952e+13;
// const double single_lower_d =  7.43771014e+20;

// const unsigned int single_lower_num_elems = 4;
// const double single_lower[] = { 7.0e-03, 6.13329193e+06, 
// 				-8.75648952e+13, 7.43771014e+20 };

// const double single_upper_a =  1.70299864e-02;
// const double single_upper_b =  6.71604177e+06;
// const double single_upper_c = -4.03671469e+13;
// const double single_upper_d = -1.93651537e+20;

// const unsigned int single_upper_num_elems = 4;
// const double single_upper[] = { 1.8e-02, 6.71604177e+06, 
// 				-4.03671469e+13, -1.93651537e+20 };


// Single-Mode tinitial = 1.5

// const double single_lower_a = -1.84789885e-03;
// const double single_lower_b =  6.33734185e+06;
// const double single_lower_c = -8.75422870e+13;
// const double single_lower_d =  7.08325309e+20;

const unsigned int single_lower_num_elems = 4;
const double single_lower[] = { -1.9e-03, 6.33734185e+06, 
				-8.75422870e+13, 7.08325309e+20 };

// const double single_upper_a =  6.37274073e-03
// const double single_upper_b =  7.07826704e+06
// const double single_upper_c = -6.34113186e+13
// const double single_upper_d =  2.54394281e+20

const unsigned int single_upper_num_elems = 4;
const double single_upper[] = { 6.5e-03, 7.07826704e+06, 
				-6.34113186e+13, 2.54394281e+20 };


// Double Mode tinitial = 0.0

// const double double_lower_a =  7.59735880e-03;
// const double double_lower_b =  5.95425811e+06;
// const double double_lower_c = -7.57874769e+13;
// const double double_lower_d =  5.34742737e+20;

const unsigned int double_lower_num_elems = 4;
const double double_lower[] = { 7.4e-03, 5.95425811e+06,
				-7.57874769e+13, 5.34742737e+20  };

// const double double_upper_a =  1.75142951e-02;
// const double double_upper_b =  6.64159562e+06;
// const double double_upper_c = -4.31708314e+13;
// const double double_upper_d = -7.74470832e+19;

const unsigned int double_upper_num_elems = 4;
const double double_upper[] = { 1.9e-02, 6.64159562e+06, 
				-4.31708314e+13, -7.74470832e+19  };


const MovingBoundary single_lower_bounds(single_lower_num_elems, single_lower);
const MovingBoundary single_upper_bounds(single_upper_num_elems, single_upper);

const MovingBoundary double_lower_bounds(double_lower_num_elems, double_lower);
const MovingBoundary double_upper_bounds(double_upper_num_elems, double_upper);


// Hard-code the boundary to be used ... ack!!

const MovingBoundary & active_lower_bounds = single_lower_bounds;
const MovingBoundary & active_upper_bounds = single_upper_bounds;


void build_filename(const std::string & filename_base,
		    const unsigned int num_digits, const unsigned int index,
		    std::string & filename)
  {
  std::stringstream name_stream;

  name_stream << filename_base;

  name_stream.width(num_digits);
  name_stream.fill('0');

  name_stream << index;

  filename = name_stream.str();
  }


int main(int argc, char * argv[])
  {
  using std::cout;
  using std::cerr;
  using std::endl;

  if (argc < 7)
    {
    cerr << "Using: " << argv[0] << " filename_base num_digits start_index "
	 << "end_index trace_flow_file poly_species1 [ poly_species2 ... ]" 
	 << endl;
    return 1;
    }

  int argPtr = 1;

  const std::string filename_base = argv[argPtr++];

  const unsigned int num_digits 
    = static_cast<unsigned int>(atoi(argv[argPtr++]));

  const unsigned int start_index
    = static_cast<unsigned int>(atoi(argv[argPtr++]));

  const unsigned int end_index
    = static_cast<unsigned int>(atoi(argv[argPtr++]));

  if (start_index >= end_index)
    {
    cerr << "Error: start_index must be less than end_index" << endl;
    return -1;
    }

  const std::string trace_flow_file = argv[argPtr++];

  std::vector<std::string> poly_species_names;

  while (argPtr < argc)
    poly_species_names.push_back(argv[argPtr++]);

  const unsigned int num_poly_species = poly_species_names.size();

  // Load the particle map

  const PartMap partmap(trace_flow_file);

  const double ref_time = partmap.get_map_time();

  cerr << "Reference time " << ref_time << endl;

  const double ref_interface_pos = partmap.get_initial_interface_pos();

  // Run through the files

  cout << "bin_index" << " " << "bin_center" << " " << "bin_center_align" 
       << " " << "map_bin_center" << " " << "map_bin_center_align" << " "
       << "num_entries" 
       << " " << "bin_weight" << " " << "bin_mean" << " " << "bin_stddev" 
       << " " << "bin_min" << " " << "bin_max" << endl;

  std::string filename;

  for (unsigned int file_index = start_index; file_index <= end_index;
       file_index++)
    {
    // Load the file

    build_filename(filename_base, num_digits, file_index, filename);

    const QuickFlash::File::DataFile df(filename, buffer_size);

    const QuickFlash::File::MeshInfo & meshinfo = df.get_mesh_info();

    const unsigned int dims = meshinfo.get_dims();

    const double time = meshinfo.get_sim_time();

    cerr << "Processing file [ " << filename << " ] time [ " << time << " ]" 
	 << endl;

    const double current_interface = partmap.get_interface_position(time);


    // Build a CylDomain object (needs dimensions)

    std::vector<double> domain_center(dims);

    domain_center[0] = 0.0;

    for (unsigned int i = 1; i < dims; i++)
      domain_center[i] = 0.045;

    const CylDomain cyldom(domain_center, domain_rad);


    // Determine the bin layout

    const double lower_coord = active_lower_bounds[time];

    const double default_domain_width 
      = active_upper_bounds[time] - lower_coord;

    const double fl_num_bins = default_domain_width / bin_width;

    const unsigned int num_bins = static_cast<unsigned int>(fl_num_bins) + 1;

    const double domain_width = static_cast<double>(num_bins) * bin_width;
    const double upper_coord = lower_coord + domain_width;

    Stats::BinArray binarray(num_bins, lower_coord, upper_coord);

    // Load the data sets

    const QuickFlash::File::Dataset & dens = df.get_dataset(dens_name);

    std::vector<const QuickFlash::File::Dataset *> poly_datasets;

    for (unsigned int poly_index = 0; poly_index < num_poly_species; 
	 poly_index++)
      {
      const QuickFlash::File::Dataset & dset 
	= df.get_dataset(poly_species_names[poly_index]);

      poly_datasets.push_back(&dset);
      }

    // Run through the blocks

    const unsigned int num_blocks = meshinfo.get_num_blocks();

    QuickFlash::Block::BlockData<double> dens_block;

    std::vector< QuickFlash::Block::BlockData<double> > 
      poly_blocks(num_poly_species);

    for (unsigned int block_index = 0; block_index < num_blocks; block_index++)
      {
      if (meshinfo.is_leaf(block_index))
	{
	// See if block (x-coordinate) falls within the bin domain

	const double block_center = meshinfo.get_block_center(block_index)[0];
	const double block_width = meshinfo.get_block_width(block_index)[0];

	const double block_halfwidth = block_width * 0.5;

	const double block_min = block_center - block_halfwidth;
	const double block_max = block_center + block_halfwidth;

	if ((block_min < upper_coord) && (block_max > lower_coord))
	  {
	  // Load the block information

	  const QuickFlash::Block::BlockInfo & block_info
	    = meshinfo.get_block_info(block_index);

	  // Load the block data

	  dens.get_block_data(block_index, dens_block);

	  for (unsigned int poly_index = 0; poly_index < num_poly_species;
	       poly_index++)
	    poly_datasets[poly_index]->get_block_data(block_index, 
						      poly_blocks[poly_index]);

	  const std::vector<double> & cell_width = block_info.get_cell_width();

	  const double cell_x_width = cell_width[0];
	  const double cell_x_halfwidth = cell_x_width * 0.5;

	  const double cell_volume = block_info.get_cell_volume();

	  const unsigned int num_cells = block_info.get_num_cells();

	  std::vector<double> cell_center;

	  for (unsigned int cell_index = 0; cell_index < num_cells; 
	       cell_index++)
	    {
	    // Get the center coordinates

	    block_info.get_cell_center(cell_index, cell_center);

	    const double cell_x_center = cell_center[0];

	    const double cell_x_min = cell_x_center - cell_x_halfwidth;
	    const double cell_x_max = cell_x_center + cell_x_halfwidth;

	    if ((cell_x_min < upper_coord) && (cell_x_max > lower_coord))
	      {
	      if (cyldom.in_domain(cell_center))
		{
		const double dens_value = dens_block[cell_index];

		double poly_value = 0.0;

		for (unsigned int poly_index = 0; 
		     poly_index < num_poly_species; poly_index++)
		  poly_value += poly_blocks[poly_index][cell_index];

		const double cell_mass = cell_volume * dens_value;
		// const double poly_mass = poly_value * cell_mass;

		binarray.add_data_spread(cell_x_center, poly_value, 
					 cell_x_width, cell_mass);
		}
	      }
	    }
	  }
	}
      }

    // Report on the findings

    cout << "[ File " << filename << " Time " << time << " Interface " 
	 << current_interface << " ReferenceTime " << ref_time << " ]" << endl;

    for (unsigned int bin_index = 0; bin_index < num_bins; bin_index++)
      {
      const Stats::Bin & bin = binarray.get_bin(bin_index);

      const double bin_center = binarray.get_bin_center(bin_index);

      const double bin_center_align = bin_center - current_interface;

      // See if this position can be mapped back to the reference time

      double map_bin_center = 0.0;

      bool bin_ok = true;

      try
	{
	map_bin_center = partmap.get_initial_position(time, bin_center);
	}
      catch (QuickFlash::Except ex)
	{
	cerr << "Unable to map position [ " << bin_center << " ] at time [ "
	     << time << " ]" << endl;
	bin_ok = false;
	}

      // If the mapping is OK, print the bin's data

      if (bin_ok)
	{
	const double map_bin_center_align = map_bin_center - ref_interface_pos;

	const unsigned int num_entries = bin.get_num_entries();

	const double bin_weight = bin.get_weight_sum();
	const double bin_mean = bin.get_mean();
	const double bin_stddev = bin.get_pop_stddev();

	const double bin_min = bin.get_min_value();
	const double bin_max = bin.get_max_value();

	cout << bin_index << " " << bin_center << " " << bin_center_align 
	     << " " << map_bin_center << " " << map_bin_center_align << " " 
	     << num_entries << " "
	     << bin_weight << " " << bin_mean << " " << bin_stddev << " " 
	     << bin_min << " " << bin_max << endl;
	}
      }
    }

  return 0;
  }
