// C++ program file mixing_layer.cpp

/*
  By Nathan C. Hearn
     May 16, 2007

  Classes for storing mixing layer information for each timestep.
*/


#include "mixing_layer.hpp"
#include <string>
#include <vector>
#include <iostream>
#include "stat_bins.hpp"
#include "quickflash_file_datafile.hpp"
#include "quickflash_file_meshinfo.hpp"
#include "quickflash_file_dataset.hpp"
#include "quickflash_block_blockinfo.hpp"
#include "quickflash_block_blockdata.hpp"


const std::string dens_name = "dens";
const std::string foam_name = "foam";


void get_mixing_layer(const std::string & flash_filename, 
		      const double min_x, const double max_x, 
		      const double bin_spacing, 
		      const double min_poly, const double max_poly,
		      MixingLayer & mixing_layer, 
		      const unsigned int buffer_size)
  {
  using std::cerr;
  using std::endl;

  const QuickFlash::File::DataFile df(flash_filename, buffer_size, true);

  const QuickFlash::File::MeshInfo & meshinfo = df.get_mesh_info();

  const double time = meshinfo.get_sim_time();

  cerr << "Processing file [ " << flash_filename << " ] time [ " << time 
       << " ]" << endl;

  // Determine the bin layout

  const double lower_coord = min_x;

  const double default_domain_width = max_x - lower_coord;

  const double fl_num_bins = default_domain_width / bin_spacing;

  const unsigned int num_bins = static_cast<unsigned int>(fl_num_bins) + 1;

  const double domain_width = static_cast<double>(num_bins) * bin_spacing;
  const double upper_coord = lower_coord + domain_width;

  Stats::BinArray binarray(num_bins, lower_coord, upper_coord);

  // Load the data sets

  const QuickFlash::File::Dataset & dens_set = df.get_dataset(dens_name);

  const QuickFlash::File::Dataset & foam_set = df.get_dataset(foam_name);

  // Run through the blocks

  const unsigned int num_blocks = meshinfo.get_num_blocks();

  std::vector<double> block_center_vect;
  std::vector<double> block_width_vect;
  
  QuickFlash::Block::BlockData<double> dens_block;
  QuickFlash::Block::BlockData<double> foam_block;

  for (unsigned int block_index = 0; block_index < num_blocks; block_index++)
    {
    if (meshinfo.is_leaf(block_index))
      {
      // See if block (x-coordinate) falls within the bin domain

      meshinfo.get_block_center(block_index, block_center_vect);
      meshinfo.get_block_width(block_index, block_width_vect);

      const double block_center = block_center_vect[0];
      const double block_width = block_width_vect[0];

      const double block_halfwidth = block_width * 0.5;

      const double block_min = block_center - block_halfwidth;
      const double block_max = block_center + block_halfwidth;

      if ((block_min < upper_coord) && (block_max > lower_coord))
	{
	// Load the block information

	const QuickFlash::Block::BlockInfo & block_info
	  = meshinfo.get_block_info(block_index);

	// Load the block data

	dens_set.get_block_data(block_index, dens_block);
	foam_set.get_block_data(block_index, foam_block);

	const std::vector<double> & cell_width = block_info.get_cell_width();

	const double cell_x_width = cell_width[0];
	const double cell_x_halfwidth = cell_x_width * 0.5;

	const double cell_volume = block_info.get_cell_volume();

	const unsigned int num_cells = block_info.get_num_cells();

	std::vector<double> cell_center;

	for (unsigned int cell_index = 0; cell_index < num_cells; cell_index++)
	  {
	  // Get the center coordinates

	  block_info.get_cell_center(cell_index, cell_center);

	  const double cell_x_center = cell_center[0];

	  const double cell_x_min = cell_x_center - cell_x_halfwidth;
	  const double cell_x_max = cell_x_center + cell_x_halfwidth;

	  if ((cell_x_min < upper_coord) && (cell_x_max > lower_coord))
	    {
//          const double dens_value = dens_block[cell_index];

	    const double foam_value = foam_block[cell_index];

	    const double poly_value = 1.0 - foam_value;

//          const double cell_mass = cell_volume * dens_value;
	    // const double poly_mass = poly_value * cell_mass;

	    // WEIGHT BY CELL VOLUME INSTEAD OF MASS?????

// 	    binarray.add_data_spread(cell_x_center, poly_value, 
// 				     cell_x_width, cell_mass);
	    binarray.add_data_spread(cell_x_center, poly_value, 
				     cell_x_width, cell_volume);
	    }
	  }
	}
      }
    }

  // Find the bins where the mean poly value crosses the min_x threshhold

  const double layer_min_pos 
    = binarray.get_transition_pos_min(min_poly, true);

  const double layer_max_pos 
    = binarray.get_transition_pos_max(max_poly, false);

  // Reset the MixingLayer object

  mixing_layer.reset(time, layer_min_pos, layer_max_pos, 0.0);
  }


void get_mixing_layers(const std::vector<std::string> & flash_filename_list,
		       const double min_x, const double max_x, 
		       const double bin_spacing, 
		       const double min_poly, const double max_poly,
		       std::vector<MixingLayer> & mixing_layer_list,
		       const unsigned int buffer_size)
  { 
  const unsigned int num_files = flash_filename_list.size();

  mixing_layer_list.resize(num_files);

  for (unsigned int index = 0; index < num_files; index++)
    get_mixing_layer(flash_filename_list[index], min_x, max_x,
		     bin_spacing, min_poly, max_poly, 
		     mixing_layer_list[index], buffer_size);
  }


void get_mixing_layer_minmax(const std::string & flash_filename, 
			     const double min_poly, const double max_poly,
			     MixingLayer & mixing_layer, 
			     const unsigned int buffer_size)
  {
  using std::cerr;
  using std::endl;

  const QuickFlash::File::DataFile df(flash_filename, buffer_size, true);

  const QuickFlash::File::MeshInfo & meshinfo = df.get_mesh_info();

  const double time = meshinfo.get_sim_time();

  cerr << "Processing file [ " << flash_filename << " ] time [ " << time 
       << " ]" << endl;

  // Determine the bin layout

  double layer_min = meshinfo.get_volume_maxbounds()[0];  // Pull min back
  double layer_max = meshinfo.get_volume_minbounds()[0];  // Push max forward

  // Load the data sets

  const QuickFlash::File::Dataset & foam_set = df.get_dataset(foam_name);

  // Run through the blocks

  const unsigned int num_blocks = meshinfo.get_num_blocks();
  
  QuickFlash::Block::BlockData<double> foam_block;

  for (unsigned int block_index = 0; block_index < num_blocks; block_index++)
    {
    if (meshinfo.is_leaf(block_index))
      {
      // Load the block information

      const QuickFlash::Block::BlockInfo & block_info
	= meshinfo.get_block_info(block_index);

      // Load the block data
      
      foam_set.get_block_data(block_index, foam_block);

      // Parse the cells

      const unsigned int num_cells = block_info.get_num_cells();

      std::vector<double> cell_center;

      for (unsigned int cell_index = 0; cell_index < num_cells; cell_index++)
	{
	// Get the center coordinates

	block_info.get_cell_center(cell_index, cell_center);

	const double cell_x = cell_center[0];

	const double foam_value = foam_block[cell_index];

	const double poly_value = 1.0 - foam_value;

	if ((poly_value < min_poly) && (cell_x < layer_min))
	  layer_min = cell_x;

	if ((poly_value > max_poly) && (cell_x > layer_max))
	  layer_max = cell_x;
	}
      }
    }

  // Find the bins where the mean poly value crosses the min_x threshhold

  // Reset the MixingLayer object

  mixing_layer.reset(time, layer_min, layer_max, 0.0);
  }


void get_mixing_layers_minmax(const 
			      std::vector<std::string> & flash_filename_list,
			      const double min_poly, const double max_poly,
			      std::vector<MixingLayer> & mixing_layer_list,
			      const unsigned int buffer_size)
  { 
  const unsigned int num_files = flash_filename_list.size();

  mixing_layer_list.resize(num_files);

  for (unsigned int index = 0; index < num_files; index++)
    get_mixing_layer_minmax(flash_filename_list[index], min_poly, max_poly, 
		     mixing_layer_list[index], buffer_size);
  }


