// C++ program file

/*
  By Nathan C. Hearn
     January 20, 2007
*/


#include "part_map.hpp"
#include <string>
#include <fstream>
#include "quickflash.hpp"


// Class PartList

void PartList::reset(const double time, const double interface_pos,
		     const std::vector<unsigned int> & part_ids,
		     const std::vector<double> & part_pos)
  {
  reset();

  const unsigned int num_parts = part_ids.size();

  if (part_pos.size() != num_parts)
    throw QuickFlash::Except("Vector length mismatch", __FILE__, __LINE__);

  t = time;
  interface = interface_pos;

  part_list.resize(num_parts);

  for (unsigned int part_index = 0; part_index < num_parts; part_index++)
    part_list[part_index].reset(part_ids[part_index], part_pos[part_index]);

  // Sort the list by position

  std::sort(part_list.begin(), part_list.end());
  }


void PartList::find_nearest_parts(const double pos, unsigned int & low_index,
				  unsigned int & high_index, 
				  double & interval_size,
				  double & pos_low_delta) const
  {
  const unsigned int num_parts = part_list.size();

  if (num_parts < 2)
    throw QuickFlash::Except("Insufficient particles in list", __FILE__, __LINE__);

  unsigned int low = 0;
  unsigned int high = num_parts - 1;

  double low_pos = part_list[low].get_pos();
  double high_pos = part_list[high].get_pos();

  if ((pos < low_pos) || (!(pos < high_pos)))
    throw QuickFlash::Except("Position out of bounds", __FILE__, __LINE__);

  while (low < (high - 1))
    {
    const unsigned int temp = (low + high) / 2;

    const double temp_pos = part_list[temp].get_pos();

    if (temp_pos > pos)
      high = temp;
    else
      low = temp;
    }

  // ERROR CHECKING???
  
  low_pos = part_list[low].get_pos();
  high_pos = part_list[high].get_pos();

  low_index = low;
  high_index = high;

  interval_size = high_pos - low_pos;
  pos_low_delta = pos - low_pos;
  }


// Class PartMap

void PartMap::reset()
  {
  map_time = 0.0;

  map_low_state_index = 0;
  map_high_state_index = 0;

  map_low_state_weight = 0.0;
  map_high_state_weight = 0.0;

  map_interface = 0.0;

  states.clear(); 
  }


void PartMap::reset(const std::string & filename)
  {
  /*
    Loads in the flow tracer file.  The file is assumed to be a stream of
    numbers.  The first two entries are the number of timesteps and the
    number of tracer positions, respectively.  The data for each timestep
    then follows, beginning with the simulation time and the interface 
    position, followed by the position of each point.

    Template:

      num_times num_parts
      time0 interface_pos0 part0_pos part1_pos part2_pos ...
      time1 interface_pos1 part0_pos part1_pos part2_pos ...
      time2 interface_pos2 part0_pos ...

    NOTE: From timestep to timestep, the data must maintain a consistent order
    for the tracers.  (i.e. the data for particle i always appears at the
    same index in each timestep).
  */

  reset();

  // Open the file

  std::ifstream in_file(filename.c_str());

  unsigned int num_times = 0;
  unsigned int num_parts = 0;

  in_file >> num_times;
  in_file >> num_parts;

  states.resize(num_times);

  std::vector<unsigned int> part_ids(num_parts);

  for (unsigned int part_index = 0; part_index < num_parts; part_index++)
    part_ids[part_index] = part_index;

  for (unsigned int time_index = 0; time_index < num_times; time_index++)
    {
    double step_time = 0.0;
    double interface_pos = 0.0;

    in_file >> step_time;
    in_file >> interface_pos;

    std::vector<double> current_positions(num_parts);

    for (unsigned int part_index = 0; part_index < num_parts; part_index++)
      in_file >> current_positions[part_index];

    states[time_index].reset(step_time, interface_pos, part_ids, 
			     current_positions);
    }

  if (num_times > 0)
    set_map_time(states[0].get_time());
  }


void PartMap::reset(const PartMap & source)
  {
  if (&source != this)
    {
    map_time = source.map_time;

    map_low_state_index = source.map_low_state_index;
    map_high_state_index = source.map_high_state_index;

    map_low_state_weight = source.map_low_state_weight;
    map_high_state_weight = source.map_high_state_weight;

    map_interface = source.map_interface;

    states = source.states;
    }
  }


void PartMap::set_map_time(const double time) const
  {
  map_time = time;

  double low_state_high_state_delta = 0.0;
  double low_state_map_time_delta = 0.0;

  find_nearest_times(map_time, map_low_state_index, map_high_state_index,
		     low_state_high_state_delta, low_state_map_time_delta);

  double high_state_weight = 0.0;

  if (low_state_high_state_delta > 0.0)
    {
    high_state_weight = low_state_map_time_delta / low_state_high_state_delta;

    if (high_state_weight < 0.0)
      high_state_weight = 0.0;
    else if (high_state_weight > 1.0)
      high_state_weight = 1.0;
    }

  map_low_state_weight = 1.0 - high_state_weight;
  map_high_state_weight = high_state_weight;

  map_interface = get_interface_position(time);
  }


void PartMap::find_nearest_times(const double time, unsigned int & low_index,
				 unsigned int & high_index, 
				 double & interval_time,
				 double & time_low_delta) const
  {
  const unsigned int num_states = states.size();

  if (num_states < 2)
    throw QuickFlash::Except("Insufficient states in list", __FILE__, __LINE__);

  unsigned int low = 0;
  unsigned int high = num_states - 1;

  double low_time = states[low].get_time();
  double high_time = states[high].get_time();

  if ((time < low_time) || (!(time < high_time)))
    throw QuickFlash::Except("Time out of bounds", __FILE__, __LINE__);

  while (low < (high - 1))
    {
    const unsigned int temp = (low + high) / 2;

    const double temp_time = states[temp].get_time();

    if (temp_time > time)
      high = temp;
    else
      low = temp;
    }

  // ERROR CHECKING???
  
  low_time = states[low].get_time();
  high_time = states[high].get_time();

  low_index = low;
  high_index = high;

  interval_time = high_time - low_time;
  time_low_delta = time - low_time;
  }


double PartMap::get_mapped_position(const double current_time, 
				    const double current_pos) const
  {
  // Determine the two neighboring states

  unsigned int low_state_index = 0;
  unsigned int high_state_index = 0;

  double interval_time = 0.0;
  double time_low_delta = 0.0;

  find_nearest_times(current_time, low_state_index, high_state_index, 
		     interval_time, time_low_delta);

  // Determine the appropriate weights for each state based on intervals

  double low_state_weight = 1.0;
  double high_state_weight = 0.0;

  if (interval_time > 0.0)
    {
    high_state_weight = time_low_delta / interval_time;

    if (high_state_weight > 1.0)
      high_state_weight = 1.0;
    else if (high_state_weight < 0.0)
      high_state_weight = 0.0;

    low_state_weight = 1.0 - high_state_weight;
    }

  // ELSE ERROR CHECKING???

  // In each state determine the mapping of current_pos to the reference state

  const double low_state_avg_pos 
    = get_interp_ref_pos(current_pos, low_state_index);

  const double high_state_avg_pos
    = get_interp_ref_pos(current_pos, high_state_index);

  const double avg_pos = (low_state_weight * low_state_avg_pos)
    + (high_state_weight * high_state_avg_pos);

  return avg_pos;
  }


void PartMap::get_mapped_position_list(const double current_time, 
				 const std::vector<double> & current_pos_list,
				 std::vector<double> & initial_pos_list) const
  {
  // Determine the two neighboring states

  unsigned int low_state_index = 0;
  unsigned int high_state_index = 0;

  double interval_time = 0.0;
  double time_low_delta = 0.0;

  find_nearest_times(current_time, low_state_index, high_state_index, 
		     interval_time, time_low_delta);

  // Determine the appropriate weights for each state based on intervals

  double low_state_weight = 1.0;
  double high_state_weight = 0.0;

  if (interval_time > 0.0)
    {
    high_state_weight = time_low_delta / interval_time;

    if (high_state_weight > 1.0)
      high_state_weight = 1.0;
    else if (high_state_weight < 0.0)
      high_state_weight = 0.0;

    low_state_weight = 1.0 - high_state_weight;
    }

  // ELSE ERROR CHECKING???

  // Run through the positions

  const unsigned int num_pos = current_pos_list.size();

  initial_pos_list.resize(num_pos);

  for (unsigned int pos_index = 0; pos_index < num_pos; pos_index++)
    {
    const double current_pos = current_pos_list[pos_index];

    // In each state, map current_pos to the reference state

    const double low_state_avg_pos 
      = get_interp_ref_pos(current_pos, low_state_index);

    const double high_state_avg_pos
      = get_interp_ref_pos(current_pos, high_state_index);

    // Store the weighted average of these two positions

    initial_pos_list[pos_index] = (low_state_weight * low_state_avg_pos)
      + (high_state_weight * high_state_avg_pos);
    }
  }


double PartMap::get_interface_position(const double time) const
  {
  // Determine the two neighboring states

  unsigned int low_state_index = 0;
  unsigned int high_state_index = 0;

  double interval_time = 0.0;
  double time_low_delta = 0.0;

  find_nearest_times(time, low_state_index, high_state_index, 
		     interval_time, time_low_delta);

  // Get interface positions at both states

  const double low_interface = states[low_state_index].get_interface_pos();
  const double high_interface = states[high_state_index].get_interface_pos();

  // Determine the interpolated interface position

  double low_weight = 0.0;
  double high_weight = 1.0;

  if (interval_time > 0.0)
    {
    high_weight = time_low_delta / interval_time;

    if (high_weight < 0.0)
      high_weight = 0.0;
    else if (high_weight > 1.0)
      high_weight = 1.0;

    low_weight = 1.0 - high_weight;
    }

  const double interface_pos 
    = (low_weight * low_interface) + (high_weight * high_interface);

  return interface_pos;
  }


double PartMap::get_interp_ref_pos(const double current_pos, 
				   const unsigned int current_state_index)
  const
  {
  // Call get_interp_ref_pos for each map state, and compute the average

  const double low_pos = get_interp_ref_pos(current_pos, current_state_index,
					    map_low_state_index);

  const double high_pos = get_interp_ref_pos(current_pos, current_state_index,
					     map_high_state_index);

  const double avg_pos = (map_low_state_weight * low_pos)
    + (map_high_state_weight * high_pos);

  return avg_pos;
  }


double PartMap::get_interp_ref_pos(const double current_pos, 
				   const unsigned int current_state_index,
				   const unsigned int ref_state_index) const
  {
  // Determine the two neighboring particles in the current state

  unsigned int low_part_index = 0;
  unsigned int high_part_index = 0;

  double current_high_low_delta = 0.0;
  double current_pos_low_delta = 0.0;

  states[current_state_index].find_nearest_parts(current_pos, low_part_index,
						 high_part_index,
						 current_high_low_delta,
						 current_pos_low_delta);

  // Determine weights of the particles in the current state based on deltas

  double low_part_weight = 1.0;
  double high_part_weight = 0.0;

  if (current_high_low_delta > 0.0)
    {
    high_part_weight = current_pos_low_delta / current_high_low_delta;

    if (high_part_weight > 1.0)
      high_part_weight = 1.0;
    else if (high_part_weight < 0.0)
      high_part_weight = 0.0;

    low_part_weight = 1.0 - high_part_weight;
    }

  // ELSE ERROR CHECKING???

  // Extract the positions from the reference state

  const double low_part_pos 
    = states[ref_state_index].get_pos(low_part_index);

  const double high_part_pos 
    = states[ref_state_index].get_pos(high_part_index);

  const double avg_pos 
    = (low_part_weight * low_part_pos) + (high_part_weight * high_part_pos);

  return avg_pos;
  }
