// C++ header file part_map.hpp

/*
  By Nathan C. Hearn
     January 20, 2007

  Reader for trace_flow output with mapping facilities.
*/


#ifndef PART_MAP_HPP
#define PART_MAP_HPP


#include <string>
#include <vector>


class PartInfo
  {
  public:
    PartInfo() : id(0), pos(0.0) { }
    PartInfo(const unsigned int part_id) : id(0), pos(0.0) { reset(part_id); }
    PartInfo(const unsigned int part_id, const double part_pos) :
      id(0), pos(0.0)
      { reset(part_id, part_pos); }
    PartInfo(const PartInfo & source) : id(0), pos(0.0) { reset(source); }
    ~PartInfo() { }

    PartInfo operator=(const PartInfo & source)
      {
      reset(source);
      return *this;
      }

    bool operator<(const double ref_pos) const { return (pos < ref_pos); }
    bool operator<(const PartInfo & rhs) const { return (pos < rhs.pos); }

    void reset(const unsigned int part_id=0, const double part_pos=0.0)
      {
      id = part_id;
      pos = part_pos;
      }

    void reset(const PartInfo & source)
      {
      if (&source !=this)
	{
	id = source.id;
	pos = source.pos;
	}
      }

    void set_id(const unsigned int part_id) { id = part_id; }
    void set_pos(const double part_pos) { pos = part_pos; }

    unsigned int get_id() const { return id; }
    double get_pos() const { return pos; }

  private:
    unsigned int id;
    double pos;
  };


class PartList
  {
  public:
    PartList() : t(0.0), interface(0.0), part_list() { }
    PartList(const double time, const double interface_pos,
	     const std::vector<unsigned int> & part_ids,
	     const std::vector<double> & part_pos) : 
      t(0.0), interface(0.0), part_list()
      { reset(time, interface_pos, part_ids, part_pos); }
    PartList(const PartList & source) : 
      t(0.0), interface(0.0), part_list() 
      { reset(source); }
    ~PartList() { }

    PartList & operator=(const PartList & source)
      {
      reset(source);
      return *this;
      }

    bool operator<(const double ref_time) const { return (t < ref_time); }
    bool operator<(const PartList & rhs) const { return (t < rhs.t); }

    double get_time() const { return t; }

    double get_interface_pos() const { return interface; }

    unsigned int get_id(const unsigned int index) const 
      { return part_list[index].get_id(); }

    double get_pos(const unsigned int index) const
      { return part_list[index].get_pos(); }

    PartInfo & operator[](const unsigned int index) 
      { return part_list[index]; }

    const PartInfo & operator[](const unsigned int index) const
      { return part_list[index]; }

    void reset()
      {	
      t = 0.0;
      interface = 0.0;
      part_list.clear();
      }

    void reset(const double time, const double interface_pos,
	       const std::vector<unsigned int> & part_ids,
	       const std::vector<double> & part_pos);

    void reset(const PartList & source)
      {
      if (&source != this)
	{
	t = source.t;
	interface = source.interface;
	part_list = source.part_list;
	}
      }

    void find_nearest_parts(const double pos, unsigned int & low_index,
			    unsigned int & high_index, double & interval_size,
			    double & low_delta) const;

  private:
    double t;
    double interface;

    std::vector<PartInfo> part_list;
  };


class PartMap
  {
  public:
    PartMap() : 
      map_time(0.0), map_low_state_index(0), map_high_state_index(0),
      map_low_state_weight(0.0), map_high_state_weight(0.0),
      map_interface(0.0), states() 
      { }

    PartMap(const std::string & filename) : 
      map_time(0.0), map_low_state_index(0), map_high_state_index(0),
      map_low_state_weight(0.0), map_high_state_weight(0.0),
      map_interface(0.0), states() 
      { reset(filename); }

    PartMap(const PartMap & source) : 
      map_time(0.0), map_low_state_index(0), map_high_state_index(0),
      map_low_state_weight(0.0), map_high_state_weight(0.0),
      map_interface(0.0), states() 
      { reset(source); }

    ~PartMap() { }

    PartMap & operator=(const PartMap & source)
      {
      reset(source);
      return *this;
      }

    void reset();

    void reset(const std::string & filename);

    void reset(const PartMap & source);

    void set_map_time(const double time) const;

    void find_nearest_times(const double time, unsigned int & low_index,
			    unsigned int & high_index, double & interval_time,
			    double & time_low_delta) const;
    
    double get_map_time() const { return map_time; }

    double get_mapped_interface_pos() const { return map_interface; }

    double get_mapped_position(const double current_time, 
			       const double current_pos) const;

    void get_mapped_position_list(const double current_time, 
				 const std::vector<double> & current_pos_list,
				 std::vector<double> & initial_pos_list) const;

    double get_interface_position(const double time) const;

  private:
    double get_interp_ref_pos(const double current_pos, 
			      const unsigned int current_state_index) const;

    double get_interp_ref_pos(const double current_pos, 
			      const unsigned int current_state_index,
			      const unsigned int ref_state_index) const;

  private:
    mutable double map_time;

    mutable unsigned int map_low_state_index;
    mutable unsigned int map_high_state_index;

    mutable double map_low_state_weight;
    mutable double map_high_state_weight;

    mutable double map_interface;

    std::vector<PartList> states;
  };



#endif  // PART_MAP_HPP
