#!/usr/bin/env python


try :
    import psyco
except ImportError :
    pass

from sys import argv, exit, stderr, stdout


index_label = "Index"
time_label = "Time"
raw_min_label = "raw_min"
raw_max_label = "raw_max"
raw_width_label = "raw_width"
rt_min_label = "rt_min"
rt_max_label = "rt_max"
rt_width_label = "rt_width"


# out_file_mode = "PS"
# out_file_suffix = "ps"
out_file_mode = "SVG"
out_file_suffix = "svg"


def read_file(filename) :

    f = file(filename, "r")

    # Find the label line

    labels = None

    end_reached = False

    while ((not end_reached) and (labels is None)) :

        line = f.readline()

        if (line == "") :
            end_reached = True
        else :
            labels = line.partition("#")[0].split()

            if (len(labels) < 1) :
                labels = None

    if (labels is None) :
        raise ValueError("Labels not found in file" + filename)

    num_columns = len(labels)

    data_columns = list()

    for idx in range(num_columns) :
        data_columns.append(list())
    
    finished = False

    while (not finished) :

        line = f.readline()

        if (line == "") :
            finished = True
            
        else :
            
            words = line.partition("#")[0].split()

            num_words = len(words)

            if (num_words != num_columns) :
                raise ValueError("Incorrect formatting in file")

            for idx in range(num_words) :
                data_columns[idx].append(float(words[idx]))

    # Build the dictionary

    data = dict()

    for idx in range(num_columns) :
        data[labels[idx]] = data_columns[idx]

    return data


def fit_agt2(time_data, width_data, start_time=None, end_time=None) :

    from numarray import zeros, Float
    from numarray.linear_algebra import linear_least_squares
    
    orig_data_len = len(time_data)

    # Find the start and stop times

    data_start = 0
    data_end = orig_data_len - 1

    if (start_time is not None) :

        found_it = False

        while ((not found_it) and (data_start < data_end)) :

            cur_time = time_data[data_start]

            if (cur_time < start_time) :
                data_start += 1
            else :
                found_it = True

        if (not found_it) :
            raise ValueError("Unable to find start time")

    if (end_time is not None) :

        found_it = False

        while ((not found_it) and (data_end >= data_start)) :

            cur_time = time_data[data_end]

            if (cur_time > end_time) :
                data_end -= 1
            else :
                found_it = True

        if (not found_it) :
            raise ValueError("Unable to find end time")

    data_len = (data_end + 1) - data_start

    if (data_len < 2) :
        raise ValueError("Insufficient data for least-squares fit")
    
    # Build the matrices

    time_mat = zeros((data_len, 1), Float)
    data_mat = zeros(data_len, Float)

    for idx in range(data_len) :

        data_idx = idx + data_start

        cur_time = time_data[data_idx]

        time_mat[idx] = cur_time * cur_time

        data_mat[idx] = width_data[data_idx]

    # Perform least squares fit

    coef_mat, sqr_resid, rank_a, sing_a \
              = linear_least_squares(time_mat, data_mat)

    # Build a fitted solution

    fit_times = zeros(orig_data_len, Float)
    fit_data = zeros(orig_data_len, Float)

    for idx in range(orig_data_len) :

        # data_idx = idx + data_start

        data_idx = idx

        cur_time = time_data[data_idx]

        fit_times[idx] = cur_time

        fit_data[idx] = coef_mat[0] * cur_time * cur_time

    return coef_mat, fit_times, fit_data
        
    

def main() :
    
    argc = len(argv)

    if (argc < 5) :
        stderr.write("Usage: " + argv[0] + " filename time_offset_ns "
                     + " fit_time_start_ns fit_time_end_ns " + "\n"
                     + "    [ output_file_prefix ]" + "\n\n");
        exit(1)


    argptr = int(1)

    filename = str(argv[argptr])
    argptr += 1

    time_offset_ns = float(argv[argptr])
    argptr += 1

    fit_time_start = float(argv[argptr])
    argptr += 1

    fit_time_end = float(argv[argptr])
    argptr += 1

    out_file_prefix = None
    use_out_file = False

    if (argptr < argc) :
        out_file_prefix = str(argv[argptr])
        argptr += 1

        use_out_file = True


    # Get the data

    datadict = read_file(filename)


    # Set up matplotlib

    import matplotlib

    if (use_out_file) :
        matplotlib.use(out_file_mode)
    else :
        matplotlib.use("wxAgg")

    import pylab


    # Prepare the data

    time_list_sec = datadict[time_label]

    num_times = len(time_list_sec)

    time_list_ns = list()

    for idx in range(num_times) :
        time_list_ns.append((time_list_sec[idx] * 1.0e+09) - time_offset_ns)

    raw_width_list = datadict[raw_width_label]
    rt_width_list = datadict[rt_width_label]

    rt_fit_coeff, rt_fit_times, rt_fit_width \
                  = fit_agt2(time_list_ns, rt_width_list, fit_time_start,
                             fit_time_end)

    rt_min_list = datadict[rt_min_label]
    rt_max_list = datadict[rt_max_label]

    raw_min_list = datadict[raw_min_label]
    raw_max_list = datadict[raw_max_label]

    raw_bound_x = pylab.concatenate((time_list_ns, time_list_ns[::-1]))
    raw_bound_y = pylab.concatenate((raw_min_list,raw_max_list[::-1]))

    rt_bound_x = pylab.concatenate((time_list_ns, time_list_ns[::-1]))
    rt_bound_y = pylab.concatenate((rt_min_list,rt_max_list[::-1]))

    # Create some graphs

    figure1 = pylab.figure(1)

    width_comp_plot = figure1.add_subplot(2, 1, 1)

    width_comp_plot.set_title("Mixing Layer Width")
    # width_comp_plot.set_xlabel("Time (ns)")
    width_comp_plot.set_ylabel("Width (cm)")

    width_comp_plot.scatter(time_list_ns, raw_width_list, label="Full Width",
                            facecolor="red", linewidth=0, s=4)
    width_comp_plot.scatter(time_list_ns, rt_width_list,
                            label="RT Width (1D Expansion Removed)",
                            facecolor="blue", linewidth=0, s=4)
    width_comp_plot.plot(rt_fit_times, rt_fit_width,
                         label=("Least Squares RT Fit - " + str(fit_time_start)
                                + " to " + str(fit_time_end) + " ns"), 
                         c="magenta")
    
    width_comp_plot.legend(loc="upper left", pad=0.025)
    width_comp_plot.grid()
    width_comp_plot.set_xlim(0.0, 40.0)
    width_comp_plot.set_ylim(-0.005, 0.045)

    boundary_comp_plot = figure1.add_subplot(2, 1, 2)

    boundary_comp_plot.set_title("Mixing Layer Domain")
    boundary_comp_plot.set_xlabel("Time Since Interface-Shock Interaction "
                                  + "(ns)")
    boundary_comp_plot.set_ylabel("Position (cm)")

    boundary_comp_plot.fill(raw_bound_x, raw_bound_y, facecolor="green",
                            alpha=0.5, label="Full Mixing Layer", linewidth=0)

    boundary_comp_plot.fill(rt_bound_x, rt_bound_y, facecolor="purple",
                            alpha=0.5, label="RT Mixing Layer", linewidth=0)

    boundary_comp_plot.legend(loc="upper left", pad=0.025)
    boundary_comp_plot.grid()
    boundary_comp_plot.set_xlim(0.0, 40.0)
    
    
    if (use_out_file) :
        pylab.savefig(out_file_prefix + "_width_comp." + out_file_suffix)



    if (not use_out_file) :
        pylab.show()


if (__name__ == "__main__") :
    main()
