// C++ program file trace_flow.cpp

/*
  By Nathan C. Hearn
     January 20, 2007

  Trajectory reconstruction program for 1D Flash simulations.
*/


#include <iostream>
#include <string>
#include <sstream>
#include <algorithm>
#include "quickflash.hpp"


const std::string vel_dataset_name = "velx";
const std::string poly_dataset_name = "poly";
const unsigned int buffer_size = 4000;


class Part
  {
  public:
    Part() : pos(0.0), vel(0.0), poly(0.0) { }
    Part(const double position) : 
      pos(0.0), vel(0.0), poly(0.0) 
      { reset(position); }
    Part(const double position, const double velocity) :
      pos(0.0), vel(0.0), poly(0.0)
      { reset(position, velocity); }
    Part(const double position, const double velocity, 
	 const double poly_massfract) :
      pos(0.0), vel(0.0), poly(0.0)
      { reset(position, velocity, poly_massfract); }
    Part(const Part & source) : 
      pos(0.0), vel(0.0), poly(0.0) 
      { reset(source); }
    ~Part() { }

    Part & operator=(const Part & source)
      {
      reset(source);
      return *this;
      }

    void reset(const double position=0.0, const double velocity=0.0,
	       const double poly_massfract=0.0)
      {
      pos = position;
      vel = velocity;
      poly = poly_massfract;
      }

    void reset(const Part & source)
      {
      if (&source != this)
	{
	pos = source.pos;
	vel = source.vel;
	poly = source.poly;
	}
      }

    bool operator<(const Part & rhs) const
      {
      bool ret_val = false;

      if (pos < rhs.pos)
	ret_val = true;

      return ret_val;
      }

    double get_pos() const { return pos; }
    double get_vel() const { return vel; }
    double get_poly() const {return poly; }

    void set_pos(const double position) { pos = position; }
    void set_vel(const double velocity) { vel = velocity; }
    void set_poly(const double poly_massfract) { poly = poly_massfract; }

  private:
    double pos;
    double vel;
    double poly;
  }; 
      

void build_filename(const std::string & filename_base,
		    const unsigned int num_digits, const unsigned int index,
		    std::string & filename)
  {
  std::stringstream name_stream;

  name_stream << filename_base;

  name_stream.width(num_digits);
  name_stream.fill('0');

  name_stream << index;

  filename = name_stream.str();
  }


void get_particles(const std::string & filename, const double min_x,
		   const double max_x, std::vector<Part> & parts,
		   double & current_time, double & min_read_pos, 
		   double & max_read_pos)
  {
  QuickFlash::File::DataFile dfile(filename, buffer_size, true);

  const QuickFlash::File::MeshInfo & meshinfo = dfile.get_sim_info();

  const unsigned int num_blocks = meshinfo.get_num_blocks();
  const unsigned int dims = meshinfo.get_dims();

  if (dims != 1)
    throw QuickFlash::Except("Only one spatial dimension supported", __FILE__, 
			__LINE__);

  current_time = meshinfo.get_sim_time();

  const unsigned int block_cells = meshinfo.get_num_block_cells();

  parts.resize(0);

  const QuickFlash::File::Dataset & vel_set 
    = dfile.get_dataset(vel_dataset_name);

  const QuickFlash::File::Dataset & poly_set 
    = dfile.get_dataset(poly_dataset_name);
  
  std::vector<double> pos_data;

  QuickFlash::Block::BlockData<double> vel_data;
  QuickFlash::Block::BlockData<double> poly_data;

  // Run through the blocks (and the cells of the blocks) to set positions

  for (unsigned int block_index = 0; block_index < num_blocks; block_index++)
    if (meshinfo.is_leaf(block_index))
      {
      const QuickFlash::Block::BlockInfo & block_info 
	= meshinfo.get_block_info(block_index);

      vel_set.get_block_data(block_index, vel_data);
      poly_set.get_block_data(block_index, poly_data);

      for (unsigned int cell_index = 0; cell_index < block_cells; cell_index++)
	{
	block_info.get_cell_center(cell_index, pos_data);

	const double pos = pos_data[0];

	if ((!(pos < min_x)) && (pos < max_x))
	  {
	  const double vel = vel_data[cell_index];
	  const double poly = poly_data[cell_index];

	  parts.push_back(Part(pos, vel, poly));
	  }
	}
      }

  const unsigned int num_parts = parts.size();

  if (num_parts < 1)
    throw QuickFlash::Except("No positions found", __FILE__, __LINE__);

  // Sort the particles by position

  std::sort(parts.begin(), parts.end());

  min_read_pos = parts[0].get_pos();
  max_read_pos = parts[num_parts - 1].get_pos();
  }


void get_velocities(const std::string & filename, 
		    const std::vector<Part> & parts,
		    std::vector<double> & vel, double & current_time)
  {
  const unsigned int num_parts = parts.size();

  vel.resize(num_parts);

  QuickFlash::File::DataFile dfile(filename, buffer_size, true);

  const QuickFlash::File::MeshInfo & meshinfo = dfile.get_sim_info();

  const unsigned int dims = meshinfo.get_dims();

  if (dims != 1)
    throw QuickFlash::Except("Only one spatial dimension supported", __FILE__, 
			__LINE__);

  current_time = meshinfo.get_sim_time();

  const QuickFlash::File::Dataset & vel_set 
    = dfile.get_dataset(vel_dataset_name);

  bool block_set = false;

  unsigned int current_block = 0;

  QuickFlash::Block::BlockData<double> vel_data;

  std::vector<double> pos_vect(1, 0.0);

  // Run through the particles to find the velocities of the nearest grid cells

  for (unsigned int part_index = 0; part_index < num_parts; part_index++)
    {
    const double pos = parts[part_index].get_pos();

    pos_vect[0] = pos;

    // Get the block index and cell index for this position

    const unsigned int block_index = meshinfo.get_block_index(pos_vect);

    if ((!block_set) || (current_block != block_index))
      {
      vel_set.get_block_data(block_index, vel_data);

      current_block = block_index;
      block_set = true;
      }

    const QuickFlash::Block::BlockInfo & block_info 
      = meshinfo.get_block_info(block_index);

    const unsigned int cell_index = block_info.get_nearest_cell(pos_vect);

    vel[part_index] = vel_data[cell_index];
    }
  }


void get_poly(const std::string & filename, const std::vector<Part> & parts,
	      std::vector<double> & poly)
  {
  const unsigned int num_parts = parts.size();

  poly.resize(num_parts);

  QuickFlash::File::DataFile dfile(filename, buffer_size, true);

  const QuickFlash::File::MeshInfo & meshinfo = dfile.get_sim_info();

  const unsigned int dims = meshinfo.get_dims();

  if (dims != 1)
    throw QuickFlash::Except("Only one spatial dimension supported", __FILE__, 
			__LINE__);

  const QuickFlash::File::Dataset & poly_set 
    = dfile.get_dataset(poly_dataset_name);

  bool block_set = false;

  unsigned int current_block = 0;

  QuickFlash::Block::BlockData<double> poly_data;

  std::vector<double> pos_vect(1, 0.0);

  // Run through the particles to find the velocities of the nearest grid cells

  for (unsigned int part_index = 0; part_index < num_parts; part_index++)
    {
    const double pos = parts[part_index].get_pos();

    pos_vect[0] = pos;

    // Get the block index and cell index for this position

    const unsigned int block_index = meshinfo.get_block_index(pos_vect);

    if ((!block_set) || (current_block != block_index))
      {
      poly_set.get_block_data(block_index, poly_data);

      current_block = block_index;
      block_set = true;
      }

    const QuickFlash::Block::BlockInfo & block_info 
      = meshinfo.get_block_info(block_index);

    const unsigned int cell_index = block_info.get_nearest_cell(pos_vect);

    poly[part_index] = poly_data[cell_index];
    }
  }


void set_poly(const std::string & filename, std::vector<Part> & parts)
  {
  std::vector<double> poly_values;

  get_poly(filename, parts, poly_values);

  const unsigned int num_parts = parts.size();

  for (unsigned int index = 0; index < num_parts; index++)
    parts[index].set_poly(poly_values[index]);
  }


void move_parts(const double dt, std::vector<Part> & parts)
  {
  const unsigned int num_parts = parts.size();

  for (unsigned int index = 0; index < num_parts; index++)
    {
    const double current_pos = parts[index].get_pos();
    const double current_vel = parts[index].get_vel();

    const double new_pos 
      = current_pos + (dt * current_vel);

    parts[index].set_pos(new_pos);
    }
  }


void move_parts(const std::string & next_filename, std::vector<Part> & parts,
		const double prev_time, double & next_time)
  {
  std::vector<double> vel;

  get_velocities(next_filename, parts, vel, next_time);

  const double dt = next_time - prev_time;

  if (!(dt > 0.0))
    throw QuickFlash::Except("Non-positive time delta", __FILE__, __LINE__);

  // Insert velocities

  const unsigned int num_parts = parts.size();

  for (unsigned int index = 0; index < num_parts; index++)
    parts[index].set_vel(vel[index]);

  move_parts(dt, parts);
  }


void print_parts(const double sim_time, const double interface_pos,
		 const std::vector<Part> & parts)
  {
  using std::cout;
  using std::endl;

  // Print particles to stdout

  cout << sim_time << " " << interface_pos;

  const unsigned int num_parts = parts.size();

  for (unsigned int part_index = 0; part_index < num_parts; part_index++)
    cout << " " << parts[part_index].get_pos();

  cout << endl;
  }


double find_interface(const std::vector<Part> & parts)
  {
  const double interface_poly = 0.5;  // 50%

  // Work with a backup

  std::vector<Part> work_list = parts;

  // Sort by position

  std::sort(work_list.begin(), work_list.end());

  // Find the interface (50% poly), starting from the left

  const unsigned int num_parts = work_list.size();

  if (num_parts < 2)
    throw QuickFlash::Except("To few points to find interface", __FILE__, 
			   __LINE__);

  bool found_interface = false;

  unsigned int part_index = 0;

  while ((!found_interface) && (part_index < num_parts))
    {
    const double poly = work_list[part_index].get_poly();

    if (poly < interface_poly)
      found_interface = true;
    else
      part_index++;
    }

  if ((part_index == 0) || (!found_interface))
    throw QuickFlash::Except("Unable to locate interface in sample set", 
			   __FILE__, __LINE__);

  const unsigned int low_index = part_index - 1;
  const unsigned int high_index = part_index;

  const double low_pos = work_list[low_index].get_pos();
  const double low_poly = work_list[low_index].get_poly();

  const double high_pos = work_list[high_index].get_pos();
  const double high_poly = work_list[high_index].get_poly();

  const double pos_interval = high_pos - low_pos;  // Positive value

  if (!(pos_interval > 0.0))
    throw QuickFlash::Except("Error in position ordering", __FILE__, __LINE__);

  const double poly_interval = high_poly - low_poly;  // Negative value

  if (!(poly_interval < 0.0))
    throw QuickFlash::Except("Error in interface search", __FILE__, __LINE__);

  const double poly_delta = interface_poly - low_poly;

  const double pos_delta = (poly_delta / poly_interval) * pos_interval;

  return (low_pos + pos_delta);
  }


void check_pos(const double current_time, const std::vector<Part> & part_list,
	       double & min_pos, double & max_pos)
  {
  using std::cerr;
  using std::endl;

  // Make sure sample points are still in the correct order

  const unsigned int num_parts = part_list.size();

  if (num_parts > 1)
    {
    std::vector<unsigned int> misplaced_ids;

    double lastpos = part_list[0].get_pos();

    double min_x = lastpos;
    double max_x = lastpos;

    for (unsigned int index = 1; index < num_parts; index++)
      {
      const double currentpos = part_list[index].get_pos();

      if (!(currentpos > lastpos))
	misplaced_ids.push_back(index);
      else
	lastpos = currentpos;

      if (currentpos < min_x)
	min_x = currentpos;
      else if (currentpos > max_x)
	max_x = currentpos;
      }

    min_pos = min_x;
    max_pos = max_x;

    const unsigned int num_misplaced = misplaced_ids.size();

    if (num_misplaced > 0)
      {
      cerr << "Warning: Paths crossed at time " << current_time 
	   << " for particles [ ";

      for (unsigned int index = 0; index < num_misplaced; index++)
	cerr << misplaced_ids[index] << " ";

      cerr << "]" << endl;
      }
    }
  }
  

int main(int argc, char * argv[])
  {
  using std::cout;
  using std::cerr;
  using std::endl;

  const int req_args = 7;

  if (argc != req_args)
    {
    cerr << "Usage: " << argv[0] << " filename_base num_digits start_index "
	 << "end_index min_x max_x" << endl;
    return 1;
    }

  int argPtr = 1;

  const std::string filename_base = argv[argPtr++];

  const unsigned int num_digits 
    = static_cast<unsigned int>(atoi(argv[argPtr++]));

  const unsigned int start_index
    = static_cast<unsigned int>(atoi(argv[argPtr++]));

  const unsigned int end_index
    = static_cast<unsigned int>(atoi(argv[argPtr++]));

  if (start_index >= end_index)
    {
    cerr << "Error: start_index must be less than end_index" << endl;
    return -1;
    }

  const double min_x = atof(argv[argPtr++]);
  const double max_x = atof(argv[argPtr++]);

  if (!(min_x < max_x))
    {
    cerr << "Error: min_x must be less than max_x" << endl;
    return -1;
    }

  // Use the first file to generate the particles

  std::string filename;

  build_filename(filename_base, num_digits, start_index, filename);

  std::vector<Part> part_list;
  double sim_time = 0.0;
  double min_read_pos = 0.0;
  double max_read_pos = 0.0;

  cerr << "Loading initial file [ " << filename << " ]" << endl;

  get_particles(filename, min_x, max_x, part_list, sim_time, min_read_pos,
		max_read_pos);

  // Print the header - num_steps num_parts

  const unsigned int num_parts = part_list.size();

  cout << (end_index + 1) - start_index << " " << num_parts << endl;

  // Print the initial set of particles

  const double init_interface = find_interface(part_list);

  print_parts(sim_time, init_interface, part_list);

  // Run through the remaining files

  for (unsigned int file_index = start_index + 1; file_index <= end_index;
       file_index++)
    {
    build_filename(filename_base, num_digits, file_index, filename);

    double new_sim_time = 0.0;

    cerr << "Loading file [ " << filename << " ]" << endl;

    move_parts(filename, part_list, sim_time, new_sim_time);

    double min_check_pos = 0.0;
    double max_check_pos = 0.0;

    check_pos(new_sim_time, part_list, min_check_pos, max_check_pos);

    // Find the polyimide mass fraction for the new positions

    if (file_index < end_index)
      {
      std::string next_filename;

      build_filename(filename_base, num_digits, (file_index + 1), 
		     next_filename);

      set_poly(next_filename, part_list);
      }

    const double current_interface = find_interface(part_list);

    print_parts(new_sim_time, current_interface, part_list);

    sim_time = new_sim_time;
    }

  cout << endl;

  return 0;
  }
