// C++ header file velocity_field.hpp

/*
  By Nathan C. Hearn
     May 24, 2006

  Velocity field derived from 1D Flash simulations.
*/


#ifndef VELOCITY_FIELD_HPP
#define VELOCITY_FIELD_HPP


#include <string>
#include <vector>


class VelocityPoint
  {
  public:
    VelocityPoint() : pos(0.0), vel(0.0) { }

    VelocityPoint(const double position, const double velocity) :
      pos(0.0), vel(0.0)
      { reset(position, velocity); }

    VelocityPoint(const VelocityPoint & source) :
      pos(0.0), vel(0.0)
      { reset(source); }

    ~VelocityPoint() { }

    VelocityPoint & operator=(const VelocityPoint & source)
      {
      reset(source);
      return *this;
      }

    bool operator<(const VelocityPoint & rs) const { return (pos < rs.pos); }

    void reset()
      {
      pos = 0.0;
      vel = 0.0;
      }

    void reset(const double position, const double velocity)
      {
      pos = position;
      vel = velocity;
      }

    void reset(const VelocityPoint & source)
      {
      if (&source != this)
	{
	pos = source.pos;
	vel = source.vel;
	}
      }

    double get_pos() const { return pos; }

    double get_vel() const { return vel; }

  private:
    double pos;
    double vel;
  };


class VelocityField
  {
  public:
    VelocityField() : time(0.0), min_pos(0.0), max_pos(0.0), point_array() { }

    VelocityField(const std::string & flash_datafile, 
		  const unsigned int buffer_size=0) :
      time(0.0), min_pos(0.0), max_pos(0.0), point_array()
      { reset(flash_datafile, buffer_size); }

    VelocityField(const VelocityField & source) :
      time(0.0), min_pos(0.0), max_pos(0.0), point_array()
      { reset(source); }

    ~VelocityField() { }

    VelocityField & operator=(const VelocityField & source)
      {
      reset(source);
      return *this;
      }

    const VelocityPoint & operator[](const unsigned int index) const
      { return point_array[index]; }

    void reset();

    void reset(const std::string & flash_datafile, 
	       const unsigned int buffer_size=0);

    void reset(const VelocityField & source);

    double get_time() const { return time; }

    double get_min_pos() const { return min_pos; }
    double get_max_pos() const { return max_pos; }

    double get_interp_vel(const double pos) const;

  private:
    void load_file(const std::string & flash_datafile, 
		   const unsigned int buffer_size=0);

    void get_nearest_points(const double pos, unsigned int & prev_cell_index,
			    unsigned int & next_cell_index, 
			    double & delta_pos_prev,
			    double & delta_next_prev) const;

  private:
    double time;

    double min_pos;
    double max_pos;

    std::vector<VelocityPoint> point_array;
  };


class SimVelocityFields
  {
  public:
    SimVelocityFields() : vel_fields() { }

    SimVelocityFields(const std::string & filename_base, 
		      const unsigned int start_index, 
		      const unsigned int end_index,
		      const unsigned int num_digits=4,
		      const unsigned int buffer_size=1000) :
      vel_fields()
      { 
      reset(filename_base, start_index, end_index, num_digits, buffer_size); 
      }

    SimVelocityFields(const SimVelocityFields & source) :
      vel_fields()
      { reset(source); }

    ~SimVelocityFields() { }

    SimVelocityFields & operator=(const SimVelocityFields & source)
      {
      reset(source);
      return *this;
      }

    const VelocityField & operator[](const unsigned int index) const
      { return vel_fields[index]; }

    void reset() { vel_fields.clear(); }

    void reset(const std::string & filename_base, 
	       const unsigned int start_index, 
	       const unsigned int end_index, 
	       const unsigned int num_digits=4,
	       const unsigned int buffer_size=1000);

    void reset(const SimVelocityFields & source);

    double get_interp_vel(const double time, const double pos) const;
    double get_interp_acc(const double time, const double pos) const;

  private:
    void find_nearest_times(const double time, unsigned int & low_index,
			    unsigned int & high_index, double & delta_time_low,
			    double & delta_high_low) const;

  private:
    std::vector<VelocityField> vel_fields;
  };


#endif  // VELOCITY_FIELD_HPP
