#!/usr/bin/env python

# generate_report.py

# By Nathan C. Hearn
#    November 22, 2007


default_flame_slicer = "flame_slicer"
default_drawimage = "../../utils/drawimage/drawimage.py"

default_cross_resolution = int(256)

option_list_separator = ","


def main() :

    from optparse import OptionParser
    from sys import exit, stderr
    from os import system, path, makedirs
    from glob import glob

    from metadata_reader import parse_metadata
    from composite import SliceComposite

    # Set up the command line parser

    usage = str("%prog [ options ] input_files ...")
    version = str("%prog 0.1")

    argparser = OptionParser(usage, version=version)

    use_ascii_data_str = str("use-ascii-data")
    argparser.add_option("--" + use_ascii_data_str, action="store_true",
                         dest="use_ascii_data",
                         help="Use intermediate CSV files")
    argparser.set_defaults(use_ascii_data=False)

    flame_slicer_str = str("flame-slicer")
    argparser.add_option("--" + flame_slicer_str, type="string",
                         dest="flame_slicer", help="flame_slicer executable")
    argparser.set_defaults(flame_slicer=default_flame_slicer)

    cross_resolution_str = str("cross-resolution")
    argparser.add_option("--" + cross_resolution_str, type="int",
                         dest="cross_resolution",
                         help="Lateral resolution for slices")
    argparser.set_defaults(cross_resolution=default_cross_resolution)

    skip_images_str = str("skip-images")
    argparser.add_option("--" + skip_images_str, action="store_true",
                         dest="skip_images", help="Do not convert ASCII data")
    argparser.set_defaults(skip_images=False)

    drawimage_str = str("drawimage")
    argparser.add_option("--" + drawimage_str, type="string",
                         dest="drawimage",
                         help="drawimage program for converting CSV files")
    argparser.set_defaults(drawimage=default_drawimage)

    colormap_str = str("colormap")
    argparser.add_option("--" + colormap_str, type="string",
                         dest="colormap", help="Colormap file for drawimage")
    argparser.set_defaults(colormap=None)

    skip_report_str = str("skip-report")
    argparser.add_option("--" + skip_report_str, action="store_true",
                         dest="skip_report", help="Skip report generation")
    argparser.set_defaults(skip_report=False)

    skip_slicer_str = str("skip-slicer")
    argparser.add_option("--" + skip_slicer_str, action="store_true",
                         dest="skip_slicer", help="Skip flame_slicer")
    argparser.set_defaults(skip_slicer=False)

    output_path_str = str("output-path")
    argparser.add_option("--" + output_path_str, type="string",
                         dest="output_path",
                         help="Directory path for output files")
    argparser.set_defaults(output_path=None)

    user_datasets_str = str("datasets")
    argparser.add_option("--" + user_datasets_str, type="string",
                         dest="user_datasets",
                         help="Comma-separated list of datasets to process")
    argparser.set_defaults(user_datasets=None)
    

    # Process the command line input

    opts, args = argparser.parse_args()

    use_ascii_data = opts.use_ascii_data

    flame_slicer = opts.flame_slicer

    cross_resolution = opts.cross_resolution
    
    drawimage = opts.drawimage

    colormap_file = opts.colormap

    skip_images = opts.skip_images

    skip_report = opts.skip_report
    skip_slicer = opts.skip_slicer

    output_path = opts.output_path

    user_datasets = opts.user_datasets

    min_args = int(1)

    argc = len(args)

    if (argc < min_args) :
        argparser.print_help(file=stderr)
        exit(-1)

    argptr = int(0)

    file_list = list()
    
    while (argptr < argc) :
        file_list.append(args[argptr])
        argptr += 1


    # Basic checks

    if (skip_slicer and skip_report) :
        stderr.write("\nError: No operation requested\n\n")
        exit(-1)

    if (cross_resolution < 1) :
        stderr.write("\nError: Cross resolution must be a positive integer\n"
                     + "\n")
        exit(-1)


    # Make sure the directory exists

    if (output_path is not None) :
        if (not path.exists(output_path)) :
            makedirs(output_path)


    # A few basic settings

    drawimage_opts = ""

    if (colormap_file is not None) :
        drawimage_opts += "--colormap=" + colormap_file + " "

    
    # Perform the operation

    for filename in file_list :

        stderr.write("\nProcessing file [ " + filename + " ]\n\n")

        metadata_file = None

        base_file_name = path.split(filename)[1]

        if (base_file_name == "") :
            stderr.write("\nError: Unable to find base file name in "
                         + filename + "\n\n")
            raise RuntimeError("Unable to find base file name")

        current_output_prefix = base_file_name

        if (output_path is not None) :
            current_output_prefix = path.join(output_path, base_file_name)

        stderr.write("  Output file prefix [ " + current_output_prefix
                     + " ]\n\n")

        if (not skip_slicer) :

            slicer_options = filename \
                             + " --output-prefix=" + current_output_prefix \
                             + " --cross-resolution=" + str(cross_resolution)

            if (not use_ascii_data) :
                slicer_options += " --image-only"

            if (user_datasets is not None) :
                slicer_options += " --datasets=" + user_datasets

            slicer_cmd = flame_slicer + " " + slicer_options

            stderr.write("Running slicer command: " + slicer_cmd + "\n")

            system(slicer_cmd)  ## Check for errors???

            stderr.write("\n")

            metadata_file = current_output_prefix + "_metadata.out"

            stderr.write("Metadata file: " + metadata_file + "\n")

        else :

            metadata_file = filename

            # Determine the original data file name from metadata_file

            orig_file_end_index = metadata_file.rfind("_metadata.out")

            if (orig_file_end_index < 1) :
                stderr.write("\nError: Unable to parse filename " + filename
                             + " for base file name\n\n")
                exit(-2)

            current_output_prefix = metadata_file[0:orig_file_end_index]

            stderr.write("Using metadata file " + metadata_file + "\n")

        ## Generate the report

        if (not skip_report) :

            stderr.write("\n")

            file_md, full_dataset_md_list = parse_metadata(metadata_file)

            sim_file_name = path.split(file_md.get_group_name())[1]

            sim_time = file_md["Time"]

            flame_pos = file_md["Flame_Pos"]

            slice_min_coords = file_md["Grid_Min_Coords"]
            slice_max_coords = file_md["Grid_Max_Coords"]

            xsect_min_coords = file_md["XSect_Grid_Min_Coords"]
            xsect_max_coords = file_md["XSect_Grid_Max_Coords"]

            xsect_bottom_pos = file_md["XSect_Bottom_Pos"]
            xsect_flame_front_pos = file_md["XSect_Flame_Front_Pos"]
            xsect_top_pos = file_md["XSect_Top_Pos"]

            slice_report = SliceComposite(sim_file_name, sim_time, flame_pos,
                                          slice_min_coords, slice_max_coords,
                                          xsect_bottom_pos,
                                          xsect_flame_front_pos, xsect_top_pos,
                                          xsect_min_coords, xsect_max_coords)

            ## Limit the datasets to those chosen by the user; set up file name

            dataset_md_list = None

            slice_report_file = None

            if (user_datasets is not None) :

                slice_report_file = current_output_prefix
                
                user_dataset_list = user_datasets.split(option_list_separator)

                dataset_md_list = list()

                for dataset_md in full_dataset_md_list :

                    dataset_name = dataset_md.get_group_name()

                    if (dataset_name in user_dataset_list) :
                        
                        dataset_md_list.append(dataset_md)

                        slice_report_file += "_" + dataset_name

                if (len(dataset_md_list) < 1) :
                    raise RuntimeError("No user-selected datasets present")

                slice_report_file += "_slice.svg"
                
            else :
                
                dataset_md_list = full_dataset_md_list

                slice_report_file = current_output_prefix + "_slice.svg"

            ## Run through the datasets

            for dataset_md in dataset_md_list :

                dataset_name = dataset_md.get_group_name()

                dataset_min_value = dataset_md["Min_Value"]
                dataset_max_value = dataset_md["Max_Value"]

                xy_image_md = None
                xy_image_file = None

                xz_image_md = None
                xz_image_file = None

                xsect_bottom_image_md = None
                xsect_bottom_image_file = None

                xsect_flame_front_image_md = None
                xsect_flame_front_image_file = None

                xsect_top_image_md = None
                xsect_top_image_file = None

                if (use_ascii_data) :
                    
                    image_md_list \
                        = dataset_md.get_keyword_group_list(group_type="Data")

                    ## Find the xy- and xz-slices

                    num_images = len(image_md_list)

                    index = 0

                    while ((index < num_images)
                           and ((xy_image_md is None)
                                or (xz_image_md is None))) :
                        
                        image_md = image_md_list[index]

                        plane_id = image_md.get_group_name()

                        if (plane_id == "xy") :
                            xy_image_md = image_md
                        elif (plane_id == "xz") :
                            xz_image_md = image_md

                        index += 1

                    if (xy_image_md is None) :
                        raise RuntimeError("xy Plane image not found")

                    if (xz_image_md is None) :
                        raise RuntimeError("xz Plane image not found")

                    xy_data_file = xy_image_md["File"]
                    xz_data_file = xz_image_md["File"]

                    # Find the cross-sections

                    xsect_md_list \
                        = dataset_md.get_keyword_group_list("XSect_Data")

                    num_xsect_md = len(xsect_md_list)

                    index = 0

                    while ((index < num_xsect_md)
                           and ((xsect_bottom_image_md is None)
                                or (xsect_flame_front_image_md is None)
                                or (xsect_top_image_md is None))) :
                        
                        xsect_md = xsect_md_list[index]

                        xsect_type = xsect_md["Type"]
                            
                        if (xsect_type == "bottom") :
                            xsect_bottom_image_md = xsect_md
                        elif (xsect_type == "flame") :
                            xsect_flame_front_image_md = xsect_md
                        elif (xsect_type == "top") :
                            xsect_top_image_md = xsect_md

                        index += 1

                    if (xsect_bottom_image_md is None) :
                        raise RuntimeError("Bottom cross-section data "
                                           + "not found")

                    if (xsect_flame_front_image_md is None) :
                        raise RuntimeError("Flame front cross-section data "
                                           + "not found")

                    if (xsect_top_image_md is None) :
                        raise RuntimeError("Top cross-section data "
                                           + "not found")

                    xsect_bottom_data_file = xsect_bottom_image_md["File"]
                    
                    xsect_flame_front_data_file \
                        = xsect_flame_front_image_md["File"]

                    xsect_top_data_file = xsect_top_image_md["File"]

                    # Convert the image if necessary

                    if (not skip_images) :
                        
                        xy_drawimage_cmd = drawimage + " " + drawimage_opts \
                                           + " " + xy_data_file
                        xz_drawimage_cmd = drawimage + " " + drawimage_opts \
                                           + " " + xz_data_file

                        xsect_bottom_drawimage_cmd \
                            = drawimage + " " + drawimage_opts + " " \
                              + xsect_bottom_data_file
                        
                        xsect_flame_front_drawimage_cmd \
                            = drawimage + " " + drawimage_opts + " " \
                              + xsect_flame_front_data_file
                        
                        xsect_top_drawimage_cmd \
                            = drawimage + " " + drawimage_opts + " " \
                              + xsect_top_data_file

                        ## CHECK THE FOLLOWING SYSTEM CALLS FOR ERRORS???
                        
                        stderr.write("Running convert command: "
                                     + xy_drawimage_cmd + "\n")
                        
                        system(xy_drawimage_cmd)

                        stderr.write("Running convert command: "
                                     + xz_drawimage_cmd + "\n")
                        
                        system(xz_drawimage_cmd)

                        stderr.write("Running convert command: "
                                     + xsect_bottom_drawimage_cmd + "\n")
                        
                        system(xsect_bottom_drawimage_cmd)

                        stderr.write("Running convert command: "
                                     + xsect_flame_front_drawimage_cmd + "\n")
                        
                        system(xsect_flame_front_drawimage_cmd)

                        stderr.write("Running convert command: "
                                     + xsect_top_drawimage_cmd + "\n")
                        
                        system(xsect_top_drawimage_cmd)

                    xy_image_file = xy_data_file + ".png"
                    xz_image_file = xz_data_file + ".png"

                    xsect_bottom_image_file = xsect_bottom_data_file + ".png"

                    xsect_flame_front_image_file \
                        = xsect_flame_front_data_file + ".png"

                    xsect_top_image_file = xsect_top_data_file + ".png"

                else :

                    image_md_list \
                        = dataset_md.get_keyword_group_list(group_type="Image")

                    ## Find the xy- and xz-slices

                    num_images = len(image_md_list)

                    index = 0

                    while ((index < num_images) and
                           ((xy_image_md is None) or (xz_image_md is None))) :
                        
                        image_md = image_md_list[index]

                        plane_id = image_md.get_group_name()

                        if (plane_id == "xy") :
                            xy_image_md = image_md
                        elif (plane_id == "xz") :
                            xz_image_md = image_md

                        index += 1

                    if (xy_image_md is None) :
                        raise RuntimeError("xy Plane image not found")

                    if (xz_image_md is None) :
                        raise RuntimeError("xz Plane image not found")

                    # Find the cross-sections

                    xsect_md_list \
                        = dataset_md.get_keyword_group_list("XSect_Image")

                    num_xsect_md = len(xsect_md_list)

                    index = 0

                    while ((index < num_xsect_md)
                           and ((xsect_bottom_image_md is None)
                                or (xsect_flame_front_image_md is None)
                                or (xsect_top_image_md is None))) :
                        
                        xsect_md = xsect_md_list[index]

                        xsect_type = xsect_md["Type"]

                        if (xsect_type == "bottom") :
                            xsect_bottom_image_md = xsect_md
                        elif (xsect_type == "flame") :
                            xsect_flame_front_image_md = xsect_md
                        elif (xsect_type == "top") :
                            xsect_top_image_md = xsect_md

                        index += 1

                    if (xsect_bottom_image_md is None) :
                        raise RuntimeError("Bottom cross-section data "
                                           + "not found")

                    if (xsect_flame_front_image_md is None) :
                        raise RuntimeError("Flame front cross-section data "
                                           + "not found")

                    if (xsect_top_image_md is None) :
                        raise RuntimeError("Top cross-section data "
                                           + "not found")

                    xy_image_file = xy_image_md["File"]
                    xz_image_file = xz_image_md["File"]

                    xsect_bottom_image_file = xsect_bottom_image_md["File"]

                    xsect_flame_front_image_file \
                        = xsect_flame_front_image_md["File"]

                    xsect_top_image_file = xsect_top_image_md["File"]

                xy_slice_min_value = xy_image_md["Min_Value"]
                xy_slice_max_value = xy_image_md["Max_Value"]

                xz_slice_min_value = xz_image_md["Min_Value"]
                xz_slice_max_value = xz_image_md["Max_Value"]

                xsect_bottom_min_value = xsect_bottom_image_md["Min_Value"]
                xsect_bottom_max_value = xsect_bottom_image_md["Max_Value"]

                xsect_flame_front_min_value \
                    = xsect_flame_front_image_md["Min_Value"]

                xsect_flame_front_max_value \
                    = xsect_flame_front_image_md["Max_Value"]

                xsect_top_min_value = xsect_top_image_md["Min_Value"]
                xsect_top_max_value = xsect_top_image_md["Max_Value"]

                xy_image_file_nopath = path.split(xy_image_file)[1]
                xz_image_file_nopath = path.split(xz_image_file)[1]

                xsect_bottom_image_file_nopath \
                     = path.split(xsect_bottom_image_file)[1]

                xsect_flame_front_image_file_nopath \
                     = path.split(xsect_flame_front_image_file)[1]

                xsect_top_image_file_nopath \
                     = path.split(xsect_top_image_file)[1]

                slice_report.add_dataset(dataset_name,
                                         dataset_min_value, dataset_max_value,
                                         xy_image_file_nopath,
                                         xy_slice_min_value,
                                         xy_slice_max_value,
                                         xz_image_file_nopath,
                                         xz_slice_min_value,
                                         xz_slice_max_value,
                                         xsect_bottom_image_file_nopath,
                                         xsect_bottom_min_value,
                                         xsect_bottom_max_value,
                                         xsect_flame_front_image_file_nopath,
                                         xsect_flame_front_min_value,
                                         xsect_flame_front_max_value,
                                         xsect_top_image_file_nopath,
                                         xsect_top_min_value,
                                         xsect_top_max_value)

            slice_svg_code = slice_report.generate_svg(1000)

            slice_f = file(slice_report_file, "w")

            for line in slice_svg_code :
                slice_f.write(line + "\n")

            slice_f.write("\n")

            stderr.write("Wrote report file " + slice_report_file + "\n")

    stderr.write("\n")


if (__name__ == "__main__") :
    main()
