# metadata_reader.py

# By Nathan C. Hearn
#    November 21, 2007
#
# Reader for metadata files from flame_slicer.


from sys import stderr


class KeywordGroup :

    def __init__(self, group_name, group_type=None) :

        self.__name = group_name
        self.__type = group_type

        self.__sort_key = int(-1)  # Default is higher precedence than zero

        self.__values = dict()


    def get_group_name(self) :
        return self.__name


    def get_group_type(self) :
        return self.__type


    def set_sort_key(self, key) :
        self.__sort_key = key


    def get_sort_key(self) :
        return self.__sort_key


    def __cmp__(self, other) :
        return cmp(self.__sort_key, other.get_sort_key())


    def get_string(self, line_prefix="") :

        string = line_prefix + "Name [ " + str(self.__name) + " ]\n"

        if (self.__type is not None) :
            string += line_prefix + "  Type [ " + str(self.__type) + " ]\n"

        string += line_prefix + "  Sort key [ " + str(self.__sort_key) + " ]\n"

        for key in self.__values :
            string += line_prefix + "  Key [ " + str(key) + " ] Value [ " \
                      + str(self.__values[key]) + " ]\n"

        return string


    def __str__(self) :
        return self.get_string()
    

    def add_keyword_value(self, keyword, value=None) :

        if (self.__values.has_key(keyword)) :
            stderr.write("\nDuplicate keyword [ " + str(keyword) + " ]\n\n")
            raise RuntimeError("Duplicate keyword")

        self.__values[keyword] = value
        

    def keyword_present(self, keyword) :
        return self.__values.has_key(keyword)


    def get_value(self, keyword) :
        return self.__values[keyword]


    def __getitem__(self, keyword) :
        return self.get_value(keyword)


    def get_keyword_list(self) :
        return self.__values.keys()


class DataGroup(KeywordGroup) :

    def __init__(self, group_name, group_type=None) :

        KeywordGroup.__init__(self, group_name, group_type)

        self.__keyword_group_list = list()


    def get_string(self, line_prefix="") :

        string = KeywordGroup.get_string(self) + "\n"

        for group in self.__keyword_group_list :
            string += group.get_string(line_prefix + "  ") + "\n"
            
        return string


    def __str__(self) :
        return self.get_string()


    def add_keyword_group(self, keyword_group) :
        
        sort_key = keyword_group.get_sort_key()

        # Find the sorted location for this group

        num_groups = len(self.__keyword_group_list)

        index = (0)

        found_it = False

        while ((not found_it) and (index < num_groups)) :
            
            if (self.__keyword_group_list[index].get_sort_key() > sort_key) :
                found_it = True
            else :
                index += 1

        if (not found_it) :
            self.__keyword_group_list.append(keyword_group)
        else :
            self.__keyword_group_list.insert(index, keyword_group)


    def get_keyword_group_list(self, group_type=None) :

        keyword_group_list = list()

        for group in self.__keyword_group_list :

            include_group = False

            if (group_type is None) :
                include_group = True
            elif (group.get_group_type() == group_type) :
                include_group = True
                
            if (include_group) :
                keyword_group_list.append(group)

        return keyword_group_list


def split_vector(vector_string) :

    ## Split on both "," and " "

    split_words = vector_string.split()

    vector_elems = list()

    for word in split_words :
        comma_words = word.split(",")

        for elem in comma_words :
            vector_elems.append(elem)

    return vector_elems


def remove_comments(line) :
    return line.split("#")[0].strip()


def read_next_line(input_file) :

    ret_val = None  # Return None if at end of file

    finished = False

    while (not finished) :
        
        current_line = input_file.readline()

        if (current_line == "") :

            finished = True  # End of file
            
        else :
            
            work_line = remove_comments(current_line)

            if (work_line != "") :
                ret_val = work_line
                finished = True

    return ret_val


def parse_file_line(line) :

    work_line = line.strip()

    keyword_values = dict()

    split_pairs = work_line.split("]")

    num_split_pairs= len(split_pairs)

    # Parse the elements

    for pair in split_pairs :

        pair_elems = pair.split("[")

        num_elems = len(pair_elems)

        if ((num_elems > 2) or (num_elems < 1)) :
            stderr.write("\nError parsing metadata [ " + work_line + " ]\n\n")
            raise RuntimeError("Error parsing metadata")

        keyword = None
        value = None

        keyword = pair_elems[0].strip()

        temp_value = None

        if (num_elems > 1) :
            temp_value = pair_elems[1].strip()

        include_pair = True

        if (keyword == "") :
            if ((num_elems < 2) or (temp_value == "")) :
                include_pair = False
            else :
                stderr.write("\nError: Missing keyword [ " + work_line
                             + " ]\n\n")
                raise RuntimeError("Missing keyword")

        if (include_pair) :
        
            if (num_elems > 1) :
                if (temp_value != "") :
                    value = split_vector(temp_value)

                if (len(value) == 1) :
                    value = value[0]
                elif (len(value) < 1) :
                    value = None

            keyword_values[keyword] = value

    return keyword_values


def parse_metadata(filename) :

    datafile_group = None
    
    dataset_list = list()

    current_line = None

    f = file(filename, "r")

    # Read the file metadata

    current_line = read_next_line(f)

    if (current_line is None) :
        raise RuntimeError("EOF reached early")

    current_metadata = parse_file_line(current_line)

    if (len(current_metadata) != 1) :
        raise RuntimeError("Improper flash filename entry")

    flash_file_name = current_metadata["Flash file"]

    datafile_group = KeywordGroup(flash_file_name)
    
    finished_header = False

    while (not finished_header) :

        current_line = read_next_line(f)

        if (current_line is None) :

            finished_header = True  # End of file
            current_metadata = None

        else :

            current_metadata = parse_file_line(current_line)

            if (current_metadata.has_key("Dataset")) :
                finished_header = True
            else :

                if (current_metadata.has_key("Time")) :
                    
                    time = float(current_metadata["Time"])
                    datafile_group.add_keyword_value("Time", time)
                    
                elif (current_metadata.has_key("Flame pos")) :
                    
                    pos = float(current_metadata["Flame pos"])
                    datafile_group.add_keyword_value("Flame_Pos", pos)
                    
                elif (current_metadata.has_key("Grid min coords")) :

                    min_coords_str = current_metadata["Grid min coords"]
                    min_coords = list()

                    for elem in min_coords_str :
                        min_coords.append(float(elem))

                    datafile_group.add_keyword_value("Grid_Min_Coords",
                                                     min_coords)

                elif (current_metadata.has_key("Grid max coords")) :

                    max_coords_str = current_metadata["Grid max coords"]
                    max_coords = list()

                    for elem in max_coords_str :
                        max_coords.append(float(elem))

                    datafile_group.add_keyword_value("Grid_Max_Coords",
                                                     max_coords)

                elif (current_metadata.has_key("Grid center")) :

                    center_coords_str = current_metadata["Grid center"]
                    center_coords = list()

                    for elem in center_coords_str :
                        center_coords.append(float(elem))

                    datafile_group.add_keyword_value("Grid_Center",
                                                     center_coords)

                elif (current_metadata.has_key("Grid width")) :

                    axis_width_str = current_metadata["Grid width"]
                    axis_width = list()

                    for elem in axis_width_str :
                        axis_width.append(float(elem))

                    datafile_group.add_keyword_value("Grid_Width", axis_width)

                elif (current_metadata.has_key("Grid pixels")) :

                    grid_pixels_str = current_metadata["Grid pixels"]
                    grid_pixels = list()

                    for elem in grid_pixels_str :
                        grid_pixels.append(int(elem))

                    datafile_group.add_keyword_value("Grid_Pixels",
                                                     grid_pixels)

                elif (current_metadata.has_key("XSect grid min coords")) :

                    min_coords_str = current_metadata["XSect grid min coords"]
                    min_coords = list()

                    for elem in min_coords_str :
                        min_coords.append(float(elem))

                    datafile_group.add_keyword_value("XSect_Grid_Min_Coords",
                                                     min_coords)

                elif (current_metadata.has_key("XSect grid max coords")) :

                    max_coords_str = current_metadata["XSect grid max coords"]
                    max_coords = list()

                    for elem in max_coords_str :
                        max_coords.append(float(elem))

                    datafile_group.add_keyword_value("XSect_Grid_Max_Coords",
                                                     max_coords)

                elif (current_metadata.has_key("XSect grid center")) :

                    center_coords_str = current_metadata["XSect grid center"]
                    center_coords = list()

                    for elem in center_coords_str :
                        center_coords.append(float(elem))

                    datafile_group.add_keyword_value("XSect_Grid_Center",
                                                     center_coords)

                elif (current_metadata.has_key("XSect grid width")) :

                    axis_width_str = current_metadata["XSect grid width"]
                    axis_width = list()

                    for elem in axis_width_str :
                        axis_width.append(float(elem))

                    datafile_group.add_keyword_value("XSect_Grid_Width",
                                                     axis_width)

                elif (current_metadata.has_key("XSect grid pixels")) :

                    grid_pixels_str = current_metadata["XSect grid pixels"]
                    grid_pixels = list()

                    for elem in grid_pixels_str :
                        grid_pixels.append(int(elem))

                    datafile_group.add_keyword_value("XSect_Grid_Pixels",
                                                     grid_pixels)

                elif (current_metadata.has_key("XSect bottom x-pos")) :

                    xsect_bottom_pos \
                        = float(current_metadata["XSect bottom x-pos"])

                    datafile_group.add_keyword_value("XSect_Bottom_Pos",
                                                     xsect_bottom_pos)

                elif (current_metadata.has_key("XSect flame-front x-pos")) :

                    xsect_flame_front_pos \
                        = float(current_metadata["XSect flame-front x-pos"])

                    datafile_group.add_keyword_value("XSect_Flame_Front_Pos",
                                                     xsect_flame_front_pos)

                elif (current_metadata.has_key("XSect top x-pos")) :

                    xsect_top_pos \
                        = float(current_metadata["XSect top x-pos"])

                    datafile_group.add_keyword_value("XSect_Top_Pos",
                                                     xsect_top_pos)

    # Process datasets

    finished_datasets = False

    while (not finished_datasets) :

        if (current_metadata is None) :

            finished_datasets = True

        else :

            if (not current_metadata.has_key("Dataset")) :
                raise RuntimeError("Unrecognized file line")

            dataset_name = current_metadata["Dataset"]
            
            dataset_group = DataGroup(dataset_name, "Dataset")

            min_value = float(current_metadata["Min_Value"])
            max_value = float(current_metadata["Max_Value"])

            dataset_group.add_keyword_value("Min_Value", min_value)
            dataset_group.add_keyword_value("Max_Value", max_value)
            
            # Read the Image, data, and cross section info

            dataset_complete = False

            while (not dataset_complete) :

                current_line = read_next_line(f)

                if (current_line is None) :
                    
                    dataset_complete = True
                    current_metadata = None

                else :

                    current_metadata = parse_file_line(current_line)

                    if (current_metadata.has_key("Dataset")) :
                        
                        dataset_complete = True

                    else :

                        if (current_metadata.has_key("Image")) :

                            image_group \
                                = KeywordGroup(current_metadata["Image"],
                                               "Image")

                            filename = current_metadata["File"]

                            image_group.add_keyword_value("File", filename)

                            min_coords = list()
                            max_coords = list()

                            for elem in current_metadata["Min_Coords"] :
                                min_coords.append(float(elem))
                            
                            for elem in current_metadata["Max_Coords"] :
                                max_coords.append(float(elem))

                            image_group.add_keyword_value("Min_Coords",
                                                          min_coords)

                            image_group.add_keyword_value("Max_Coords",
                                                          max_coords)

                            min_val = float(current_metadata["Min_Value"])
                            max_val = float(current_metadata["Max_Value"])

                            image_group.add_keyword_value("Min_Value", min_val)
                            image_group.add_keyword_value("Max_Value", max_val)

                            dataset_group.add_keyword_group(image_group)
                    
                        elif (current_metadata.has_key("Data")) :

                            image_group \
                                = KeywordGroup(current_metadata["Data"],
                                               "Data")

                            filename = current_metadata["File"]

                            image_group.add_keyword_value("File", filename)

                            min_coords = list()
                            max_coords = list()

                            for elem in current_metadata["Min_Coords"] :
                                min_coords.append(float(elem))
                            
                            for elem in current_metadata["Max_Coords"] :
                                max_coords.append(float(elem))

                            image_group.add_keyword_value("Min_Coords",
                                                          min_coords)

                            image_group.add_keyword_value("Max_Coords",
                                                          max_coords)

                            min_val = float(current_metadata["Min_Value"])
                            max_val = float(current_metadata["Max_Value"])

                            image_group.add_keyword_value("Min_Value", min_val)
                            image_group.add_keyword_value("Max_Value", max_val)

                            dataset_group.add_keyword_group(image_group)

                        elif (current_metadata.has_key("XSect")) :

                            xsect_type = current_metadata["XSect"]

                            group_type = "XSect_" + xsect_type

                            image_group \
                                = KeywordGroup(current_metadata["File"],
                                               group_type)

                            filename = current_metadata["File"]

                            image_group.add_keyword_value("File", filename)

                            if (current_metadata.has_key("Index")) :
                                xsect_index = int(current_metadata["Index"])
                                image_group.set_sort_key(xsect_index)

                            xsect_type = str(current_metadata["Type"])
                            image_group.add_keyword_value("Type", xsect_type)

                            x_pos = float(current_metadata["Pos"])
                            image_group.add_keyword_value("Pos", x_pos)

                            min_coords = list()
                            max_coords = list()

                            for elem in current_metadata["Min_Coords"] :
                                min_coords.append(float(elem))
                            
                            for elem in current_metadata["Max_Coords"] :
                                max_coords.append(float(elem))

                            image_group.add_keyword_value("Min_Coords",
                                                          min_coords)

                            image_group.add_keyword_value("Max_Coords",
                                                          max_coords)

                            min_val = float(current_metadata["Min_Value"])
                            max_val = float(current_metadata["Max_Value"])

                            image_group.add_keyword_value("Min_Value", min_val)
                            image_group.add_keyword_value("Max_Value", max_val)

                            dataset_group.add_keyword_group(image_group)


            dataset_list.append(dataset_group)

    return datafile_group, dataset_list
