// C++ program file flame_dist.cpp

/*
  By Nathan C. Hearn
     May 6, 2008

  Distribution of material within flame.
*/


#include "flame_dist.hpp"
#include <vector>
#include <algorithm>
#include "quickflash_file_datafile.hpp"
#include "quickflash_file_dataset.hpp"
#include "quickflash_file_meshinfo.hpp"
#include "quickflash_block_blockinfo.hpp"
#include "quickflash_block_blockdata.hpp"
#include "quickflash_except.hpp"
#include "quickflash_stats_bins.hpp"
#include "quickflash_tree_block.hpp"
#include "file_defs.hpp"


// Class FlameDistribution

FlameDistribution::FlameDistribution() :
  total_mass(0.0), total_burned_mass(0.0), total_unburned_mass(0.0),
  bin_array_min_coord(0.0), bin_array_max_coord(0.0), num_bins(0),
  bin_width(0.0), bin_centers(), bin_volumes(),
  mass_bins_total(), dens_bins_total_mean(), dens_bins_total_dispersion(),
  mass_bins_burned(), dens_bins_burned_mean(), dens_bins_burned_dispersion(),
  mass_bins_unburned(), dens_bins_unburned_mean(), 
  dens_bins_unburned_dispersion(), mass_fract_bins_burned(),
  mass_fract_center_pos(0.0), mass_center_pos(0.0), flame_min_pos(0.0),
  flame_max_pos(0.0)
  { }


FlameDistribution::FlameDistribution(const QuickFlash::File::DataFile & df, 
				     const unsigned int search_axis,
				     const double bin_array_min_pos,
				     const unsigned int bin_count, 
				     const double sample_bin_width,
				     const double min_flame_value, 
				     const double max_flame_value) :
  total_mass(0.0), total_burned_mass(0.0), total_unburned_mass(0.0),
  bin_array_min_coord(0.0), bin_array_max_coord(0.0), num_bins(0),
  bin_width(0.0), bin_centers(), bin_volumes(),
  mass_bins_total(), dens_bins_total_mean(), dens_bins_total_dispersion(),
  mass_bins_burned(), dens_bins_burned_mean(), dens_bins_burned_dispersion(),
  mass_bins_unburned(), dens_bins_unburned_mean(), 
  dens_bins_unburned_dispersion(), mass_fract_bins_burned(),
  mass_fract_center_pos(0.0), mass_center_pos(0.0), flame_min_pos(0.0),
  flame_max_pos(0.0)
  { 
  reset(df, search_axis, bin_array_min_pos, bin_count, sample_bin_width,
	min_flame_value, max_flame_value); 
  }


FlameDistribution::FlameDistribution(const FlameDistribution & source) :
  total_mass(0.0), total_burned_mass(0.0), total_unburned_mass(0.0),
  bin_array_min_coord(0.0), bin_array_max_coord(0.0), num_bins(0),
  bin_width(0.0), bin_centers(), bin_volumes(),
  mass_bins_total(), dens_bins_total_mean(), dens_bins_total_dispersion(),
  mass_bins_burned(), dens_bins_burned_mean(), dens_bins_burned_dispersion(),
  mass_bins_unburned(), dens_bins_unburned_mean(), 
  dens_bins_unburned_dispersion(), mass_fract_bins_burned(),
  mass_fract_center_pos(0.0), mass_center_pos(0.0), flame_min_pos(0.0),
  flame_max_pos(0.0)
  { reset(source); }


void FlameDistribution::reset()
  {
  total_mass = 0.0;
  total_burned_mass = 0.0;
  total_unburned_mass = 0.0;

  bin_array_min_coord = 0.0;
  bin_array_max_coord = 0.0;

  num_bins = 0;

  bin_width = 0.0;

  bin_centers.clear();
  bin_volumes.clear();

  mass_bins_total.clear();
  dens_bins_total_mean.clear();
  dens_bins_total_dispersion.clear();

  mass_bins_burned.clear();
  dens_bins_burned_mean.clear();
  dens_bins_burned_dispersion.clear();

  mass_bins_unburned.clear();
  dens_bins_unburned_mean.clear();
  dens_bins_unburned_dispersion.clear();

  mass_fract_bins_burned.clear();

  mass_fract_center_pos = 0.0;
  mass_center_pos = 0.0;

  flame_min_pos = 0.0;
  flame_max_pos = 0.0;
  }


void FlameDistribution::reset(const QuickFlash::File::DataFile & df, 
			      const unsigned int search_axis,
			      const double bin_array_min_pos,
			      const unsigned int bin_count, 
			      const double sample_bin_width,
			      const double min_flame_value, 
			      const double max_flame_value)
  {
  reset();

  if (!(sample_bin_width > 0.0))
    throw QuickFlash::Except("Bin width must be positive", __FILE__, __LINE__);

  if (bin_count < 1)
    throw QuickFlash::Except("At least one bin required", __FILE__, __LINE__);

  bin_width = sample_bin_width;

  num_bins = bin_count;

  // Get the simulation info

  const QuickFlash::File::MeshInfo & meshinfo = df.get_mesh_info();

  const unsigned int dims = meshinfo.get_dims();

  if (search_axis >= dims)
    throw QuickFlash::Except("Search axis index out of range", __FILE__, 
			   __LINE__);

  // Get the datasets

  const QuickFlash::File::Dataset & dens_set = df.get_dataset(Density_Set_Name);
  const QuickFlash::File::Dataset & flame_set = df.get_dataset(Flame_Set_Name);

  // Set up the bins

  const double array_width = static_cast<float>(bin_count) * sample_bin_width;

  bin_array_min_coord = bin_array_min_pos;
  bin_array_max_coord = bin_array_min_coord + array_width;

  QuickFlash::Stats::BinArray total_dens_array(bin_array_min_pos,
					     bin_count, sample_bin_width);

  QuickFlash::Stats::BinArray burned_dens_array(bin_array_min_pos,
					      bin_count, sample_bin_width);

  QuickFlash::Stats::BinArray unburned_dens_array(bin_array_min_pos,
						bin_count, sample_bin_width);

  const double array_min_coord = total_dens_array.get_min_boundary();
  const double array_max_coord = total_dens_array.get_max_boundary();

  bin_array_min_coord = array_min_coord;
  bin_array_max_coord = array_max_coord;

  // Run through the data

  QuickFlash::Block::BlockData<double> dens_data;
  QuickFlash::Block::BlockData<double> flame_data;

  std::vector<double> cell_center;

  QuickFlash::Tree::ConstTreeIter node_iter = meshinfo.get_tree_iter();

  while (node_iter.valid_node())
    {
    const QuickFlash::Block::BlockInfo & block_info = node_iter.get_block_info();

    const double block_min_pos = block_info.get_min_coords()[search_axis];
    const double block_max_pos = block_info.get_max_coords()[search_axis];

    bool skip_branch = false;

    if ((array_min_coord > block_max_pos) || (array_max_coord < block_min_pos))
      skip_branch = true;

    if ((!skip_branch) && node_iter.is_leaf())
      {
      const unsigned int block_index = node_iter.get_block_index();

      const double cell_volume = block_info.get_cell_volume();  // CARTESIAN!!
      const double cell_axis_width = block_info.get_cell_width()[search_axis];

      const unsigned int num_cells = block_info.get_num_cells();

      // Load the data

      dens_set.get_block_data(block_index, dens_data);
      flame_set.get_block_data(block_index, flame_data);

      // Process the cells

      for (unsigned int cell_index = 0; cell_index < num_cells; cell_index++)
	{
	block_info.get_cell_center(cell_index, cell_center);

	const double search_axis_center = cell_center[search_axis];

	const double cell_dens = dens_data[cell_index];

	const double cell_burn_fract = flame_data[cell_index];

	const double cell_unburn_fract 
	  = std::max(0.0, std::min(1.0, 1.0 - cell_burn_fract));

	const double burned_dens = cell_burn_fract * cell_dens;
	const double unburned_dens = cell_unburn_fract * cell_dens;

	// WE SHOULD CLIP THE CELLS CROSSING MIN_COORD OR MAX_COORD!!!

	total_dens_array.add_data_spread(search_axis_center, cell_dens,
					 cell_axis_width, cell_volume);
					   
	burned_dens_array.add_data_spread(search_axis_center, burned_dens,
					  cell_axis_width, cell_volume);

	unburned_dens_array.add_data_spread(search_axis_center, unburned_dens,
					    cell_axis_width, cell_volume);
	}
      }

    if (skip_branch)
      node_iter.next_branch();
    else
      node_iter.next();
    }

  // Build the member arrays

  bin_centers = total_dens_array.get_bin_centers();

  bin_volumes.resize(bin_count, 0.0);

  mass_bins_total.resize(bin_count, 0.0);
  dens_bins_total_mean.resize(bin_count, 0.0);
  dens_bins_total_dispersion.resize(bin_count, 0.0);

  mass_bins_burned.resize(bin_count, 0.0);
  dens_bins_burned_mean.resize(bin_count, 0.0);
  dens_bins_burned_dispersion.resize(bin_count, 0.0);

  mass_bins_unburned.resize(bin_count, 0.0);
  dens_bins_unburned_mean.resize(bin_count, 0.0);
  dens_bins_unburned_dispersion.resize(bin_count, 0.0);

  mass_fract_bins_burned.resize(bin_count, 0.0);

  double array_sum_total_mass = 0.0;
  double array_sum_burned_mass = 0.0;
  double array_sum_unburned_mass = 0.0;

  for (unsigned int bin_index = 0; bin_index < bin_count; bin_index++)
    {
    const QuickFlash::Stats::Bin & total_bin = total_dens_array[bin_index];

    const QuickFlash::Stats::Bin & burned_bin = burned_dens_array[bin_index];
    const QuickFlash::Stats::Bin & unburned_bin = unburned_dens_array[bin_index];

    // The total mass is the mean value times the weight (volume) of the bin

    const double bin_vol = total_bin.get_weight_sum();

    const double bin_mass = total_bin.get_weight_value_sum();

    double bin_dens = 0.0;
    double bin_dens_disp = 0.0;

    if (bin_mass > 0.0)
      {
      bin_dens = total_bin.get_mean();
      bin_dens_disp = total_bin.get_pop_stddev();
      }

    const double bin_burned_mass = burned_bin.get_weight_value_sum();

    double burned_dens = 0.0;
    double burned_dens_disp = 0.0;

    if (bin_burned_mass > 0.0)
      {
      burned_dens = burned_bin.get_mean();
      burned_dens_disp = burned_bin.get_pop_stddev();
      }

    const double bin_unburned_mass = unburned_bin.get_weight_value_sum();

    double unburned_dens = 0.0;
    double unburned_dens_disp = 0.0;

    if (bin_unburned_mass > 0.0)
      {
      unburned_dens = unburned_bin.get_mean();
      unburned_dens_disp = unburned_bin.get_pop_stddev();
      }

    const double bin_flame_fract 
      = (bin_mass > 0.0) ? (bin_burned_mass / bin_mass) : 0.0;
  
    // Store the values

    bin_volumes[bin_index] = bin_vol;

    mass_bins_total[bin_index] = bin_mass;
    dens_bins_total_mean[bin_index] = bin_dens;
    dens_bins_total_dispersion[bin_index] = bin_dens_disp;

    mass_bins_burned[bin_index] = bin_burned_mass;
    dens_bins_burned_mean[bin_index] = burned_dens;
    dens_bins_burned_dispersion[bin_index] = burned_dens_disp;

    mass_bins_unburned[bin_index] = bin_unburned_mass;
    dens_bins_unburned_mean[bin_index] = unburned_dens;
    dens_bins_unburned_dispersion[bin_index] = unburned_dens_disp;

    mass_fract_bins_burned[bin_index] = bin_flame_fract;

    // Add to accumulation variables

    array_sum_total_mass += bin_mass;

    array_sum_burned_mass += bin_burned_mass;
    array_sum_unburned_mass += bin_unburned_mass;
    }

  total_mass = array_sum_total_mass;
  total_burned_mass = array_sum_burned_mass;
  total_unburned_mass = array_sum_unburned_mass;

  if (find_flame_center_mass_fract() < 0)
    throw QuickFlash::Except("Unable to find mass fraction center", __FILE__,
			   __LINE__);

  if (find_flame_center_mass() < 0)
    throw QuickFlash::Except("Unable to find mass center", __FILE__, __LINE__);

  if (find_flame_extremes(min_flame_value, max_flame_value) < 0)
    throw QuickFlash::Except("Unable to find flame boundaries", __FILE__,
			   __LINE__);
  }


void FlameDistribution::reset(const FlameDistribution & source)
  {
  if (&source != this)
    {
    total_mass = source.total_mass;
    total_burned_mass = source.total_burned_mass;
    total_unburned_mass = source.total_unburned_mass;

    bin_array_min_coord = source.bin_array_min_coord;
    bin_array_max_coord = source.bin_array_max_coord;

    num_bins = source.num_bins;

    bin_width = source.bin_width;

    bin_centers = source.bin_centers;
    bin_volumes = source.bin_volumes;

    mass_bins_total = source.mass_bins_total;
    dens_bins_total_mean = source.dens_bins_total_mean;
    dens_bins_total_dispersion = source.dens_bins_total_dispersion;

    mass_bins_burned = source.mass_bins_burned;
    dens_bins_burned_mean = source.dens_bins_burned_mean;
    dens_bins_burned_dispersion = source.dens_bins_burned_dispersion;

    mass_bins_unburned = source.mass_bins_unburned;
    dens_bins_unburned_mean = source.dens_bins_unburned_mean;
    dens_bins_unburned_dispersion = source.dens_bins_unburned_dispersion;

    mass_fract_bins_burned = source.mass_fract_bins_burned;

    mass_fract_center_pos = source.mass_fract_center_pos;
    mass_center_pos = source.mass_center_pos;

    flame_min_pos = source.flame_min_pos;
    flame_max_pos = source.flame_max_pos;
    }
  }


int FlameDistribution::find_flame_center_mass_fract()
  {
  // NOTE: The bin arrays must be set up BEFORE calling this function

  // <x> = sum_i(x_i * burned_fract_i * unburned_fract_i) 
  //       / sum_i(burned_fract_i * unburned_fract_i)

  int ret_val = -1;  // Error state

  const unsigned int bin_count = num_bins;

  if (bin_count > 0)
    {
    double accum_weight_pos = 0.0;
    double accum_weight = 0.0;

    for (unsigned int bin_index = 0; bin_index < bin_count; bin_index++)
      {
      const double bin_pos = bin_centers[bin_index];
      const double mass_fract = mass_fract_bins_burned[bin_index];

      const double weight = mass_fract * (1.0 - mass_fract);

      accum_weight_pos += bin_pos * weight;
      accum_weight += weight;
      }

    if (accum_weight > 0.0)
      {
      mass_fract_center_pos = accum_weight_pos / accum_weight;

      ret_val = 0;  // OK
      }
    }

  return ret_val;
  }


int FlameDistribution::find_flame_center_mass()
  {
  // NOTE: The bin arrays must be set up BEFORE calling this function

  // <x> = sum_i(x_i * dens_fuel_i * dens_ash_i) 
  //       / sum_i(dens_fuel_i * dens_ash_i)

  int ret_val = -1;  // Error state

  const unsigned int bin_count = num_bins;

  if (bin_count > 0)
    {
    double accum_weight_pos = 0.0;
    double accum_weight = 0.0;

    for (unsigned int bin_index = 0; bin_index < bin_count; bin_index++)
      {
      const double bin_pos = bin_centers[bin_index];
      const double bin_dens = dens_bins_total_mean[bin_index];
      const double mass_fract = mass_fract_bins_burned[bin_index];

      const double weight 
	= ((1.0 - mass_fract) * bin_dens) * (mass_fract * bin_dens);

      accum_weight_pos += bin_pos * weight;
      accum_weight += weight;
      }

    if (accum_weight > 0.0)
      {
      mass_center_pos = accum_weight_pos / accum_weight;

      ret_val = 0;  // OK
      }
    }

  return ret_val;
  }


int FlameDistribution::find_flame_extremes(const double min_flame_value, 
					   const double max_flame_value)
  {
  // NOTE: The bin arrays must be set up BEFORE calling this function

  // Search for changes in mass fraction at ends of arrays

  int ret_val = -1;  // Error state

  const unsigned int bin_count = num_bins;

  if ((bin_count > 1) && (min_flame_value < max_flame_value))
    {
    bool found_flame = false;

    double min_pos = 0.0;
    double max_pos = 0.0;

    for (unsigned int bin_index = 0; bin_index < bin_count; bin_index++)
      {
      const double bin_pos = bin_centers[bin_index];
      const double mass_fract = mass_fract_bins_burned[bin_index];

      if ((!(mass_fract < min_flame_value)) 
	  && (!(mass_fract > max_flame_value)))
	{
	if (!found_flame)
	  {
	  min_pos = bin_pos;
	  max_pos = bin_pos;
	  
	  found_flame = true;
	  }
	else
	  max_pos = bin_pos;  // We are stepping through increasing positions
	}
      }

    if (found_flame)
      {
      flame_min_pos = min_pos;
      flame_max_pos = max_pos;

      ret_val = 0;  // OK
      }
    }

  return ret_val;
  }



