// C++ program file mpi_utils.cpp

/*
  By Nathan C. Hearn
     February 10, 2008

  MPI-related utilities.
*/


#ifdef USE_MPI

// Include the MPI headers first to make MPI/xlC on BG/P happy ... (grrrr....)

#include <mpi.h>

#include "mpi_utils.hpp"
#include <vector>
#include "quickflash_except.hpp"


void send_vect(const int dest_rank, const int mpi_tag,
	       const std::vector<unsigned int> & data,
	       const MPI_Comm mpi_comm)
  {
  const unsigned int num_elems = data.size();

  unsigned int * data_array = new unsigned int[num_elems];

  for (unsigned int i = 0; i < num_elems; i++)
    data_array[i] = data[i];

  MPI_Request mpi_req;

  MPI_Isend(data_array, num_elems, MPI_UNSIGNED, dest_rank, mpi_tag, mpi_comm,
	    &mpi_req);

  MPI_Status status;

  MPI_Wait(&mpi_req, &status);

  delete[] data_array;
  data_array = 0;
  }


void recv_vect(const int src_rank, const int mpi_tag,
	       std::vector<unsigned int> & data,
	       const MPI_Comm mpi_comm)
  {
  MPI_Status status;

  MPI_Probe(src_rank, mpi_tag, mpi_comm, &status);

  int int_message_length = 0;

  MPI_Get_count(&status, MPI_UNSIGNED, &int_message_length);

  if (int_message_length >= 0)
    {
    const unsigned int message_length 
      = static_cast<unsigned int>(int_message_length);

    unsigned int * data_array = new unsigned int[message_length];

    MPI_Request mpi_req;

    MPI_Irecv(data_array, int_message_length, MPI_UNSIGNED, src_rank, mpi_tag,
	      mpi_comm, &mpi_req);

    MPI_Status recv_status;

    MPI_Wait(&mpi_req, &recv_status);

    data.resize(message_length);

    for (unsigned int i = 0; i < message_length; i++)
      data[i] = data_array[i];

    delete[] data_array;
    data_array = 0;
    }
  else
    throw QuickFlash::Except("Unrecognized message length", __FILE__, __LINE__);
  }


void send_vect(const int dest_rank, const int mpi_tag,
	       const std::vector<double> & data,
	       const MPI_Comm mpi_comm)
  {
  const unsigned int num_elems = data.size();

  double * data_array = new double[num_elems];

  for (unsigned int i = 0; i < num_elems; i++)
    data_array[i] = data[i];

  MPI_Request mpi_req;

  MPI_Isend(data_array, num_elems, MPI_DOUBLE, dest_rank, mpi_tag, mpi_comm,
	    &mpi_req);

  MPI_Status status;

  MPI_Wait(&mpi_req, &status);

  delete[] data_array;
  data_array = 0;
  }



void recv_vect(const int src_rank, const int mpi_tag,
	       std::vector<double> & data,
	       const MPI_Comm mpi_comm)
  {
  MPI_Status status;

  MPI_Probe(src_rank, mpi_tag, mpi_comm, &status);

  int int_message_length = 0;

  MPI_Get_count(&status, MPI_DOUBLE, &int_message_length);

  if (int_message_length >= 0)
    {
    const unsigned int message_length 
      = static_cast<unsigned int>(int_message_length);

    double * data_array = new double[message_length];

    MPI_Request mpi_req;

    MPI_Irecv(data_array, int_message_length, MPI_DOUBLE, src_rank, mpi_tag,
	      mpi_comm, &mpi_req);

    MPI_Status recv_status;

    MPI_Wait(&mpi_req, &recv_status);

    data.resize(message_length);

    for (unsigned int i = 0; i < message_length; i++)
      data[i] = data_array[i];

    delete[] data_array;
    data_array = 0;
    }
  else
    throw QuickFlash::Except("Unrecognized message length", __FILE__, __LINE__);
  }


void bcast_vect_server(const std::vector<unsigned int> & src_data,
		       const MPI_Comm mpi_comm)
  {
  int mpi_rank = 0;

  MPI_Comm_rank(mpi_comm, &mpi_rank);

  unsigned int num_elems = src_data.size();

  MPI_Bcast(&num_elems, 1, MPI_UNSIGNED, mpi_rank, mpi_comm);

  if (num_elems > 0)
    {
    unsigned int * src_array = new unsigned int[num_elems];

    for (unsigned int index = 0; index < num_elems; index++)
      src_array[index] = src_data[index];

    MPI_Bcast(src_array, num_elems, MPI_UNSIGNED, mpi_rank, mpi_comm);

    delete[] src_array;
    src_array = 0;
    }
  }


void bcast_vect_client(const int src_rank, 
		       std::vector<unsigned int> & recv_data, 
		       const MPI_Comm mpi_comm)
  {
  int mpi_rank =  0;

  MPI_Comm_rank(mpi_comm, &mpi_rank);

  if (mpi_rank == src_rank)
    throw QuickFlash::Except("Client scatter call from source process", 
			   __FILE__, __LINE__);

  unsigned int num_elems = 0;
    
  MPI_Bcast(&num_elems, 1, MPI_UNSIGNED, src_rank, mpi_comm);

  if (num_elems > 0)
    {
    unsigned int * recv_array = new unsigned int[num_elems];

    MPI_Bcast(recv_array, num_elems, MPI_UNSIGNED, src_rank, mpi_comm);

    recv_data.resize(num_elems);

    for (unsigned int index = 0; index < num_elems; index++)
      recv_data[index] = recv_array[index];

    delete[] recv_array;
    recv_array = 0;
    }
  }


void bcast_vect_server(const std::vector<double> & src_data,
		       const MPI_Comm mpi_comm)
  {
  int mpi_rank = 0;

  MPI_Comm_rank(mpi_comm, &mpi_rank);

  unsigned int num_elems = src_data.size();

  MPI_Bcast(&num_elems, 1, MPI_UNSIGNED, mpi_rank, mpi_comm);

  if (num_elems > 0)
    {
    double * src_array = new double[num_elems];

    for (unsigned int index = 0; index < num_elems; index++)
      src_array[index] = src_data[index];

    MPI_Bcast(src_array, num_elems, MPI_DOUBLE, mpi_rank, mpi_comm);

    delete[] src_array;
    src_array = 0;
    }
  }


void bcast_vect_client(const int src_rank, std::vector<double> & recv_data, 
		       const MPI_Comm mpi_comm)
  {
  int mpi_rank =  0;

  MPI_Comm_rank(mpi_comm, &mpi_rank);

  if (mpi_rank == src_rank)
    throw QuickFlash::Except("Client scatter call from source process", 
			   __FILE__, __LINE__);

  unsigned int num_elems = 0;
    
  MPI_Bcast(&num_elems, 1, MPI_UNSIGNED, src_rank, mpi_comm);

  if (num_elems > 0)
    {
    double * recv_array = new double[num_elems];

    MPI_Bcast(recv_array, num_elems, MPI_DOUBLE, src_rank, mpi_comm);

    recv_data.resize(num_elems);

    for (unsigned int index = 0; index < num_elems; index++)
      recv_data[index] = recv_array[index];

    delete[] recv_array;
    recv_array = 0;
    }
  }


void sum_vect_master(const std::vector<unsigned int> & local_data,
		     std::vector<unsigned int> & global_sum_data,
		     const MPI_Comm mpi_comm)
  {
  int mpi_rank = 0;

  MPI_Comm_rank(mpi_comm, &mpi_rank);

  unsigned int num_elems = local_data.size();

  MPI_Bcast(&num_elems, 1, MPI_UNSIGNED, mpi_rank, mpi_comm);

  if (num_elems > 0)
    {
    unsigned int * local_array = new unsigned int[num_elems];

    for (unsigned int index = 0; index < num_elems; index++)
      local_array[index] = local_data[index];

    unsigned int * global_array = new unsigned int[num_elems];

    MPI_Reduce(local_array, global_array, num_elems, MPI_UNSIGNED, MPI_SUM,
	       mpi_rank, mpi_comm);

    delete[] local_array;
    local_array = 0;

    global_sum_data.resize(num_elems);

    for (unsigned int index = 0; index < num_elems; index++)
      global_sum_data[index] = global_array[index];

    delete[] global_array;
    global_array = 0;
    }
  }


void sum_vect_worker(const int dest_rank, 
		     const std::vector<unsigned int> & local_data,
		     const MPI_Comm mpi_comm)
  {
  int mpi_rank =  0;

  MPI_Comm_rank(mpi_comm, &mpi_rank);

  if (mpi_rank == dest_rank)
    throw QuickFlash::Except("Client MPI call from source process", 
			   __FILE__, __LINE__);

  unsigned int num_elems = 0;
    
  MPI_Bcast(&num_elems, 1, MPI_UNSIGNED, dest_rank, mpi_comm);

  const unsigned int local_elems = local_data.size();

  if ((local_elems > 0) && (local_elems != num_elems))
    throw QuickFlash::Except("Incompatible local data vector", __FILE__, 
			   __LINE__);

  if (num_elems > 0)
    {
    unsigned int * local_array = new unsigned int[num_elems];

    if (local_elems > 0)
      for (unsigned int index = 0; index < num_elems; index++)
	local_array[index] = local_data[index];
    else
      for (unsigned int index = 0; index < num_elems; index++)
	local_array[index] = 0;

    MPI_Reduce(local_array, 0, num_elems, MPI_UNSIGNED, MPI_SUM,
		dest_rank, mpi_comm);

    delete[] local_array;
    local_array = 0;
    }
  }


void sum_vect_master(const std::vector<double> & local_data,
		     std::vector<double> & global_sum_data,
		     const MPI_Comm mpi_comm)
  {
  int mpi_rank = 0;

  MPI_Comm_rank(mpi_comm, &mpi_rank);

  unsigned int num_elems = local_data.size();

  MPI_Bcast(&num_elems, 1, MPI_UNSIGNED, mpi_rank, mpi_comm);

  if (num_elems > 0)
    {
    double * local_array = new double[num_elems];

    for (unsigned int index = 0; index < num_elems; index++)
      local_array[index] = local_data[index];

    double * global_array = new double[num_elems];

    MPI_Reduce(local_array, global_array, num_elems, MPI_DOUBLE, MPI_SUM,
	       mpi_rank, mpi_comm);

    delete[] local_array;
    local_array = 0;

    global_sum_data.resize(num_elems);

    for (unsigned int index = 0; index < num_elems; index++)
      global_sum_data[index] = global_array[index];

    delete[] global_array;
    global_array = 0;
    }
  }


void sum_vect_worker(const int dest_rank, 
		     const std::vector<double> & local_data,
		     const MPI_Comm mpi_comm)
  {
  int mpi_rank =  0;

  MPI_Comm_rank(mpi_comm, &mpi_rank);

  if (mpi_rank == dest_rank)
    throw QuickFlash::Except("Client MPI call from source process", 
			   __FILE__, __LINE__);

  unsigned int num_elems = 0;
    
  MPI_Bcast(&num_elems, 1, MPI_UNSIGNED, dest_rank, mpi_comm);

  const unsigned int local_elems = local_data.size();

  if ((local_elems > 0) && (local_elems != num_elems))
    throw QuickFlash::Except("Incompatible local data vector", __FILE__, 
			   __LINE__);

  if (num_elems > 0)
    {
    double * local_array = new double[num_elems];

    if (local_elems > 0)
      for (unsigned int index = 0; index < num_elems; index++)
	local_array[index] = local_data[index];
    else
      for (unsigned int index = 0; index < num_elems; index++)
	local_array[index] = 0.0;

    MPI_Reduce(local_array, 0, num_elems, MPI_DOUBLE, MPI_SUM,
		dest_rank, mpi_comm);

    delete[] local_array;
    local_array = 0;
    }
  }


#endif  // USE_MPI
