// C++ program file structure.cpp

/*
  By Nathan C. Hearn
     February 1, 2008

  Functions for computing structure functions.

  For right now, it is assumed that periodic boundary conditions exist.
*/


// Include the MPI headers first to make MPI/xlC on BG/P happy ... (grrrr....)

#ifdef USE_MPI
#include <mpi.h>
#include "decomp_utils.hpp"
#include "mpi_utils.hpp"
#endif  // USE_MPI

#include "structure.hpp"
#include <cmath>
#include <vector>
#include "quickflash_block_blockinfo.hpp"
#include "quickflash_block_blockdata.hpp"
#include "quickflash_except.hpp"
#include "quickflash_counters.hpp"
#include "quickflash_geometry.hpp"
#include "quickflash_stats_bins.hpp"


void get_scalar_struct_fn(const QuickFlash::Block::BlockInfo & grid_info,
			  const GridVector & vel_data, 
			  const double bin_width, 
			  const double max_radius, 
			  const std::vector<unsigned int> & fn_order_list,
			  std::vector<double> & r_bin_radii,
			  std::vector<unsigned int> & r_bin_cell_count,
			  std::vector< std::vector<double> > 
			    & fn_values_r_list,
			  std::vector< std::vector<double> > 
			    & fn_stddev_r_list,
			  std::vector<double> & y_bin_positions,
			  std::vector<unsigned int> & y_bin_cell_count,
			  std::vector< std::vector<double> > 
			    & fn_values_y_list,
			  std::vector< std::vector<double> >
			    & fn_stddev_y_list,
			  std::vector<double> & z_bin_positions,
			  std::vector<unsigned int> & z_bin_cell_count,
			  std::vector< std::vector<double> > 
			    & fn_values_z_list,
			  std::vector< std::vector<double> >
			    & fn_stddev_z_list)
  {
  const unsigned int num_orders = fn_order_list.size();

  if (num_orders < 1)
    throw QuickFlash::Except("No function orders specified", __FILE__, __LINE__);

  const unsigned int num_cells = grid_info.get_num_cells();

  std::vector<QuickFlash::Stats::BinArray> function_bins_r_list;
  std::vector<QuickFlash::Stats::BinArray> function_bins_y_list;
  std::vector<QuickFlash::Stats::BinArray> function_bins_z_list;

  get_scalar_struct_fn(grid_info, vel_data, bin_width, max_radius, 
		       fn_order_list, 0, num_cells, function_bins_r_list,
		       function_bins_y_list, function_bins_z_list);

  const QuickFlash::Stats::BinArray & prototype_r_bins = function_bins_r_list[0];

  r_bin_radii = prototype_r_bins.get_bin_centers();

  const unsigned int num_r_bins = r_bin_radii.size();

  r_bin_cell_count.resize(num_r_bins);

  for (unsigned int bin_index = 0; bin_index < num_r_bins; bin_index++)
    r_bin_cell_count[bin_index] 
      = prototype_r_bins[bin_index].get_num_entries();

  fn_values_r_list.resize(num_orders);
  fn_stddev_r_list.resize(num_orders);

  for (unsigned int order_index = 0; order_index < num_orders; order_index++)
    {
    const QuickFlash::Stats::BinArray & function_bins_r
      = function_bins_r_list[order_index];

    std::vector<double> & fn_values = fn_values_r_list[order_index];
    std::vector<double> & stddev_values = fn_stddev_r_list[order_index];

    fn_values.resize(num_r_bins);
    stddev_values.resize(num_r_bins);

    for (unsigned int bin_index = 0; bin_index < num_r_bins; bin_index++)
      {
      fn_values[bin_index] = function_bins_r[bin_index].get_mean();
      stddev_values[bin_index] = function_bins_r[bin_index].get_pop_stddev();
      }
    }

  const QuickFlash::Stats::BinArray & prototype_y_bins = function_bins_y_list[0];

  y_bin_positions = prototype_y_bins.get_bin_centers();

  const unsigned int num_y_bins = y_bin_positions.size();

  y_bin_cell_count.resize(num_y_bins);

  for (unsigned int bin_index = 0; bin_index < num_y_bins; bin_index++)
    y_bin_cell_count[bin_index] 
      = prototype_y_bins[bin_index].get_num_entries();

  fn_values_y_list.resize(num_orders);
  fn_stddev_y_list.resize(num_orders);

  for (unsigned int order_index = 0; order_index < num_orders; order_index++)
    {
    const QuickFlash::Stats::BinArray & function_bins_y
      = function_bins_y_list[order_index];

    std::vector<double> & fn_values = fn_values_y_list[order_index];
    std::vector<double> & stddev_values = fn_stddev_y_list[order_index];

    fn_values.resize(num_y_bins);
    stddev_values.resize(num_y_bins);

    for (unsigned int bin_index = 0; bin_index < num_y_bins; bin_index++)
      {
      fn_values[bin_index] = function_bins_y[bin_index].get_mean();
      stddev_values[bin_index] = function_bins_y[bin_index].get_pop_stddev();
      }
    }

  const QuickFlash::Stats::BinArray & prototype_z_bins = function_bins_z_list[0];

  z_bin_positions = prototype_z_bins.get_bin_centers();

  const unsigned int num_z_bins = z_bin_positions.size();

  z_bin_cell_count.resize(num_z_bins);

  for (unsigned int bin_index = 0; bin_index < num_z_bins; bin_index++)
    z_bin_cell_count[bin_index] 
      = prototype_z_bins[bin_index].get_num_entries();

  fn_values_z_list.resize(num_orders);
  fn_stddev_z_list.resize(num_orders);

  for (unsigned int order_index = 0; order_index < num_orders; order_index++)
    {
    const QuickFlash::Stats::BinArray & function_bins_z
      = function_bins_z_list[order_index];

    std::vector<double> & fn_values = fn_values_z_list[order_index];
    std::vector<double> & stddev_values = fn_stddev_z_list[order_index];

    fn_values.resize(num_z_bins);
    stddev_values.resize(num_z_bins);

    for (unsigned int bin_index = 0; bin_index < num_z_bins; bin_index++)
      {
      fn_values[bin_index] = function_bins_z[bin_index].get_mean();
      stddev_values[bin_index] = function_bins_z[bin_index].get_pop_stddev();
      }
    }
  }


void get_scalar_struct_fn(const QuickFlash::Block::BlockInfo & grid_info,
			  const GridVector & vel_data, const double bin_width, 
			  const double max_radius,
			  const std::vector<unsigned int> & fn_order_list,
			  const unsigned int start_proc_cell_index,
			  const unsigned int num_proc_cells,
			  std::vector<QuickFlash::Stats::BinArray>
			    & function_bins_r_list,
			  std::vector<QuickFlash::Stats::BinArray>
			    & function_bins_y_list,
			  std::vector<QuickFlash::Stats::BinArray>
			    & function_bins_z_list)
  {
  if (num_proc_cells > 0)
    {
    const unsigned int num_orders = fn_order_list.size();

    if (num_orders < 1)
      throw QuickFlash::Except("At least one function order required", __FILE__,
			     __LINE__);

    // Check the input and collect parameters

    const unsigned int dims = grid_info.get_space_dims();

    if (dims < 1)
      throw QuickFlash::Except("At least one dimension required", __FILE__,
			     __LINE__);

    if (vel_data.size() != dims)
      throw QuickFlash::Except("Need one velocity component per grid dimension",
			   __FILE__, __LINE__);

    const unsigned int num_cells = grid_info.get_num_cells();

    if (num_cells < 1)
      throw QuickFlash::Except("No cells present", __FILE__, __LINE__);

    const unsigned int end_proc_cell_index 
      = (start_proc_cell_index + num_proc_cells) - 1;

    if (end_proc_cell_index >= num_cells)
      throw QuickFlash::Except("Requested grid indexes out of range", __FILE__,
			     __LINE__);

    const std::vector<unsigned int> & grid_dims = grid_info.get_block_dims();

    for (unsigned int set_index = 0; set_index < dims; set_index++)
      {
      if (vel_data[set_index] == 0)
	throw QuickFlash::Except("Non-existent velocity data", __FILE__, 
			       __LINE__);

      if (vel_data[set_index]->get_block_dims() != grid_dims)
	throw QuickFlash::Except("Incompatible grid dimensions", __FILE__, 
			       __LINE__);
      }

    if (!(bin_width > 0.0))
      throw QuickFlash::Except("Bin width must be positive", __FILE__, __LINE__);

    if (!(max_radius > 0.0))
      throw QuickFlash::Except("Maximum radius must be positive", __FILE__, 
			     __LINE__);

    const std::vector<double> & grid_width = grid_info.get_block_width();

    // Determine the number of bins -- use same radius, width for all

    const double fl_min_num_bins = ((max_radius / bin_width) + 0.25);

    const unsigned int num_bins 
      = static_cast<unsigned int>(fl_min_num_bins) + 1;
  
    const double function_max_radius 
      = static_cast<double>(num_bins) * bin_width;

    function_bins_r_list.resize(num_orders);

    function_bins_y_list.resize(num_orders);

    if (dims > 1)
      function_bins_z_list.resize(num_orders);
    else
      function_bins_z_list.clear();

    for (unsigned int order_index = 0; order_index < num_orders; order_index++)
      {
      function_bins_r_list[order_index].reset(0.0, function_max_radius, 
					      num_bins);

      function_bins_y_list[order_index].reset(0.0, function_max_radius, 
					      num_bins);

      if (dims > 1)
	function_bins_z_list[order_index].reset(0.0, function_max_radius, 
						num_bins);
      }

    const double max_sq_radius = function_max_radius * function_max_radius;

    // Set up the coordinate counters

    std::vector<double> position_coords(dims);
    std::vector<double> source_coords(dims);

    std::vector<double> position_vel(dims);
    std::vector<double> source_vel(dims);

    std::vector<double> vel_diff(dims);

    // For now, assume one position alias along each cardinal direction

    const std::vector<unsigned int> alias_counts(dims, 3);

    QuickFlash::VectorCounter alias_counter(alias_counts);

    const std::vector<int> zero_int_vect(dims, 0);

    // Declare space for alias

    std::vector<int> alias_index(dims);
    std::vector<double> alias_pos(dims);

    std::vector<double> alias_displacement(dims);
    std::vector<double> alias_disp_unit_vector(dims);

    
    // Run through the requested positions

    std::vector<unsigned int> position_index_vector(dims);
    QuickFlash::VectorCounter source_counter(grid_dims);

    for (unsigned int position_index = start_proc_cell_index; 
	 position_index <= end_proc_cell_index; position_index++)
      {
      grid_info.get_cell_center(position_index, position_coords);

      grid_info.get_vector_index(position_index, position_index_vector);

      for (unsigned int i = 0; i < dims; i++)
	position_vel[i] = (*vel_data[i])[position_index];

      // Loop through the source cells

      source_counter.reset_counter();

      while (source_counter.in_bounds())
	{
	const unsigned int source_index 
	  = grid_info.get_scalar_index(source_counter.get_state());

	bool is_myself = false;

	if (source_index == position_index)
	  is_myself = true;

	grid_info.get_cell_center(source_index, source_coords);

	for (unsigned int i = 0; i < dims; i++)
	  source_vel[i] = (*vel_data[i])[source_index];

	for (unsigned int i = 0; i < dims; i++)
	  vel_diff[i] = source_vel[i] - position_vel[i];

	// Run through the aliases

	alias_counter.reset_counter();

	while (alias_counter.in_bounds())
	  {
	  const std::vector<unsigned int> & alias_state 
	    = alias_counter.get_state();

	  for (unsigned int i = 0; i < dims; i++)
	    {
	    // Center alias index around zero

	    alias_index[i] = static_cast<int>(alias_state[i]) - 1;
	    }

	  // Don't process myself (but my aliases are OK ...)

	  bool try_alias = true;

	  if (is_myself)
	    if (alias_index == zero_int_vect)
	      try_alias = false;

	  if (try_alias)
	    {
	    alias_pos = source_coords;

	    for (unsigned int i = 0; i < dims; i++)
	      {
	      if (alias_index[i] > 0)
		alias_pos[i] += grid_width[i];
	      else if (alias_index[i] < 0)
		alias_pos[i] -= grid_width[i];
	      }

	    for (unsigned int i = 0; i < dims; i++)
	      alias_displacement[i] = alias_pos[i] - position_coords[i];


	    // Process the axis-parallel velocities

	    // NEED TO GENERALIZE FOR ARBITRARY DIMENSIONS!!!!

	    bool process_y_axis_velocity = true;

	    for (unsigned int i = 0; i < dims; i++)
	      if ((i != 0) && (position_index_vector[i] != source_counter[i]))
		process_y_axis_velocity = false;

	    if (process_y_axis_velocity)
	      {
	      for (unsigned int i = 0; i < dims; i++)
		{
		// Accept only aliases along y-axis (alias_index[!y] == 0)

		if ((i != 0) && (alias_index[i] != 0))
		  process_y_axis_velocity = false;
		}
	      }

	    if (process_y_axis_velocity)
	      {
	      const double abs_y_disp = std::fabs(alias_displacement[0]);

	      if (abs_y_disp < function_max_radius)
		{
		const double abs_vel_y_diff = std::fabs(vel_diff[0]);

		for (unsigned int order_index = 0; order_index < num_orders; 
		     order_index++)
		  {
		  const unsigned int fn_order = fn_order_list[order_index];

		  double fn_element = 1.0;

		  for (unsigned int mult_index = 0; mult_index < fn_order; 
		       mult_index++)
		    fn_element *= abs_vel_y_diff;

		  // For right now, just add point data
	      
		  function_bins_y_list[order_index].add_data(abs_y_disp, 
							     fn_element);
		  }
		}
	      }

	    bool process_z_axis_velocity = true;

	    if (dims < 2)
	      process_z_axis_velocity = false;
	    else 
	      {
	      for (unsigned int i = 0; i < dims; i++)
		if ((i != 1) 
		    && (position_index_vector[i] != source_counter[i]))
		  process_z_axis_velocity = false;

	      if (process_z_axis_velocity)
		{
		for (unsigned int i = 0; i < dims; i++)
		  {
		  // Accept only aliases along z-axis (alias_index[!z] == 0)

		  if ((i != 1) && (alias_index[i] != 0))
		    process_z_axis_velocity = false;
		  }
		}
	      }

	    if (process_z_axis_velocity)
	      {
	      const double abs_z_disp = std::fabs(alias_displacement[1]);

	      if (abs_z_disp < function_max_radius)
		{
		const double abs_vel_z_diff = std::fabs(vel_diff[1]);

		for (unsigned int order_index = 0; order_index < num_orders; 
		     order_index++)
		  {
		  const unsigned int fn_order = fn_order_list[order_index];

		  double fn_element = 1.0;

		  for (unsigned int mult_index = 0; mult_index < fn_order; 
		       mult_index++)
		    fn_element *= abs_vel_z_diff;
		    
		  // For right now, just add point data
	      
		  function_bins_z_list[order_index].add_data(abs_z_disp, 
							     fn_element);
		  }
		}
	      }

	    // Process the radial velocities
	    
	    double sq_dist = 0.0;

	    for (unsigned int i = 0; i < dims; i++)
	      {
	      const double axis_disp = alias_displacement[i];

	      sq_dist += axis_disp * axis_disp;
	      }

	    if (sq_dist < max_sq_radius)
	      {
	      const double dist = std::sqrt(sq_dist);

	      const double oneOver_dist = 1.0 / dist;

	      for (unsigned int i = 0; i < dims; i++)
		alias_disp_unit_vector[i] 
		  = alias_displacement[i] * oneOver_dist;

	      double vel_r_diff = 0.0;

	      for (unsigned int i = 0; i < dims; i++)
		vel_r_diff += vel_diff[i] * alias_disp_unit_vector[i];

	      const double abs_vel_r_diff = std::fabs(vel_r_diff);

	      for (unsigned int order_index = 0; order_index < num_orders; 
		   order_index++)
		{
		const unsigned int fn_order = fn_order_list[order_index];

		double fn_element = 1.0;

		for (unsigned int mult_index = 0; mult_index < fn_order; 
		   mult_index++)
		  fn_element *= abs_vel_r_diff;

		// For right now, just add point data
	      
		function_bins_r_list[order_index].add_data(dist, fn_element);
		}
	      }
	    }

	  alias_counter.increment();
	  }

	source_counter.increment();
	}
      }
    }
  }


#ifdef USE_MPI

const int mpi_tag_start_proc_cell_index = 100;
const int mpi_tag_fn_order = 200;
const int mpi_tag_num_proc_cells = 300;
const int mpi_tag_bin_width = 400;
const int mpi_tag_max_radius = 500;
const int mpi_tag_grid_dims = 600;
const int mpi_tag_min_bounds = 700;
const int mpi_tag_max_bounds = 800;
const int mpi_tag_vel_data = 900;
const int mpi_tag_value_weights = 1000;
const int mpi_tag_weights = 1100;


void collect_stats_master(const QuickFlash::Stats::BinArray & bin_array,
			  std::vector<unsigned int> & bin_data_count,
			  std::vector<double> & bin_means,
			  std::vector<double> & bin_stddev,
			  const MPI_Comm mpi_comm)
  {
  // Get MPI-related values

  const int root_rank = 0;

  int mpi_rank = 0;
  int mpi_size = 0;

  MPI_Comm_rank(mpi_comm, &mpi_rank);
  MPI_Comm_size(mpi_comm, &mpi_size);

  if (mpi_rank != root_rank)
    throw QuickFlash::Except("Server function can only be called by local root",
			   __FILE__, __LINE__);

  // Set up data structures

  const unsigned int num_bins = bin_array.get_num_bins();

  std::vector<unsigned int> accum_bin_counts(num_bins);
  std::vector<double> accum_value_weights(num_bins);
  std::vector<double> accum_sqValue_weights(num_bins);
  std::vector<double> accum_weights(num_bins);

  // Get local data

  for (unsigned int bin_index = 0; bin_index < num_bins; bin_index++)
    {
    const QuickFlash::Stats::Bin & bin = bin_array[bin_index];

    accum_bin_counts[bin_index] = bin.get_num_entries();

    accum_value_weights[bin_index] = bin.get_weight_value_sum();
    accum_sqValue_weights[bin_index] = bin.get_weight_sqValue_sum();

    accum_weights[bin_index] = bin.get_weight_sum();
    }

  // Collect data from other processes

  if (mpi_size > 1)
    {
    std::vector<unsigned int> global_bin_counts(num_bins);
    std::vector<double> global_value_weights(num_bins);
    std::vector<double> global_sqValue_weights(num_bins);
    std::vector<double> global_weights(num_bins);
   
    sum_vect_master(accum_bin_counts, global_bin_counts, mpi_comm);

    sum_vect_master(accum_value_weights, global_value_weights, mpi_comm);
    sum_vect_master(accum_sqValue_weights, global_sqValue_weights, mpi_comm);

    sum_vect_master(accum_weights, global_weights, mpi_comm);

    // Move global data to accum structures using swap

    accum_bin_counts.swap(global_bin_counts);
    accum_value_weights.swap(global_value_weights);
    accum_sqValue_weights.swap(global_sqValue_weights);
    accum_weights.swap(global_weights);
    }

  // Process the data

  bin_data_count = accum_bin_counts;

  bin_means.resize(num_bins);
  bin_stddev.resize(num_bins);

  for (unsigned int bin_index = 0; bin_index < num_bins; bin_index++)
    {
    const double weight = accum_weights[bin_index];

    double mean_val = 0.0;
    double stddev_val = 0.0;

    if (weight > 0.0)
      {
      const double oneOver_weight = 1.0 / weight;

      const double value_weight = accum_value_weights[bin_index];
      const double sqValue_weight = accum_value_weights[bin_index];

      mean_val = value_weight * oneOver_weight;

      // Use sigma^2 = [ SUM(w_i * x_i^2) / SUM(w_i) ] - (N * MEAN(x_i)^2)

      const double sq_stddev 
	= (sqValue_weight * oneOver_weight) - (mean_val * mean_val);

      if (sq_stddev > 0.0)
	stddev_val = std::sqrt(sq_stddev);
      }

    bin_means[bin_index] = mean_val;
    bin_stddev[bin_index] = stddev_val;
    }
  }


void collect_stats_worker(const QuickFlash::Stats::BinArray & bin_array,
			  const MPI_Comm mpi_comm)
  {
  const int root_rank = 0;

  int mpi_rank = 0;
  int mpi_size = 0;

  MPI_Comm_rank(mpi_comm, &mpi_rank);
  MPI_Comm_size(mpi_comm, &mpi_size);

  if (mpi_rank == root_rank)
    throw QuickFlash::Except("Worker function must not be called by local root",
			   __FILE__, __LINE__);

  // Get data

  const unsigned int num_bins = bin_array.get_num_bins();

  std::vector<unsigned int> local_bin_counts(num_bins);
  std::vector<double> local_value_weights(num_bins);
  std::vector<double> local_sqValue_weights(num_bins);
  std::vector<double> local_weights(num_bins);

  // Get global data

  for (unsigned int bin_index = 0; bin_index < num_bins; bin_index++)
    {
    const QuickFlash::Stats::Bin & bin = bin_array[bin_index];

    local_bin_counts[bin_index] = bin.get_num_entries();

    local_value_weights[bin_index] = bin.get_weight_value_sum();
    local_sqValue_weights[bin_index] = bin.get_weight_sqValue_sum();

    local_weights[bin_index] = bin.get_weight_sum();
    }

  // Send the data to root

  sum_vect_worker(root_rank, local_bin_counts, mpi_comm);

  sum_vect_worker(root_rank, local_value_weights, mpi_comm);
  sum_vect_worker(root_rank, local_sqValue_weights, mpi_comm);

  sum_vect_worker(root_rank, local_weights, mpi_comm);
  }


void collect_stats_worker_empty(const MPI_Comm mpi_comm)
  {
  const int root_rank = 0;

  int mpi_rank = 0;
  int mpi_size = 0;

  MPI_Comm_rank(mpi_comm, &mpi_rank);
  MPI_Comm_size(mpi_comm, &mpi_size);

  if (mpi_rank == root_rank)
    throw QuickFlash::Except("Worker function must not be called by local root",
			   __FILE__, __LINE__);

  // I have no data, so just send a null vector

  const std::vector<double> null_vect;

  sum_vect_worker(root_rank, null_vect, mpi_comm);

  sum_vect_worker(root_rank, null_vect, mpi_comm);
  sum_vect_worker(root_rank, null_vect, mpi_comm);

  sum_vect_worker(root_rank, null_vect, mpi_comm);
  }


void get_scalar_struct_fn_master(const QuickFlash::Block::BlockInfo & grid_info,
				 const GridVector & vel_data, 
				 const double bin_width, 
				 const double max_radius, 
				 const std::vector<unsigned int> 
				   & fn_order_list,
				 std::vector<double> & r_bin_radii,
				 std::vector<unsigned int> & r_bins_cell_count,
				 std::vector< std::vector<double> > 
				   & fn_values_r_list,
				 std::vector< std::vector<double> > 
				   & fn_stddev_r_list,
				 std::vector<double> & y_bin_positions,
				 std::vector<unsigned int> & y_bins_cell_count,
				 std::vector< std::vector<double> > 
				   & fn_values_y_list,
				 std::vector< std::vector<double> >
				   & fn_stddev_y_list,
				 std::vector<double> & z_bin_positions,
				 std::vector<unsigned int> & z_bins_cell_count,
				 std::vector< std::vector<double> > 
				   & fn_values_z_list,
				 std::vector< std::vector<double> >
				   & fn_stddev_z_list,
				 const MPI_Comm mpi_comm)
  {
  const unsigned int dims = grid_info.get_space_dims();

  const std::vector<unsigned int> & grid_dims = grid_info.get_block_dims();

  for (unsigned int i = 0; i < dims; i++)
    {
    if (vel_data[i] == 0)
      throw QuickFlash::Except("Non-existant velocity data", __FILE__, __LINE__);

    if (vel_data[i]->get_block_dims() != grid_dims)
      throw QuickFlash::Except("Incompatible velocity data dimensions", __FILE__,
			     __LINE__);
    }

  const unsigned int num_orders = fn_order_list.size();

  if (num_orders < 1)
    throw QuickFlash::Except("At least one order must be specified", __FILE__,
			   __LINE__);

  // Get MPI-related values

  const int root_rank = 0;

  int mpi_rank = 0;
  int mpi_size = 0;

  MPI_Comm_rank(mpi_comm, &mpi_rank);
  MPI_Comm_size(mpi_comm, &mpi_size);

  if (mpi_rank != root_rank)
    throw QuickFlash::Except("Server function can only be called by local root",
			   __FILE__, __LINE__);

  const unsigned int mpi_size_unsigned = static_cast<unsigned int>(mpi_size);

  // Get some basic data about the space

  const std::vector<double> & min_bounds = grid_info.get_min_coords();
  const std::vector<double> & max_bounds = grid_info.get_max_coords();

  const unsigned int num_cells = grid_info.get_num_cells();

  // Determine the distribution of cells to be processed

  std::vector<unsigned int> start_proc_cell_indexes;
  std::vector<unsigned int> num_proc_cells;

  simple_linear_decomp(num_cells, mpi_size, start_proc_cell_indexes, 
		       num_proc_cells);

  const unsigned int my_start_cell_index = start_proc_cell_indexes[0];
  const unsigned int my_num_cells = num_proc_cells[0];

  // Send the data and layout to everyone else

  for (unsigned int proc_id = 1; proc_id < mpi_size_unsigned; proc_id++)
    {
    const unsigned int start_cell_index_proc
      = start_proc_cell_indexes[proc_id];

    const unsigned int num_cells_proc
      = num_proc_cells[proc_id];

    int start_cell_index_proc_mpi = static_cast<int>(start_cell_index_proc);
    int num_cells_proc_mpi = static_cast<int>(num_cells_proc);

    MPI_Send(&start_cell_index_proc_mpi, 1, MPI_UNSIGNED, proc_id, 
	     mpi_tag_start_proc_cell_index, mpi_comm);

    MPI_Send(&num_cells_proc_mpi, 1, MPI_UNSIGNED, proc_id, 
	     mpi_tag_num_proc_cells, mpi_comm);
    }

  // Send the global data

  double bin_width_mpi = bin_width;
  double max_radius_mpi = max_radius;

  MPI_Bcast(&bin_width_mpi, 1, MPI_DOUBLE, mpi_rank, mpi_comm);

  MPI_Bcast(&max_radius_mpi, 1, MPI_DOUBLE, mpi_rank, mpi_comm);

  bcast_vect_server(fn_order_list, mpi_comm);

  bcast_vect_server(grid_dims, mpi_comm);

  bcast_vect_server(min_bounds, mpi_comm);
  bcast_vect_server(max_bounds, mpi_comm);

  for (unsigned int i = 0; i < dims; i++)
    bcast_vect_server(vel_data[i]->get_data(), mpi_comm);

  // Do my work

  std::vector<QuickFlash::Stats::BinArray> function_bins_r_list;
  std::vector<QuickFlash::Stats::BinArray> function_bins_y_list;
  std::vector<QuickFlash::Stats::BinArray> function_bins_z_list;

  get_scalar_struct_fn(grid_info, vel_data, bin_width, max_radius, 
		       fn_order_list, my_start_cell_index, my_num_cells, 
		       function_bins_r_list, function_bins_y_list,
		       function_bins_z_list);

  // Collect the statistics

  r_bin_radii = function_bins_r_list[0].get_bin_centers();
  
  fn_values_r_list.resize(num_orders);
  fn_stddev_r_list.resize(num_orders);

  for (unsigned int order_index = 0; order_index < num_orders; order_index++)
    collect_stats_master(function_bins_r_list[order_index], r_bins_cell_count,
			 fn_values_r_list[order_index], 
			 fn_stddev_r_list[order_index], mpi_comm);

  y_bin_positions = function_bins_y_list[0].get_bin_centers();
  
  fn_values_y_list.resize(num_orders);
  fn_stddev_y_list.resize(num_orders);

  for (unsigned int order_index = 0; order_index < num_orders; order_index++)
    collect_stats_master(function_bins_y_list[order_index], y_bins_cell_count,
			 fn_values_y_list[order_index], 
			 fn_stddev_y_list[order_index], mpi_comm);

  z_bin_positions = function_bins_z_list[0].get_bin_centers();
  
  fn_values_z_list.resize(num_orders);
  fn_stddev_z_list.resize(num_orders);

  for (unsigned int order_index = 0; order_index < num_orders; order_index++)
    collect_stats_master(function_bins_z_list[order_index], z_bins_cell_count,
			 fn_values_z_list[order_index], 
			 fn_stddev_z_list[order_index], mpi_comm);
  }
  

void get_scalar_struct_fn_worker(const MPI_Comm mpi_comm)
  {
  MPI_Status status;

  const int root_rank = 0;

  int mpi_rank = 0;
  int mpi_size = 0;

  MPI_Comm_rank(mpi_comm, &mpi_rank);
  MPI_Comm_size(mpi_comm, &mpi_size);

  if (mpi_rank == root_rank)
    throw QuickFlash::Except("Worker function must not be called by local root",
			   __FILE__, __LINE__);

  unsigned int start_proc_cell_index = 0;
  unsigned int num_proc_cells = 0;

  MPI_Recv(&start_proc_cell_index, 1, MPI_UNSIGNED, root_rank, 
	   mpi_tag_start_proc_cell_index, mpi_comm, &status);
  
  MPI_Recv(&num_proc_cells, 1, MPI_UNSIGNED, root_rank, 
	   mpi_tag_num_proc_cells, mpi_comm, &status);


  // Get information about the binning

  double bin_width = 0.0;
  double max_radius = 0.0;

  MPI_Bcast(&bin_width, 1, MPI_DOUBLE, root_rank, mpi_comm);

  MPI_Bcast(&max_radius, 1, MPI_DOUBLE, root_rank, mpi_comm);


  // Get the function orders

  std::vector<unsigned int> fn_order_list;

  bcast_vect_client(root_rank, fn_order_list, mpi_comm);


  // Get the grid information

  std::vector<unsigned int> grid_dims;

  bcast_vect_client(root_rank, grid_dims, mpi_comm);

  std::vector<double> min_bounds;
  std::vector<double> max_bounds;

  bcast_vect_client(root_rank, min_bounds, mpi_comm);
  bcast_vect_client(root_rank, max_bounds, mpi_comm);

  QuickFlash::Block::BlockInfo gridinfo(min_bounds, max_bounds, grid_dims,
					QuickFlash::Geometry::Cartesian);

  const unsigned int dims = gridinfo.get_space_dims();

  const unsigned int num_cells = gridinfo.get_num_cells();

  // Set up the velocity data

  std::vector< QuickFlash::Block::BlockData<double> > vel_data(dims);

  std::vector<double> axis_vel_data_array;

  for (unsigned int i = 0; i < dims; i++)
    {
    QuickFlash::Block::BlockData<double> & axis_vel_data = vel_data[i];

    axis_vel_data.reset(grid_dims);

    bcast_vect_client(root_rank, axis_vel_data_array, mpi_comm);

    if (axis_vel_data_array.size() != num_cells)
      throw QuickFlash::Except("Incompatible velocity data array", __FILE__,
			     __LINE__);

    axis_vel_data.set_data(axis_vel_data_array);
    }


  // Get the local statistics

  if (num_cells > 0)
    {
    GridVector vel_data_ptrs(dims);

    for (unsigned int i = 0; i < dims; i++)
      vel_data_ptrs[i] = &vel_data[i];

    std::vector<QuickFlash::Stats::BinArray> function_bins_r_list;
    std::vector<QuickFlash::Stats::BinArray> function_bins_y_list;
    std::vector<QuickFlash::Stats::BinArray> function_bins_z_list;
  
    get_scalar_struct_fn(gridinfo, vel_data_ptrs, bin_width, max_radius,
			 fn_order_list, start_proc_cell_index, num_proc_cells, 
			 function_bins_r_list, function_bins_y_list,
			 function_bins_z_list);

    // Get the weights and return to root; assume root has correct bin centers

    const unsigned int num_orders = fn_order_list.size();

    for (unsigned int order_index = 0; order_index < num_orders; order_index++)
      collect_stats_worker(function_bins_r_list[order_index], mpi_comm);

    for (unsigned int order_index = 0; order_index < num_orders; order_index++)
      collect_stats_worker(function_bins_y_list[order_index], mpi_comm);

    for (unsigned int order_index = 0; order_index < num_orders; order_index++)
      collect_stats_worker(function_bins_z_list[order_index], mpi_comm);
    }
  else
    {
    // I have no data, so just send zeros

    const unsigned int num_orders = fn_order_list.size();

    // r-list

    for (unsigned int order_index = 0; order_index < num_orders; order_index++)
      collect_stats_worker_empty(mpi_comm);

    // y-list

    for (unsigned int order_index = 0; order_index < num_orders; order_index++)
      collect_stats_worker_empty(mpi_comm);

    // z-list

    for (unsigned int order_index = 0; order_index < num_orders; order_index++)
      collect_stats_worker_empty(mpi_comm);
    }
  }


#endif  // USE_MPI
