// C++ program file fft.cpp

/*
  By Nathan C. Hearn
     October 30, 2007

  Interface to GSL FFT routines.
*/


#include "fft.hpp"
#include <math.h>
#include <vector>
#include <complex>
#include <gsl/gsl_fft_complex.h>



int get_power(const std::vector< std::complex<double> > & fft_values,
	      std::vector<double> & spectral_power)
  {
  int ret_val = -1;  // Error state

  const unsigned int num_fft_values = fft_values.size();

  if (num_fft_values > 0)
    {
    const unsigned int num_power_values 
      = get_power_vector_size(num_fft_values);

    spectral_power.resize(num_power_values);

    const double const_component = fft_values[0].real();

    spectral_power[0] = const_component * const_component;

    unsigned int last_power_index = num_power_values - 1;

    if ((num_fft_values % 2) == 0)
      {
      const unsigned int fft_index = last_power_index;

      const double last_component = fft_values[fft_index].real();

      spectral_power[last_power_index] = last_component * last_component;

      last_power_index--;
      }

    for (unsigned int power_index = 1; power_index <= last_power_index; 
	 power_index++)
      {
      const unsigned int reflect_index = num_fft_values - power_index;

      // Use the real component of the negative-positive frequency product

      std::complex<double> cplx_square 
	= fft_values[power_index] * fft_values[reflect_index];

      spectral_power[power_index] = fabs(cplx_square.real());
      }

    ret_val = 0;  // OK
    }

  return ret_val;
  }


unsigned int get_power_vector_size(const unsigned int num_fft_values)
  { return (num_fft_values / 2) + 1; }


// Class FFT

FFT::FFT() : num_elems(0), workspace(0), wavetable(0) { }


FFT::~FFT()  { reset(); }


void FFT::reset() const
  {
  num_elems = 0;

  if (wavetable != 0)
    {
    gsl_fft_complex_wavetable_free(wavetable);
    wavetable = 0;
    }

  if (workspace != 0)
    {
    gsl_fft_complex_workspace_free(workspace);
    workspace = 0;
    }
  }


void FFT::reset(const unsigned int real_vector_size) const
  {
  if (real_vector_size != num_elems)
    {
    reset();

    num_elems = real_vector_size;

    if (num_elems > 0)
      {
      workspace = gsl_fft_complex_workspace_alloc(num_elems);
      wavetable = gsl_fft_complex_wavetable_alloc(num_elems);
      }
    }
  }


int FFT::compute_fft(const std::vector<double> & data_vector,
		     std::vector< std::complex<double> > & trans_vector) 
  const
  {
  int ret_val = -1;  // Error state

  const unsigned int real_vector_size = data_vector.size();

  if (real_vector_size > 0)
    {
    if (real_vector_size != num_elems)
      reset(real_vector_size);

    // Compose the packed real-imaginary vector for real data

    double * work_vector = new double[real_vector_size * 2];

    for (unsigned int i = 0; i < real_vector_size; i++)
      {
      const unsigned int work_vector_index = i * 2;

      work_vector[work_vector_index] = data_vector[i];
      work_vector[work_vector_index + 1] = 0.0;  // Real input data
      }

    gsl_fft_complex_forward(work_vector, 1, num_elems, wavetable, workspace);

    // Unpack the transformed vector into trans_vector

    trans_vector.resize(real_vector_size);

    for (unsigned int i = 0; i < real_vector_size; i++)
      {
      const unsigned int work_vector_index = i * 2;

      trans_vector[i].real() = work_vector[work_vector_index];
      trans_vector[i].imag() = work_vector[work_vector_index + 1];
      }

    delete[] work_vector;
    work_vector = 0;

    ret_val = 0;  // OK
    }

  return ret_val;
  }
