// C++ program file layer_detection.cpp

/*
  By Nathan C. Hearn
     October 31, 2007

  Layer detection utilities.
*/


#include "layer_detection.hpp"
#include <vector>
#include "quickflash_except.hpp"


int find_bubble_trough_poly_dens_power(const std::vector<double> & cell_pos,
				       const std::vector<double> & power_ratio,
				       const double target_power_ratio,
				       double & bubble_position)
  {
  int ret_val = -1;  // Error state

  // Starting from the lower positions, look for the first rise in the bubble
  // power above the target power ratio

  const unsigned int num_pos = cell_pos.size();

  if (num_pos < 2)
    throw QuickFlash::Except("At least two positions required");

  if (power_ratio.size() != num_pos)
    throw QuickFlash::Except("Incompatible power ratio vector");

  unsigned int search_index = 0;

  bool found_bubble = false;

  while ((!found_bubble) && (search_index < num_pos))
    {
    if (power_ratio[search_index] > target_power_ratio)
      found_bubble = true;
    else
      search_index++;
    }

  if (found_bubble)
    {
    bubble_position = cell_pos[search_index];
    ret_val = 0;  // OK
    }

  return ret_val;
  }


int find_spike_tip_poly_dens_wavenumber(const std::vector<double> & cell_pos,
					const std::vector<double> & avg_poly,
					const double start_poly,
			 const std::vector<unsigned int> & dominant_wavenumber,
					const unsigned int target_wavenumber,
					double & start_poly_pos,
					double & spike_pos)
  {
  int ret_val = -1;  // Error state

  // Start from the higher positions where avg_poly exceeds start_poly.  Look 
  // for the first peak in either avg_poly_dens or power_ratio.

  const unsigned int num_pos = cell_pos.size();

  if (num_pos < 2)
    throw QuickFlash::Except("At least two positions required");

  if (avg_poly.size() != num_pos)
    throw QuickFlash::Except("Incompatible average poly vector");

  if (dominant_wavenumber.size() != num_pos)
    throw QuickFlash::Except("Incompatible dominant wavenumber vector");

  // Find the starting index

  unsigned int search_index = num_pos - 1;

  bool found_poly_start = false;

  while ((!found_poly_start) && (search_index > 0))
    {
    if (avg_poly[search_index] > start_poly)
      found_poly_start = true;
    else
      search_index--;
    }

  if (found_poly_start)
    {
    const double init_poly_pos = cell_pos[search_index];

    start_poly_pos = init_poly_pos;

    bool found_spike = false;

    while ((!found_spike) && (search_index > 0))
      {
      if (dominant_wavenumber[search_index] == target_wavenumber)
	found_spike = true;
      else
	search_index--;
      }


    if (found_spike)
      {
      spike_pos = cell_pos[search_index];
      ret_val = 0;  // OK
      }
    else
      {
      spike_pos = init_poly_pos;
      ret_val = 1;  // No peak found
      }
    }

  return ret_val;
  }


int find_spike_tip_poly_dens(const std::vector<double> & cell_pos,
			     const std::vector<double> & avg_poly,
			     const double start_poly,
			     const std::vector<double> & avg_poly_dens,
			     const std::vector<double> & power_ratio,
			     double & start_poly_pos, double & spike_pos)
  {
  int ret_val = -1;  // Error state

  // Start from the higher positions where avg_poly exceeds start_poly.  Look 
  // for the first peak in either avg_poly_dens or power_ratio.

  const unsigned int num_pos = cell_pos.size();

  if (num_pos < 2)
    throw QuickFlash::Except("At least two positions required");

  if (avg_poly.size() != num_pos)
    throw QuickFlash::Except("Incompatible average poly vector");

  if (avg_poly_dens.size() != num_pos)
    throw QuickFlash::Except("Incompatible average poly dens vector");

  if (power_ratio.size() != num_pos)
    throw QuickFlash::Except("Incompatible power ratio vector");

  // Find the starting index

  unsigned int search_index = num_pos - 1;

  bool found_poly_start = false;

  while ((!found_poly_start) && (search_index > 0))
    {
    if (avg_poly[search_index] > start_poly)
      found_poly_start = true;
    else
      search_index--;
    }

  if (found_poly_start)
    {
    const double init_poly_pos = cell_pos[search_index];

    start_poly_pos = init_poly_pos;

    double last_poly_dens = avg_poly_dens[search_index];
    double last_power_ratio = power_ratio[search_index];

    bool found_rising_poly_dens = false;
    bool found_rising_power_ratio = false;

    bool found_spike = false;

    search_index--;

    while ((!found_spike) && (search_index > 0))
      {
      const double current_poly_dens = avg_poly_dens[search_index];
      const double current_power_ratio = power_ratio[search_index];

      if (!found_rising_poly_dens)
	{
	if (current_poly_dens > last_poly_dens)
	  found_rising_poly_dens = true;
	}
      else if (current_poly_dens < last_poly_dens)
	found_spike = true;

      if (!found_rising_power_ratio)
	{
	if (current_power_ratio > last_power_ratio)
	  found_rising_power_ratio = true;
	}
      else if (current_power_ratio < last_power_ratio)
	found_spike = true;

      last_poly_dens = current_poly_dens;
      last_power_ratio = current_power_ratio;

      if (!found_spike)
	search_index--;
      }

    if (found_spike)
      {
      spike_pos = cell_pos[search_index];
      ret_val = 0;  // OK
      }
    else
      {
      spike_pos = init_poly_pos;
      ret_val = 1;  // No peak found
      }
    }

  return ret_val;
  }
