// C++ program file layer_distrib.cpp

/*
  By Nathan C. Hearn
     October 22, 2007

  Output of layer distribution information.
*/


#include <stdlib.h>
#include <iostream>
#include <string>
#include <vector>
#include <fstream>
#include "quickflash_file_datafile.hpp"
#include "quickflash_except.hpp"
#include "quickflash_block_blockdata.hpp"
#include "mixing_layer.hpp"
#include "layer_detection.hpp"



template<typename T>
void print_grid_data(const std::string & filename,
		     const QuickFlash::Block::BlockData<T> & grid_data)
  {
  using std::endl;

  const std::vector<unsigned int> & grid_dims = grid_data.get_block_dims();

  if (grid_dims.size() != 2)
    throw QuickFlash::Except("Only 2D grids supported at this time", __FILE__,
			   __LINE__);

  const unsigned int num_rows = grid_dims[0];
  const unsigned int num_cols = grid_dims[1];

  std::ofstream f(filename.c_str());

  std::vector<unsigned int> cell_index(2);

  for (unsigned int m = 0; m < num_rows; m++)
    {
    cell_index[0] = m;

    for (unsigned int n = 0; n < num_cols; n++)
      {
      cell_index[1] = n;

      f << grid_data[cell_index] << " ";
      }

    f << endl;
    }

  f << endl;
  }



int main(int argc, char * argv[])
  {
  using std::cout;
  using std::cerr;
  using std::endl;

  const int min_args = 6;

  if (argc < min_args)
    {
    cerr << endl << "Usage: " << argv[0] 
	 << " filename_prefix min_index max_index output_prefix dims "
	 << endl
	 << "<min_coord> <max_coord> <sample_spacing>" << endl << endl;
    return 1;
    }

  int req_args = min_args;

  int argptr = 1;

  const std::string filename_prefix = argv[argptr++];

  const unsigned int min_file_index 
    = static_cast<unsigned int>(atoi(argv[argptr++]));

  const unsigned int max_file_index 
    = static_cast<unsigned int>(atoi(argv[argptr++]));

  if (max_file_index < min_file_index)
    {
    cerr << "Incompatible file index bounds" << endl;
    return -1;
    }

  const std::string output_prefix = argv[argptr++];

  const unsigned int dims = static_cast<unsigned int>(atoi(argv[argptr++]));

  req_args += 3 * dims;

  if (argc < req_args)
    {
    cerr << endl << "Error: Insufficient vector elements" << endl << endl;
    return -1;
    }

  std::vector<double> min_coord(dims);

  for (unsigned int i = 0; i < dims; i++)
    min_coord[i] = atof(argv[argptr++]);

  std::vector<double> max_coord(dims);

  for (unsigned int i = 0; i < dims; i++)
    max_coord[i] = atof(argv[argptr++]);

  bool coords_ok = true;

  for (unsigned int i = 0; i < dims; i++)
    if (!(min_coord[i] < max_coord[i]))
      coords_ok = false;

  if (!coords_ok)
    {
    cerr << endl << "Error: Min coords must be less than max coords" << endl
	 << endl;
    return -1;
    }

  std::vector<double> sample_spacing(dims);

  bool spacing_ok = true;

  for (unsigned int i = 0; i < dims; i++)
    {
    const double axis_spacing = atof(argv[argptr++]);

    if (!(axis_spacing > 0.0))
      spacing_ok = false;

    sample_spacing[i] = axis_spacing;
    }

  if (!spacing_ok)
    {
    cerr << endl << "Error: Sample spacing must be positive" << endl << endl;
    return -1; 
    }

  cout << "file_index" << "\t" << "sim_time" << "\t" << "bubble_pos" << "\t"
       << "start_poly_pos" << "\t" << "spike_pos_wavenumber" << "\t"
       << "spike_pos_poly" << "\t" << "selected_spike_pos" << endl;

  for (unsigned int file_index = min_file_index; file_index <= max_file_index;
       file_index++)
    {
    // Open the file

    std::stringstream indexstream;

    indexstream.width(4);
    indexstream.fill('0');

    indexstream << file_index;

    const std::string index_string = indexstream.str();

    const std::string datafile_name = filename_prefix + index_string;

    QuickFlash::File::DataFile datafile(datafile_name, 1000, true);

    const QuickFlash::File::SimInfo & siminfo = datafile.get_sim_info();

    const double sim_time = siminfo.get_sim_time();

    // Get the mixing layer

    MixingLayer mix_layer(datafile, min_coord, max_coord, sample_spacing);

    const std::vector<double> & bin_centers 
      = mix_layer.get_bin_array_centers();

    std::vector<double> bin_masses;
    std::vector<double> poly_masses;

    mix_layer.get_bin_masses(bin_masses);
    mix_layer.get_bin_poly_masses(poly_masses);

    std::vector<double> mean_dens;
    std::vector<double> mean_poly_dens;
    std::vector<double> mean_poly;

    mix_layer.get_vol_avg_dens(mean_dens);
    mix_layer.get_vol_avg_poly_dens(mean_poly_dens);
    mix_layer.get_vol_avg_poly(mean_poly);

    std::vector<unsigned int> bin_entries;

    mix_layer.get_bin_entries(bin_entries);

    std::vector<unsigned int> secondary_frequencies_dens;
    std::vector<double> secondary_power_dens;

    mix_layer.get_secondary_power_ratio_dens(secondary_power_dens, 
					     secondary_frequencies_dens);

    std::vector<unsigned int> secondary_frequencies_poly_dens;
    std::vector<double> secondary_power_poly_dens;

    mix_layer.get_secondary_power_ratio_poly_dens(secondary_power_poly_dens, 
					     secondary_frequencies_poly_dens);

    const unsigned int num_bins = bin_centers.size();

    const std::string output_basename = output_prefix + "_" + index_string;

    const std::string array_filename = output_basename + "_array.out";

    std::ofstream outfile(array_filename.c_str());

    // ERROR CHECKING???

    outfile << "#Bin" << "\t" << "center_x" << "\t" << "bin_entries" << "\t" 
	    << "mass" << "\t" << "poly_mass"
	    << "\t" << "mean_dens" << "\t" << "mean_poly_dens" << "\t" 
	    << "mean_poly" << "\t" << "secondary_power_dens" << "\t"
	    << "secondary_freq_dens" << "\t" << "secondary_power_poly_dens"
	    << "\t" << "secondary_freq_poly_dens" << endl;

    for (unsigned int index = 0; index < num_bins; index++)
      outfile << index << "\t" << bin_centers[index] << "\t" 
	      << bin_entries[index] << "\t" << bin_masses[index] 
	      << "\t" << poly_masses[index] << "\t" << mean_dens[index] 
	      << "\t" 
	      << mean_poly_dens[index] << "\t" << mean_poly[index] << "\t"
	      << secondary_power_dens[index] << "\t" 
	      << secondary_frequencies_dens[index] << "\t"
	      << secondary_power_poly_dens[index] << "\t" 
	      << secondary_frequencies_poly_dens[index] << endl;

    outfile << endl;

    QuickFlash::Block::BlockData<unsigned int> grid_bin_entries;
    QuickFlash::Block::BlockData<double> grid_dens;
    QuickFlash::Block::BlockData<double> grid_poly_dens;

    mix_layer.get_bin_entries(grid_bin_entries);
    mix_layer.get_vol_avg_dens(grid_dens);
    mix_layer.get_vol_avg_poly_dens(grid_poly_dens);

    const std::string grid_bin_entries_filename 
      = output_basename + "_grid_entries.out";

    const std::string grid_dens_filename = output_basename + "_grid_dens.out";

    const std::string grid_poly_dens_filename 
      = output_basename + "_grid_poly_dens.out";

    print_grid_data(grid_bin_entries_filename, grid_bin_entries);
    print_grid_data(grid_dens_filename, grid_dens);
    print_grid_data(grid_poly_dens_filename, grid_poly_dens);

    QuickFlash::Block::BlockData<double> grid_fft_dens;
    QuickFlash::Block::BlockData<double> grid_fft_poly_dens;

    mix_layer.get_lateral_spectral_power_dens(grid_fft_dens);
    mix_layer.get_lateral_spectral_power_poly_dens(grid_fft_poly_dens);

    // const std::string grid_fft_dens_filename 
    //   = output_basename + "_grid_dens_fft.out";

    // const std::string grid_fft_poly_dens_filename 
    //   = output_basename + "_grid_poly_dens_fft.out";

    // print_grid_data(grid_fft_dens_filename, grid_fft_dens);
    // print_grid_data(grid_fft_poly_dens_filename, grid_fft_poly_dens);

    // Perform the detection

    double bubble_pos = min_coord[0];

    const double target_bubble_power_ratio = 0.01;

    if (find_bubble_trough_poly_dens_power(bin_centers, 
					   secondary_power_poly_dens,
					   target_bubble_power_ratio,
					   bubble_pos) < 0)
      cerr << endl << "ERROR: Unable to find bubble position" << endl << endl;

    const double start_poly = 0.15;
    const unsigned int target_wavenumber = 13;

    double spike_pos_wavenumber = 0.0;
    double start_poly_pos = 0.0;

    if (find_spike_tip_poly_dens_wavenumber(bin_centers, mean_poly, start_poly,
					    secondary_frequencies_poly_dens,
					    target_wavenumber, start_poly_pos,
					    spike_pos_wavenumber) < 0)
      cerr << endl << "ERROR: Unable to find spike position (wavenumber) - "
	   << "Index " << file_index << endl << endl;

    double spike_pos_poly = 0.0;
    
    if (find_spike_tip_poly_dens(bin_centers, mean_poly, start_poly,
				 mean_poly_dens, secondary_power_poly_dens,
				 start_poly_pos, spike_pos_poly) < 0)
      cerr << endl << "ERROR: Unable to find spike position (poly) - "
	   << "Index " << file_index << endl << endl;
    
    double selected_spike_pos = spike_pos_poly;

    if (!(selected_spike_pos > bubble_pos))
      selected_spike_pos = spike_pos_wavenumber;

    if (!(selected_spike_pos > bubble_pos))
      selected_spike_pos = start_poly_pos;

    cout << file_index << "\t" << sim_time << "\t" << bubble_pos << "\t"
	 << start_poly_pos << "\t" << spike_pos_wavenumber << "\t"
	 << spike_pos_poly << "\t" << selected_spike_pos << endl;
    }

  return 0;
  }
