// C++ program file mixing_layer.cpp

/*
  By Nathan C. Hearn
     May 16, 2007

  Classes for storing mixing layer information for each timestep.
*/


#include "mixing_layer.hpp"
#include <vector>
#include <iostream>
#include "quickflash_except.hpp"
#include "quickflash_file_datafile.hpp"
#include "quickflash_file_meshinfo.hpp"
#include "quickflash_file_dataset.hpp"
#include "stat_bins.hpp"
#include "grid_fft.hpp"


const char MixingLayer::dens_name[] = "dens";
const char MixingLayer::foam_name[] = "foam";

const unsigned int MixingLayer::default_buffer_size = 1000;


MixingLayer::MixingLayer() :
  dims(0), sim_time(0.0), min_coord(), max_coord(), bin_spacing(), 
  dens_array(), poly_dens_array(), poly_array(), dens_grid(), poly_dens_grid()
  { }


MixingLayer::MixingLayer(const QuickFlash::File::DataFile & datafile, 
			 const std::vector<double> & min_sample_coord,
			 const std::vector<double> & max_sample_coord,
			 const std::vector<double> & sample_spacing,
			 const unsigned int buffer_size) :
  dims(0), sim_time(0.0), min_coord(), max_coord(), bin_spacing(), 
  dens_array(), poly_dens_array(), poly_array(), dens_grid(), poly_dens_grid()
  { 
  reset(datafile, min_sample_coord, max_sample_coord, sample_spacing,
	buffer_size); 
  }


MixingLayer::MixingLayer(const MixingLayer & source) :
  dims(0), sim_time(0.0), min_coord(), max_coord(), bin_spacing(), 
  dens_array(), poly_dens_array(), poly_array(), dens_grid(), poly_dens_grid()
  { reset(source); }


void MixingLayer::reset()
  {
  dims = 0;

  sim_time = 0.0;

  min_coord.clear();
  max_coord.clear();

  bin_spacing.clear();

  // int_pos = 0.0;

  dens_array.reset();
  poly_dens_array.reset();
  poly_array.reset();

  dens_grid.reset();
  poly_dens_grid.reset();
  }


void MixingLayer::reset(const QuickFlash::File::DataFile & datafile, 
			const std::vector<double> & min_sample_coord, 
			const std::vector<double> & max_sample_coord, 
			const std::vector<double> & sample_spacing,
			const unsigned int buffer_size)
  {
  reset();

  dims = min_sample_coord.size();

  if (dims < 1)
    throw QuickFlash::Except("At least one spatial dimension required",
			   __FILE__, __LINE__);

  if (max_sample_coord.size() != dims)
    throw QuickFlash::Except("Incompatible boundary coordinate vectors",
			   __FILE__, __LINE__);

  for (unsigned int i = 0; i < dims; i++)
    if (!(max_sample_coord[i] > min_sample_coord[i]))
      throw QuickFlash::Except("Min coords must be less than max coords",
			     __FILE__, __LINE__);

  min_coord = min_sample_coord;
  max_coord = max_sample_coord;

  bin_spacing = sample_spacing;

  const unsigned int bin_spacing_size = bin_spacing.size();

  if (bin_spacing_size == 1)
    {
    const double spacing = bin_spacing[0];

    for (unsigned int i = 1; i < dims; i++)
      bin_spacing[i] = spacing;
    }
  else if (bin_spacing_size != dims)
    throw QuickFlash::Except("Incompatible bin spacing vector");

  for (unsigned int i = 0; i < dims; i++)
    if (!(bin_spacing[i] > 0.0))
      throw QuickFlash::Except("Bin spacing must be positive", __FILE__, 
			     __LINE__);

  load_data(datafile, buffer_size);
  }


void MixingLayer::reset(const MixingLayer & source)
  {
  if (&source != this)
    {
    dims = source.dims;

    sim_time = source.sim_time;

    min_coord = source.min_coord;
    max_coord = source.max_coord;
	
    bin_spacing = source.bin_spacing;

    // int_pos = source.int_pos;

    dens_array = source.dens_array;
    poly_dens_array = source.poly_dens_array;
    poly_array = source.poly_array;

    dens_grid = source.dens_grid;
    poly_dens_grid = source.poly_dens_grid;
    }
  }


void MixingLayer::get_lateral_spectral_power_dens(const unsigned int bin_index,
						  std::vector<double> & power) 
  const
  {
  if (dims != 2)
    throw QuickFlash::Except("Only 2D grids supported at this time", __FILE__,
			   __LINE__);

  DoubleBlock temp_data;

  get_vol_avg_dens(temp_data);

  if (get_power_spectra_row(temp_data, bin_index, power) < 0)
    throw QuickFlash::Except("Unable to compute spectral power", __FILE__, 
			   __LINE__);
  }


void MixingLayer::get_lateral_spectral_power_dens(DoubleBlock & power) 
  const
  {
  if (dims != 2)
    throw QuickFlash::Except("Only 2D grids supported at this time", __FILE__,
			   __LINE__);

  DoubleBlock temp_data;

  get_vol_avg_dens(temp_data);

  if (get_power_spectra_row(temp_data, power) < 0)
    throw QuickFlash::Except("Unable to compute spectral power", __FILE__, 
			   __LINE__);
  }


void MixingLayer::get_lateral_spectral_power_poly_dens(
					      const unsigned int bin_index,
					      std::vector<double> & power) 
  const
  {
  if (dims != 2)
    throw QuickFlash::Except("Only 2D grids supported at this time", __FILE__,
			   __LINE__);

  DoubleBlock temp_data;

  get_vol_avg_poly_dens(temp_data);

  if (get_power_spectra_row(temp_data, bin_index, power) < 0)
    throw QuickFlash::Except("Unable to compute spectral power", __FILE__, 
			   __LINE__);
  }


void MixingLayer::get_lateral_spectral_power_poly_dens(DoubleBlock & power) 
  const
  {
  if (dims != 2)
    throw QuickFlash::Except("Only 2D grids supported at this time", __FILE__,
			   __LINE__);

  DoubleBlock temp_data;

  get_vol_avg_poly_dens(temp_data);

  if (get_power_spectra_row(temp_data, power) < 0)
    throw QuickFlash::Except("Unable to compute spectral power", __FILE__, 
			   __LINE__);
  }


void MixingLayer::get_secondary_power_ratio_dens(std::vector<double> 
						 & bin_power_ratio,
						 std::vector<unsigned int>
						 & secondary_freq) const
  {
  DoubleBlock power_data;

  get_lateral_spectral_power_dens(power_data);

  const std::vector<unsigned int> & grid_dims = power_data.get_block_dims();

  const unsigned int num_rows = grid_dims[0];
  const unsigned int num_cols = grid_dims[1];

  bin_power_ratio.resize(num_rows);
  secondary_freq.resize(num_rows);

  std::vector<unsigned int> grid_index(2);

  for (unsigned int row_index = 0; row_index < num_rows; row_index++)
    {
    grid_index[0] = row_index;
    grid_index[1] = 0;

    const double primary_power = power_data[grid_index];

    unsigned int freq_index = 0;
    double secondary_power = -1.0;  // Invalid power value

    double power_ratio = 0.0;

    if (primary_power > 0.0)
      {
      for (unsigned int col_index = 2; col_index < num_cols; col_index++)
	{
	grid_index[1] = col_index;

	const double current_power = power_data[grid_index];

	if (current_power > secondary_power)
	  {
	  secondary_power = current_power;
	  freq_index = col_index;
	  }
	}

      power_ratio = secondary_power / primary_power;
      }

    bin_power_ratio[row_index] = power_ratio;
    secondary_freq[row_index] = freq_index;
    }
  }


void MixingLayer::get_secondary_power_ratio_poly_dens(std::vector<double> 
						      & bin_power_ratio,
						      std::vector<unsigned int>
						      & secondary_freq) const
  {
  DoubleBlock power_data;

  get_lateral_spectral_power_poly_dens(power_data);

  const std::vector<unsigned int> & grid_dims = power_data.get_block_dims();

  const unsigned int num_rows = grid_dims[0];
  const unsigned int num_cols = grid_dims[1];

  bin_power_ratio.resize(num_rows);
  secondary_freq.resize(num_rows);

  std::vector<unsigned int> grid_index(2);

  for (unsigned int row_index = 0; row_index < num_rows; row_index++)
    {
    grid_index[0] = row_index;
    grid_index[1] = 0;

    const double primary_power = power_data[grid_index];

    unsigned int freq_index = 0;
    double secondary_power = -1.0;  // Invalid power value

    double power_ratio = -1.0;  // Invalid ratio value

    if (primary_power > 0.0)
      {
      for (unsigned int col_index = 2; col_index < num_cols; col_index++)
	{
	grid_index[1] = col_index;

	const double current_power = power_data[grid_index];

	if (current_power > secondary_power)
	  {
	  secondary_power = current_power;
	  freq_index = col_index;
	  }
	}

      power_ratio = secondary_power / primary_power;
      }

    bin_power_ratio[row_index] = power_ratio;
    secondary_freq[row_index] = freq_index;
    }
  }


void MixingLayer::load_data(const QuickFlash::File::DataFile & datafile,
			    const unsigned int buffer_size)
  {
  unsigned int active_buffer_size = buffer_size;

  if (active_buffer_size < 1)
    active_buffer_size = default_buffer_size;

  const QuickFlash::File::MeshInfo & meshinfo = datafile.get_mesh_info();

  sim_time = meshinfo.get_sim_time();

  const unsigned int file_dims = meshinfo.get_dims();

  if (file_dims != dims)
    throw QuickFlash::Except("Incompatible file dimensions", __FILE__, __LINE__);

  // Determine the bin layout

  grid_dims.resize(dims);

  for (unsigned int i = 0; i < dims; i++)
    {
    const double axis_bin_spacing = bin_spacing[i];

    const double lower_coord = min_coord[i];

    const double default_domain_width = max_coord[i] - lower_coord;

    const double fl_num_bins = default_domain_width / axis_bin_spacing;

    const unsigned int num_axis_bins 
      = static_cast<unsigned int>(fl_num_bins) + 1;

    const double domain_width 
      = static_cast<double>(num_axis_bins) * axis_bin_spacing;

    max_coord[i] = lower_coord + domain_width;

    grid_dims[i] = num_axis_bins;
    }

  const unsigned int num_bins_x = grid_dims[0];

  const double min_coord_x = min_coord[0];
  const double max_coord_x = max_coord[0];

  // Set up the arrays

  dens_array.reset(num_bins_x, min_coord_x, max_coord_x);
  poly_dens_array.reset(num_bins_x, min_coord_x, max_coord_x);
  poly_array.reset(num_bins_x, min_coord_x, max_coord_x);

  dens_grid.reset(grid_dims, min_coord, max_coord);
  poly_dens_grid.reset(grid_dims, min_coord, max_coord);

  // Load the data sets

  const QuickFlash::File::Dataset & dens_set = datafile.get_dataset(dens_name);

  const QuickFlash::File::Dataset & foam_set = datafile.get_dataset(foam_name);

  // Run through the blocks

  const unsigned int num_blocks = meshinfo.get_num_blocks();
  
  QuickFlash::Block::BlockData<double> dens_block;
  QuickFlash::Block::BlockData<double> foam_block;

  std::vector<double> cell_center;

  std::vector<double> cell_min;
  std::vector<double> cell_max;

  for (unsigned int block_index = 0; block_index < num_blocks; block_index++)
    {
    if (meshinfo.is_leaf(block_index))
      {
      const QuickFlash::Block::BlockInfo & block_info
	= meshinfo.get_block_info(block_index);

      // See if block (x-coordinate) falls within the bin domain

      const std::vector<double> & block_min = block_info.get_min_coords();
      const std::vector<double> & block_max = block_info.get_max_coords();

      bool block_grid_overlap = true;

      for (unsigned int i = 0; i < dims; i++)
	if ((!(block_min[i] < max_coord[i])) 
	    || (!(block_max[i] > min_coord[i])))
	  block_grid_overlap = false;

      if (block_grid_overlap)
	{
	// Load the block data

	dens_set.get_block_data(block_index, dens_block);
	foam_set.get_block_data(block_index, foam_block);

	const std::vector<double> & cell_width = block_info.get_cell_width();

	const double cell_width_x = cell_width[0];

	const double cell_volume = block_info.get_cell_volume();

	const unsigned int num_cells = block_info.get_num_cells();

	for (unsigned int cell_index = 0; cell_index < num_cells; cell_index++)
	  {
	  // Get the center coordinates

	  block_info.get_cell_center(cell_index, cell_center);

	  const double cell_center_x = cell_center[0];

	  block_info.get_cell_bounds(cell_index, cell_min, cell_max);

	  bool cell_grid_overlap = true;

	  for (unsigned int i = 0; i < dims; i++)
	    if ((!(cell_min[i] < max_coord[i])) 
		|| (!(cell_max[i] > min_coord[i])))
	      cell_grid_overlap = false;

	  if (cell_grid_overlap)
	    {
	    const double dens_value = dens_block[cell_index];

	    const double foam_value = foam_block[cell_index];

	    const double poly_value = 1.0 - foam_value;

	    const double poly_dens_value = poly_value * dens_value;

	    // WEIGHT BY CELL VOLUME INSTEAD OF MASS?????

	    dens_array.add_data_spread(cell_center_x, dens_value,
				       cell_width_x, cell_volume);

	    poly_dens_array.add_data_spread(cell_center_x, poly_dens_value,
					    cell_width_x, cell_volume);

	    poly_array.add_data_spread(cell_center_x, poly_value,
				       cell_width_x, cell_volume);

	    dens_grid.add_data_spread(cell_center, dens_value,
				      cell_width, cell_volume);

	    poly_dens_grid.add_data_spread(cell_center, poly_dens_value,
					   cell_width, cell_volume);
	    }
	  }
	}
      }
    }
  }
