// C++ program file flash_stat_bins.cpp

/*
  By Nathan C. Hearn
     September 20, 2006

  Statistical binning classes.
*/


#include "stat_bins.hpp"
#include <vector>
#include "quickflash_except.hpp"
#include "quickflash_geometry.hpp"


namespace Stats
{

// Class Bin

void Bin::reset()
  {
  num_entries = 0;

  sum_weight = 0.0;

  sum_weight_values = 0.0;
  sum_weight_sqValues = 0.0;

  min_value = 0.0;
  max_value = 0.0;
  }


void Bin::reset(const Bin & source)
  {
  if (&source != this)
    {
    num_entries = source.num_entries;

    sum_weight = source.sum_weight;

    sum_weight_values = source.sum_weight_values;
    sum_weight_sqValues = source.sum_weight_sqValues;
    }
  }


void Bin::add_bin(const Bin & source)
  {
  num_entries += source.num_entries;

  sum_weight += source.sum_weight;

  sum_weight_values += source.sum_weight_values;
  sum_weight_sqValues += source.sum_weight_sqValues;

  const double source_min_val = source.min_value;
  const double source_max_val = source.max_value;

  if (source_min_val < min_value)
    min_value = source_min_val;

  if (source_max_val > max_value)
    max_value = source_max_val;
  }


void Bin::add_entry(const double value, const double weight)
  {
  // Should we allow zero weights???

  sum_weight += weight;

  const double weight_value = weight * value;

  sum_weight_values += weight_value;

  sum_weight_sqValues += weight_value * value;

  if (num_entries > 0)
    {
    if (value < min_value)
      min_value = value;
    else if (value > max_value)
      max_value = value;
    }
  else
    {
    min_value = value;
    max_value = value;
    }

  num_entries++;
  }


double Bin::get_pop_sqStddev() const
  {
  // Implement sigma^2 = [ SUM(w_i * x_i^2) / SUM(w_i) ] - (N * MEAN(x_i)^2)

  // const double fl_num_entries = static_cast<double>(num_entries);

  // const double oneOver_mean_weight = fl_num_entries / sum_weight;

  double sqStddev = 0.0;

  if (sum_weight > 0.0)
    {
    const double oneOver_sum_weight = 1.0 / sum_weight;

    const double mean_val = get_mean();

    // double var = (sum_weight_sqValues * oneOver_mean_weight) 
    //   - (fl_num_entries * mean_val * mean_val);

    sqStddev = (sum_weight_sqValues * oneOver_sum_weight) 
      - (mean_val * mean_val);

    if (sqStddev < 0.0)
      sqStddev = 0.0;
    }

  return sqStddev;
  }


// Class BinArray


void BinArray::reset()
  {
  num_bins = 0;

  bin_min_bounds.clear();
  bin_max_bounds.clear();

  bin_center_coords.clear();

  bins.clear();
  }


void BinArray::reset(const unsigned int num_bins, const double min_boundary, 
		     const double max_boundary)
  {
  if (num_bins < 1)
    throw QuickFlash::Except("No bins requested", __FILE__, __LINE__);

  if (!(min_boundary < max_boundary))
    throw QuickFlash::Except("Incompatible boundaries", __FILE__, __LINE__);

  // Build the arrays of boundaries

  std::vector<double> bin_min_boundaries(num_bins);
  std::vector<double> bin_max_boundaries(num_bins);

  const double total_width = max_boundary - min_boundary;

  const double bin_width = total_width / static_cast<double>(num_bins);

  // Make sure boundaries are consistent between bins

  double current_min = min_boundary;

  for (unsigned int i = 0; i < num_bins; i++)
    {
    bin_min_boundaries[i] = current_min;

    current_min += bin_width;

    bin_max_boundaries[i] = current_min;
    }

  bin_max_boundaries[num_bins-1] = max_boundary;  // For good measure

  // Set up the bins

  reset(bin_min_boundaries, bin_max_boundaries);
  }


void BinArray::reset(const std::vector<double> bin_min_boundaries, 
		     const std::vector<double> bin_max_boundaries)
  {
  reset();

  num_bins = bin_min_boundaries.size();

  // Make sure arrays are compatible

  if (bin_max_boundaries.size() != num_bins)
    throw QuickFlash::Except("Incompatible bin boundary arrays", __FILE__, 
			__LINE__);

  // Set up the structures

  bin_min_bounds = bin_min_boundaries;
  bin_max_bounds = bin_max_boundaries;

  bin_center_coords.resize(num_bins);
  bin_widths.resize(num_bins);

  bins.resize(num_bins);

  // Fill in arrays and check the boundaries ... do not allow overlaps??

  for (unsigned int bin_index = 0; bin_index < num_bins; bin_index++)
    {
    const double bin_min = bin_min_boundaries[bin_index];
    const double bin_max = bin_max_boundaries[bin_index];

    if (!(bin_min < bin_max))
      throw QuickFlash::Except("Incompatible min/max bin boundaries", __FILE__, 
			  __LINE__);

    bin_center_coords[bin_index] = (bin_min + bin_max) * 0.5;

    bin_widths[bin_index] = bin_max - bin_min;
    }
  }


void BinArray::reset(const BinArray & source)
  {
  if (&source != this)
    {
    num_bins = source.num_bins;

    bin_min_bounds = source.bin_min_bounds;
    bin_max_bounds = source.bin_max_bounds;

    bin_center_coords = source.bin_center_coords;
    bin_widths = source.bin_widths;

    bins = source.bins;
    }
  }


void BinArray::add_data(const double coordinate, const double value, 
			const double weight)
  {
  // Run through the bins adding the weight and value to any overlapping bin

  for (unsigned int bin_index = 0; bin_index < num_bins; bin_index++)
    {
    if (in_bin(bin_index, coordinate))
      bins[bin_index].add_entry(value, weight);
    }
  }


void BinArray::add_data_spread(const double coordinate, const double value, 
			       const double spread, const double weight)
  {
  // Run through the bins, assigning the fraction of the weight to each bin
  // ... according to the fraction in which the bin covers the spread

  for (unsigned int bin_index = 0; bin_index < num_bins; bin_index++)
    {
    double bin_weight = weight;

    if (spread > 0.0)
      bin_weight *= get_overlap_fract(bin_index, coordinate, spread);	
    else if (!(in_bin(bin_index, coordinate)))
      bin_weight = 0.0;

    if (bin_weight > 0.0)
      bins[bin_index].add_entry(value, bin_weight);
    }
  }


bool BinArray::in_bin(const unsigned int bin_index, const double coordinate, 
		      const double spread) const
  {
  // Determine if this position fits in the specified bin

  // CHECK VALUE FOR SPREAD??

  // Get the boundaries of the bin

  const double bin_min = bin_min_bounds[bin_index];
  const double bin_max = bin_max_bounds[bin_index];

  bool inside = false;

  if (spread > 0.0)
    {
    const double half_spread = spread * 0.5;

    const double coord_min = coordinate - half_spread;
    const double coord_max = coordinate + half_spread;

    // See if any of the range [min_coord,max_coord) falls in [bin_min,bin_max)

    if ((coord_min < bin_max) && (coord_max > bin_min))
      inside = true;
    }
  else if ((!(coordinate < bin_min)) && (coordinate < bin_max))
    inside = true;

  return inside;
  }


double BinArray::get_overlap_fract(const unsigned int bin_index, 
				   const double coordinate,
				   const double spread) const
  {
  /*
    Returns the fraction of the range 

        [ coordinate-(spread/2) , coordinate+(spread/2) ) 

    that falls within the specified bin.

    The value returned is in the range [0.0,1.0], i.e., "none of the spread 
    is within the bin" up to "all of the spread is in the bin."
  */

  double ret_fract = 1.0;  // Spread is completely within block

  if (spread > 0.0)
    {
    const double half_spread = spread * 0.5;

    const double coord_min = coordinate - half_spread;
    const double coord_max = coordinate + half_spread;

    // Determine if the spread is not completely within the block

    const double bin_min = bin_min_bounds[bin_index];
    const double bin_max = bin_max_bounds[bin_index];

    if (!(coord_max > bin_min))
      ret_fract = 0.0;
    else if (!(coord_min < bin_max))
      ret_fract = 0.0;
    else if ((coord_min > bin_min) || (coord_max < bin_max))
      {
      // Compute partial coverage

      double overlap_width = spread;

      const double max_excess = coord_max - bin_max;

      if (max_excess > 0.0)
	overlap_width -= max_excess;

      const double min_excess = bin_min - coord_min;

      if (min_excess > 0.0)
	overlap_width -= min_excess;

      // Determine the fraction

      ret_fract = overlap_width / spread;

      if (ret_fract < 0.0)
	ret_fract = 0.0;
      }
    else
      ret_fract = 1.0;  // Full coverage
    }
  else if (!(in_bin(bin_index, coordinate)))
    ret_fract = 0.0;  // No overlap
  else
    ret_fract = 1.0;  // Full overlap

  return ret_fract;
  }


void BinArray::get_means(std::vector<double> & bin_data) const
  {
  bin_data.resize(num_bins);

  for (unsigned int i = 0; i < num_bins; i++)
    bin_data[i] = bins[i].get_mean();
  }


void BinArray::get_sum_weights(std::vector<double> & bin_data) const
  {
  bin_data.resize(num_bins);

  for (unsigned int i = 0; i < num_bins; i++)
    bin_data[i] = bins[i].get_sum_weights();
  }


void BinArray::get_sum_weight_values(std::vector<double> & bin_data) const
  {
  bin_data.resize(num_bins);

  for (unsigned int i = 0; i < num_bins; i++)
    bin_data[i] = bins[i].get_sum_weight_values();
  }


void BinArray::get_bin_entries(std::vector<unsigned int> & bin_data) const
  {
  bin_data.resize(num_bins);

  for (unsigned int i = 0; i < num_bins; i++)
    bin_data[i] = bins[i].get_num_entries();
  }


// Class BinGrid


BinGrid::BinGrid() : space_dims(0), grid_info(), grid_data() { }


BinGrid::BinGrid(const std::vector<unsigned int> & dims, 
		 const std::vector<double> & min_coords,
		 const std::vector<double> & max_coords) :
  space_dims(0), grid_info(), grid_data()
  { reset(dims, min_coords, max_coords); }


BinGrid::BinGrid(const BinGrid & source) :
  space_dims(0), grid_info(), grid_data()
  { reset(source); }


void BinGrid::reset()
  {
  space_dims = 0;

  grid_info.reset();
  grid_data.reset();
  }


void BinGrid::reset(const std::vector<unsigned int> & dims, 
		    const std::vector<double> & min_coords,
		    const std::vector<double> & max_coords)
  {
  grid_info.reset(min_coords, max_coords, dims, QuickFlash::Geometry::Cartesian);

  grid_data.reset(dims);

  space_dims = grid_info.get_space_dims();

  num_bins = grid_info.get_num_cells();

  // Reset the grid data

  for (unsigned int i = 0; i < num_bins; i++)
    grid_data[i].reset();
  }


void BinGrid::reset(const BinGrid & source)
  {
  if (&source != this)
    {
    space_dims = source.space_dims;

    num_bins = source.num_bins;

    grid_info = source.grid_info;
    grid_data = source.grid_data;
    }
  }


void BinGrid::add_data(const std::vector<double> & coords, const double value,
		       const double weight)
  {
  if (coords.size() != space_dims)
    throw QuickFlash::Except("Incompatible coordinate vector", __FILE__, 
			   __LINE__);

  if (grid_info.in_block(coords))
    {
    const unsigned int bin_index = grid_info.get_nearest_cell(coords);

    grid_data[bin_index].add_entry(value, weight);
    }
  }


void BinGrid::add_data_spread(const std::vector<double> & coords,
			      const double value,
			      const std::vector<double> &  spread, 
			      const double weight)
  {
  if (coords.size() != space_dims)
    throw QuickFlash::Except("Incompatible coordinate vector", __FILE__, 
			   __LINE__);

  if (spread.size() != space_dims)
    throw QuickFlash::Except("Incompatible spread vector", __FILE__, 
			   __LINE__);

  // Determine the boundaries of the spreaded data

  std::vector<double> min_coords = coords;
  std::vector<double> max_coords = coords;

  for (unsigned int i = 0; i < space_dims; i++)
    {
    double axis_spread = spread[i];

    if (axis_spread < 0.0)
      axis_spread = 0.0;

    const double half_spread = 0.5 * axis_spread;

    min_coords[i] -= half_spread;
    max_coords[i] += half_spread;
    }

  // Determine the range of bin indexes

  bool spread_in_grid = true;

  const std::vector<unsigned int> & grid_dims = grid_info.get_block_dims();

  std::vector<unsigned int> min_bin_index(space_dims);
  std::vector<unsigned int> max_bin_index(space_dims);

  const std::vector<double> & min_grid_coords = grid_info.get_min_coords();
  const std::vector<double> & max_grid_coords = grid_info.get_max_coords();

  for (unsigned int i = 0; i < space_dims; i++)
    {
    const double grid_min = min_grid_coords[i];
    const double grid_max = max_grid_coords[i];

    const double spread_min = min_coords[i];
    const double spread_max = max_coords[i];

    if ((!(spread_max > grid_min)) || (!(spread_min < grid_max)))
      spread_in_grid = false;
    else
      {
      const unsigned int axis_size = grid_dims[i];

      min_bin_index[i] = 0;
      max_bin_index[i] = axis_size - 1;

      if (grid_min < spread_min)
	min_bin_index[i] 
	  = grid_info.get_nearest_cell_axis_index(i, spread_min);

      if (grid_max > spread_max)
	max_bin_index[i] 
	  = grid_info.get_nearest_cell_axis_index(i, spread_max);
      }
    }

  // Distribute the data

  if (spread_in_grid)
    {
    std::vector<unsigned int> bin_index_range_size(space_dims);

    for (unsigned int i = 0; i < space_dims; i++)
      bin_index_range_size[i] = (max_bin_index[i] + 1) - min_bin_index[i];

    QuickFlash::VectorCounter counter(bin_index_range_size);

    std::vector<unsigned int> bin_index(space_dims);

    while (counter.in_bounds())
      {
      for (unsigned int i = 0; i < space_dims; i++)
	bin_index[i] = min_bin_index[i] + counter[i];

      const double overlap_fract 
	= get_overlap_fract(bin_index, coords, spread);

      const double bin_weight = weight * overlap_fract;

      if (bin_weight > 0.0)
	grid_data[bin_index].add_entry(value, bin_weight);

      counter.increment();
      }
    }
  }


void BinGrid::get_means(QuickFlash::Block::BlockData<double> & bin_data) const
  {
  bin_data.reset(grid_data.get_block_dims());

  const unsigned int num_bins = grid_data.get_num_cells();

  for (unsigned int i = 0; i < num_bins; i++)
    bin_data[i] = grid_data[i].get_mean();
  }


void BinGrid::get_sum_weights(QuickFlash::Block::BlockData<double> & bin_data) 
  const
  {
  bin_data.reset(grid_data.get_block_dims());

  const unsigned int num_bins = grid_data.get_num_cells();

  for (unsigned int i = 0; i < num_bins; i++)
    bin_data[i] = grid_data[i].get_sum_weights();
  }


void BinGrid::get_sum_weight_values(QuickFlash::Block::BlockData<double> 
				    & bin_data) const
  {
  bin_data.reset(grid_data.get_block_dims());

  const unsigned int num_bins = grid_data.get_num_cells();

  for (unsigned int i = 0; i < num_bins; i++)
    bin_data[i] = grid_data[i].get_sum_weight_values();
  }


void BinGrid::get_bin_entries(QuickFlash::Block::BlockData<unsigned int> 
			      & bin_data) const
  {
  bin_data.reset(grid_data.get_block_dims());

  const unsigned int num_bins = grid_data.get_num_cells();

  for (unsigned int i = 0; i < num_bins; i++)
    bin_data[i] = grid_data[i].get_num_entries();
  }


double BinGrid::get_overlap_fract(const std::vector<unsigned int> & bin_index, 
				  const std::vector<double> & coords,
				  const std::vector<double> & spread) const
  {
  /*
    Returns the fraction of the area of coords +/- (spread / 2) that falls 
    within the specified bin.

    The value returned is in the range [0.0,1.0], i.e., "none of the spread 
    is within the bin" up to "all of the spread is in the bin."
  */

  // ERROR CHECKING???

  double ret_fract = 1.0;  // Spread is completely within block

  std::vector<double> bin_min_coords;
  std::vector<double> bin_max_coords;

  grid_info.get_cell_bounds(bin_index, bin_min_coords, bin_max_coords);

  for (unsigned int i = 0; i < space_dims; i++)
    {
    const double axis_coord = coords[i];
    double axis_spread = spread[i];

    if (axis_spread < 0.0)
      axis_spread = 0.0;

    const double half_spread = 0.5 * axis_spread;

    const double coord_min = axis_coord - half_spread;
    const double coord_max = axis_coord + half_spread;

    // Determine if the spread is not completely within the block

    const double bin_min = bin_min_coords[i];
    const double bin_max = bin_max_coords[i];

    if ((!(coord_max > bin_min)) || (!(coord_min < bin_max)))
      ret_fract = 0.0;
    else if (axis_spread > 0.0)
      {
      if ((coord_min > bin_min) || (coord_max < bin_max))
	{
        // Compute partial coverage

	double axis_overlap_width = axis_spread;

	const double max_excess = coord_max - bin_max;

	if (max_excess > 0.0)
	  axis_overlap_width -= max_excess;

	const double min_excess = bin_min - coord_min;

	if (min_excess > 0.0)
	  axis_overlap_width -= min_excess;

	// Determine the fraction

	ret_fract *= axis_overlap_width / axis_spread;

	if (ret_fract < 0.0)
	  ret_fract = 0.0;
	}
      }
    }

  return ret_fract;
  }


}  // End namespace Stats
