// C++ program file map_uniform.cpp

/*
  By Nathan C. Hearn
     July 6, 2007

  Program for producing a 2D uniform (Cartesian) density mesh from a 
  Flash data file.
*/


#include <iostream>
#include <stdlib.h>
#include <string>
#include <vector>
#include "quickflash_file_datafile.hpp"
#include "quickflash_file_meshinfo.hpp"
#include "quickflash_file_dataset.hpp"
#include "quickflash_block_blockinfo.hpp"
#include "quickflash_block_blockdata.hpp"
#include "accum_grid.hpp"


const unsigned int buffer_size = 1000;

const std::string dens_set_name = "dens";


int main(int argc, char * argv[])
  {
  using std::cout;
  using std::cerr;
  using std::endl;

  const int min_args = 8;

  if (argc < min_args)
    {
    cerr << "Usage: " << argv[0] << " min_x max_x min_y max_y x_cells y_cells "
	 << "flash_file [ window_x_size window_y_size ]" << endl;
    cerr << endl;
    return 1;
    }

  int argptr = 1;

  int req_args = min_args;

  const double min_x = atof(argv[argptr++]);
  const double max_x = atof(argv[argptr++]);

  if (!(min_x < max_x))
    {
    cerr << "Error: Incompatible x bounds" << endl << endl;
    return -1;
    }

  const double min_y = atof(argv[argptr++]);
  const double max_y = atof(argv[argptr++]);

  if (!(min_y < max_y))
    {
    cerr << "Error: Incompatible y bounds" << endl << endl;
    return -1;
    }

  const unsigned int x_cells = static_cast<unsigned int>(atoi(argv[argptr++]));
  const unsigned int y_cells = static_cast<unsigned int>(atoi(argv[argptr++]));

  if ((x_cells < 1) || (y_cells < 1))
    {
    cerr << "Error: Invalid cell count" << endl << endl;
    return -1; 
    }

  const std::string filename = argv[argptr++];

  bool window_size_specified = false;

  std::vector<double> window_size;

  if (argptr < argc)
    {
    req_args += 2;

    if (argc < req_args)
      {
      cerr << "Error: Must specify window size along both axes" << endl
	   << endl;
      return -1;
      }

    const double window_x_size = atof(argv[argptr++]);
    const double window_y_size = atof(argv[argptr++]);

    if ((!(window_x_size > 0.0)) || (!(window_y_size > 0.0)))
      {
      cerr << "Error: Sampling window sizes must be positive" << endl << endl;
      return -1;
      }

    window_size.resize(2);

    window_size[0] = window_x_size;
    window_size[1] = window_y_size;

    window_size_specified = true;
    }

  // Report on input

  cerr << "x-range [ " << min_x << " - " << max_x << " ] cells [ " << x_cells 
       << " ]" << endl;
  cerr << "y-range [ " << min_y << " - " << max_y << " ] cells [ " << y_cells 
       << " ]" << endl;

  cerr << "Flash file [ " << filename << " ]" << endl;

  if (window_size_specified)
    cerr << "Window size [ " << window_size[0] << " " << window_size[1] 
	 << " ]" << endl;

  // Set up the grid

  std::vector<double> min_grid_coords(2);
  std::vector<double> max_grid_coords(2);

  min_grid_coords[0] = min_x;
  min_grid_coords[1] = min_y;

  max_grid_coords[0] = max_x;
  max_grid_coords[1] = max_y;

  std::vector<unsigned int> grid_dims(2);

  grid_dims[0] = x_cells;
  grid_dims[1] = y_cells;

  AccumGrid grid(min_grid_coords, max_grid_coords, grid_dims, window_size);

  // Load the mesh and fill the grid

  QuickFlash::File::DataFile dfile(filename, buffer_size);

  const QuickFlash::File::MeshInfo & meshinfo = dfile.get_mesh_info();

  const QuickFlash::File::Dataset & dens_set 
    = dfile.get_dataset(dens_set_name);

  const unsigned int num_blocks = meshinfo.get_num_blocks();

  QuickFlash::Block::BlockData<double> dens_data;

  std::vector<double> cell_center;
  std::vector<double> cell_width;

  for (unsigned int block_index = 0; block_index < num_blocks; block_index++)
    if (meshinfo.is_leaf(block_index))
      {
      dens_set.get_block_data(block_index, dens_data);

      const QuickFlash::Block::BlockInfo & block_info 
	= meshinfo.get_block_info(block_index);

      const unsigned int num_cells = block_info.get_num_cells();

      const double cell_vol = block_info.get_cell_volume();  // Cartesian

      cell_width = block_info.get_cell_width();
      cell_width.resize(2, 0.0);  // Project up/down

      for (unsigned int cell_index = 0; cell_index < num_cells; cell_index++)
	{
	const double cell_mass = dens_data[cell_index] * cell_vol;

	// Get cell coords - project up/down if necessary

	block_info.get_cell_center(cell_index, cell_center);

	cell_center.resize(2, 0.0);

	grid.distribute_value(cell_center, cell_width, cell_mass);
	}
      }

  // Output the grid

  const double cell_volume = grid.get_cell_volume();

  const double oneOver_cell_volume = 1.0 / cell_volume;

  cout << "# Grided densities from file [ " << filename << " ]" << endl;
  cout << "# x-range [ " << min_x << " - " << max_x << " ] cells [ " 
       << x_cells << " ]" << endl;
  cout << "# y-range [ " << min_y << " - " << max_y << " ] cells [ " 
       << y_cells << " ]" << endl;

  if (window_size_specified)
    cout << "# Window size [ " << window_size[0] << " " << window_size[1] 
	 << " ]" << endl;

  std::vector<unsigned int> cell_index(2);
	
  for (unsigned int i = 0; i < x_cells; i++)
    {
    cell_index[0] = i;

    for (unsigned int j = 0; j < y_cells; j++)
      {
      cell_index[1] = j;

      const double cell_mass = grid[cell_index];

      const double cell_dens = cell_mass * oneOver_cell_volume;

      cout << cell_dens << " ";
      }

    cout << endl;
    }

  cout << endl;

  return 0;
  }
