// C++ program file mass_bins.cpp

/*
  By Nathan C. Hearn
     July 2, 2008

  Bin mass based on ranges in values for a given dataset.
*/


#include <iostream>
#include <string>
#include <vector>
#include <list>
#include <cstdlib>
#include <fstream>
#include <cmath>
#include "quickflash_utils_commandline.hpp"
#include "quickflash_utils_file.hpp"
#include "quickflash_utils_text.hpp"
#include "quickflash_file_datafile.hpp"
#include "quickflash_file_meshinfo.hpp"
#include "quickflash_file_dataset.hpp"
#include "quickflash_block_blockdata.hpp"
#include "quickflash_block_blockinfo.hpp"
#include "quickflash_stats_bins.hpp"


const char default_comment_marker[] = "# ";

const unsigned int buffer_size = 500;

const char dens_name[] = "dens";


int main(int argc, char * argv[])
  {
  using std::cerr;
  using std::endl;

  std::vector<std::string> argument_list;
  std::list<QuickFlash::Utils::Option> option_list;

  QuickFlash::Utils::parse_command_line(argc, argv, argument_list, option_list);

  const unsigned int num_args = argument_list.size();

  const unsigned int min_args = 5;

  if (num_args < min_args)
    {
    const std::string full_exec = argv[0];

    std::string exec_path;
    std::string exec_name;

    QuickFlash::Utils::split_name_path_file(full_exec, exec_path, exec_name);
    
    cerr << endl << "Usage: " << exec_name << " dataset min_value max_value "
	 << "num_bins"
	 << endl << "\t"
	 << "[ --output-file=FILENAME ] [ --add-comment-marker[=MARKER] ]" 
	 << endl << "\t"
	 << "[ --logscale ] [ --compute-radii ]"
	 << endl << "\t"
	 << "filename1 [ filename2 ... ]" 
	 << endl << endl;

    return 1;
    }


  // Read the arguments

  unsigned int argptr = 0;

  const std::string dataset_name = argument_list[argptr++];

  const double min_value = std::atof(argument_list[argptr++].c_str());
  const double max_value = std::atof(argument_list[argptr++].c_str());

  const int int_num_bins = std::atoi(argument_list[argptr++].c_str());

  std::list<std::string> datafile_name_list;

  while (argptr < num_args)
    datafile_name_list.push_back(argument_list[argptr++]);


  // Parse the options

  bool use_output_file = false;

  std::string output_file_name;

  std::string comment_marker = "";

  bool use_log_scale = false;

  bool compute_radii = false;

  std::list<QuickFlash::Utils::Option>::const_iterator option_iter 
    = option_list.begin();

  while (option_iter != option_list.end())
    {
    const QuickFlash::Utils::Option & current_option = *option_iter;

    const std::string & keyword = current_option.get_key();

    if (keyword == "output-file")
      {
      if (!current_option.value_present())
	{
	cerr << endl << "Error: Output file option requires file name"
	     << endl << endl;

	return -1;
	}

      use_output_file = true;

      output_file_name = current_option.get_value();
      }
    else if (keyword == "add-comment-marker")
      {
      if (current_option.value_present())
	comment_marker = current_option.get_value() + " ";
      else
	comment_marker = default_comment_marker;
      }
    else if (keyword == "logscale")
      {
      use_log_scale = true;
      }
    else if (keyword == "compute-radii")
      {
      compute_radii = true;
      }
    else
      {
      cerr << endl << "Error: Unrecognized option " << keyword << endl 
	   << endl;

      return -1;
      }
    
    ++option_iter;
    }


  // Perform some error checking and get derived quantities

  if (!(min_value < max_value))
    {
    cerr << endl << "Error: Min value must be less than max value" << endl 
	 << endl;
    return -1;
    }

  if (use_log_scale)
    if (!(min_value > 0.0))
      {
      cerr << endl << "Error: Positive values required for log scale" << endl;
      return -1;
      }

  if (int_num_bins < 1)
    {
    cerr << endl << "Error: Number of bins must be positive" << endl << endl;
    return -1;
    }

  const unsigned int num_bins = static_cast<unsigned int>(int_num_bins);

  // Set up the bins

  QuickFlash::Stats::BinArray mass_bins(min_value, max_value, num_bins, 
				      use_log_scale);

  const std::vector<double> & bin_centers = mass_bins.get_bin_centers();

  QuickFlash::Stats::BinArray radii_bins;

  if (compute_radii)
    radii_bins.reset(min_value, max_value, num_bins, use_log_scale);

  // Set up the output object

  std::ostream * f = &(std::cout);

  if (use_output_file)
    f = new std::ofstream(output_file_name.c_str());

  (*f) << comment_marker << "Dataset " << dataset_name << endl;
  (*f) << comment_marker << "MinValue " << min_value << endl;
  (*f) << comment_marker << "MaxValue " << max_value << endl;
  (*f) << comment_marker << "NumBins " << num_bins << endl;

  (*f) << comment_marker << "LogScale ";

  if (use_log_scale)
    (*f) << "true";
  else
    (*f) << "false";

  (*f) << endl;

  (*f) << comment_marker << "BinMinValues " << mass_bins.get_bin_min_bounds() 
       << endl;

  (*f) << comment_marker << "BinValueWidths " << mass_bins.get_bin_widths() 
       << endl;

  (*f) << comment_marker << "BinCenterValues " << bin_centers << endl;

  (*f) << comment_marker << "ComputeRadii ";

  if (compute_radii)
    (*f) << "true";
  else
    (*f) << "false";

  (*f) << endl;

  (*f) << endl;

  (*f) << comment_marker << "Data" << endl;

  (*f) << endl;

  (*f) << "File_Name" << "\t" << "Sim_Time" << "\t" << "Total_Mass " 
       << "\t" << "Mass_Below_Range" << "\t" << "Mass_Above_Range";

  for (unsigned int bin_index = 0; bin_index < num_bins; bin_index++)
    (*f) << "\t" << "Mass_Bin_" << bin_index;

  if (compute_radii)
    for (unsigned int bin_index = 0; bin_index < num_bins; bin_index++)
      {
      (*f) << "\t" << "Min_Radius_Bin_" << bin_index
	   << "\t" << "Max_Radius_Bin_" << bin_index;
      }

  (*f) << endl;
  
  // Run through the files

  std::list<std::string>::const_iterator filename_iter 
    = datafile_name_list.begin();

  while (filename_iter != datafile_name_list.end())
    {
    const std::string & filename = *filename_iter;

    double total_mass = 0.0;

    double mass_below = 0.0;
    double mass_above = 0.0;

    mass_bins.reset_bins();

    if (compute_radii)
      radii_bins.reset_bins();

    // Open the file and get some data

    QuickFlash::File::DataFile df(filename, buffer_size, true);

    const QuickFlash::File::MeshInfo & meshinfo = df.get_mesh_info();

    const unsigned int num_blocks = meshinfo.get_num_blocks();

    const double sim_time = meshinfo.get_sim_time();

    const unsigned int dims = meshinfo.get_dims();

    // Open the datasets and set up storage

    const QuickFlash::File::Dataset & dens_set = df.get_dataset(dens_name);
    const QuickFlash::File::Dataset & value_set = df.get_dataset(dataset_name);

    QuickFlash::Block::BlockData<double> dens_data;
    QuickFlash::Block::BlockData<double> value_data;

    std::vector<double> cell_center_pos;

    for (unsigned int block_index = 0; block_index < num_blocks; block_index++)
      {
      if (meshinfo.is_leaf(block_index))
	{
	dens_set.get_block_data(block_index, dens_data);
	value_set.get_block_data(block_index, value_data);

	const QuickFlash::Block::BlockInfo & block_info 
	  = meshinfo.get_block_info(block_index);

	const unsigned int num_cells = block_info.get_num_cells();

	for (unsigned int cell_index = 0; cell_index < num_cells; cell_index++)
	  {
	  const double cell_vol = block_info.get_cell_volume(cell_index);
	  const double cell_dens = dens_data[cell_index];

	  const double cell_mass = cell_vol * cell_dens;

	  const double cell_value = value_data[cell_index];

	  double cell_radius = 0.0;

	  if (compute_radii)
	    {
	    block_info.get_cell_center(cell_index, cell_center_pos);

	    double sq_cell_rad = 0.0;

	    for (unsigned int i = 0; i < dims; i++)
	      {
	      const double axis_coord = cell_center_pos[i];

	      sq_cell_rad += axis_coord * axis_coord;
	      }

	    cell_radius = std::sqrt(sq_cell_rad);
	    }

	  if (mass_bins.in_bounds(cell_value))
	    {
	    mass_bins.add_data(cell_value, cell_dens, cell_vol);

	    if (compute_radii)
	      radii_bins.add_data(cell_value, cell_radius, cell_mass);
	    }
	  else
	    {
	    if (cell_value < min_value)
	      mass_below += cell_mass;
	    else
	      mass_above += cell_mass;
	    }

	  total_mass += cell_mass;
	  }
	}
      }

    // Report on results

    (*f) << filename << "\t" << sim_time << "\t" << total_mass << "\t"
	 << mass_below << "\t" << mass_above;

    for (unsigned int bin_index = 0; bin_index < num_bins; bin_index++)
      {
      const QuickFlash::Stats::Bin & bin = mass_bins[bin_index];

      (*f) << "\t" << bin.get_weight_value_sum();
      }

    if (compute_radii)
      for (unsigned int bin_index = 0; bin_index < num_bins; bin_index++)
	{
	const QuickFlash::Stats::Bin & bin = radii_bins[bin_index];

	if (bin.entry_present())
	  (*f) << "\t" << bin.get_min_value() << "\t" << bin.get_max_value();
	else
	  (*f) << "\t" << "NA" << "\t" << "NA";
	}

    (*f) << endl;

    ++filename_iter;
    }

  (*f) << endl;

  if (use_output_file)
    {
    delete f;
    f = &(std::cout);
    }

  return 0;
  }
