#!/usr/bin/env python

# generate_distrib_plot.py

# By Nathan C. Hearn
#    October 29, 2009
#
# Plotting program for mass distribution files

# Make sure the QuickFlash/utils directory is in PYTHONPATH!!!


try :
    import psyco
except ImportError :
    pass


def main() :

    from sys import exit, stderr
    from os import path
    from optparse import OptionParser

    from plot_gen import read_keywords_table, Plot
    from metadata_reader import parse_distrib_metadata


    ## Set up the command line parser

    usage = str("%prog [options] data_file1[,legend] "
                + "[ data_file2[,legend] ... ]")
    version = str("%prog 0.1")

    argparser = OptionParser(usage, version=version)

    output_prefix_str = str("output-prefix")
    argparser.add_option("--" + output_prefix_str, type="string",
                         dest="output_prefix",
                         help="Prefix for output plot files")
    argparser.set_defaults(output_prefix=None)

    plot_title_str = str("plot-title")
    argparser.add_option("--" + plot_title_str, type="string",
                         dest="plot_title", help="Plot title")
    argparser.set_defaults(plot_title=None)

    output_format_str = str("output-format")
    argparser.add_option("--" + output_format_str, type="string",
                         dest="output_format",
                         help="Name of matplotlib backend")
    argparser.set_defaults(output_format=None)

    background_vel_str = str("background-vel")
    argparser.add_option("--" + background_vel_str, type="string",
                         dest="background_vel",
                         help=("Background velocity (cm / s) "
                               + "with optional comma-separated time offset "
                               + "(s)"))
    argparser.set_defaults(background_vel=None)


    ## Process the command line input

    opts, args = argparser.parse_args()

    argc = len(args)

    min_args = int(1)

    if (argc < min_args) :
        stderr.write("\n")
        argparser.print_help(file=stderr)
        stderr.write("\n")
        exit(1)


    ## Get the required arguments

    argptr = int(0)

    data_file_list = list()
    legend_text_list = list()

    while (argptr < argc) :
        
        current_elem_str = args[argptr]
        argptr += 1

        elem_words = current_elem_str.split(",")

        num_words = len(elem_words)

        if ((num_words < 1) or (num_words > 2)) :
            stderr.write("\nError: Improper format for data file "
                         + "specification [ " + current_elem_str + " ]\n\n")
            exit(-1)

        data_file = elem_words[0]

        legend_text = data_file

        if (num_words > 1) :
            legend_text = elem_words[1]

        data_file_list.append(data_file)
        legend_text_list.append(legend_text)

    num_files = len(data_file_list)


    ## Process the options

    output_prefix = opts.output_prefix

    if (output_prefix is None) :
        output_prefix = ""
    elif (len(path.split(output_prefix)[1]) > 0) :
        output_prefix += "_"

    plot_title = opts.plot_title

    if (plot_title is None) :
        plot_title = output_file

    background_vel = float(0.0)
    background_vel_time_offset = float(0.0)

    if (opts.background_vel is not None) :
        background_vel_words = opts.background_vel.split(",")

        num_background_vel_words = len(background_vel_words)

        if ((num_background_vel_words < 1) or (num_background_vel_words > 2)) :
            stderr.write("\nError: Background velocity requires one or two "
                         + "comma-separated entries\n\n")
            exit(-1)

        background_vel = float(background_vel_words[0])

        if (num_background_vel_words == 2) :
            background_vel_time_offset = float(background_vel_words[1])


    ## Build the plot

    plot = Plot(plot_title, "Bin Center Position (microns)", "Bin Mass (mg)")

    for file_index in range(num_files) :

        ## Get the simulation time and distribution filename

        metadata_filename = data_file_list[file_index]

        print "***", metadata_filename

        metadata, slice_info = parse_distrib_metadata(metadata_filename)

        sim_time = metadata.get_value("SimTime")
        
        distrib_file = metadata.get_value("LayerDistribFile")

        ## Read the data

        keyword_values, plot_data \
            = read_keywords_table(distrib_file, table_start_line="BinData")

        ## Prepare the data

        pos_offset = (-1.0) * (sim_time - background_vel_time_offset) \
                     * background_vel

        orig_bin_centers = plot_data["Bin_Center"]
        orig_bin_masses = plot_data["Bin_Mass"]
        
        num_entries = len(orig_bin_centers)

        bin_centers = list()
        bin_masses = list()
        
        for entry_index in range(num_entries) :

            new_bin_center = (orig_bin_centers[entry_index] + pos_offset) \
                             * 1.0e+04

            new_bin_mass = orig_bin_masses[entry_index] * 1.0e+03  ## g -> mg
            
            bin_centers.append(new_bin_center)
            bin_masses.append(new_bin_mass)


        ## Plot the data

        dataset_title = legend_text_list[file_index]

        plot.add_plot(bin_centers, bin_masses, legend_text=dataset_title)

    output_filename = output_prefix + "distrib.svg"

    plot.show_plot(output_filename)


if (__name__ == "__main__") :
    main()
