// C++ program file layer_detection.cpp

/*
  By Nathan C. Hearn
     April 3, 2008

  Layer detection routines for analyzing simulations of Omega RT experiments.
*/


#include "layer_detection.hpp"
#include <string>
#include <vector>
#include <cmath>
#include "quickflash_file_datafile.hpp"
#include "quickflash_file_meshinfo.hpp"
#include "quickflash_file_dataset.hpp"
#include "quickflash_except.hpp"
#include "quickflash_block_blockdata.hpp"
#include "quickflash_tree_block.hpp"
#include "quickflash_block_blockinfo.hpp"


int find_maximal_envelope(const QuickFlash::File::DataFile & df,
			  const std::string & dataset_name,
			  const unsigned int search_axis,
			  const double min_data_value,
			  const double max_data_value,
			  double & min_coord, double & max_coord,
			  const bool limit_sample_radius,
			  const double max_sample_radius)
  {
  int ret_val = -1;  // Error state

  const QuickFlash::File::MeshInfo & meshinfo = df.get_mesh_info();

  const unsigned int dims = meshinfo.get_dims();

  if (search_axis >= dims)
    throw QuickFlash::Except("Axis index out of bounds", __FILE__, __LINE__);

  std::vector<double> domain_center;

  const double sq_max_sample_radius = max_sample_radius * max_sample_radius;

  if (limit_sample_radius)
    {
    if (!(max_sample_radius > 0.0))
      throw QuickFlash::Except("Sample radius must be a positive value", 
			     __FILE__, __LINE__);

    const std::vector<double> & min_bounds = meshinfo.get_volume_minbounds();
    const std::vector<double> & max_bounds = meshinfo.get_volume_maxbounds();

    domain_center.resize(dims);

    for (unsigned int i = 0; i < dims; i++)
      domain_center[i] = 0.5 * (min_bounds[i] + max_bounds[i]);
    }

  const QuickFlash::File::Dataset & dataset = df.get_dataset(dataset_name);

  QuickFlash::Block::BlockData<double> block_data;

  std::vector<double> cell_pos;

  bool found_interior = false;

  double envelope_min_pos = 0.0;
  double envelope_max_pos = 0.0;

  QuickFlash::Tree::ConstTreeIter node_iter = meshinfo.get_tree_iter();

  while (node_iter.valid_node())
    {
    const QuickFlash::Block::BlockInfo & block_info 
      = node_iter.get_block_info();

    const double block_min_pos = block_info.get_min_coords()[search_axis];
    const double block_max_pos = block_info.get_max_coords()[search_axis];

    bool skip_branch = false;

    if (found_interior)
      {
      if ((block_min_pos > envelope_min_pos) 
	  && (block_max_pos < envelope_max_pos))
	skip_branch = true;
      }

    if ((!skip_branch) && node_iter.is_leaf())
      {
      const unsigned int block_index = node_iter.get_block_index();

      dataset.get_block_data(block_index, block_data);

      const unsigned int num_cells = block_info.get_num_cells();

      // Find a cell in the desired range

      for (unsigned int cell_index = 0; cell_index < num_cells; cell_index++)
	{
	const double cell_value = block_data[cell_index];

	if ((cell_value > min_data_value) && (cell_value < max_data_value))
	  {
	  block_info.get_cell_center(cell_index, cell_pos);

	  bool use_cell = true;

	  if (limit_sample_radius)
	    {
	    double sq_cell_pos_radius = 0.0;

	    for (unsigned int i = 0; i < dims; i++)
	      if (i != search_axis)
		{
		const double axis_comp = cell_pos[i] - domain_center[i];

		sq_cell_pos_radius += axis_comp * axis_comp;
		}

	    if (sq_cell_pos_radius > sq_max_sample_radius)
	      use_cell = false;
	    }

	  if (use_cell)
	    {
	    const double axis_cell_pos = cell_pos[search_axis];

	    if (found_interior)
	      {
	      if (axis_cell_pos < envelope_min_pos)
		envelope_min_pos = axis_cell_pos;
	      else if (axis_cell_pos > envelope_max_pos)
		envelope_max_pos = axis_cell_pos;
	      }
	    else
	      {
	      envelope_min_pos = axis_cell_pos;
	      envelope_max_pos = axis_cell_pos;

	      found_interior = true;
	      }
	    }
	  }
	}
      }
    
    if (skip_branch)
      node_iter.next_branch();
    else
      node_iter.next();
    }

  if (found_interior)
    {
    min_coord = envelope_min_pos;
    max_coord = envelope_max_pos;

    ret_val = 0;  // OK
    }

  return ret_val;
  }


int find_layer_bounds_product(const std::vector<double> & bin_centers,
			      const double peak_product_fraction,
			      const std::vector<double> & poly_mass_bins,
			      const std::vector<double> & chbr_mass_bins,
			      const std::vector<double> & foam_mass_bins,
			      bool & found_min_layer_coord,
			      double & min_layer_coord, 
			      bool & found_max_layer_coord,
			      double & max_layer_coord)
  {
  int ret_val = -1;  // Error state

  const unsigned int num_bins = bin_centers.size();

  if (num_bins < 3)
    throw QuickFlash::Except("At least three bins required", __FILE__, __LINE__);

  if ((!(peak_product_fraction > 0.0)) || (!(peak_product_fraction < 1.0)))
    throw QuickFlash::Except("Peak fraction must be between zero and one", 
			   __FILE__, __LINE__);

  if (poly_mass_bins.size() != num_bins)
    throw QuickFlash::Except("Incompatible poly mass array", __FILE__, __LINE__);

  if (chbr_mass_bins.size() != num_bins)
    throw QuickFlash::Except("Incompatible CHBr mass array", __FILE__, __LINE__);

  if (foam_mass_bins.size() != num_bins)
    throw QuickFlash::Except("Incompatible foam mass array", __FILE__, __LINE__);

  // Determine the peak value of mass_(poly+chbr) * mass_foam

  double peak_value = 0.0;  // Positive-definite

  std::vector<double> product_list(num_bins);

  for (unsigned int bin_index = 0; bin_index < num_bins; bin_index++)
    {
    const double poly_mass 
      = poly_mass_bins[bin_index] + chbr_mass_bins[bin_index];

    const double foam_mass = foam_mass_bins[bin_index];

    const double prod = poly_mass * foam_mass;

    product_list[bin_index] = prod;

    if (prod > peak_value)
      peak_value = prod;
    }

  if (peak_value > 0.0)
    {
    const double threshold_value = peak_value * peak_product_fraction;
    
    // Find the location where the product crosses this value from both ends

    double lo_pos_coord = 0.0;
    bool found_lo_pos = false;

    double hi_pos_coord = 0.0;
    bool found_hi_pos = false;

    unsigned int prev_index = 0;
    double prev_next_fract = 0.0;

    if (find_threshold_trans(product_list, threshold_value, prev_index,
			     prev_next_fract, true) >= 0)
      {
      const double prev_coord = bin_centers[prev_index];

      const double next_prev_dist = bin_centers[prev_index + 1] - prev_coord;

      lo_pos_coord = prev_coord + (next_prev_dist * prev_next_fract);

      found_lo_pos = true;
      }

    if (find_threshold_trans(product_list, threshold_value, prev_index,
			     prev_next_fract, false) >= 0)
      {
      const double prev_coord = bin_centers[prev_index];

      const double next_prev_dist = bin_centers[prev_index + 1] - prev_coord;

      hi_pos_coord = prev_coord + (next_prev_dist * prev_next_fract);

      found_hi_pos = true;
      }

    // Check the results

    found_min_layer_coord = found_lo_pos;
    found_max_layer_coord = found_hi_pos;

    if (found_lo_pos)
      min_layer_coord = lo_pos_coord;
    else
      min_layer_coord = bin_centers[0];

    if (found_hi_pos)
      max_layer_coord = hi_pos_coord;
    else
      max_layer_coord = bin_centers[num_bins - 1];

    if (min_layer_coord < max_layer_coord)
      {
      if (found_lo_pos && found_hi_pos)
	ret_val = 0;
      else
	ret_val = 1;  // Possible error
      }
    }

  return ret_val;
  }


int find_threshold_trans(const std::vector<double> & value_list,
			 const double threshold_value,
			 unsigned int & prev_index, 
			 double & thresh_prev_next_fract,
			 const bool forward)
  {
  int ret_val = -1;  // Error state

  const unsigned int num_values = value_list.size();

  if (num_values < 2)
    throw QuickFlash::Except("At least two values required", __FILE__, __LINE__);

  const unsigned int max_index = num_values - 1;

  bool found_transition = false;

  unsigned int lo_index = forward ? 0 : max_index - 1;

  bool index_in_range = true;

  double interval_fraction = 0.0;

  while ((!found_transition) && index_in_range)
    {
    const unsigned int hi_index = lo_index + 1;

    const double lo_value = value_list[lo_index];
    const double hi_value = value_list[hi_index];

    const double threshold_lo_diff = threshold_value - lo_value;
    const double threshold_hi_diff = threshold_value - hi_value;

    const double diff_prod = threshold_lo_diff * threshold_hi_diff;

    if (!(diff_prod > 0.0))
      {
      if (diff_prod < 0.0)
	{
	const double hi_lo_diff = hi_value - lo_value;

	interval_fraction = threshold_lo_diff / hi_lo_diff;
	}
      else if (std::fabs(threshold_lo_diff) > 0.0)
	interval_fraction = 1.0;  // Transition at hi_index
      else if (std::fabs(threshold_hi_diff) > 0.0)
	interval_fraction = 0.0;  // Transition at lo_index
      else
	interval_fraction = 0.5;  // Place half-way

      found_transition = true;
      }

    if (!found_transition)
      {
      // Increment indexes

      if (forward)
	{
	lo_index++;

	if (lo_index >= max_index)
	  index_in_range = false;
	}
      else
	{
	if (lo_index == 0)
	  index_in_range = false;
	else
	  lo_index--;
	}
      }
    }

  if (found_transition)
    {
    prev_index = lo_index;
    thresh_prev_next_fract = interval_fraction;

    ret_val = 0;  // OK
    }

  return ret_val;
  }



	 
