// C++ program file layer_distrib.cpp

/*
  By Nathan C. Hearn
     April 4, 2008

  Program for getting mass distribution of mixing layers.
*/


#include <iostream>
#include <cstdlib>
#include <string>
#include <fstream>
#include <sstream>
#include <algorithm>
#include "quickflash_utils_commandline.hpp"
#include "quickflash_utils_file.hpp"
#include "quickflash_utils_text.hpp"
#include "quickflash_file_datafile.hpp"
#include "quickflash_file_meshinfo.hpp"
#include "quickflash_block_blockinfo.hpp"
#include "quickflash_block_blockdata.hpp"
#include "layer_distribution.hpp"
#include "layer_detection.hpp"
#include "layer_slices.hpp"
#include "shock_front.hpp"


const unsigned int buffer_size = 400;
const unsigned int search_axis = 0;  // x-axis
const unsigned int slice_axis_2d = 1;  // y-axis

const unsigned int slice_dims_3d = 2;

const char default_comment_marker[] = "# ";

const char pres_name[] = "pres";


int main(int argc, char * argv[])
  {
  using std::cerr;
  using std::endl;

  std::vector<std::string> argument_list;
  std::list<QuickFlash::Utils::Option> option_list;

  QuickFlash::Utils::parse_command_line(argc, argv, argument_list, option_list);

  const unsigned int num_args = argument_list.size();

  const unsigned int min_args = 9;

  if (num_args < min_args)
    {
    cerr << endl << "Usage: " << argv[0] << " envelope_dataset "
	 << endl << "\t"
	 << "envelope_min_value envelope_max_value distrib_bin_width"
	 << endl << "\t"
	 << "layer_peak_value_fract min_slice_spacing "
	 << endl << "\t"
	 << "slice_width slice_bin_width"
	 << endl << "\t"
	 << "[ --max-sample-radius=RADIUS ] [ --output-prefix=PREFIX ]"
	 << endl << "\t"
	 << "[ --add-comment-marker[=MARKER] ] [ --max-slices=NUMBER ]"
	 << endl << "\t"
	 << "[ --find-shock-front ]"
	 << endl << "\t"
	 << "datafile_1 [ datafile_2 ... ]" << endl << endl;
    return 1;
    }


  // Read the arguments

  unsigned int argptr = 0;

  const std::string envelope_dataset = argument_list[argptr++];

  const double envelope_min_value = std::atof(argument_list[argptr++].c_str());
  const double envelope_max_value = std::atof(argument_list[argptr++].c_str());

  const double distrib_bin_width = std::atof(argument_list[argptr++].c_str());

  const double layer_peak_value_fract 
    = std::atof(argument_list[argptr++].c_str());

  const double min_slice_spacing = std::atof(argument_list[argptr++].c_str());
  const double slice_width = std::atof(argument_list[argptr++].c_str());

  const double slice_bin_width = std::atof(argument_list[argptr++].c_str());

  std::list<std::string> datafile_list;

  while (argptr < num_args)
    datafile_list.push_back(argument_list[argptr++]);


  // Parse the options

  bool limit_sample_radius = false;
  double max_sample_radius = 0.0;

  bool use_output_prefix = false;
  std::string output_prefix;

  bool add_comment_marker = false;
  std::string comment_marker = "";

  bool limit_num_slices = false;
  unsigned int max_slices = 0;

  bool find_shock_front = false;

  std::list<QuickFlash::Utils::Option>::const_iterator option_iter
    = option_list.begin();

  while (option_iter != option_list.end())
    {
    const QuickFlash::Utils::Option & current_option = *option_iter;

    const std::string & keyword = current_option.get_key();

    if (keyword == "max-sample-radius")
      {
      if (!current_option.value_present())
        {
        cerr << endl << "Error: Max sample radius requires a floating point "
             "argument" << endl << endl;
	return -1;
	}

      current_option.get_value(max_sample_radius);

      if (!(max_sample_radius > 0.0))
	{
	cerr << endl << "Error: Max sample radius must be positive"
	     << endl << endl;
	return -1;
	}

      limit_sample_radius = true;
      }
    else if (keyword == "output-prefix")
      {
      if (!current_option.value_present())
	{
	cerr << endl << "Error: Output prefix requires an argument" << endl
	     << endl;
	return -1;
	}

      output_prefix = current_option.get_value();

      use_output_prefix = true;
      }
    else if (keyword == "add-comment-marker")
      {
      add_comment_marker = true;

      if (current_option.value_present())
	comment_marker = current_option.get_value() + " ";
      else
	comment_marker = default_comment_marker;
      }
    else if (keyword == "max-slices")
      {
      if (!current_option.value_present())
        {
        cerr << endl << "Error: Max slice count requires an integer argument "
             << endl << endl;
	return -1;
	}

      current_option.get_value(max_slices);

      if (max_slices < 1)
	{
	cerr << endl << "Error: Max slice count must be positive"
	     << endl << endl;
	return -1;
	}

      limit_num_slices = true;
      }
    else if (keyword == "find-shock-front")
      {
      find_shock_front = true;
      }
    else
      {
      cerr << endl << "Error: Unrecognized option " << keyword << endl << endl;
      return -1;
      }

    ++option_iter;
    }


  // Perform some checks

  if (!(distrib_bin_width > 0.0))
    {
    cerr << endl << "Error: Distribution bin width must be positive" << endl 
	 << endl;
    return -1;
    }

  if (!(envelope_min_value < envelope_max_value))
    {
    cerr << endl << "Error: Envelope min value must be less than max value"
	 << endl << endl;
    return -1;
    }

  if (!(min_slice_spacing > 0.0))
    {
    cerr << endl << "Error: Slice spacing must be positive" << endl << endl;
    return -1;
    }

  if (!(slice_width > 0.0))
    {
    cerr << endl << "Error: Slice width must be positive" << endl << endl;
    return -1;
    }

  if (!(slice_bin_width > 0.0))
    {
    cerr << endl << "Error: Slice bin width must be positive" << endl << endl;
    return -1;
    }


  // Run through the files

  std::list<std::string>::const_iterator datafile_iter = datafile_list.begin();

  while (datafile_iter != datafile_list.end())
    {
    const std::string & datafile_name = *datafile_iter;

    // Set up the filenames

    std::string datafile_path;
    std::string unqualified_datafile_name;

    QuickFlash::Utils::split_name_path_file(datafile_name, datafile_path,
					  unqualified_datafile_name);

    std::string output_filename_prefix = "";
 
    if (use_output_prefix)
      output_filename_prefix += output_prefix;

    output_filename_prefix += unqualified_datafile_name;

    std::string output_filename_path = "";
    std::string unqualified_output_filename_prefix = "";

    QuickFlash::Utils::split_name_path_file(output_filename_prefix, 
					  output_filename_path,
					  unqualified_output_filename_prefix);


    // Open the file

    cerr << "** Processing file " << unqualified_datafile_name << endl;

    QuickFlash::File::DataFile df(datafile_name, buffer_size, true);

    const QuickFlash::File::MeshInfo & meshinfo = df.get_mesh_info();

    const std::vector<double> & max_bounds = meshinfo.get_volume_maxbounds();

    const double sim_time = meshinfo.get_sim_time();

    const unsigned int file_dims = meshinfo.get_dims();

    if ((file_dims != 2) && (file_dims != 3))
      {
      cerr << endl << "Error: Only 2D and 3D datasets supported" << endl 
	   << endl;
      return -2;
      }


    // Open the metadata file

    const std::string meta_filename 
      = output_filename_prefix + "_metadata.out";

    std::ofstream meta(meta_filename.c_str(), std::ios::trunc);

    meta << "DataFile " << unqualified_datafile_name << endl;

    meta << "SimTime " << sim_time << endl;
    meta << "Dimensions " << file_dims << endl;

    meta << "EnvelopeDataset " << envelope_dataset << endl;

    if (limit_sample_radius)
      meta << "MaxSampleRadius " << max_sample_radius << endl;

    meta << "DistribBinWidth " << distrib_bin_width << endl;

    if (limit_num_slices)
      meta << "MaxSlices " << max_slices << endl;


    // Find the envelope size

    double min_envelope_pos = 0.0;
    double max_envelope_pos = 0.0;

    if (find_maximal_envelope(df, envelope_dataset, search_axis,
			      envelope_min_value, envelope_max_value,
			      min_envelope_pos, max_envelope_pos,
			      limit_sample_radius, max_sample_radius) < 0)
      {
      cerr << endl << "Error: Unable to find envelope" << endl << endl;
      return -2;
      }

    // Pad the envelope with twice distrib_bin_width

    min_envelope_pos -= 2.0 * distrib_bin_width;
    max_envelope_pos += 2.0 * distrib_bin_width;

    meta << "MinEnvelopePos " << min_envelope_pos << endl;
    meta << "MaxEnvelopePos " << max_envelope_pos << endl;

    const double envelope_width = max_envelope_pos - min_envelope_pos;

    meta << "EnvelopeWidth " << envelope_width << endl;


    // Get the mass distributions

    LayerDistribution layer_distrib(df, search_axis, distrib_bin_width,
				    min_envelope_pos, max_envelope_pos,
				    limit_sample_radius, max_sample_radius);

    const double total_mass = layer_distrib.get_total_mass();
    const double poly_mass = layer_distrib.get_poly_mass();
    const double chbr_mass = layer_distrib.get_chbr_mass();
    const double foam_mass = layer_distrib.get_foam_mass();

    const double min_array_coord = layer_distrib.get_array_min_coord();
    const double max_array_coord = layer_distrib.get_array_max_coord();

    meta << "MinArrayPos " << min_array_coord << endl;
    meta << "MaxArrayPos " << max_array_coord << endl;

    meta << "ArrayTotalMass " << total_mass << endl;
    meta << "ArrayPolyMass " << poly_mass << endl;
    meta << "ArrayCHBrMass " << chbr_mass << endl;
    meta << "ArrayFoamMass " << foam_mass << endl;

    const unsigned int num_bins = layer_distrib.get_num_bins();

    const std::vector<double> & bin_centers = layer_distrib.get_bin_centers();

    const std::vector<double> & total_mass_bins 
      = layer_distrib.get_total_mass_bins();

    const std::vector<double> & poly_mass_bins 
      = layer_distrib.get_poly_mass_bins();

    const std::vector<double> & chbr_mass_bins 
      = layer_distrib.get_chbr_mass_bins();

    const std::vector<double> & foam_mass_bins 
      = layer_distrib.get_foam_mass_bins();


    // Report on the results

    const std::string distf_filename = output_filename_prefix + "_distrib.csv";
    const std::string unqualified_distf_filename 
      = unqualified_output_filename_prefix + "_distrib.csv";

    meta << "LayerDistribFile " << unqualified_distf_filename << endl;

    std::ofstream distf(distf_filename.c_str(), std::ios::trunc);

    distf << comment_marker << "DataFile " << unqualified_datafile_name 
	  << endl;
    distf << comment_marker << "EnvelopeDataset " << envelope_dataset << endl;

    if (limit_sample_radius)
      distf << comment_marker << "MaxSampleRadius " << max_sample_radius 
	    << endl;

    distf << comment_marker << "BinWidth " << distrib_bin_width << endl;

    distf << comment_marker << "MinEnvelopePos " << min_envelope_pos << endl;
    distf << comment_marker << "MaxEnvelopePos " << max_envelope_pos << endl;

    distf << comment_marker << "MinArrayPos " << min_array_coord << endl;
    distf << comment_marker << "MaxArrayPos " << max_array_coord << endl;

    distf << comment_marker << "ArrayTotalMass " << total_mass << endl;
    distf << comment_marker << "ArrayPolyMass " << poly_mass << endl;
    distf << comment_marker << "ArrayCHBrMass " << chbr_mass << endl;
    distf << comment_marker << "ArrayFoamMass " << foam_mass << endl;

    distf << endl;

    distf << comment_marker << "BinData" << endl;

    distf << endl;

    distf << "Bin_Center" << "\t" << "Bin_Mass" << "\t" << "Bin_Poly_Mass" 
	  << "\t" << "Bin_CHBr_Mass" << "\t" << "Bin_Foam_Mass" << endl;

    for (unsigned int bin_index = 0; bin_index < num_bins; bin_index++)
      distf << bin_centers[bin_index] << "\t" << total_mass_bins[bin_index]
	<< "\t" << poly_mass_bins[bin_index] 
	<< "\t" << chbr_mass_bins[bin_index]
	<< "\t" << foam_mass_bins[bin_index] << endl;

    distf << endl;


    // Find the layer boundaries

    double min_layer_bound = 0.0;
    double max_layer_bound = 0.0;

    bool found_min_layer_bound = false;
    bool found_max_layer_bound = false;

    if (find_layer_bounds_product(bin_centers, layer_peak_value_fract,
				  poly_mass_bins, chbr_mass_bins,
				  foam_mass_bins, found_min_layer_bound,
				  min_layer_bound, found_max_layer_bound,
				  max_layer_bound) < 0)
      {
      meta << "FoundLayerMin false" << endl;
      meta << "FoundLayerMax false" << endl;

      // Use envelope positions

      min_layer_bound = min_envelope_pos;
      max_layer_bound = max_envelope_pos;
      }
    else
      {
      if (found_min_layer_bound)
	meta << "FoundLayerMin true" << endl;
      else
	meta << "FoundLayerMin false" << endl;

      if (found_max_layer_bound)
	meta << "FoundLayerMax true" << endl;
      else
	meta << "FoundLayerMax false" << endl;
      }

    meta << "LayerPeakValueFract " << layer_peak_value_fract << endl;

    meta << "LayerMinPos " << min_layer_bound << endl;
    meta << "LayerMaxPos " << max_layer_bound << endl;

    const double layer_width = max_layer_bound - min_layer_bound;

    meta << "LayerWidth " << layer_width << endl;

    
    // Find the shock front

    if (find_shock_front)
      {
      double shock_front_pos = 0.0;

      bool found_shock_front = false;

      if (find_axis_peak(df, pres_name, search_axis, shock_front_pos,
			 limit_sample_radius, max_sample_radius,
			 true, max_layer_bound, max_bounds[search_axis]) >= 0)
	found_shock_front = true;

      if (found_shock_front)
	{
	meta << "FoundShockFront true" << endl;
	meta << "ShockFrontPos " << shock_front_pos << endl;
	}
      else
	meta << "FoundShockFront false" << endl;
      }


    // Determine the proper slice separation

    double slice_spacing = min_slice_spacing;
    
    if (limit_num_slices)
      {
      const double limited_slice_spacing 
	= layer_width / static_cast<double>(max_slices);

      if (limited_slice_spacing > slice_spacing)
	slice_spacing = limited_slice_spacing;
      }

    meta << "SliceSpacing " << slice_spacing << endl;
    meta << "SliceWidth " << slice_width << endl;
    meta << "SliceBinWidth " << slice_bin_width << endl;

    const unsigned int unlimited_num_slices 
      = static_cast<unsigned int>(layer_width / slice_spacing) + 1;

    const unsigned int num_slices 
      = limit_num_slices 
        ? std::max(unlimited_num_slices, max_slices) : unlimited_num_slices;


    // Perform the slices

    double current_slice_center = min_layer_bound + (0.5 * slice_spacing);

    for (unsigned int slice_index = 0; slice_index < num_slices; slice_index++)
      {
      std::stringstream sstream;

      sstream << slice_index;

      if (file_dims == 2)
	{
	const std::string slice_filename_suffix 
	  = "_slice_" + sstream.str() + ".csv";

	const std::string slice_filename 
	  = output_filename_prefix + slice_filename_suffix;

	const std::string unqualified_slice_filename 
	  = unqualified_output_filename_prefix + slice_filename_suffix;

	meta << "Slice1D " << slice_index << " " << current_slice_center 
	     << " " << unqualified_slice_filename << endl;

	std::ofstream slicef(slice_filename.c_str(), std::ios::trunc);

	slicef << comment_marker << "DataFile " << unqualified_datafile_name
	       << endl;

	slicef << comment_marker << "DataFileDims " << file_dims << endl;

	slicef << comment_marker << "SliceCenter " << current_slice_center
	       << endl;
	slicef << comment_marker << "SliceWidth " << slice_width << endl;
	slicef << comment_marker << "SliceBinWidth " << slice_bin_width
	       << endl;

	LayerSlice1D slice(df, search_axis, slice_axis_2d,
			   current_slice_center, slice_width, slice_bin_width,
			   limit_sample_radius, max_sample_radius);

	slicef << comment_marker << "SliceMass " << slice.get_total_mass()
	       << endl;

	slicef << comment_marker << "SlicePolyMass " << slice.get_poly_mass()
	       << endl;

	slicef << comment_marker << "SliceCHBrMass " << slice.get_chbr_mass()
	       << endl;

	slicef << comment_marker << "SliceFoamMass " << slice.get_foam_mass()
	       << endl;

	slicef << comment_marker << "SliceData" << endl;

	const unsigned int num_slice_bins = slice.get_num_bins();

	const std::vector<double> & bin_centers = slice.get_bin_centers();

	const std::vector<double> & total_mass_bins 
	  = slice.get_total_mass_bins();

	const std::vector<double> & poly_mass_bins
	  = slice.get_poly_mass_bins();

	const std::vector<double> & chbr_mass_bins
	  = slice.get_chbr_mass_bins();

	const std::vector<double> & foam_mass_bins
	  = slice.get_foam_mass_bins();

	slicef << "Bin_Center" << "\t" << "Bin_Mass" << "\t" 
	       << "Bin_Poly_Mass" << "\t" << "Bin_CHBr_Mass" << "\t" 
	       << "Bin_Foam_Mass" << endl;

	for (unsigned int bin_index = 0; bin_index < num_slice_bins; 
	     bin_index++)
	  slicef << bin_centers[bin_index] << "\t"
		 << total_mass_bins[bin_index] << "\t"
		 << poly_mass_bins[bin_index] << "\t"
		 << chbr_mass_bins[bin_index] << "\t"
		 << foam_mass_bins[bin_index] << endl;

	slicef << endl;
	}
      else
	{	
	const std::string slice_filename_suffix 
	  = "_slice_" + sstream.str() + ".csv";

	const std::string slice_total_mass_filename 
	  = output_filename_prefix + "_mass" + slice_filename_suffix;

	const std::string unqualified_slice_total_mass_filename 
	  = unqualified_output_filename_prefix + "_mass" 
	    + slice_filename_suffix;

	const std::string slice_poly_mass_filename 
	  = output_filename_prefix + "_poly_mass" + slice_filename_suffix;

	const std::string unqualified_slice_poly_mass_filename 
	  = unqualified_output_filename_prefix + "_poly_mass" 
	    + slice_filename_suffix;

	const std::string slice_chbr_mass_filename 
	  = output_filename_prefix + "_chbr_mass" + slice_filename_suffix;

	const std::string unqualified_slice_chbr_mass_filename 
	  = unqualified_output_filename_prefix + "_chbr_mass" 
	    + slice_filename_suffix;

	const std::string slice_foam_mass_filename 
	  = output_filename_prefix + "_foam_mass" + slice_filename_suffix;

	const std::string unqualified_slice_foam_mass_filename 
	  = unqualified_output_filename_prefix + "_foam_mass" 
	    + slice_filename_suffix;


	meta << "Slice2D " << slice_index << " " << current_slice_center 
	     << " Total_Mass " << unqualified_slice_total_mass_filename 
	     << endl;

	meta << "Slice2D " << slice_index << " " << current_slice_center 
	     << " Poly_Mass " << unqualified_slice_poly_mass_filename 
	     << endl;

	meta << "Slice2D " << slice_index << " " << current_slice_center 
	     << " CHBr_Mass " << unqualified_slice_chbr_mass_filename 
	     << endl;

	meta << "Slice2D " << slice_index << " " << current_slice_center 
	     << " Foam_Mass " << unqualified_slice_foam_mass_filename 
	     << endl;


	std::ofstream stotalf(slice_total_mass_filename.c_str(), 
			      std::ios::trunc);

	stotalf << comment_marker << "Slice Total_Mass" << endl;

	stotalf << comment_marker << "DataFile " << unqualified_datafile_name
		<< endl;

	stotalf << comment_marker << "SimTime " << sim_time << endl;

	stotalf << comment_marker << "DataFileDims " << file_dims << endl;

	stotalf << comment_marker << "SliceCenter " << current_slice_center
		<< endl;
	stotalf << comment_marker << "SliceWidth " << slice_width << endl;
	stotalf << comment_marker << "SliceBinWidth " << slice_bin_width
		<< endl;


	std::ofstream spolyf(slice_poly_mass_filename.c_str(), 
			     std::ios::trunc);

	spolyf << comment_marker << "Slice Poly_Mass" << endl;

	spolyf << comment_marker << "DataFile " << unqualified_datafile_name
	       << endl;

	spolyf << comment_marker << "SimTime " << sim_time << endl;

	spolyf << comment_marker << "DataFileDims " << file_dims << endl;

	spolyf << comment_marker << "SliceCenter " << current_slice_center
	       << endl;
	spolyf << comment_marker << "SliceWidth " << slice_width << endl;
	spolyf << comment_marker << "SliceBinWidth " << slice_bin_width
	       << endl;


	std::ofstream schbrf(slice_chbr_mass_filename.c_str(), 
			     std::ios::trunc);

	schbrf << comment_marker << "Slice CHBr_Mass" << endl;

	schbrf << comment_marker << "DataFile " << unqualified_datafile_name
	       << endl;

	schbrf << comment_marker << "SimTime " << sim_time << endl;

	schbrf << comment_marker << "DataFileDims " << file_dims << endl;

	schbrf << comment_marker << "SliceCenter " << current_slice_center
	       << endl;
	schbrf << comment_marker << "SliceWidth " << slice_width << endl;
	schbrf << comment_marker << "SliceBinWidth " << slice_bin_width
	       << endl;


	std::ofstream sfoamf(slice_foam_mass_filename.c_str(), 
			     std::ios::trunc);

	sfoamf << comment_marker << "Slice Foam_Mass" << endl;

	sfoamf << comment_marker << "DataFile " << unqualified_datafile_name
	       << endl;

	sfoamf << comment_marker << "SimTime " << sim_time << endl;

	sfoamf << comment_marker << "DataFileDims " << file_dims << endl;

	sfoamf << comment_marker << "SliceCenter " << current_slice_center
	       << endl;
	sfoamf << comment_marker << "SliceWidth " << slice_width << endl;
	sfoamf << comment_marker << "SliceBinWidth " << slice_bin_width
	       << endl;


	std::vector<double> accum_bin_width(slice_dims_3d, slice_bin_width);

	LayerSlice2D slice(df, search_axis, current_slice_center, slice_width,
			   accum_bin_width, limit_sample_radius,
			   max_sample_radius);

	const QuickFlash::Block::BlockInfo & grid_info = slice.get_grid_info();

	const std::vector<unsigned int> & grid_dims 
	  = grid_info.get_block_dims();

	stotalf << comment_marker << "GridDims " << grid_dims << endl;
	spolyf << comment_marker << "GridDims " << grid_dims << endl;
	schbrf << comment_marker << "GridDims " << grid_dims << endl;
	sfoamf << comment_marker << "GridDims " << grid_dims << endl;

	const std::vector<double> & grid_min_bounds 
	  = grid_info.get_min_coords();

	const std::vector<double> & grid_max_bounds 
	  = grid_info.get_max_coords();

	stotalf << comment_marker << "GridMinBounds " << grid_min_bounds 
		<< endl;
	stotalf << comment_marker << "GridMaxBounds " << grid_max_bounds 
		<< endl;

	spolyf << comment_marker << "GridMinBounds " << grid_min_bounds 
	       << endl;
	spolyf << comment_marker << "GridMaxBounds " << grid_max_bounds 
	       << endl;

	schbrf << comment_marker << "GridMinBounds " << grid_min_bounds 
	       << endl;
	schbrf << comment_marker << "GridMaxBounds " << grid_max_bounds 
	       << endl;

	sfoamf << comment_marker << "GridMinBounds " << grid_min_bounds 
	       << endl;
	sfoamf << comment_marker << "GridMaxBounds " << grid_max_bounds 
	       << endl;

	// Write out the data

	stotalf << comment_marker << "SliceData" << endl;
	spolyf << comment_marker << "SliceData" << endl;
	schbrf << comment_marker << "SliceData" << endl;
	sfoamf << comment_marker << "SliceData" << endl;

	const QuickFlash::Block::BlockData<double> & total_mass_grid 
	  = slice.get_total_mass_bins();

	const QuickFlash::Block::BlockData<double> & poly_mass_grid 
	  = slice.get_poly_mass_bins();

	const QuickFlash::Block::BlockData<double> & chbr_mass_grid 
	  = slice.get_chbr_mass_bins();

	const QuickFlash::Block::BlockData<double> & foam_mass_grid 
	  = slice.get_foam_mass_bins();

	const unsigned int i_dims = grid_dims[0];
	const unsigned int j_dims = grid_dims[1];

	std::vector<unsigned int> pixel_index(2);

	for (unsigned int i = 0; i < i_dims; i++)
	  {
	  pixel_index[0] = i;

	  for (unsigned int j = 0; j < j_dims; j++)
	    {
	    pixel_index[1] = j;

	    const double total_mass_value = total_mass_grid[pixel_index];
	    const double poly_mass_value = poly_mass_grid[pixel_index];
	    const double chbr_mass_value = chbr_mass_grid[pixel_index];
	    const double foam_mass_value = foam_mass_grid[pixel_index];

	    stotalf << total_mass_value << " ";
	    spolyf << poly_mass_value << " ";
	    schbrf << chbr_mass_value << " ";
	    sfoamf << foam_mass_value << " ";
	    }

	  stotalf << endl;
	  spolyf << endl;
	  schbrf << endl;
	  sfoamf << endl;
	  }

	stotalf << endl;
	spolyf << endl;
	schbrf << endl;
	sfoamf << endl;
	}

      current_slice_center += slice_spacing;
      }

    ++datafile_iter;
    }

  return 0;
  }


    

    

    
