// C++ header file layer_slices.hpp

/*
  By Nathan C. Hearn
     April 5, 2008

  All-inclusive 2D/3D data slicing tools for Omega laser experiment 
  simulations.
*/


#ifndef LAYER_SLICES_HPP
#define LAYER_SLICES_HPP


#include <vector>
#include "quickflash_file_datafile.hpp"
#include "quickflash_block_blockinfo.hpp"
#include "quickflash_block_blockdata.hpp"


class LayerSlice1D
  {
  public :

    LayerSlice1D();

    LayerSlice1D(const QuickFlash::File::DataFile & df,
		 const unsigned int search_axis,
		 const unsigned int slice_axis,
		 const double search_axis_center_coord,
		 const double slice_width,
		 const double accum_bin_width,
		 const bool limit_sample_radius=false,
		 const double max_sample_radius=0.0);

    LayerSlice1D(const LayerSlice1D & source);

    ~LayerSlice1D() { }

    LayerSlice1D & operator=(const LayerSlice1D & source)
      {
      reset(source);
      return *this;
      }

    void reset();

    void reset(const QuickFlash::File::DataFile & df,
	       const unsigned int search_axis,
	       const unsigned int slice_axis,
	       const double search_axis_center_coord,
	       const double slice_width,
	       const double accum_bin_width,
	       const bool limit_sample_radius=false,
	       const double max_sample_radius=0.0);

    void reset(const LayerSlice1D & source);

    double get_total_mass() const { return total_mass; }
    double get_poly_mass() const { return poly_mass; }
    double get_chbr_mass() const { return chbr_mass; }
    double get_foam_mass() const { return foam_mass; }

    double get_array_min_coord() const { return bin_array_min_coord; }
    double get_array_max_coord() const { return bin_array_max_coord; }

    unsigned int get_num_bins() const { return num_bins; }

    const std::vector<double> & get_bin_centers() const { return bin_centers; }

    const std::vector<double> & get_total_mass_bins() const 
      { return total_mass_bins; }

    const std::vector<double> & get_poly_mass_bins() const 
      { return poly_mass_bins; }

    const std::vector<double> & get_chbr_mass_bins() const 
      { return chbr_mass_bins; }

    const std::vector<double> & get_foam_mass_bins() const 
      { return foam_mass_bins; }

    const std::vector<double> & get_total_dens_variance_bins() const 
      { return total_dens_variance_bins; }

    const std::vector<double> & get_poly_dens_variance_bins() const 
      { return poly_dens_variance_bins; }

    const std::vector<double> & get_chbr_dens_variance_bins() const 
      { return chbr_dens_variance_bins; }

    const std::vector<double> & get_foam_dens_variance_bins() const 
      { return foam_dens_variance_bins; }

  private :

    double total_mass;
    double poly_mass;
    double chbr_mass;
    double foam_mass;

    double bin_array_min_coord;
    double bin_array_max_coord;

    unsigned int num_bins;

    std::vector<double> bin_centers;

    std::vector<double> total_mass_bins;
    std::vector<double> poly_mass_bins;
    std::vector<double> chbr_mass_bins;
    std::vector<double> foam_mass_bins;

    std::vector<double> total_dens_variance_bins;
    std::vector<double> poly_dens_variance_bins;
    std::vector<double> chbr_dens_variance_bins;
    std::vector<double> foam_dens_variance_bins;
  };


// Class LayerSlice2D

class LayerSlice2D
  {
  public :

    LayerSlice2D();

    LayerSlice2D(const QuickFlash::File::DataFile & df, 
		 const unsigned int search_axis,
		 const double search_axis_center_coord,
		 const double slice_width,
		 const std::vector<double> & accum_bin_width,
		 const bool limit_sample_radius=false,
		 const double max_sample_radius=0.0);

    LayerSlice2D(const LayerSlice2D & source);

    ~LayerSlice2D() { }

    LayerSlice2D & operator=(const LayerSlice2D & source)
      {
      reset(source);
      return *this;
      }

    void reset();

    void reset(const QuickFlash::File::DataFile & df, 
	       const unsigned int search_axis,
	       const double search_axis_center_coord,
	       const double slice_width,
	       const std::vector<double> & accum_bin_width,
	       const bool limit_sample_radius=false,
	       const double max_sample_radius=0.0);

    void reset(const LayerSlice2D & source);

    double get_total_mass() const { return total_mass; }
    double get_poly_mass() const { return poly_mass; }
    double get_chbr_mass() const { return chbr_mass; }
    double get_foam_mass() const { return foam_mass; }

    const std::vector<double> get_grid_min_coords() const 
      { return grid_info.get_min_coords(); }

    const std::vector<double> get_grid_max_coords() const 
      { return grid_info.get_max_coords(); }

    unsigned int get_num_bins() const { return grid_info.get_num_cells(); }

    const QuickFlash::Block::BlockInfo & get_grid_info() const 
      { return grid_info; }

    const std::vector<unsigned int> & get_slice_to_file_index_map() const
      { return slice_to_file_index_map; }

    const QuickFlash::Block::BlockData<double> & get_total_mass_bins() const 
      { return total_mass_bins; }

    const QuickFlash::Block::BlockData<double> & get_poly_mass_bins() const 
      { return poly_mass_bins; }

    const QuickFlash::Block::BlockData<double> & get_chbr_mass_bins() const 
      { return chbr_mass_bins; }

    const QuickFlash::Block::BlockData<double> & get_foam_mass_bins() const 
      { return foam_mass_bins; }


  private :

    double total_mass;
    double poly_mass;
    double chbr_mass;
    double foam_mass;

    QuickFlash::Block::BlockInfo grid_info;

    std::vector<unsigned int> slice_to_file_index_map;

    QuickFlash::Block::BlockData<double> total_mass_bins;
    QuickFlash::Block::BlockData<double> poly_mass_bins;
    QuickFlash::Block::BlockData<double> chbr_mass_bins;
    QuickFlash::Block::BlockData<double> foam_mass_bins;
  };


#endif  // LAYER_SLICES_HPP
