// C++ program file part_data_extractor.cpp

/*
  By Nathan C. Hearn
     November 21, 2008

  Class for handling particle extraction from multiple files.
*/


#include "part_data_extractor.hpp"
#include <vector>
#include <string>
#include "quickflash_except.hpp"
#include "quickflash_file_datafile.hpp"
#include "quickflash_particles_data.hpp"


// Class PartDataExtractor


PartDataExtractor::PartDataExtractor() :
  num_files(0), data_file_list(), num_elems(0), file_elem_name_indexes()
  { }


PartDataExtractor::PartDataExtractor(const std::vector<std::string> 
				       & file_name_list,
				     const std::vector<std::string> 
				       & elem_name_list) :
  num_files(0), data_file_list(), num_elems(0), file_elem_name_indexes()
  { reset(file_name_list, elem_name_list); }


void PartDataExtractor::reset()
  {
  // Delete the pointers

  for (unsigned int file_index = 0; file_index < num_files; file_index++)
    {
    delete data_file_list[file_index];
    data_file_list[file_index] = 0;
    }

  num_files = 0;

  data_file_list.clear();

  num_elems = 0;

  file_elem_name_indexes.clear();
  }


void PartDataExtractor::reset(const std::vector<std::string> & file_name_list,
			      const std::vector<std::string> & elem_name_list)
  {
  reset();

  const unsigned int local_num_files = file_name_list.size();

  if (local_num_files > 0)
    {
    data_file_list.resize(local_num_files, 0);

    for (unsigned int file_index = 0; file_index < local_num_files; 
	 file_index++)
      {
      const std::string & filename = file_name_list[file_index];

      data_file_list[file_index] 
	= new QuickFlash::File::DataFile(filename, 0, false, true);
      }

    num_files = local_num_files;

    setup_elems(elem_name_list);
    }
  }


void PartDataExtractor::add_part_data(const unsigned int particle_tag, 
				      std::vector<double> & times,
				      std::vector< std::vector<double> > 
				        & elem_data) const
  {
  if (elem_data.size() == 0)
    elem_data.resize(num_elems);
  else if (elem_data.size() != num_elems)
    throw QuickFlash::Except("Element data vector size mismatch", __FILE__,
			     __LINE__);

  std::vector<double> full_part_data;

  for (unsigned int file_index = 0; file_index < num_files; file_index++)
    {
    const QuickFlash::File::DataFile & current_file 
      = *(data_file_list[file_index]);

    const QuickFlash::Particles::PartData & part_data 
      = current_file.get_part_data();

    if (part_data.particle_present(particle_tag))
      {
      const double sim_time = current_file.get_sim_info().get_sim_time();

      times.push_back(sim_time);

      part_data.get_particle_data_tag(particle_tag, full_part_data);

      const std::vector<unsigned int> & elem_map 
	= file_elem_name_indexes[file_index];

      for (unsigned int elem_index = 0; elem_index < num_elems; elem_index++)
	{
	const unsigned int local_elem_index = elem_map[elem_index];

	const double elem_value = full_part_data[local_elem_index];

	elem_data[elem_index].push_back(elem_value);
	}
      }
    }
  }


void PartDataExtractor::setup_elems(const std::vector<std::string> 
				      & elem_name_list)
  {
  // NOTE: Set up data files in data_file_list before calling this function

  file_elem_name_indexes.resize(num_files);

  num_elems = elem_name_list.size();

  for (unsigned int file_index = 0; file_index < num_files; file_index++)
    {
    std::vector<unsigned int> & elem_index_array 
      = file_elem_name_indexes[file_index];

    elem_index_array.resize(num_elems, 0);

    const QuickFlash::Particles::PartData & part_data 
      = data_file_list[file_index]->get_part_data();

    for (unsigned int elem_index = 0; elem_index < num_elems; elem_index++)
      {
      const std::string & elem_name = elem_name_list[elem_index];

      if (!(part_data.variable_present(elem_name)))
	throw QuickFlash::Except("Unable to find particle data element",
				 __FILE__, __LINE__);

      elem_index_array[elem_index] = part_data.get_variable_index(elem_name);
      }
    }
  }


  



