#!/usr/bin/env python

# drawimage.py

# By Nathan C. Hearn
#    November 6, 2006

try :
    import psyco
except ImportError :
    pass  # Oh, well ...

numeric_zeros = None
numeric_int = None
numeric_float = None

try :
    import numpy
    numeric_zeros = numpy.zeros
    numeric_int = numpy.int
    numeric_float = numpy.float
except ImportError :
    import Numeric
    numeric_zeros = Numeric.zeros
    numeric_int = Numeric.Int
    numeric_float = Numeric.Float


png_suffix = str(".png")
csv_suffix = str(".csv")


def remove_comments(line) :

    # Remove comment characters "#" and all following characters

    clean_line = line.split("#")[0]

    return clean_line


def read_file(filename) :

    # Numeric package already imported

    rows = list()

    f = file(filename, "r")

    file_lines = f.readlines()

    cols_counted = False

    num_cols = int(0)

    for line in file_lines :

        clean_line = remove_comments(line)
    
        words = clean_line.split()

        if (len(words) > 0) :
            
            cols = list()

            for num in words :
                cols.append(float(num))

            if (cols_counted) :
                if (len(cols) != num_cols) :
                    raise ValueError("Incorrect number of columns")
            else :
                num_cols = len(cols)
                cols_counted = True

            rows.append(cols)

    num_rows = len(rows)

    data = numeric_zeros((num_rows, num_cols), numeric_float)

    for i in range(num_rows) :

        row_list = rows[i]
        
        for j in range(num_cols) :

            data[i][j] = row_list[j]

    return data


def compute_symmetry(symmetry_axis, data) :

    if ((symmetry_axis < 0) or (symmetry_axis > 1)) :
        raise ValueError("Improper symmetry axis")

    # Numeric package already imported

    size = data.shape

    if (len(size) != 2) :
        raise RuntimeError("Data must be two-dimensional")

    m_size = size[0]
    n_size = size[1]

    max_m = m_size - 1
    max_n = n_size - 1

    sum_data = numeric_zeros(size, numeric_float)
    diff_data = numeric_zeros(size, numeric_float)

    for m in range(m_size) :
        for n in range(n_size) :

            symm_m = m
            symm_n = n

            if (symmetry_axis == 0) :
                symm_n = max_n - n
            else :
                symm_m = max_m - m

            primary_value = data[m][n]
            reflect_value = data[symm_m][symm_n]

            sum_data[m][n] = 0.5 * (primary_value + reflect_value)
            diff_data[m][n] = 0.5 * (primary_value - reflect_value)

    return sum_data, diff_data


def save_data(data, filename=None, comment=None) :

    from sys import stdout

    size = data.shape

    if (len(size) != 2) :
        raise RuntimeError("Data must be two-dimensional")

    m_size = size[0]
    n_size = size[1]

    output = stdout

    if (filename is not None) :
        output = file(filename, "w")

    if (comment is not None) :
        output.write("# " + comment + "\n")

    for m in range(m_size) :
        
        for n in range(n_size) :
            output.write(str(data[m][n]) + " ")

        output.write("\n")


def save_image(filename, data, min_value=None, max_value=None) :

    from PIL import Image

    size = data.shape

    if (len(size) != 2) :
        raise ValueError("Data must be two-dimensional")

    i_size = size[0]
    j_size = size[1]

    # Find the max and min data values

    image_min_val = min_value
    image_max_val = max_value

    if ((image_min_val is None) or (image_max_val is None)) :

        data_max_val = data[0][0]
        data_min_val = data_max_val

        for i in range(i_size) :
            for j in range(j_size) :
            
                val = data[i][j]

                if (val < data_min_val) :
                    data_min_val = val
                elif (val > data_max_val) :
                    data_max_val = val

        if (image_min_val is None) :
            image_min_val = data_min_val

        if (image_max_val is None) :
            image_max_val = data_max_val

    if (not (image_min_val < image_max_val)) :
        raise ValueError("Error: Min value -- max value mismatch")

    # Store the data in an array

    pixel_count = i_size * j_size

    pixel_array = numeric_zeros(pixel_count, numeric_int)

    offset = image_min_val
    factor = 255.0 / (image_max_val - image_min_val)

    ptr = int(0)

    for i in range(i_size) :
        for j in range(j_size) :
            
            pixel = int((data[i][j] - offset) * factor)

            if (pixel < 0) :
                pixel = 0
            elif (pixel > 255) :
                pixel = 255

            pixel_array[ptr] = pixel
            ptr += 1

    new_im = Image.new("L", (j_size, i_size))  # Transpose back

    new_im.putdata(pixel_array)

    new_im.save(filename)


def main() :

    from optparse import OptionParser
    from sys import exit, stderr

    # Set up the command line parser

    usage = str("%prog input_file [options]")
    version = str("%prog 0.1")

    argparser = OptionParser(usage, version=version)

    output_prefix_str = str("output-prefix")
    argparser.add_option("--" + output_prefix_str, type="string",
                         dest="output_prefix", help="Output file prefix")
    argparser.set_defaults(output_prefix=None)

    min_value_str = str("min-value")
    argparser.add_option("--" + min_value_str, type="float",
                         dest="min_value", help="Image minimum data value")
    argparser.set_defaults(min_value=None)

    max_value_str = str("max-value")
    argparser.add_option("--" + max_value_str, type="float",
                         dest="max_value", help="Image maximum data value")
    argparser.set_defaults(max_value=None)

    compute_symmetry_str = str("compute-symmetry")
    argparser.add_option("--" + compute_symmetry_str, type="int",
                         dest="symmetry_axis",
                         help="Specify symmetry axis (0 or 1)")
    argparser.set_defaults(symmetry_axis=None)

    write_data_str = str("write-data")
    argparser.add_option("--" + write_data_str, action="store_true",
                         dest="write_data",
                         help="Write ASCII data instead of image")
    argparser.set_defaults(write_data=False)
    

    # Process the command line input

    opts, args = argparser.parse_args()

    output_prefix = opts.output_prefix

    min_value = opts.min_value
    max_value = opts.max_value

    symmetry_axis = opts.symmetry_axis

    write_data = opts.write_data


    min_args = int(1)

    argc = len(args)

    if (argc < min_args) :
        stderr.write("Error: Improper number of required arguments "
                     + "(see --help)\n\n")
        exit(-1)

    argptr = int(0)

    input_file = str(args[argptr])
    argptr += 1


    # Perform some basic checks

    if ((output_prefix is None) and (not write_data)) :
        stderr.write("Error: Image output requires an output prefix\n\n")
        exit(-1)


    if ((min_value is not None) and (max_value is not None)) :
        if (not (min_value < max_value)) :
            stderr.write("Error: Min value [ " + str(min_value)
                         + " ] must be less than max [ " + str(max_value)
                         + " ]\n\n")
            exit(-1)


    # Read the data

    data = read_file(input_file)


    # Output the image

    if (not write_data) :
        output_file = output_prefix + png_suffix
        
        save_image(output_file, data, min_value, max_value)
    else :
        output_file = None

        if (output_prefix is not None) :
            output_file = output_prefix + csv_suffix

        file_comment = "Data from file [ " + input_file + " ]"

        save_data(data, filename=output_file, comment=file_comment)


    # Output symmetry information

    if (symmetry_axis is not None) :

        sum_data, diff_data = compute_symmetry(symmetry_axis, data)

        if (write_data) :

            sum_filename = None
            diff_filename = None

            if (output_prefix is not None) :
                sum_filename = output_prefix + "_sum" + csv_suffix
                diff_filename = output_prefix + "_diff" + csv_suffix
            
            sum_file_comment = "Sum data from file [ " + input_file \
                               + " ] with symmetry axis [ " \
                               + str(symmetry_axis) + " ]"

            diff_file_comment = "Difference data from file [ " + input_file \
                                + " ] with symmetry axis [ " \
                                + str(symmetry_axis) + " ]"

            save_data(sum_data, filename=sum_filename,
                      comment=sum_file_comment)
            save_data(diff_data, filename=diff_filename,
                      comment=diff_file_comment)

        else :

            sum_filename = output_prefix + "_sum" + png_suffix
            diff_filename = output_prefix + "_diff" + png_suffix

            save_image(sum_filename, sum_data)
            save_image(diff_filename, diff_data)


if (__name__ == "__main__") :
    main()
