// C++ program file materials.hpp

/*
  By Nathan C. Hearn
     June 7, 2007

*/


#include "materials.hpp"
#include <vector>
#include "quickflash_except.hpp"


// Utility functions

const char BacklighterType_Ti_label[] = "Ti";
const char BacklighterType_Sc_label[] = "Sc";
const char BacklighterType_V_label[] = "V";


int read_backlighter_type(const std::string & backlighter_label, 
			  BacklighterType & backlighter_type)
  {
  int ret_val = -1;  // Error state

  if (backlighter_label == BacklighterType_Ti_label)
    {
    backlighter_type = Backlighter_Ti;
    ret_val = 0;  // OK
    }
  else if (backlighter_label == BacklighterType_Sc_label)
    {
    backlighter_type = Backlighter_Sc;
    ret_val = 0;  // OK
    }
  else if (backlighter_label == BacklighterType_V_label)
    {
    backlighter_type = Backlighter_V;
    ret_val = 0;  // OK
    }

  return ret_val;
  }


int get_backlighter_label(const BacklighterType & backlighter_type,
			  std::string & backlighter_label)
  {
  int ret_val = -1;  // Error state

  switch(backlighter_type)
    {
    case Backlighter_Ti :

      backlighter_label = BacklighterType_Ti_label;
      ret_val = 0;  // OK

      break;

    case Backlighter_Sc :

      backlighter_label = BacklighterType_Sc_label;
      ret_val = 0;  // OK

      break;

    case Backlighter_V :

      backlighter_label = BacklighterType_V_label;
      ret_val = 0;  // OK

      break;
    }

  return ret_val;
  }


const double Ti_Poly_AbsCoeff = 3.01e+01;
const double Ti_CHBr_AbsCoeff = 1.00e+02;
const double Ti_Foam_AbsCoeff = 2.48e+01;

const double Sc_Poly_AbsCoeff = 3.96e+01;
const double Sc_CHBr_AbsCoeff = 1.30e+02;
const double Sc_Foam_AbsCoeff = 3.12e+01;

const double V_Poly_AbsCoeff = 2.25e+01;
const double V_CHBr_AbsCoeff = 7.85e+01;
const double V_Foam_AbsCoeff = 1.63e+01;


int get_absorb_coefficients(const BacklighterType & backlighter_type,
			    double & poly_coeff, double & chbr_coeff,
			    double & foam_coeff)
  {
  int ret_val = -1;  // Error state

  switch(backlighter_type)
    {
    case Backlighter_Ti :

      poly_coeff = Ti_Poly_AbsCoeff;
      chbr_coeff = Ti_CHBr_AbsCoeff;
      foam_coeff = Ti_Foam_AbsCoeff;

      ret_val = 0;  // OK
      break;

    case Backlighter_Sc :

      poly_coeff = Sc_Poly_AbsCoeff;
      chbr_coeff = Sc_CHBr_AbsCoeff;
      foam_coeff = Sc_Foam_AbsCoeff;

      ret_val = 0;  // OK
      break;

    case Backlighter_V :

      poly_coeff = V_Poly_AbsCoeff;
      chbr_coeff = V_CHBr_AbsCoeff;
      foam_coeff = V_Foam_AbsCoeff;

      ret_val = 0;  // OK
      break;
    }

  return ret_val;
  }


// Class Absorption

void Absorption::reset(const BacklighterType backlighter_type)
  {
  // Set up working absorption coefficient values

  backlighter = backlighter_type;

  if (get_absorb_coefficients(backlighter, poly_abs_coeff, chbr_abs_coeff, 
			      foam_abs_coeff) < 0)
    throw QuickFlash::Except("Unrecognized backlighter", __FILE__, __LINE__);
  }


double Absorption::get_trans_fract(const double sample_length, 
				   const std::vector<double> & poly_density,
				   const std::vector<double> & chbr_density,
				   const std::vector<double> & foam_density) 
  const
  {
  double poly_areal_density = 0.0;
  double chbr_areal_density = 0.0;
  double foam_areal_density = 0.0;

  // Compute polyimide areal density

  const unsigned int num_poly = poly_density.size();

  for (unsigned int idx = 0; idx < num_poly; idx++)
    poly_areal_density += poly_density[idx];

  poly_areal_density *= sample_length;

  // Compute CHBr areal density

  const unsigned int num_chbr = chbr_density.size();

  for (unsigned int idx = 0; idx < num_chbr; idx++)
    chbr_areal_density += chbr_density[idx];

  chbr_areal_density *= sample_length;

  // Compute foam areal density

  const unsigned int num_foam = foam_density.size();

  for (unsigned int idx = 0; idx < num_foam; idx++)
    foam_areal_density += foam_density[idx];

  foam_areal_density *= sample_length;

  // Compute the fraction transmitted

  return get_trans_fract(poly_areal_density, chbr_areal_density, 
			 foam_areal_density);
  }


double Absorption::get_trans_fract(const std::vector<double> & sample_length, 
				   const std::vector<double> & poly_density,
				   const std::vector<double> & chbr_density,
				   const std::vector<double> & foam_density) 
  const
  {
  double poly_areal_density = 0.0;
  double chbr_areal_density = 0.0;
  double foam_areal_density = 0.0;

  // Check the input

  const unsigned int num_points = sample_length.size();

  if (poly_density.size() != num_points)
    throw QuickFlash::Except("Incompatible length for poly_density", __FILE__,
			   __LINE__);

  if (chbr_density.size() != num_points)
    throw QuickFlash::Except("Incompatible length for chbr_density", __FILE__,
			   __LINE__);

  if (foam_density.size() != num_points)
    throw QuickFlash::Except("Incompatible length for foam_density", __FILE__,
			   __LINE__);

  // Compute the areal densities

  for (unsigned int idx = 0; idx < num_points; idx++)
    {
    const double length = sample_length[idx];

    poly_areal_density += poly_density[idx] * length;
    chbr_areal_density += chbr_density[idx] * length;
    foam_areal_density += foam_density[idx] * length;
    }

  // Compute the fraction transmitted

  return get_trans_fract(poly_areal_density, chbr_areal_density, 
			 foam_areal_density);
  }
