#!/usr/bin/python

# pytracer_partask.py

# By Nathan C. Hearn
#    July 19, 2007
#
# ParTask-based queuing system for PyTracer


# Species names

dens_name = str("dens")

poly_name = "poly"
chbr_name = "chbr"
foam_name = "foam"




# Default values -- for reference

detector_x_default = 0.1
axis_y_default = 0.0
axis_z_default = 0.0

sample_spacing_default = 0.0005

detector_azimuth_default = 45.0

source_type_default = "Sc"  # "Across strip" view

source_distance_default = 1.2
source_radius_default = 20.0e-04

film_distance_default = 22.8

film_width_default = 5.0
film_height_default = 5.0

width_pixels_default = 2273
height_pixels_default = 2273

rays_per_pixel_default = int(10)

buffer_size_default = int(250)
cache_size_default = int(5000)

num_threads_default = int(0)


## RayTask

from pytracer import PyTracerCore
from ParTask import UserTask


raytracer = PyTracerCore()  # Initialize before use!!!  # NOT THREADSAFE!!!!

# COULD THREAD-RELATED PROBLEMS BE AVOIDED BY PUTTING raytracer IN RayTask???

class RayTask(UserTask) :

    def __init__(self, task=None) :

        UserTask.__init__(self, task)

    def run_task(self) :

        command_str = self.get_command_string()

        words = command_str.split()

        if (len(words) != 2) :
            raise RuntimeError("Insufficient parameters in command [ "
                               + command_str + " ]")

        start_pixel = int(words[0])
        num_pixels = int(words[1])

        end_pixel = start_pixel + (num_pixels - 1)

        pixel_values = list()

        for index in range(start_pixel, (start_pixel + num_pixels)) :
            value = raytracer.get_pixel_value(index)
            pixel_values.append(str(value))

        # Set up return string : start_pixel pixel1 [ pixel2 ...]

        ret_str = "Start " + str(start_pixel) + " " + str(pixel_values)
        
        self.set_command_result(ret_str)

        raytracer.report_cache_stats()

        return ret_str


# Command line parser

def parse_cli() :

    from sys import exit, stderr
    from optparse import OptionParser

    usage = str("%prog input_task_list output_task_list data_file [options]")
    version = str("%prog 0.1")

    argParser = OptionParser(usage, version=version)

    # Detector x-coordinate
    
    detector_x_str = str("detector-x")

    argParser.add_option("--" + detector_x_str, type="float",
                         dest="detector_x",
                         help="Detector axis x-coordinate (cm)")

    argParser.set_defaults(detector_x=detector_x_default)

    # Cylinder axis y-coordinate
    
    axis_y_str = str("axis-y")

    argParser.add_option("--" + axis_y_str, type="float",
                         dest="axis_y",
                         help="Cylinder axis y-coordinate (cm)")

    argParser.set_defaults(axis_y=axis_y_default)

    # Cylinder axis z-coordinate
    
    axis_z_str = str("axis-z")

    argParser.add_option("--" + axis_z_str, type="float",
                         dest="axis_z",
                         help="Cylinder axis z-coordinate (cm)")

    argParser.set_defaults(axis_z=axis_z_default)

    # Sample spacing
    
    sample_spacing_str = str("sample-spacing")

    argParser.add_option("--" + sample_spacing_str, type="float",
                         dest="sample_spacing",
                         help="Spacing between samples on ray (cm)")

    argParser.set_defaults(sample_spacing=sample_spacing_default)

    # Detector azimuth
    
    detector_azimuth_str = str("detector-azim")

    argParser.add_option("--" + detector_azimuth_str, type="float",
                         dest="detector_azimuth",
                         help="Azimuthal angle of detector axis (deg)")

    argParser.set_defaults(detector_azimuth=detector_azimuth_default)

    # Source type

    source_type_str = str("source-type")

    argParser.add_option("--" + source_type_str, type="string",
                         dest="source_type",
                         help="X-ray source material type (Ti, Sc, V)")
    
    argParser.set_defaults(source_type=source_type_default)

    # Custom polyimide x-ray absorption coefficient

    poly_absorption_str = str("poly-absorption")

    argParser.add_option("--" + poly_absorption_str, type="float",
                         dest="poly_absorption",
                         help=("Custom x-ray absorption coefficient for "
                               + "polyimide (cm^2 / g)"))

    argParser.set_defaults(poly_absorption=None)

    # Custom CHBr x-ray absorption coefficient

    chbr_absorption_str = str("chbr-absorption")

    argParser.add_option("--" + chbr_absorption_str, type="float",
                         dest="chbr_absorption",
                         help=("Custom x-ray absorption coefficient for "
                               + "CHBr (cm^2 / g)"))

    argParser.set_defaults(chbr_absorption=None)

    # Custom foam x-ray absorption coefficient

    foam_absorption_str = str("foam-absorption")

    argParser.add_option("--" + foam_absorption_str, type="float",
                         dest="foam_absorption",
                         help=("Custom x-ray absorption coefficient for "
                               + "foam (cm^2 / g)"))

    argParser.set_defaults(foam_absorption=None)

    # Source distance
    
    source_distance_str = str("source-dist")

    argParser.add_option("--" + source_distance_str, type="float",
                         dest="source_distance",
                         help="Distance from axis to x-ray source (cm)")

    argParser.set_defaults(source_distance=source_distance_default)

    # Source radius
    
    source_radius_str = str("source-rad")

    argParser.add_option("--" + source_radius_str, type="float",
                         dest="source_radius",
                         help="Radius of x-ray source (cm)")

    argParser.set_defaults(source_radius=source_radius_default)

    # Film distance
    
    film_distance_str = str("film-dist")

    argParser.add_option("--" + film_distance_str, type="float",
                         dest="film_distance",
                         help="Distance from axis to film (cm)")

    argParser.set_defaults(film_distance=film_distance_default)

    # Film width
    
    film_width_str = str("film-width")

    argParser.add_option("--" + film_width_str, type="float",
                         dest="film_width",
                         help="Width of film (cm)")

    argParser.set_defaults(film_width=film_width_default)

    # Film height
    
    film_height_str = str("film-height")

    argParser.add_option("--" + film_height_str, type="float",
                         dest="film_height",
                         help="Height of film (cm)")

    argParser.set_defaults(film_height=film_height_default)

    # Film pixels along width
    
    width_pixels_str = str("film-width-pixels")

    argParser.add_option("--" + width_pixels_str, type="int",
                         dest="width_pixels",
                         help="Pixel count along film width")

    argParser.set_defaults(width_pixels=width_pixels_default)

    # Film pixels along height
    
    height_pixels_str = str("film-height-pixels")

    argParser.add_option("--" + height_pixels_str, type="int",
                         dest="height_pixels",
                         help="Pixel count along film height")

    argParser.set_defaults(height_pixels=height_pixels_default)

    # Rays per pixel
    
    rays_per_pixel_str = str("pixel-rays")

    argParser.add_option("--" + rays_per_pixel_str, type="int",
                         dest="rays_per_pixel",
                         help="Number of rays per pixel")

    argParser.set_defaults(rays_per_pixel=rays_per_pixel_default)

    # Buffer size

    buffer_size_str = str("buffer-size")

    argParser.add_option("--" + buffer_size_str, type="int",
                         dest="buffer_size",
                         help="Read-ahead buffer size")

    argParser.set_defaults(buffer_size=buffer_size_default)
    
    # Cache size

    cache_size_str = str("cache-size")

    argParser.add_option("--" + cache_size_str, type="int",
                         dest="cache_size",
                         help="Read cache size")

    argParser.set_defaults(cache_size=cache_size_default)

    # Number of threads

    num_threads_str = str("threads")

    argParser.add_option("--" + num_threads_str, type="int",
                         dest="num_threads",
                         help="Number of threads (zero deactivates threads)")

    argParser.set_defaults(num_threads=num_threads_default)


    # Parse the command line

    opts, args = argParser.parse_args()

    argc = len(args)

    min_args = int(3)

    if (argc < min_args) :
        stderr.write("\n")
        argParser.print_help(file=stderr)
        stderr.write("\n")
        exit(1)

    argptr = int(0)

    input_task_list = str(args[argptr])
    argptr += 1

    output_task_list = str(args[argptr])
    argptr += 1

    filename = str(args[argptr])
    argptr += 1

    return input_task_list, output_task_list, filename, opts


# Main program

def main() :

    from sys import stderr
    from ParTask import MPI_Active, MPI_Rank, MPI_Size, MPI_Root

    import materials

    # Read the command line arguments

    input_task_list, output_task_list, filename, opts = parse_cli()

    # ERROR CHECKING???

    detector_x = opts.detector_x

    axis_y = opts.axis_y
    axis_z = opts.axis_z

    sample_spacing = opts.sample_spacing

    axis_azimuth = opts.detector_azimuth

    source_type = opts.source_type

    source_distance = opts.source_distance
    source_radius = opts.source_radius
    
    film_distance = opts.film_distance

    film_width = opts.film_width
    film_height = opts.film_height

    width_pixels = opts.width_pixels
    height_pixels = opts.height_pixels

    rays_per_pixel = opts.rays_per_pixel

    buffer_size = opts.buffer_size
    cache_size = opts.cache_size

    num_threads = opts.num_threads

    # Get material properties

    poly_opacity = materials.get_poly_opacity(source_type)
    chbr_opacity = materials.get_chbr_opacity(source_type)
    foam_opacity = materials.get_foam_opacity(source_type)

    if (opts.poly_absorption is not None) :
        poly_opacity = opts.poly_absorption

    if (opts.chbr_absorption is not None) :
        chbr_opacity = opts.chbr_absorption

    if (opts.foam_absorption is not None) :
        foam_opacity = opts.foam_absorption

    # Determine the thread operation

    use_threads = False
    use_server_threads = False

    if (num_threads > 0) :

        use_server_threads = True

        if (num_threads > 1) :
            use_threads = True

            raise RuntimeError("Multiple threads not allowed at this time")
            
    else :
        
        num_threads = 1  # But no actual threads running


    # Initialize the raytracer

    raytracer.reset(filename, dens_name, detector_x, axis_y, axis_z,
                    buffer_size, cache_size)

    raytracer.set_source_detector(axis_azimuth, source_distance, source_radius,
                                  sample_spacing, rays_per_pixel,
                                  film_distance, film_width, film_height,
                                  width_pixels, height_pixels)

    raytracer.add_species(poly_name, poly_opacity)
    raytracer.add_species(chbr_name, chbr_opacity)
    raytracer.add_species(foam_name, foam_opacity)


    # Run the queue

    my_role = None

    if (MPI_Active) :
        stderr.write("MPI active -- process [ " + str(MPI_Rank) + " ] of [ "
                     + str(MPI_Size) + " ]\n")


    if ((MPI_Root == MPI_Rank) or (not MPI_Active)) :
        input_file = "mytasks_mpi.pre"
        output_file = "mytasks_mpi.post"

        # I am the server

        from ParTask import Server

        my_role = Server(input_task_list, output_task_list, num_threads,
                         user_task=RayTask, use_threads=use_server_threads)
    else :
        from ParTask import Client

        my_role = Client(num_threads, user_task=RayTask,
                         use_threads=use_threads)

    my_role.run()

    stderr.write("Tasks finished -- rank [ " + str(MPI_Rank) + " ]\n")

    if (MPI_Active) :
        import mpi
        mpi.barrier()


if (__name__ == "__main__") :
    main()
