// C++ header file rays.hpp

/*
  By Nathan C. Hearn
     November 2, 2006

  Generator for mesh sample points.
*/


#ifndef RAYS_HPP
#define RAYS_HPP


#include <vector>
#include <gsl/gsl_rng.h>


class Rays
  {
  public:
    Rays() :
      initialized(false), source_pos(), source_radius(0.0), cyl_y(0.0),
      cyl_z(0.0), min_bounds(), max_bounds(), use_random_source_pos(false),
      use_end_point_center(false), random_gen(0), source_parallel(),
      source_tangent()
      { }

    Rays(const std::vector<double> & light_source_pos,
	 const double light_source_radius,
	 const double cylinder_center_y,
	 const double cylinder_center_z,
	 const std::vector<double> & min_domain_bounds,
	 const std::vector<double> & max_domain_bounds,
	 const unsigned long int random_seed=0) :
      initialized(false), source_pos(), source_radius(0.0), cyl_y(0.0),
      cyl_z(0.0), min_bounds(), max_bounds(), use_random_source_pos(false),
      use_end_point_center(false), random_gen(0), source_parallel(),
      source_tangent()
      { 
      reset(light_source_pos, light_source_radius, cylinder_center_y,
	    cylinder_center_z, min_domain_bounds, max_domain_bounds, 
	    random_seed);
      }

    ~Rays() { reset(); }

    void reset();

    void reset(const std::vector<double> & light_source_pos,
	       const double light_source_radius,
	       const double cylinder_center_y,
	       const double cylinder_center_z,
	       const std::vector<double> & min_domain_bounds,
	       const std::vector<double> & max_domain_bounds,
	       const unsigned long int random_seed=0);

    void set_use_end_point_center(const bool state=true)
      { use_end_point_center = state; }

    void get_samples(const std::vector<double> & end_point_center,
		     const double end_point_box_x_width,
		     const double end_point_box_y_width,
		     const double spacing,
		     std::vector< std::vector<double> > & samples) const;

    bool in_bounds(const std::vector<double> & sample_point) const;

  private:
    void setup_source_axes();

    void get_random_end_pos(const std::vector<double> & end_point_center,
			    const double end_point_box_x_width,
			    const double end_point_box_y_width,
			    std::vector<double> & end_point) const;

    void get_random_start_pos(const std::vector<double> & end_point,
			      const double spacing,
			      std::vector<double> & start_point) const;

    void get_sample_points(const std::vector<double> & start_point,
			   const std::vector<double> & end_point,
			   const double spacing,
			   std::vector< std::vector<double> > & samples) const;
    
  private:
    static const double two_pi;

    static const unsigned int dims;

    bool initialized;

    std::vector<double> source_pos;
    double source_radius;

    double cyl_y;
    double cyl_z;

    std::vector<double> min_bounds;
    std::vector<double> max_bounds;

    bool use_random_source_pos;

    bool use_end_point_center;

    mutable gsl_rng * random_gen;

    std::vector<double> source_parallel;
    std::vector<double> source_tangent;
  };
    




#endif  // RAYS_HPP
