// C++ program file raytracer_core.cpp

/*
  By Nathan C. Hearn
     July 18, 2007

  Core raytracing class.
*/


#include "raytracer_core.hpp"
#include <math.h>
#include <string>
#include <vector>
#include <iostream>
#include "quickflash_file_datafile.hpp"
#include "quickflash_except.hpp"
#include "quickflash_file_dataset.hpp"
#include "quickflash_cache_indexcache.hpp"


// Class RaytracerCore

const unsigned int RaytracerCore::dims = 3;


RaytracerCore::RaytracerCore() :
  dfile(0), meshinfo(0), num_blocks(0), local_cache_size(0), axis_center(),
  min_bounds(), max_bounds(), blockinfo_block_index(0), blockinfo(0),
  dens_set(0), dens_cache(), dens_block_read(false), dens_block_index(0),
  dens_data(0), species_sets(), species_opacities(), species_caches(),
  species_block_read(false), species_block_index(0), species_data(),
  detector_set_up(false), detector(), spacing(0.0), num_rays(0),
  oneOver_num_rays(0.0), cache_hits(0), cache_misses(0), block_reuse_count(0)
  { reset(); }


RaytracerCore::RaytracerCore(const QuickFlash::File::DataFile & data_file,
			     const std::string & density_dataset_name,
			     const std::vector<double> & center,
			     const unsigned int cache_size) :
  dfile(0), meshinfo(0), num_blocks(0), local_cache_size(0), axis_center(),
  min_bounds(), max_bounds(), blockinfo_block_index(0), blockinfo(0),
  dens_set(0), dens_cache(), dens_block_read(false), dens_block_index(0),
  dens_data(0), species_sets(), species_opacities(), species_caches(),
  species_block_read(false), species_block_index(0), species_data(),
  detector_set_up(false), detector(), spacing(0.0), num_rays(0),
  oneOver_num_rays(0.0), cache_hits(0), cache_misses(0), block_reuse_count(0)
  { reset(data_file, density_dataset_name, center, cache_size); }


void RaytracerCore::reset()
  {
  if ((cache_hits > 0) || (cache_misses > 0))
    report_cache_stats();

  dfile = 0;  // Just a reference

  meshinfo = 0;

  num_blocks = 0;

  local_cache_size = 0;

  axis_center.clear();

  min_bounds.clear();
  max_bounds.clear();

  blockinfo_block_index = 0;
  blockinfo = 0;

  dens_set = 0;  // Just a reference

  dens_cache.reset();

  dens_block_read = false;
  dens_block_index = 0;

  dens_data = 0;

  const unsigned int num_species = species_sets.size();

  for (unsigned int index = 0; index < num_species; index++)
    species_sets[index] = 0;  // For safety ... just a reference

  species_sets.clear();

  species_opacities.clear();

  for (unsigned int index = 0; index < num_species; index++)
    {
    delete species_caches[index];  // NOT a reference
    species_caches[index] = 0;
    }

  species_caches.clear();

  species_block_read = false;
  species_block_index = 0;

  for (unsigned int index = 0; index < num_species; index++)
    species_data[index] = 0;  // For safety ... just a reference

  species_data.clear();

  detector_set_up = false;

  detector.reset();

  spacing = 0.0;
  num_rays = 0;

  oneOver_num_rays = 0.0;

  cache_hits = 0;
  cache_misses = 0;

  block_reuse_count = 0;
  }


void RaytracerCore::reset(const QuickFlash::File::DataFile & data_file,
			  const std::string & density_set_name,
			  const std::vector<double> & center,
			  const unsigned int cache_size)
  {
  reset();

  dfile = &data_file;

  meshinfo = &(dfile->get_mesh_info());

  const unsigned int file_dims = meshinfo->get_dims();

  if (file_dims != dims)
    throw QuickFlash::Except("Data file must have three dimensions", __FILE__,
			   __LINE__);

  num_blocks = meshinfo->get_num_blocks();

  local_cache_size = cache_size;

  if (local_cache_size < 1)
    local_cache_size = 1;

  if (center.size() != dims)
    throw QuickFlash::Except("Incompatible center vector size", __FILE__,
			   __LINE__);

  axis_center = center;

  min_bounds = meshinfo->get_volume_minbounds();
  max_bounds = meshinfo->get_volume_maxbounds();

  dens_set = &(dfile->get_dataset(density_set_name));
  dens_set->set_report_stats();

  dens_cache.reset(num_blocks, local_cache_size);
  }


void RaytracerCore::add_species(const std::string & dataset_name,
				const double opacity)
  {
  if (dfile == 0)
    throw QuickFlash::Except("Data file not set up", __FILE__, __LINE__);

  // Set up the new objects

  const QuickFlash::File::Dataset & dset = dfile->get_dataset(dataset_name);

  dset.set_report_stats();

  QuickFlash::Cache::IndexCache<DoubleBlock> * dset_cache 
    = new QuickFlash::Cache::IndexCache<DoubleBlock>;

  dset_cache->reset(num_blocks, local_cache_size);

  // Store the objects

  species_sets.push_back(&dset);

  species_opacities.push_back(opacity);

  species_caches.push_back(dset_cache);  // Make sure it is deleted later

  species_data.push_back(0);

  // Clear read info

  species_block_read = false;
  species_block_index = 0;
  }


void RaytracerCore::set_source_detector(const double axis_azimuth,
					const double source_distance,
					const double source_radius,
					const double sample_spacing,
					const unsigned int rays_per_pixel,
					const double film_distance, 
					const double film_width,
					const double film_height, 
					const unsigned int width_pixels,
					const unsigned int height_pixels)
  {
  std::vector<double> film_size(2);

  film_size[0] = film_width;
  film_size[1] = film_height;

  std::vector<unsigned int> film_resolution(2);

  film_resolution[0] = width_pixels;
  film_resolution[1] = height_pixels;

  detector.reset(axis_center, min_bounds, max_bounds, axis_azimuth,
		 source_distance, source_radius, film_distance, film_size,
		 film_resolution);

  if (!(sample_spacing > 0.0))
    throw QuickFlash::Except("Non-positive sample spacing", __FILE__, __LINE__);

  spacing = sample_spacing;

  if (rays_per_pixel < 1)
    throw QuickFlash::Except("Non-positive rays per pixel", __FILE__, __LINE__);

  num_rays = rays_per_pixel;

  oneOver_num_rays = 1.0 / static_cast<double>(num_rays);

  detector_set_up = true;
  }


double RaytracerCore::get_pixel_value(const unsigned int pixel_index) const
  {
  if (!set_up())
    throw QuickFlash::Except("Raytracer not set up", __FILE__, __LINE__);

  double accum_value = 0.0;

  for (unsigned int ray_index = 0; ray_index < num_rays; ray_index++)
    {
    std::vector< std::vector<double> > sample_points;

    detector.get_sample_points(pixel_index, spacing, sample_points);

    accum_value += exp(-spacing * get_opacity_integral(sample_points));
    }

  return accum_value * oneOver_num_rays;
  }


void RaytracerCore::load_data(const unsigned int block_index) const
  {
  // Make sure raytracer is set up before calling this function!!!

  if ((!dens_block_read) || (block_index != dens_block_index))
    {
    if (!(dens_cache.data_present(block_index)))
      {
      cache_misses++;

      DoubleBlock work_block;

      dens_set->get_block_data(block_index, work_block);

      dens_cache.add_data(block_index, work_block);
      }
    else
      cache_hits++;

    dens_data = &(dens_cache[block_index]);

    dens_block_read = true;
    dens_block_index = block_index;
    }
  else
    block_reuse_count++;

  if ((!species_block_read) || (block_index != species_block_index))
    {
    const unsigned int num_species = species_sets.size();

    for (unsigned int index = 0; index < num_species; index++)
      {
      if (!(species_caches[index]->data_present(block_index)))
	{
	DoubleBlock work_block;

	species_sets[index]->get_block_data(block_index, work_block);

	species_caches[index]->add_data(block_index, work_block);
	}

      species_data[index] = &(species_caches[index]->get_item(block_index));
      }

    species_block_read = true;
    species_block_index = block_index;
    }
  }


double 
  RaytracerCore::get_opacity_integral(const std::vector< std::vector<double> > 
				      & sample_points) const
  {
  // Make sure raytracer is set up before calling this function!!!

  double accum_density = 0.0;

  const unsigned int num_species = species_sets.size();

  std::vector<double> accum_spec_dens(num_species, 0.0);

  const unsigned int num_samples = sample_points.size();

  std::vector<double> spec_data(num_species);

  for (unsigned int sample_index = 0; sample_index < num_samples; 
       sample_index++)
    {
    double density = 0.0;

    get_data(sample_points[sample_index], density, spec_data);

    accum_density += density;

    for (unsigned int spec_index = 0; spec_index < num_species; spec_index++)
      accum_spec_dens[spec_index] += density * spec_data[spec_index];
    }

  double integral = 0.0;

  if (num_species > 0)
    for (unsigned int spec_index = 0; spec_index < num_species; spec_index++)
      integral += accum_spec_dens[spec_index] * species_opacities[spec_index];
  else
    integral = accum_density;

  return integral;
  }


void RaytracerCore::get_data(const std::vector<double> & pos, double & density,
			     std::vector<double> & species_mass_fractions) 
  const
  {
  // Make sure raytracer is set up before calling this function!!!

  bool read_block_info = true;

  if (blockinfo != 0)
    if (blockinfo->in_block(pos))
      read_block_info = false;

  if (read_block_info)
    {
    blockinfo_block_index = meshinfo->get_block_index(pos);
    blockinfo = &(meshinfo->get_block_info(blockinfo_block_index));
    }

  const unsigned int cell_index = blockinfo->get_nearest_cell(pos);

  load_data(blockinfo_block_index);

  density = (*dens_data)[cell_index];

  const unsigned int num_species = species_data.size();

  species_mass_fractions.resize(num_species);

  for (unsigned int index = 0; index < num_species; index++)
    species_mass_fractions[index] = (*(species_data[index]))[cell_index];
  }


void RaytracerCore::report_cache_stats() const
  {
  using std::cerr;
  using std::endl;

  // For now, just report on the density cache

  cerr << "Raytracer internal cache hits [ " << cache_hits << " ] misses [ "
       << cache_misses << " ] items [ " << dens_cache.get_num_cache_items()
       << " ] cleans [ " << dens_cache.get_clean_counter() << " ]" << endl;

  cerr << "Sequential block reuse count [ " << block_reuse_count << " ]" 
       << endl;

  // ERROR CHECKING???

  dens_set->report_all_stats();
  }
