// C++ program file raytracer_serial.cpp

/*
  By Nathan C. Hearn
     July 19, 2007

  Serial raytracer program.
*/


#include <iostream>
#include <cmath>
#include <string>
#include <fstream>
#include "quickflash_utils_commandline.hpp"
#include "quickflash_utils_file.hpp"
#include "quickflash_utils_text.hpp"
#include "raytracer_core.hpp"
#include "materials.hpp"


const double default_detector_x = 0.1;

const double default_axis_y = 0.0;
const double default_axis_z = 0.0;

const double default_sample_spacing = 5.0e-04;

const double default_detector_azimuth_deg = 0.0;

const char default_source_type_label[] = "Sc";

const double default_source_distance = 1.2;
const double default_source_radius = 20.0e-04;

const double default_film_distance = 22.8;

const double default_film_width = 5.0;
const double default_film_height = 5.0;

const unsigned int default_film_width_pixels = 2273;
const unsigned int default_film_height_pixels = 2273;

const unsigned int default_rays_per_pixel = 1;

const unsigned int default_buffer_size = 50;
const unsigned int default_cache_size = 4000;


const double deg_to_rad = (M_PI) / 180.0;

const char density_name[] = "dens";
const char poly_name[] = "poly";
const char foam_name[] = "foam";
const char chbr_name[] = "chbr";


int main(int argc, char * argv[])
  {
  using std::cerr;
  using std::endl;
  using std::flush;

  std::vector<std::string> argument_list;
  std::list<QuickFlash::Utils::Option> option_list;

  QuickFlash::Utils::parse_command_line(argc, argv, argument_list, 
					option_list);

  const unsigned int num_args = argument_list.size();

  const unsigned int min_args = 1;

  if (num_args < min_args)
    {
    const std::string full_exec = argv[0];

    std::string exec_path;
    std::string exec_name;

    QuickFlash::Utils::split_name_path_file(full_exec, exec_path, exec_name);

    cerr << endl << "Usage: " << exec_name
	 << endl << "\t"
	 << "[ --output-prefix=OUTPUT_PREFIX ]"
	 << endl << "\t"
	 << "[ --detector-x=DETECTOR_X_COORD ]"
	 << endl << "\t"
	 << "[ --axis-center=TARGET_Y_COORD,TARGET_Z_COORD ]"
	 << endl << "\t"
	 << "[ --sample-spacing=SAMPLE_SPACING ]"
	 << endl << "\t"
	 << "[ --detector-azimuth=DETECTOR_AZIMUTH_ANGLE ]"
	 << endl << "\t"
	 << "[ --source-type={Ti|Sc|V} ]"
	 << endl << "\t"
	 << "[ --source-distance=TARGET_SOURCE_DISTANCE ]"
	 << endl << "\t"
	 << "[ --source-radius=SOURCE_PINHOLE_RADIUS ]"
	 << endl << "\t"
	 << "[ --film-distance=TARGET_FILM_DISTANCE ]"
	 << endl << "\t"
	 << "[ --film-size=FILM_WIDTH,FILM_HEIGHT ]"
	 << endl << "\t"
	 << "[ --film-pixels=FILM_WIDTH_PIXELS,FILM_HEIGHT_PIXELS ]"
	 << endl << "\t"
	 << "[ --pixel-rays=RAYS_PER_PIXEL ]"
	 << endl << "\t"
	 << "[ --buffer-size=BUFFER_BLOCKS ] [ --cache-size=CACHE_BLOCKS ]"
	 << endl << "\t"
	 << "[ --quiet ] [ --start-pixel=START_PIXEL_INDEX ]"
	 << endl << "\t"
	 << "[ --num-pixels=PIXEL_COUNT ]"
	 << endl << "\t"
	 << "flash_datafile_1 [ flash_datafile_2 ... ]"
	 << endl << endl;

    return 1;
    }

  // Read the arguments

  unsigned int argptr = 0;

  const std::string dataset_name = argument_list[argptr++];

  std::list<std::string> flash_filenames;

  while (argptr < num_args)
    flash_filenames.push_back(argument_list[argptr++]);

  // Parse the options

  std::string user_output_prefix;

  double detector_x = default_detector_x;

  double axis_y = default_axis_y;
  double axis_z = default_axis_z;

  double sample_spacing = default_sample_spacing;

  double detector_azimuth_deg = default_detector_azimuth_deg;

  std::string source_type_label = default_source_type_label;

  double source_distance = default_source_distance;
  double source_radius = default_source_radius;

  double film_distance = default_film_distance;

  double film_width = default_film_width;
  double film_height = default_film_height;

  unsigned int film_width_pixels = default_film_width_pixels;
  unsigned int film_height_pixels = default_film_height_pixels;

  unsigned int rays_per_pixel = default_rays_per_pixel;

  unsigned int buffer_size = default_buffer_size;
  unsigned int cache_size = default_cache_size;

  bool quiet_console = false;

  unsigned int start_pixel_index = 0;

  bool num_pixels_given = false;

  unsigned int num_pixels = 0;

  std::list<QuickFlash::Utils::Option>::const_iterator option_iter
    = option_list.begin();

  while (option_iter != option_list.end())
    {
    const QuickFlash::Utils::Option & current_option = *option_iter;

    const std::string & keyword = current_option.get_key();

    if (keyword == "output-prefix")
      {
      if (!current_option.value_present())
	{
	cerr << endl << "Error: Output prefix is required with option"
	     << endl << endl;
	return -1;
	}

      user_output_prefix = current_option.get_value();

      // Add an underscore if prefix contains more than just a path

      std::string user_prefix_path;
      std::string user_prefix_name;

      QuickFlash::Utils::split_name_path_file(user_output_prefix, 
					      user_prefix_path,
					      user_prefix_name);

      const unsigned int user_prefix_name_len = user_prefix_name.size();

      if (user_prefix_name_len > 0)
	if (user_prefix_name[user_prefix_name_len - 1] != '_')
	  user_output_prefix += "_";
      }
    else if (keyword == "detector-x")
      {
      if (!current_option.value_present())
	{
	cerr << endl 
	     << "Error: Detector position option requires a value"
	     << endl << endl;
	return -1;
	}

      current_option.get_value(detector_x);
      }
    else if (keyword == "axis-center")
      {
      if (!current_option.value_present())
	{
	cerr << endl 
	     << "Error: Axis position option requires y- and z-coordinates"
	     << endl << endl;
	return -1;
	}

      std::vector<double> axis_pos;

      current_option.get_value(axis_pos);

      if (axis_pos.size() != 2)
	{
	cerr << endl 
	     << "Error: Axis position option requires y- and z-coordinates"
	     << endl << endl;
	return -1;
	}

      axis_y = axis_pos[0];
      axis_z = axis_pos[1];
      }
    else if (keyword == "sample-spacing")
      {
      if (!current_option.value_present())
	{
	cerr << endl 
	     << "Error: Sample spacing option requires a positive value"
	     << endl << endl;
	return -1;
	}

      current_option.get_value(sample_spacing);

      if (!(sample_spacing > 0.0))
	{
	cerr << endl 
	     << "Error: Sample spacing option requires a positive value"
	     << endl << endl;
	return -1;
	}
      }
    else if (keyword == "source-type")
      {
      if (!current_option.value_present())
	{
	cerr << endl 
	     << "Error: Source type option requires a backlighter type"
	     << endl << endl;
	return -1;
	}

      source_type_label = current_option.get_value();
      }
    else if (keyword == "source-distance")
      {
      if (!current_option.value_present())
	{
	cerr << endl 
	     << "Error: Source distance option requires a positive value"
	     << endl << endl;
	return -1;
	}

      current_option.get_value(source_distance);

      if (!(source_distance > 0.0))
	{
	cerr << endl 
	     << "Error: Source distance option requires a positive value"
	     << endl << endl;
	return -1;
	}
      }
    else if (keyword == "source-radius")
      {
      if (!current_option.value_present())
	{
	cerr << endl 
	     << "Error: Source radius option requires a positive value"
	     << endl << endl;
	return -1;
	}

      current_option.get_value(source_radius);

      if (!(source_radius > 0.0))
	{
	cerr << endl 
	     << "Error: Source radius option requires a positive value"
	     << endl << endl;
	return -1;
	}
      }
    else if (keyword == "film-distance")
      {
      if (!current_option.value_present())
	{
	cerr << endl 
	     << "Error: Film distance option requires a positive value"
	     << endl << endl;
	return -1;
	}

      current_option.get_value(film_distance);

      if (!(film_distance > 0.0))
	{
	cerr << endl 
	     << "Error: Film distance option requires a positive value"
	     << endl << endl;
	return -1;
	}
      }
    else if (keyword == "film-size")
      {
      if (!current_option.value_present())
	{
	cerr << endl 
	     << "Error: Film size option requires a width and height"
	     << endl << endl;
	return -1;
	}

      std::vector<double> film_size;

      current_option.get_value(film_size);

      if (film_size.size() != 2)
	{
	cerr << endl 
	     << "Error: Film size option requires a width and height"
	     << endl << endl;
	return -1;
	}

      film_width = film_size[0];
      film_height = film_size[1];

      if (!(film_width > 0.0))
	{
	cerr << endl << "Error: Film width must be positive" << endl << endl;
	return -1;
	}

      if (!(film_height > 0.0))
	{
	cerr << endl << "Error: Film height must be positive" << endl << endl;
	return -1;
	}
      }
    else if (keyword == "film-pixels")
      {
      if (!current_option.value_present())
	{
	cerr << endl 
	     << "Error: Film pixels option requires width and height "
	     << "pixel counts" << endl << endl;
	return -1;
	}

      std::vector<unsigned int> film_pixels;

      current_option.get_value(film_pixels);

      if (film_pixels.size() != 2)
	{
	cerr << endl 
	     << "Error: Film pixels option requires width and height "
	     << "pixel counts" << endl << endl;
	return -1;
	}

      film_width_pixels = film_pixels[0];
      film_height_pixels = film_pixels[1];

      if (film_width_pixels < 1)
	{
	cerr << endl << "Error: Film width pixel count must be positive" 
	     << endl << endl;
	return -1;
	}

      if (film_height_pixels < 1)
	{
	cerr << endl << "Error: Film height pixel count must be positive" 
	     << endl << endl;
	return -1;
	}
      }
    else if (keyword == "pixel-rays")
      {
      if (!current_option.value_present())
	{
	cerr << endl 
	     << "Error: Pixel rays option requires a positive value"
	     << endl << endl;
	return -1;
	}

      current_option.get_value(rays_per_pixel);

      if (rays_per_pixel < 1)
	{
	cerr << endl 
	     << "Error: Pixel rays option requires a positive value"
	     << endl << endl;
	return -1;
	}
      }
    else if (keyword == "buffer-size")
      {
      if (!current_option.value_present())
	{
	cerr << endl 
	     << "Error: Buffer size option requires a block count"
	     << endl << endl;
	return -1;
	}

      current_option.get_value(buffer_size);
      }
    else if (keyword == "cache-size")
      {
      if (!current_option.value_present())
	{
	cerr << endl 
	     << "Error: Cache size option requires a block count"
	     << endl << endl;
	return -1;
	}

      current_option.get_value(cache_size);
      }
    else if (keyword == "quiet")
      {
      quiet_console = true;
      }
    else if (keyword == "start-pixel")
      {
      if (!current_option.value_present())
	{
	cerr << endl 
	     << "Error: Starting pixel option requires a pixel index"
	     << endl << endl;
	return -1;
	}

      current_option.get_value(start_pixel_index);
      }
    else if (keyword == "num-pixels")
      {
      if (!current_option.value_present())
	{
	cerr << endl 
	     << "Error: Number of pixels option requires a pixel count"
	     << endl << endl;
	return -1;
	}

      current_option.get_value(num_pixels);

      if (num_pixels < 1)
	{
	cerr << endl << "Error: Pixel count must be positive" << endl << endl;
	return -1;
	}

      num_pixels_given = true;
      }
    else
      {
      cerr << endl << "Error: Unrecognized option " << keyword << endl << endl;
      return -1;
      }

    ++option_iter;
    }


  // Get derived quantities

  const double detector_azimuth = deg_to_rad * detector_azimuth_deg;

  const unsigned int total_pixels = film_width_pixels * film_height_pixels;

  if (start_pixel_index >= total_pixels)
    {
    cerr << endl << "Error: Starting pixel index [ " << start_pixel_index 
	 << " ] out of range" << endl << endl;
    return -1;
    }

  unsigned int end_pixel_index = 0;

  if (num_pixels_given)
    {
    end_pixel_index = start_pixel_index + (num_pixels - 1);

    if (end_pixel_index >= total_pixels)
      {
      cerr << endl << "Error: Number of pixels [ " << num_pixels 
	   << " ] extends beyond pixel range" << endl << endl;
      return -1;
      }
    }
  else
    {
    end_pixel_index = total_pixels - 1;
    num_pixels = (end_pixel_index + 1) - start_pixel_index;
    }


  // Get the backligher type

  BacklighterType backlighter_type = Backlighter_Sc;

  if (read_backlighter_type(source_type_label, backlighter_type) < 0)
    {
    cerr << endl << "Error: Source type [ " << source_type_label 
	 << " ] not recognized" << endl << endl;
    return -1;
    }

  double poly_opacity = 0.0;
  double chbr_opacity = 0.0;
  double foam_opacity = 0.0;

  if (get_absorb_coefficients(backlighter_type, poly_opacity, chbr_opacity, 
			      foam_opacity) < 0)
    {
    cerr << endl << "Error: Unable to get absorbtion coefficients" << endl
	 << endl;
    return -2;
    }


  // Run through the files

  std::list<std::string>::const_iterator flash_filename_iter 
    = flash_filenames.begin();

  while (flash_filename_iter != flash_filenames.end())
    {
    const std::string & filename = *flash_filename_iter;

    if (!quiet_console)
      cerr << endl << "Starting file " << filename << endl;

    // Build the output filename

    std::stringstream filename_stream;

    filename_stream << user_output_prefix << filename << "_raytrace_" 
		    << start_pixel_index << "_" << end_pixel_index << ".csv";

    const std::string output_filename = filename_stream.str();

    // Start the output

    std::ofstream f(output_filename.c_str(), std::ios_base::trunc);

    f << "File " << filename << endl;

    f << "DetectorPosition "<< detector_x << endl;
    f << "AxisCenter " << axis_y << "," << axis_z << endl;

    f << "SampleSpacing " << sample_spacing << endl;

    f << "DetectorAzimuth " << detector_azimuth_deg << endl;

    f << "SourceDistance " << source_distance << endl;
    f << "SourceRadius " << source_radius << endl;

    f << "SourceType " << source_type_label << endl;

    f << "FilmDistance " << film_distance << endl;

    f << "FilmSize " << film_width << "," << film_height << endl;
    f << "FilmPixels " << film_width_pixels << "," << film_height_pixels 
      << endl;

    f << "PixelRays " << rays_per_pixel << endl;

    f << "StartPixel " << start_pixel_index << endl;
    f << "NumPixels " << num_pixels << endl;

    f << endl;

    f << "RaytracerData" << endl;

    f << endl;

    // Open the data file

    QuickFlash::File::DataFile datafile(filename, buffer_size, true);

    // Set up the raytracer

    std::vector<double> center(3);

    center[0] = detector_x;
    center[1] = axis_y;
    center[2] = axis_z;

    RaytracerCore raytracer(datafile, density_name, center, cache_size);

    raytracer.set_source_detector(detector_azimuth, source_distance, 
				  source_radius, sample_spacing, 
				  rays_per_pixel,
				  film_distance, film_width, film_height,
				  film_width_pixels, film_height_pixels);

    raytracer.add_species(poly_name, poly_opacity);
    raytracer.add_species(chbr_name, chbr_opacity);
    raytracer.add_species(foam_name, foam_opacity);

    // Get the pixels

    unsigned int one_percent_pixels = num_pixels / 100;

    if (one_percent_pixels < 1)
      one_percent_pixels = 1;

    for (unsigned int pixel_index = start_pixel_index; 
	 pixel_index <= end_pixel_index; pixel_index++)
      {
      // Start a new line if appropriate

      if ((pixel_index != start_pixel_index) 
	  && ((pixel_index % film_width_pixels) == 0))
	f << endl;

      // Get the pixel value from the raytracer

      const double pixel_value = raytracer.get_pixel_value(pixel_index);
    
      f << pixel_value << " " << flush;

      const unsigned int relative_index = pixel_index - start_pixel_index;

      if (!quiet_console)
	if ((relative_index % one_percent_pixels) == 0)
	  {
	  cerr << "\t" << "Set [ " << start_pixel_index << " - " 
	       << end_pixel_index << " ] Progress : " 
	       << ((100 * relative_index) / num_pixels) << "%" << endl;
	  raytracer.report_cache_stats();
	  }
      }

    f << endl << endl;

    ++flash_filename_iter;
    }

  return 0;
  }
