// C++ header file stat_bins.hpp

/*
  By Nathan C. Hearn
     November 3, 2006

  Statistical binning class for multi-dimensional arrays.
*/


#ifndef STAT_BINS_HPP
#define STAT_BINS_HPP


#include <math.h>
#include <vector>
#include "quickflash_block_blockinfo.hpp"
#include "quickflash_block_blockdata.hpp"


namespace Stats
{

class Bin
  {
  public:
    Bin() :
      num_entries(0), sum_weight(0.0), sum_weight_values(0.0),
      sum_weight_sqValues(0.0), min_value(0.0), max_value(0.0)
      { }
    Bin(const Bin & source) :
      num_entries(0), sum_weight(0.0), sum_weight_values(0.0),
      sum_weight_sqValues(0.0), min_value(0.0), max_value(0.0)
      { reset(source); }
    ~Bin() { }

    Bin & operator=(const Bin & source)
      {
      reset(source);
      return *this;
      }

    Bin & operator+=(const Bin & source)
      {
      add_bin(source);
      return *this;
      }

    void reset();

    void reset(const Bin & source);

    void reset(const unsigned int entries, const double weight,
	       const double weight_value, const double weight_sqValue,
	       const double value_min, const double value_max)
      {
      num_entries = entries;

      sum_weight = weight;

      sum_weight_values = weight_value;
      sum_weight_sqValues = weight_sqValue;

      min_value = value_min;
      max_value = value_max;
      }

    void add_bin(const Bin & source);

    void add_entry(const double value, const double weight=1.0);      

    unsigned int get_num_entries() const { return num_entries; }

    bool entry_present() const
      { 
      bool present = false;
      
      if (num_entries > 0)
	present = true;

      return present;
      }

    double get_weight_sum() const { return sum_weight; }

    double get_sum_weight_values() const { return sum_weight_values; }
    double get_sum_weight_sqValues() const { return sum_weight_sqValues; }

    double get_mean() const 
      { 
      double mean_val = 0.0;

      if (sum_weight > 0.0)
	mean_val = sum_weight_values / sum_weight;

      return mean_val;
      }

    double get_pop_sqStddev() const;
    double get_pop_stddev() const { return sqrt(get_pop_sqStddev()); }

    double get_min_value() const { return min_value; }
    double get_max_value() const { return max_value; }

  private:
    unsigned int num_entries;

    double sum_weight;

    double sum_weight_values;
    double sum_weight_sqValues;

    double min_value;
    double max_value;
  };


class BinArray
  {
  public:
    BinArray() : bin_dims(), bins() { reset(); }

    BinArray(const std::vector<unsigned int> & axis_bin_count, 
	     const std::vector<double> & min_boundary, 
	     const std::vector<double> & max_boundary) :
      bin_dims(), bins()
      { reset(axis_bin_count, min_boundary, max_boundary); }

    BinArray(const BinArray & source) : bin_dims(), bins() { reset(source); }

    ~BinArray() { }

    BinArray & operator=(const BinArray & source)
      {
      reset(source);
      return *this;
      }

    const Bin & operator[](const unsigned int bin_index) const
      { return bins[bin_index]; }

    const Bin & operator[](const std::vector<unsigned int> & bin_index) const
      { return bins[bin_index]; }

    void reset();
    void reset(const std::vector<unsigned int> & axis_bin_count, 
	       const std::vector<double> & min_boundary, 
	       const std::vector<double> & max_boundary);
    void reset(const BinArray & source);

    void add_data(const unsigned int bin_index, 
		  const double value, const double weight=1.0)
      { bins[bin_index].add_entry(value, weight); }

    void add_data(const std::vector<unsigned int> & bin_index, 
		  const double value, const double weight=1.0)
      { bins[bin_index].add_entry(value, weight); }

    void add_data(const std::vector<double> & coordinate, const double value,
		  const double weight=1.0);

    void set_data(const unsigned int bin_index, const unsigned int num_entries,
		  const double sum_weight, const double sum_weight_values, 
		  const double sum_weight_sqValues, const double min_value, 
		  const double max_value)
      {
      bins[bin_index].reset(num_entries, sum_weight, sum_weight_values, 
			    sum_weight_sqValues, min_value, max_value);
      }

    unsigned int get_dims() const { return bin_dims.get_space_dims(); }

    unsigned int get_num_bins() const { return bin_dims.get_num_cells(); }

    const std::vector<unsigned int> & get_array_dims() const
      { return bin_dims.get_block_dims(); }

    bool in_bin(const unsigned int bin_index, 
		const std::vector<double> & coordinate) const
      { return bin_dims.in_cell(bin_index, coordinate); }

//     bool in_bin(const unsigned int bin_index, 
// 		const std::vector<double> & coordinate, 
// 		const std::vector<double> & spread) const;

//     double get_overlap_fract(const unsigned int bin_index, 
// 			     const double coordinate,
// 			     const std::vector<double> & spread) const;

    const Bin & get_bin(const unsigned int bin_index) const 
      { return bins[bin_index]; }

    void get_bin_center(const unsigned int bin_index, 
			std::vector<double> & coords) const
      { return bin_dims.get_cell_center(bin_index, coords); }

    const std::vector<double> & get_bin_width() const
      { return bin_dims.get_cell_width(); }

    void get_bin_bounds(const unsigned int bin_index,
			std::vector<double> & cell_min_bounds,
			std::vector<double> & cell_max_bounds) const
      { 
      bin_dims.get_cell_bounds(bin_index, cell_min_bounds, cell_max_bounds);
      }

    unsigned int get_scalar_index(const std::vector<unsigned int> & bin_index)
      const
      { return bins.get_scalar_index(bin_index); }

    void get_vector_index(const unsigned int bin_index,
			  std::vector<unsigned int> & vector_bin_index) const
      { bins.get_vector_index(bin_index, vector_bin_index); }

  private:
    QuickFlash::Block::BlockInfo bin_dims;
    QuickFlash::Block::BlockData<Bin> bins;
  };


}  // End namespace Stats


#endif  // STAT_BINS_HPP
