# image_info.py

# By Nathan C. Hearn
#    March 28, 2008
#
# Reads slice metadata file



class ColormapInfo :

    def __init__(self, colormap_info_filename=None) :

        self.__min_value = float(0.0)
        self.__max_value = float(0.0)

        self.__value_axis_length = int(0)
        self.__width_pixels = int(0)

        self.__horizontal_colormap = False

        self.__limit_legend_values = False

        self.__use_log_values = False

        self.__colormap_values = list()
        self.__colormap_data = list()

        if (colormap_info_filename is not None) :
            self.reset(colormap_info_filename)

    def reset(self, colormap_info_filename=None) :

        self.__min_value = float(0.0)
        self.__max_value = float(0.0)

        self.__value_axis_length = int(0)
        self.__width_pixels = int(0)

        self.__horizontal_colormap = False

        self.__limit_legend_values = False

        self.__use_log_values = False

        self.__colormap_values = list()
        self.__colormap_data = list()

        if (colormap_info_filename is not None) :
            self.__read_colormap_info(colormap_info_filename)        


    def get_min_value(self) :
        return self.__min_value


    def get_max_value(self) :
        return self.__max_value


    def get_value_axis_length(self) :
        return self.__value_axis_length


    def get_width_pixels(self) :
        return self.__width_pixels


    def horizontal_colormap(self) :
        return self.__horizontal_colormap


    def use_log_values(self) :
        return self.__use_log_values


    def get_colormap_values(self) :
        return self.__colormap_values


    def get_colormap_data(self) :
        return self.__colormap_data


    def __read_colormap_info(self, colormap_info_filename) :

        f = open(colormap_info_filename, "r")

        colormap_info = dict()

        eof_reached = False
        colormap_data_reached = False

        while ((not eof_reached) and (not colormap_data_reached)) :

            line = f.readline()

            if (line == "") :
                
                eof_reached = True

            else :

                work_line = line.split("#")[0].strip()

                words = work_line.split(" ")

                num_words = len(words)

                if (num_words > 0) :
                    
                    keyword = words[0]

                    if (keyword == "ColormapData") :

                        colormap_data_reached = True
                        
                    elif (keyword == "MinValue") :

                        if (num_words != 2) :
                            raise RuntimeError("Improper colormap min "
                                               + "value entry")

                        self.__min_value = float(words[1])

                    elif (keyword == "MaxValue") :

                        if (num_words != 2) :
                            raise RuntimeError("Improper colormap max "
                                               + "value entry")

                        self.__max_value = float(words[1])

                    elif (keyword == "ValueAxisLength") :

                        if (num_words != 2) :
                            raise RuntimeError("Improper colormap value "
                                               + "axis length entry")

                        self.__value_axis_length = int(words[1])

                    elif (keyword == "WidthPixels") :

                        if (num_words != 2) :
                            raise RuntimeError("Improper colormap width "
                                               + "pixels entry")

                        self.__width_pixels = int(words[1])

                    elif (keyword == "HorizontalColormap") :

                        if (num_words != 2) :
                            raise RuntimeError("Improper horizontal "
                                                   + "colormap entry")

                        state = words[1].lower()

                        if (state == "true") :
                            self.__horizontal_colormap = True
                        elif (state == "false") :
                            self.__horizontal_colormap = False
                        else :
                            raise RuntimeError("Unrecognized horizontal "
                                               "colormap state")

                    elif (keyword == "LimitLegendValues") :

                        if (num_words != 2) :
                            raise RuntimeError("Improper limit-legend-"
                                               + "values entry")

                        state = words[1].lower()

                        if (state == "true") :
                            self.__limit_legend_values = True
                        elif (state == "false") :
                            self.__limit_legend_values = False
                        else :
                            raise RuntimeError("Unrecognized limit-legend-"
                                               "values state")

                    elif (keyword == "UseLogValues") :

                        if (num_words != 2) :
                            raise RuntimeError("Improper use-log-"
                                               + "values entry")

                        state = words[1].lower()

                        if (state == "true") :
                            self.__use_log_values = True
                        elif (state == "false") :
                            self.__use_log_values = False
                        else :
                            raise RuntimeError("Unrecognized use-log-"
                                               "values state")

        if (not colormap_data_reached) :
            raise RuntimeError("Colormap data not found")

        while (not eof_reached) :

            line = f.readline()

            if (line == "") :
                
                eof_reached = True

            else :

                work_line = line.split("#")[0].strip()

                words = work_line.split(" ")

                num_words = len(words)

                if (num_words > 0) :

                    self.__colormap_values.append(float(words[0]))

                    colormap_data_list = list()

                    for word_index in range(1, num_words) :
                        colormap_data_list.append(float(words[word_index]))

                    self.__colormap_data.append(colormap_data_list)
                

class ImageInfo :

    def __init__(self, metadata_filename=None) :

        self.__datafile = None
        self.__datafile_dims = int(0)

        self.__dataset = None

        self.__sim_time = None

        self.__grid_center = None
        self.__grid_width = None
        self.__grid_pixels = None

        self.__image_width = None   ## In case of transposition
        self.__image_pixels = None  ## In case of transposition

        self.__roll_angle = float(0.0)
        self.__polar_angle = float(0.0)
        self.__azimuthal_angle = float(0.0)

        self.__pixel_subsamples = None

        self.__use_plane_subsamples = False
        self.__use_orthogonal_samples = False
        self.__orthogonal_sample_width = float(0.0)

        self.__empty_value = float(0.0)

        self.__image_file = None

        self.__colormap_image = None
        self.__colormap_info = None

        if (metadata_filename is not None) :
            self.reset(metadata_filename)


    def reset(self, metadata_filename=None) :
    
        self.__datafile = None
        self.__datafile_dims = int(0)

        self.__dataset = None

        self.__sim_time = None

        self.__grid_center = None
        self.__grid_width = None
        self.__grid_pixels = None

        self.__image_width = None   ## In case of transposition
        self.__image_pixels = None  ## In case of transposition

        self.__roll_angle = float(0.0)
        self.__polar_angle = float(0.0)
        self.__azimuthal_angle = float(0.0)

        self.__pixel_subsamples = None

        self.__use_plane_subsamples = False
        self.__use_orthogonal_samples = False
        self.__orthogonal_sample_width = float(0.0)

        self.__empty_value = float(0.0)

        self.__image_file = None

        self.__colormap_image = None
        self.__colormap_info = None

        if (metadata_filename is not None) :
            self.__read_metadata(metadata_filename)


    def get_datafile_name(self) :
        return self.__datafile


    def get_datafile_dims(self) :
        return self.__datafile_dims


    def get_dataset(self) :
        return self.__dataset


    def get_sim_time(self) :
        return self.__sim_time


    def get_grid_center(self) :
        return self.__grid_center


    def get_grid_width(self) :
        return self.__grid_width


    def get_grid_pixels(self) :
        return self.__grid_pixels


    def get_image_width(self) :
        return self.__image_width


    def get_image_pixels(self) :
        return self.__image_pixels


    def get_roll_angle(self) :
        return self.__roll_angle


    def get_polar_angle(self) :
        return self.__polar_angle


    def get_azimuthal_angle(self) :
        return self.__azimuthal_angle


    def get_pixel_subsamples(self) :
        return self.__pixel_subsamples


    def use_plane_subsamples(self) :
        return self.__use_plane_subsamples


    def use_orthogonal_samples(self) :
        return self.__use_orthogonal_subsamples


    def get_orthogonal_sample_width(self) :
        return self.__orthogonal_sample_width


    def get_empty_value(self) :
        return self.__empty_value


    def get_image_filename(self) :
        return self.__image_file


    def get_colormap_image_filename(self) :
        return self.__colormap_image


    def get_colormap_info(self) :
        return self.__colormap_info


    def __read_metadata(self, metadata_filename) :

        f = open(metadata_filename, "r")

        eof_reached = False

        transpose_image = False

        while (not eof_reached) :

            line = f.readline()

            if (line == "") :

                eof_reached = True

            else :

                work_line = line.split("#")[0].strip()

                words = work_line.split(" ")

                num_words = len(words)

                if (num_words > 0) :
                        
                    keyword = words[0]

                    if (keyword == "DataFile") :

                        if (num_words != 2) :
                            raise RuntimeError("Improper datafile entry")

                        self.__datafile = str(words[1])

                    elif (keyword == "DataFileDims") :

                        if (num_words != 2) :
                            raise RuntimeError("Improper datafile dimensions "
                                               + "entry")

                        self.__datafile_dims = int(words[1])

                        if (self.__datafile_dims < 1) :
                            raise RuntimeError("Improper dataset dimensions")

                    elif (keyword == "Dataset") :

                        if (num_words != 2) :
                            raise RuntimeError("Improper dataset entry")

                        self.__dataset = str(words[1])

                    elif (keyword == "SimTime") :

                        if (num_words != 2) :
                            raise RuntimeError("Improper sim time entry")

                        self.__sim_time = float(words[1])

                    elif (keyword == "GridCenter") :

                        if (num_words != 2) :
                            raise RuntimeError("Improper grid center entry")

                        self.__grid_center = list()

                        grid_center_words = words[1].split(",")

                        for grid_center_elem in grid_center_words :
                            self.__grid_center.append(float(grid_center_elem))

                        ## CHECK SIZE AGAINST SPATIAL DIMENSIONS???

                    elif (keyword == "GridWidth") :

                        if (num_words != 2) :
                            raise RuntimeError("Improper grid width entry")

                        self.__grid_width = list()

                        grid_width_words = words[1].split(",")

                        for grid_width_elem in grid_width_words :

                            axis_width = float(grid_width_elem)

                            if (not (axis_width > 0.0)) :
                                raise RuntimeError("Grid width elements must "
                                                   + "be positive")
                            
                            self.__grid_width.append(float(grid_width_elem))

                        if (len(self.__grid_width) != 2) :
                            raise RuntimeError("Two grid width elements "
                                               + "required")

                    elif (keyword == "GridPixels") :

                        if (num_words != 2) :
                            raise RuntimeError("Improper grid pixels entry")

                        self.__grid_pixels = list()

                        grid_pixels_words = words[1].split(",")

                        for grid_pixels_elem in grid_pixels_words :

                            axis_pixels = int(grid_pixels_elem)

                            if (axis_pixels < 1) :
                                raise RuntimeError("Grid pixel elements must "
                                                   + "be positive")
                            
                            self.__grid_pixels.append(axis_pixels)

                        if (len(self.__grid_pixels) != 2) :
                            raise RuntimeError("Two grid pixels elements "
                                               + "required")
                        
                    elif (keyword == "RollAngle") :

                        if (num_words != 2) :
                            raise RuntimeError("Improper roll angle entry")

                        self.__roll_angle = float(words[1])

                    elif (keyword == "PolarAngle") :

                        if (num_words != 2) :
                            raise RuntimeError("Improper polar angle entry")

                        self.__polar_angle = float(words[1])

                    elif (keyword == "AzimuthalAngle") :

                        if (num_words != 2) :
                            raise RuntimeError("Improper azimuthal angle "
                                               + "entry")

                        self.__azimuthal_angle = float(words[1])

                    elif (keyword == "PixelSubsamples") :

                        if (num_words != 2) :
                            raise RuntimeError("Improper pixel subsamples "
                                               + "entry")

                        self.__pixel_subsamples = list()

                        pixel_subsamples_words = words[1].split(",")

                        for pixel_subsamples_elem in pixel_subsamples_words :

                            axis_subsamples = int(pixel_subsamples_elem)

                            if (axis_pixels < 1) :
                                raise RuntimeError("Pixel subsamples must "
                                                   + "be positive")
                            
                            self.__pixel_subsamples.append(axis_pixels)

                        num_subsample_elems = len(self.__grid_pixels)

                        if (num_subsample_elems < 2) :
                            raise RuntimeError("At least two grid pixels "
                                               + "elements required")

                        for i in range(2) :
                            if (self.__pixel_subsamples[i] > 1) :
                                self.__use_plane_subsamples = True

                        if (num_subsample_elems > 2) :
                            if (self.__pixel_subsamples[2] > 1) :
                                self.__use_orthogonal_subsamples = True

                        ## CHECK NUM_SUBSAMPLE_ELEMS AGAINST DIMENSIONS???

                    elif (keyword == "OrthogonalSampleWidth") :

                        if (num_words != 2) :
                            raise RuntimeError("Improper orthogonal sample "
                                               + "region width entry")

                        self.__orthogonal_sample_width = float(words[1])

                        if (not (self.__orthogonal_sample_width > 0.0)) :
                            raise RuntimeError("Orthogonal sample width must "
                                               + "be positive")

                        ## CHECK AGAINST PIXEL_SUBSAMPLES[2] > 1???

                    elif (keyword == "EmptyValue") :

                        if (num_words != 2) :
                            raise RuntimeError("Improper empty value entry")

                        self.__empty_value = float(words[1])

                    elif (keyword == "ImageFile") :

                        if (num_words != 2) :
                            raise RuntimeError("Improper colormap image entry")

                        self.__image_file = str(words[1])

                    elif (keyword == "ColormapImage") :

                        if (num_words != 2) :
                            raise RuntimeError("Improper colormap image entry")

                        self.__colormap_image = str(words[1])

                    elif (keyword == "ColormapInfo") :

                        if (num_words != 2) :
                            raise RuntimeError("Improper colormap info entry")

                        self.__colormap_info = ColormapInfo(words[1])

                    elif (keyword == "TransposeImage") :

                        if (num_words != 2) :
                            raise RuntimeError("Improper transpose entry")

                        if (words[1] == "true") :
                            transpose_image = True

        
        if (self.__grid_width is None) :
            raise RuntimeError("Grid width not set")

        if (self.__grid_pixels is None) :
            raise RuntimeError("Grid pixels not set")

        ## Set the width and pixel count of the saved image (not slice grid)

        self.__image_width = self.__grid_width
        self.__image_pixels = self.__grid_pixels

        if (transpose_image) :

            ## Reverse the width and pixel counts

            self.__image_width \
                = [ self.__grid_width[1], self.__grid_width[0] ]

            self.__image_pixels \
                = [ self.__grid_pixels[1], self.__grid_pixels[0] ]
            
        ## ANY ADDITIONAL CHECKS???
