// C++ program file slicer.cpp

/*
  By Nathan C. Hearn
     July 14, 2007

  Takes 2D slice from a 2D or 3D Flash dataset.

  For parallel operation, define the preprocessor variable USE_MPI.
*/


#include <math.h>
#include <stdlib.h>
#include <iostream>
#include <string>
#include <vector>
#include "quickflash_file_datafile.hpp"
#include "quickflash_file_dataset.hpp"
#include "quickflash_file_meshinfo.hpp"
#include "quickflash_file_siminfo.hpp"
#include "quickflash_block_blockinfo.hpp"
#include "quickflash_block_blockdata.hpp"
#include "quickflash_except.hpp"
#include "quickflash_slice_slicegrid.hpp"
#include "quickflash_slice_slicer.hpp"
#include "quickflash_color_colormap.hpp"
#include "quickflash_file_flashdefs.hpp"
#include "quickflash_utils_text.hpp"
#include "quickflash_image_imagegrid.hpp"
#include "quickflash_image_imagemagick.hpp"
#include "quickflash_slice_output.hpp"


#ifdef USE_MPI
#include <mpi.h>
#endif  // USE_MPI


const unsigned int read_buffer_size = 1000;
const unsigned int read_cache_size = 1000;


void determine_distrib(const unsigned int num_processes,
		       const unsigned int num_pixels,
		       std::vector<unsigned int> & start_indexes,
		       std::vector<unsigned int> & pixel_counts)
  {
  if (num_processes < 1)
    throw QuickFlash::Except("Process count must be positive", __FILE__, 
			   __LINE__);

  start_indexes.resize(num_processes);
  pixel_counts.resize(num_processes);

  const unsigned int mean_pixels = num_pixels / num_processes;
  const unsigned int excess_pixels = num_pixels % num_processes;

  const unsigned int mean_pixels_plusOne = mean_pixels + 1;

  unsigned int last_index = 0;

  for (unsigned int m = 0; m < excess_pixels; m++)
    {
    start_indexes[m] = last_index;
    pixel_counts[m] = mean_pixels_plusOne;

    last_index += mean_pixels_plusOne;
    }

  for (unsigned int m = excess_pixels; m < num_processes; m++)
    {
    start_indexes[m] = last_index;
    pixel_counts[m] = mean_pixels;

    last_index += mean_pixels;
    }

  if (last_index != num_pixels)
    throw QuickFlash::Except("Unable to compute distribution", __FILE__, 
			   __LINE__);
  }
 

void populate_grid(const QuickFlash::File::DataFile & df, 
		   const std::string & dataset_name, 
		   const std::vector<unsigned int> & subsample_dims,
		   QuickFlash::Slice::SliceGrid & grid)
  {
  // Set up MPI-related constants

  // unsigned int proc_id = 0;
  // unsigned int num_procs = 1;

#ifdef USE_MPI
#endif  // USE_MPI

  // Open file info objects

  const QuickFlash::File::MeshInfo & meshinfo = df.get_mesh_info();

  const QuickFlash::File::Dataset & dset = df.get_dataset(dataset_name);

  dset.set_cache_size(read_cache_size);
  dset.set_report_stats();

  QuickFlash::Slice::slice_simple(meshinfo, dset, subsample_dims, grid);
  }


#ifdef USE_MAGICK
int write_image(const std::string & filename, 
		const QuickFlash::Slice::SliceGrid & data,
		const bool transpose_image)
  {
  int ret_val = -1;  // Error state

  // Build the colormap

  double min_val = 0.0;
  double max_val = 0.0;

  data.get_minmax_values(min_val, max_val);

  if (!(min_val < max_val))
    max_val = (2.0 * fabs(min_val)) + 1.0;  // Make something up ...

  QuickFlash::Color::Colormap colormap(min_val, max_val);

  QuickFlash::Image::ImageGrid image(data.get_grid_data(), colormap);

  ret_val = QuickFlash::Image::ImageMagick::write_image(filename, image, 
						      transpose_image);

  return ret_val;
  }
#endif  // USE_MAGICK


#ifdef USE_MAGICK
int write_image(const std::string & filename, 
		const std::string & colormap_filename, 
		const QuickFlash::Slice::SliceGrid & data,
		const bool transpose_image, const bool rescale_colormap=false)
  {
  using std::cerr;
  using std::endl;

  int ret_val = -1;  // Error state

  // Load the colormap

  QuickFlash::Color::Colormap colormap;

  if (load_colormap(colormap_filename, colormap) >= 0)
    {
    if (rescale_colormap)
      {
      double min_val = 0.0;
      double max_val = 0.0;

      data.get_minmax_values(min_val, max_val);

      // WHAT IF min_val == max_val ???

      if (min_val < max_val)
	colormap.rescale_values(min_val, max_val);
      }

    QuickFlash::Image::ImageGrid image(data.get_grid_data(), colormap);

    ret_val = QuickFlash::Image::ImageMagick::write_image(filename, image, 
							transpose_image);
    }
  else
    cerr << endl << "Error: Unable to load colormap " << colormap_filename 
	 << endl;

  return ret_val;
  }
#endif  // USE_MAGICK
    

int main(int argc, char * argv[])
  {
  using std::cerr;
  using std::endl;

  const double deg_to_rad = (M_PI) / 180.0;

  const int min_args = 8;

  if (argc < min_args)
    {
    cerr << endl << "Usage: " << argv[0] << " center_x,center_y[,center_z]"
	 << endl << "\t"
	 << "grid_x_width grid_y_width grid_x_pixels grid_y_pixels"
	 << endl << "\t"
	 << "< roll_angle polar_angle > azimuthal_angle "
	 << endl << "\t"
	 << "[ -z num_z_samples z_width ] [ -sub subsamples_x subsamples_y ]"
	 << endl << "\t"
	 << "[ -o output_prefix ] [ -symm horiz | vert ]"
         << endl << "\t"
	 << "[ -i image_suffix ] [ -cm colormap_filename ] [ -cm-rescale ]"
	 << endl << "\t"
	 << "[ -image-only ] [ -transpose-image ]"
	 << endl << "\t"
	 << " flash_filename1 [ flash_filename2 ... ]"
	 << endl << endl;
    return 1;
    }

  int argptr = 1;

  int req_args = min_args;


  const std::string center_pos_string = argv[argptr++];

  std::vector<double> grid_center;

  QuickFlash::Utils::read_vector_string(center_pos_string, grid_center);

  const unsigned int space_dims = grid_center.size();

  if ((space_dims < 2) || (space_dims > 3))
    {
    cerr << endl << "Error: Only 2D and 3D datasets supported" << endl << endl;
    return -1;
    }

  std::vector<double> grid_width(2);

  for (unsigned int m = 0; m < 2; m++)
    grid_width[m] = atof(argv[argptr++]);

  std::vector<unsigned int> grid_pixels(2);

  for (unsigned int m = 0; m < 2; m++)
    grid_pixels[m] = static_cast<unsigned int>(atoi(argv[argptr++]));

  // Only read the roll and polar angles for 3D positions

  if (space_dims > 2)
    req_args += 2;  // Additional rotation angles required

  if (argc < req_args)
    {
    cerr << endl << "Error: Insufficient arguments for 3D" << endl << endl;
    return -1;
    }

  const double roll_angle = (space_dims == 3) ? atof(argv[argptr++]) : 0.0;
  const double polar_angle = (space_dims == 3) ? atof(argv[argptr++]) : 0.0;

  const double azimuthal_angle = atof(argv[argptr++]);

  // Optional arguments

  std::vector<unsigned int> subsample_dims(space_dims, 1);

  double z_width = 0.0;

  bool output_file_prefix_given = false;
  std::string base_output_file_prefix;

  bool compute_symmetry = false;
  unsigned int symmetry_axis = 0;

  bool output_image = false;
  std::string image_suffix;

  bool use_colormap = false;
  std::string colormap_filename;

  bool rescale_colormap = false;

  bool output_data = true;

  bool transpose_image = false;

  bool finished_options = false;

  while (!finished_options)
    {
    if (argc < req_args)
      {
      cerr << "Error: Insufficient arguments" << endl << endl;
      return -1;
      }

    // Don't increment argptr unless keyword really is an option flag

    const std::string keyword = argv[argptr];  // Don't increment yet!

    if (keyword == "-sub")
      {
      argptr++;  // OK ... increment it

      req_args += 3;

      if (argc < req_args)
	{
	cerr << "Error: Subsamples must have two dimensions" << endl << endl;
	return -1;
	}

      for (unsigned int m = 0; m < 2; m++)
	{
	const unsigned int axis_dim 
	  = static_cast<unsigned int>(atoi(argv[argptr++]));

	if (axis_dim < 1)
	  {
	  cerr << "Error: Subsample counts must be positive" << endl << endl;
	  return -1;
	  }

	subsample_dims[m] = axis_dim;
	}
      }
    else if (keyword == "-z")
      {
      argptr++;  // OK ... increment it

      req_args += 3;

      if (argc < req_args)
	{
	cerr << endl 
	     << "Error: z-samples option requires sample count and width" 
	     << endl << endl;
	return -1;
	}

      if (space_dims < 3)
	{
	cerr << endl << "Error: z-samples only supported in 3D" << endl 
	     << endl;
	return -1;
	}

      subsample_dims[2] = static_cast<unsigned int>(atoi(argv[argptr++]));

      if (subsample_dims[2] < 1)
	{
	cerr << "Error: At least one z sample required" << endl << endl;
	return -1;
	}

      z_width = atof(argv[argptr++]);

      if (!(z_width > 0.0))
	{
	cerr << "Error: z-width must be positive" << endl << endl;
	return -1;
	}
      }
    else if (keyword == "-o")
      {
      argptr++;  // OK ... increment it

      req_args += 2;

      if (argc < req_args)
	{
	cerr << "Error: Output file option requires a file name prefix" << endl
	     << endl;
	return -1;
	}

      base_output_file_prefix = argv[argptr++];

      output_file_prefix_given = true;
      }
    else if (keyword == "-symm")
      {
      argptr++;  // OK ... increment it

      req_args += 2;

      if (argc < req_args)
	{
	cerr << "Error: Symmetry computation index requires an axis specifier" 
	     << endl << endl;
	return -1;
	}

      const std::string symmetry_type = argv[argptr++];

      if (symmetry_type == "horiz")
	symmetry_axis = 0;
      else if (symmetry_type == "vert")
	symmetry_axis = 1;
      else
	{
	cerr << "Error: Symmetry axis must be either horiz or vert" << endl
	     << endl;
	return -1;
	}

      compute_symmetry = true;
      }
    else if (keyword == "-i")
      {
      argptr++;  // OK ... increment it

      req_args += 2;

      if (argc < req_args)
	{
	cerr << "Error: Image output option requires a suffix (e.g., png)" 
	     << endl << endl;
	return -1;
	}

      image_suffix = argv[argptr++];

      output_image = true;
      }
    else if (keyword == "-cm")
      {
      argptr++;  // OK ... increment it

      req_args += 2;

      if (argc < req_args)
	{
	cerr << "Error: Colormap option requires a colormap file name" 
	     << endl << endl;
	return -1;
	}

      colormap_filename = argv[argptr++];

      use_colormap = true;
      }
    else if (keyword == "-cm-rescale")
      {
      argptr++;  // OK ... increment it

      req_args += 1;

      rescale_colormap = true;
      }
    else if (keyword == "-image-only")
      {
      argptr++;  // OK ... increment it

      req_args += 1;

      output_data = false;
      }
    else if (keyword == "-transpose-image")
      {
      argptr++;  // OK ... increment it

      req_args += 1;

      transpose_image = true;
      }
    else
      finished_options = true;
    }


  // Read the filenames

  std::list<std::string> flash_filenames;

  while (argptr < argc)
    flash_filenames.push_back(argv[argptr++]);

  const unsigned int num_files = flash_filenames.size();


  // Check the number of subsamples

  unsigned int num_plane_subsamples = 1;

  for (unsigned int i = 0; i < 2; i++)
    num_plane_subsamples *= subsample_dims[i];

  if (num_plane_subsamples < 1)
    {
    cerr << "Error: Non-positive subsample dimension" << endl << endl;
    return -1;
    }


  std::list<std::string>::const_iterator flash_filename_iter 
    = flash_filenames.begin();

  while (flash_filename_iter != flash_filenames.end())
    {
    const std::string & filename = *flash_filename_iter;

    // If no ouput prefix is specified, use the input filename

    std::string output_file_prefix;

    if (!output_file_prefix_given)
      output_file_prefix = filename;
    else
      {
      output_file_prefix = base_output_file_prefix;

      // If generating more than one output file, include the filename

      if (num_files > 1)
	output_file_prefix += "_" + filename;
      }

    // OTHER ERROR CHECKING???


    // Build the report strings

    std::list<std::string> comment_list;

    std::stringstream strstream;

    strstream << "Flash filename [ " << filename << " ]";

    comment_list.push_back(strstream.str());

    strstream.str("");

    strstream << "Grid center [ " << grid_center << " ]";

    comment_list.push_back(strstream.str());

    strstream.str("");

    strstream << "Grid width [";

    for (unsigned int m = 0; m < 2; m++)
      strstream << " " << grid_width[m];

    strstream << " ]";

    comment_list.push_back(strstream.str());

    strstream.str("");

    strstream << "Grid pixels [";

    for (unsigned int m = 0; m < 2; m++)
      strstream << " " << grid_pixels[m];

    strstream << " ]";

    comment_list.push_back(strstream.str());

    strstream.str("");

    if (space_dims > 2)
      {
      strstream << "Roll angle [ " << roll_angle << " ] degrees";

      comment_list.push_back(strstream.str());

      strstream.str("");

      strstream << "Polar angle [ " << polar_angle << " ] degrees";

      comment_list.push_back(strstream.str());

      strstream.str("");
      }

    strstream << "Azimuthal angle [ " << azimuthal_angle << " ] degrees";

    comment_list.push_back(strstream.str());

    strstream.str("");

    if (num_plane_subsamples > 1)
      {
      strstream << "Pixel subsample dimensions [";

      for (unsigned int m = 0; m < 2; m++)
	strstream << " " << subsample_dims[m];

      strstream << " ]";

      comment_list.push_back(strstream.str());

      strstream.str("");
      }

    if (space_dims > 2)
      if (subsample_dims[2] > 1)
	{
	strstream << "Orthogonal samples [ " << subsample_dims[2] 
		  << " ] domain width [ " << z_width << " ]" << endl;

	comment_list.push_back(strstream.str());

	strstream.str("");
	}


    // Report the setup to stderr

    std::list<std::string>::const_iterator str_iter = comment_list.begin();
    std::list<std::string>::const_iterator end_str_iter = comment_list.end();

    while (str_iter != end_str_iter)
      {
      cerr << *str_iter << endl;
      ++str_iter;
      }


    // Build radian versions of the angles

    const double roll_rad = roll_angle * deg_to_rad;
    const double polar_rad = polar_angle * deg_to_rad;
    const double azimuthal_rad = azimuthal_angle * deg_to_rad;

    // Construct the grid

    QuickFlash::Slice::SliceGrid grid(grid_center, grid_width, grid_pixels, 
				    azimuthal_rad, polar_rad, roll_rad, 
				    z_width);

    // Open the file

    QuickFlash::File::DataFile df(filename, read_buffer_size, true);

    const QuickFlash::File::SimInfo & sim_info = df.get_sim_info();

    // Get the variable names

    const std::vector<std::string> & var_names 
      = sim_info.get_mesh_variable_names();

    const unsigned int num_vars = var_names.size();

    for (unsigned int var_index = 0; var_index < num_vars; var_index++)
      {
      const std::string & dataset_name = var_names[var_index];

      std::list<std::string> current_comments = comment_list;

      const std::string dataset_comment = "Dataset [ " + dataset_name + " ]";

      cerr << "Processing dataset [ " << dataset_name << " ]" << endl;

      current_comments.push_back(dataset_comment);

      const std::string current_output_file_prefix 
	= output_file_prefix + dataset_name;

      // Fill the grid

      populate_grid(df, dataset_name, subsample_dims, grid);

      if (output_data)
	{
	const std::string output_filename 
	  = current_output_file_prefix + ".csv";

	QuickFlash::Slice::write_grid(output_filename, grid, current_comments);
	}

      if (output_image)
	{
	const std::string image_filename 
	  = current_output_file_prefix + "." + image_suffix;

	int error_state = -1;  // Error state

#ifdef USE_MAGICK
	if (use_colormap)
	  error_state = write_image(image_filename, colormap_filename, grid,
				    transpose_image, rescale_colormap);
	else
	  error_state = write_image(image_filename, grid, transpose_image);
#endif  // USE_MAGICK

	if (error_state < 0)
	  {
	  cerr << endl << "Error: Unable to write image " << image_filename 
	       << endl << endl;
	  return -2;
	  }
	}

      // Print the symmetry values -- WARNING: UGLY CODE AHEAD!!!

      if (compute_symmetry)
	{
	std::list<std::string> working_comments;

	QuickFlash::Slice::SliceGrid sum_grid;

	sum_grid.set_symmetry_sum(symmetry_axis, grid);

	QuickFlash::Slice::SliceGrid diff_grid;

	diff_grid.set_symmetry_diff(symmetry_axis, grid);

	if (output_data)
	  {
	  const std::string output_filename 
	    = current_output_file_prefix + "_sum.csv";

	  working_comments = current_comments;

	  strstream << "Symmetry summation data for axis [ " << symmetry_axis 
		    << " ]";

	  working_comments.push_back(strstream.str());

	  strstream.str("");

	  QuickFlash::Slice::write_grid(output_filename, sum_grid, 
				      working_comments);
	  }

	if (output_image)
	  {
	  const std::string image_filename 
	    = current_output_file_prefix + "_sum." + image_suffix;

	  int error_state = -1;  // Error state

#ifdef USE_MAGICK
	  error_state = write_image(image_filename, sum_grid, transpose_image);
#endif  // USE_MAGICK

	  if(error_state < 0)
	    {
	    cerr << endl << "Error: Unable to write image " << image_filename
		 << endl << endl;
	    return -2;
	    }
	  }

	if (output_data)
	  {
	  const std::string output_filename 
	    = current_output_file_prefix + "_diff.csv";

	  working_comments = current_comments;

	  strstream << "Symmetry difference data for axis [ " << symmetry_axis 
		    << " ]";
	
	  working_comments.push_back(strstream.str());

	  strstream.str("");

	  QuickFlash::Slice::write_grid(output_filename, diff_grid, 
				      working_comments);
	  }

	if (output_image)
	  {
	  const std::string image_filename 
	    = current_output_file_prefix + "_diff." + image_suffix;

	  int error_state = -1;  // Error state

#ifdef USE_MAGICK
	  error_state = write_image(image_filename, diff_grid, 
				    transpose_image);
#endif  // USE_MAGICK

	  if (error_state < 0)
	    {
	    cerr << endl << "Error: Unable to write image " << image_filename
		 << endl << endl;
	    return -2;
	    }
	  }
	}
      }

    // Next file ...

    ++flash_filename_iter;
    }

  return 0;
  }
