// C++ program file analyze_sn.cpp

/*
  By Nathan C. Hearn
     August 22, 2008

  Driver program for a suite of run-time analysis tools.
*/



#include <iostream>
#include <vector>
#include <string>
#include <list>
#include <cstdlib>
#include <cmath>
#include <fstream>
#include "quickflash_file_datafile.hpp"
#include "quickflash_file_meshinfo.hpp"
#include "quickflash_file_dataset.hpp"
#include "quickflash_tree_block.hpp"
#include "quickflash_block_blockinfo.hpp"
#include "quickflash_block_blockdata.hpp"
#include "quickflash_utils_commandline.hpp"
#include "quickflash_utils_file.hpp"
#include "quickflash_utils_text.hpp"
#include "quickflash_slice_slicegrid.hpp"
#include "quickflash_slice_slicer.hpp"
#include "quickflash_slice_output.hpp"
#include "quickflash_stats_bins.hpp"


const unsigned int buffer_size = 25;
const unsigned int cache_size = 2000;

const double yz_azim_angle = 0.5 * (M_PI);
const double yz_polar_angle = 0.5 * (M_PI);
const double yz_roll_angle = 0.0;

const double zx_azim_angle = 0.0;
const double zx_polar_angle = -0.5 * (M_PI);
const double zx_roll_angle = -0.5 * (M_PI);


int get_min_cell_width(const QuickFlash::File::MeshInfo & meshinfo, 
		       unsigned int & refine_level,
		       double & min_cell_width, 
		       const bool use_refine_level_limit=false,
		       const unsigned int refine_level_limit=1)
  {
  int ret_val = -1;  // Error level

  const unsigned int mesh_max_refine_level 
    = meshinfo.get_max_leaf_refine_level();

  const unsigned int max_refine_level 
    = (use_refine_level_limit) 
      ? std::min(mesh_max_refine_level, refine_level_limit) 
      : mesh_max_refine_level;

  if (max_refine_level > 0)
    {
    const unsigned int dims = meshinfo.get_dims();

    // NOTE: Block refinement levels start at one, not zero

    const unsigned int refine_level_mult = 1 << (max_refine_level - 1);

    const std::vector<double> & vol_min_bounds 
      = meshinfo.get_volume_minbounds();
    
    const std::vector<double> & vol_max_bounds 
      = meshinfo.get_volume_maxbounds();

    const std::vector<unsigned int> & base_blocks 
      = meshinfo.get_base_block_dims();

    const std::vector<unsigned int> & block_dims = meshinfo.get_block_dims();

    bool found_cell_width = false;

    double smallest_cell_width = 0.0;

    for (unsigned int i = 0; i < dims; i++)
      {
      const double axis_vol_width = vol_max_bounds[i] - vol_min_bounds[i];

      const unsigned int axis_num_cells 
	= block_dims[i] * base_blocks[i] * refine_level_mult;

      if (axis_num_cells > 0)
	{
	const double axis_cell_width 
	  = axis_vol_width / static_cast<double>(axis_num_cells);

	if (!found_cell_width)
	  {
	  smallest_cell_width = axis_cell_width;

	  found_cell_width = true;
	  }
	else if (axis_cell_width < smallest_cell_width)
	  smallest_cell_width = axis_cell_width;
	}
      }

    if (found_cell_width && (smallest_cell_width > 0.0))
      {
      refine_level = max_refine_level;
      min_cell_width = smallest_cell_width;

      ret_val = 0;  // OK
      }
    }
    
  return ret_val;
  }


int find_min_var_bounds(const QuickFlash::File::DataFile & df, 
			const std::string & var_name,
			const double var_min_value, 
			std::vector<double> & var_min_bounds,
			std::vector<double> & var_max_bounds)
  {
  int ret_val = -1;  // Error state

  const QuickFlash::File::MeshInfo & meshinfo = df.get_mesh_info();

  const unsigned int dims = meshinfo.get_dims();

  var_min_bounds.resize(dims);
  var_max_bounds.resize(dims);

  const QuickFlash::File::Dataset & dset = df.get_dataset(var_name);

  QuickFlash::Block::BlockData<double> block_data;

  std::vector<double> cell_min_bounds;
  std::vector<double> cell_max_bounds;

  // Run through the tree

  bool found_interior = false;

  QuickFlash::Tree::ConstTreeIter node_iter = meshinfo.get_tree_iter();

  while (node_iter.valid_node())
    {
    const QuickFlash::Block::BlockInfo & block_info 
      = node_iter.get_block_info();

    const std::vector<double> & block_min_bounds = block_info.get_min_coords();
    const std::vector<double> & block_max_bounds = block_info.get_max_coords();

    bool skip_branch = false;

    if (found_interior)
      {
      skip_branch = true;  // Unless part of block falls outside found region

      for (unsigned int i = 0; i < dims; i++)
	if ((block_min_bounds[i] < var_min_bounds[i])
	    || (block_max_bounds[i] > var_max_bounds[i]))
	  skip_branch = false;
      }

    if ((!skip_branch) && node_iter.is_leaf())
      {
      const unsigned int block_index = node_iter.get_block_index();

      dset.get_block_data(block_index, block_data);

      const unsigned int num_cells = block_info.get_num_cells();

      // Find a cell in the desired range

      for (unsigned int cell_index = 0; cell_index < num_cells; cell_index++)
	{
	const double cell_var_value = block_data[cell_index];

	if (cell_var_value > var_min_value)
	  {
	  block_info.get_cell_bounds(cell_index, cell_min_bounds, 
				     cell_max_bounds);

	  if (!found_interior)
	    {
	    var_min_bounds = cell_min_bounds;
	    var_max_bounds = cell_max_bounds;

	    found_interior = true;
	    }
	  else
	    {
	    for (unsigned int i = 0; i < dims; i++)
	      {
	      const double axis_cell_bound = cell_min_bounds[i];

	      if (axis_cell_bound < var_min_bounds[i])
		var_min_bounds[i] = axis_cell_bound;
	      }

	    for (unsigned int i = 0; i < dims; i++)
	      {
	      const double axis_cell_bound = cell_max_bounds[i];

	      if (axis_cell_bound > var_max_bounds[i])
		var_max_bounds[i] = axis_cell_bound;
	      }
	    }
	  }
	}
      }
    
    if (skip_branch)
      node_iter.next_branch();
    else
      node_iter.next();
    }

  if (found_interior)
    ret_val = 0;  // OK

  dset.clear_cache();

  return ret_val;
  }
		     

int find_min_var_bounds(const QuickFlash::File::DataFile & df, 
			const char var_name[],
			const double var_min_value, 
			std::vector<double> & var_min_bounds,
			std::vector<double> & var_max_bounds)
  {
  const std::string var_name_str = var_name;

  return find_min_var_bounds(df, var_name_str, var_min_value, var_min_bounds,
			     var_max_bounds);
  }


int find_fuel_cm(const QuickFlash::File::DataFile & df,
		 const std::string & prog_var,
		 const std::vector<double> & min_region_bounds,
		 const std::vector<double> & max_region_bounds,
		 double & fuel_mass, std::vector<double> & fuel_cm,
		 double & fuel_cm_velz)
  {
  int ret_val = -1;  // Error state

  const QuickFlash::File::MeshInfo & meshinfo = df.get_mesh_info();

  const unsigned int dims = meshinfo.get_dims();

  const QuickFlash::File::Dataset & dens_set = df.get_dataset("dens");
  const QuickFlash::File::Dataset & velz_set = df.get_dataset("velz");
  const QuickFlash::File::Dataset & prog_set = df.get_dataset(prog_var);

  QuickFlash::Block::BlockData<double> dens_data;
  QuickFlash::Block::BlockData<double> velz_data;
  QuickFlash::Block::BlockData<double> prog_data;

  std::vector<double> cell_center;

  // Run through the tree

  std::vector<double> accum_pos_mass(dims, 0.0);  // cm = accum_pos_mass / mass
  double accum_velz_mass = 0.0;  // cm_vel_z = accum_velz_mass / mass

  double accum_mass = 0.0;

  QuickFlash::Tree::ConstTreeIter node_iter = meshinfo.get_tree_iter();

  while (node_iter.valid_node())
    {
    const QuickFlash::Block::BlockInfo & block_info 
      = node_iter.get_block_info();

    const std::vector<double> & block_min_bounds = block_info.get_min_coords();
    const std::vector<double> & block_max_bounds = block_info.get_max_coords();

    bool skip_branch = false;

    for (unsigned int i = 0; i < dims; i++)
      if ((block_min_bounds[i] > max_region_bounds[i])
	  || (block_max_bounds[i] < min_region_bounds[i]))
	skip_branch = true;

    if ((!skip_branch) && node_iter.is_leaf())
      {
      const unsigned int block_index = node_iter.get_block_index();

      dens_set.get_block_data(block_index, dens_data);
      velz_set.get_block_data(block_index, velz_data);
      prog_set.get_block_data(block_index, prog_data);

      const unsigned int num_cells = block_info.get_num_cells();

      // Run through the cells

      for (unsigned int cell_index = 0; cell_index < num_cells; cell_index++)
	{
	block_info.get_cell_center(cell_index, cell_center);

	bool use_cell = true;

	for (unsigned int i = 0; i < dims; i++)
	  {
	  const double axis_cell_center = cell_center[i];

	  if ((axis_cell_center < min_region_bounds[i]) 
	      || (axis_cell_center > max_region_bounds[i]))
	    use_cell = false;
	  }

	if (use_cell)
	  {
	  const double cell_volume = block_info.get_cell_volume(cell_index);

	  const double cell_dens_value = dens_data[cell_index];
	  const double cell_velz_value = velz_data[cell_index];
	  const double cell_prog_value = prog_data[cell_index];

	  const double cell_fuel_fract = 1.0 - cell_prog_value;

	  const double cell_fuel_dens = cell_fuel_fract * cell_dens_value;

	  const double cell_fuel_mass = cell_volume * cell_fuel_dens;

	  for (unsigned int i = 0; i < dims; i++)
	    accum_pos_mass[i] += cell_center[i] * cell_fuel_mass;

	  accum_velz_mass += cell_velz_value * cell_fuel_mass;

	  accum_mass += cell_fuel_mass;
	  }
	}
      }
    
    if (skip_branch)
      node_iter.next_branch();
    else
      node_iter.next();
    }

  if (accum_mass > 0.0)
    {
    fuel_cm.resize(dims);

    const double oneOver_accum_mass = 1.0 / accum_mass;

    for (unsigned int i = 0; i < dims; i++)
      fuel_cm[i] = accum_pos_mass[i] * oneOver_accum_mass;

    fuel_cm_velz = accum_velz_mass * oneOver_accum_mass;

    fuel_mass = accum_mass;

    ret_val = 0;  // OK
    }

  dens_set.clear_cache();
  velz_set.clear_cache();
  prog_set.clear_cache();

  return ret_val;
  }


int get_radial_mass_vel_bins(const QuickFlash::File::DataFile & df,
			     const std::vector<double> & center_coords,
			     const std::vector<double> & min_region_bounds,
			     const std::vector<double> & max_region_bounds,
			     const double bin_width,
			     QuickFlash::Stats::BinArray & mass_bins,
			     QuickFlash::Stats::BinArray & rad_vel_bins,
			     double & total_mass)
  {
  int ret_val = -1;  // Error state

  const QuickFlash::File::MeshInfo & meshinfo = df.get_mesh_info();

  const unsigned int dims = meshinfo.get_dims();

  if ((dims > 0) && (dims < 4))
    {
    bool center_ok = true;  // Make sure center fits in region

    for (unsigned int i = 0; i < dims; i++)
      {
      const double axis_center_coord = center_coords[i];

      if ((!(axis_center_coord > min_region_bounds[i])) 
	  || (!(axis_center_coord < max_region_bounds[i])))
	center_ok = false;
      }

    if (center_ok && (bin_width > 0.0))
      {
      const QuickFlash::File::Dataset & dens_set = df.get_dataset("dens");

      // Load the velocities

      std::vector<const QuickFlash::File::Dataset *> vel_sets(dims, 0);

      vel_sets[0] = &(df.get_dataset("velx"));

      if (dims > 1)
	{
	vel_sets[1] = &(df.get_dataset("vely"));

	if (dims > 2)
	  vel_sets[2] = &(df.get_dataset("velz"));
	}
    
      // Set up the bins -- radius is max that fits entirely in domain

      bool found_region_radius = false;

      double max_region_radius = 0.0;

      for (unsigned int i = 0; i < dims; i++)
	{
	const double axis_center_coord = center_coords[i];

	const double center_min = axis_center_coord - min_region_bounds[i];

	const double max_center = max_region_bounds[i] - axis_center_coord;

	if (!found_region_radius)
	  {
	  max_region_radius = std::min(center_min, max_center);
	  found_region_radius = true;
	  }
	else
	  {
	  if (center_min < max_region_radius)
	    max_region_radius = center_min;

	  if (max_center < max_region_radius)
	    max_region_radius = max_center;
	  }
	}

      unsigned int num_bins = 1;

      if (max_region_radius > 0.0)
	num_bins += static_cast<unsigned int>(max_region_radius / bin_width);

      const double max_radius = bin_width * static_cast<double>(num_bins);

      mass_bins.reset(0.0, max_radius, num_bins);
      rad_vel_bins.reset(0.0, max_radius, num_bins);

      // Run through the data

      std::vector<double> cell_center(dims, 0.0);
      std::vector<double> cell_vel(dims, 0.0);

      std::vector<double> adj_cell_center(dims, 0.0);
      std::vector<double> normal_adj_cell_center(dims, 0.0);

      QuickFlash::Block::BlockData<double> dens_data;
      std::vector< QuickFlash::Block::BlockData<double> > vel_data(dims);

      double accum_mass = 0.0;

      QuickFlash::Tree::ConstTreeIter node_iter = meshinfo.get_tree_iter();

      while (node_iter.valid_node())
	{
	const QuickFlash::Block::BlockInfo & block_info 
	  = node_iter.get_block_info();

	const std::vector<double> & block_min_bounds 
	  = block_info.get_min_coords();

	const std::vector<double> & block_max_bounds 
	  = block_info.get_max_coords();

	bool skip_branch = false;

	for (unsigned int i = 0; i < dims; i++)
	  if ((block_min_bounds[i] > max_region_bounds[i])
	      || (block_max_bounds[i] < min_region_bounds[i]))
	    skip_branch = true;

	if ((!skip_branch) && node_iter.is_leaf())
	  {
	  const unsigned int block_index = node_iter.get_block_index();

	  dens_set.get_block_data(block_index, dens_data);

	  for (unsigned int i = 0; i < dims; i++)
	    vel_sets[i]->get_block_data(block_index, vel_data[i]);

	  // Run through the cells
	
	  const unsigned int num_cells = block_info.get_num_cells();

	  for (unsigned int cell_index = 0; cell_index < num_cells; 
	       cell_index++)
	    {
	    block_info.get_cell_center(cell_index, cell_center);

	    bool use_cell = true;

	    for (unsigned int i = 0; i < dims; i++)
	      {
	      const double axis_cell_center = cell_center[i];
	      
	      if ((axis_cell_center < min_region_bounds[i]) 
		  || (axis_cell_center > max_region_bounds[i]))
		use_cell = false;
	      }

	    if (use_cell)
	      {
	      const double cell_volume 
		= block_info.get_cell_volume(cell_index);

	      const double cell_dens = dens_data[cell_index];

	      // Get the cell position relateive to center_coords

	      for (unsigned int i = 0; i < dims; i++)
		adj_cell_center[i] = cell_center[i] - center_coords[i];

	      // Find the distance from the center to this cell

	      double cell_sq_dist = 0.0;

	      for (unsigned int i = 0; i < dims; i++)
		{
		const double axis_disp = adj_cell_center[i];

		cell_sq_dist += axis_disp * axis_disp;
		}

	      const double cell_dist = std::sqrt(cell_sq_dist);

	      if (mass_bins.in_bounds(cell_dist))
		{
		mass_bins.add_data(cell_dist, cell_dens, cell_volume);

		const double cell_mass = cell_dens * cell_volume;

		// Get the radial velocity and add to vel_bins

		double rad_vel_comp = 0.0;  // Define r=0 as rad_vel=0

		if (cell_dist > 0.0)
		  {
		  const double oneOver_cell_dist = 1.0 / cell_dist;

		  for (unsigned int i = 0; i < dims; i++)
		    normal_adj_cell_center[i] 
		      = adj_cell_center[i] * oneOver_cell_dist;

		  // Compute dot product of velocity and normal_adj_cell_center

		  for (unsigned int i = 0; i < dims; i++)
		    {
		    const double axis_vel = vel_data[i][cell_index];

		    rad_vel_comp += axis_vel * normal_adj_cell_center[i];
		    }
		  }

		rad_vel_bins.add_data(cell_dist, rad_vel_comp, cell_mass);
	
		accum_mass += cell_mass;
		}
	      }
	    }
	  }

	if (skip_branch)
	  node_iter.next_branch();
	else
	  node_iter.next();
	}
      
      if (accum_mass > 0.0)
	{
	total_mass = accum_mass;
	ret_val = 0;  // OK
	}

      // Clear the caches

      dens_set.clear_cache();

      for (unsigned int i = 0; i < dims; i++)
	{
	const QuickFlash::File::Dataset & vset = *(vel_sets[i]);

	vset.clear_cache();
	}
      }
    }

  return ret_val;
  }


int get_axis_dens_vel(const QuickFlash::File::DataFile & df,
		      const unsigned int axis, const bool positive_axis,
		      const std::vector<double> & center_coords,
		      const double axis_min_region_bound,
		      const double axis_max_region_bound,
		      const double sample_spacing,
		      std::vector<double> & rel_sample_positions,
		      std::vector<double> & sample_densities,
		      std::vector<double> & sample_radial_velocities)
  {
  int ret_val = -1;  // Error state

  const QuickFlash::File::MeshInfo & meshinfo = df.get_mesh_info();

  const unsigned int dims = meshinfo.get_dims();

  if ((dims > 0) && (dims < 4) && (axis < dims) && (sample_spacing > 0.0))
    {
    if ((center_coords[axis] > axis_min_region_bound)
	&& (center_coords[axis] < axis_max_region_bound))
      {
      const QuickFlash::File::Dataset & dens_set = df.get_dataset("dens");

      // Load the velocities

      std::string vel_name;

      switch (axis)
	{
	case 0 :
	  vel_name = "velx";
	  break;

	case 1 :
	  vel_name = "vely";
	  break;

	case 2 :
	  vel_name = "velz";
	  break;
	}

      const QuickFlash::File::Dataset & vel_set = df.get_dataset(vel_name);
    
      // Set up the samples -- radius is max that fits entirely in domain

      const double axis_center_coord = center_coords[axis];

      const double center_min = axis_center_coord - axis_min_region_bound;

      const double max_center = axis_max_region_bound - axis_center_coord;

      const double max_region_radius = std::min(center_min, max_center);

      unsigned int num_bins = 1;

      if (max_region_radius > 0.0)
	num_bins += static_cast<unsigned int>(max_region_radius 
					      / sample_spacing);

      rel_sample_positions.resize(num_bins);

      sample_densities.resize(num_bins);
      sample_radial_velocities.resize(num_bins);

      // Sample data at the bin centers

      std::vector<double> sample_pos = center_coords;

      bool block_loaded = false;
      unsigned int current_block = 0;

      QuickFlash::Block::BlockData<double> dens_data;
      QuickFlash::Block::BlockData<double> vel_data;

      double current_radius = 0.5 * sample_spacing;

      const double pos_axis_mult = positive_axis ? 1.0 : -1.0;
      
      for (unsigned int bin_index = 0; bin_index < num_bins; bin_index++)
	{
	const double center_adj = current_radius * pos_axis_mult;

	sample_pos[axis] = center_coords[axis] + center_adj;

	unsigned int block_index = 0;
	unsigned int cell_index = 0;

	rel_sample_positions[bin_index] = current_radius;

	if (meshinfo.in_bounds(sample_pos))
	  {
	  meshinfo.get_cell_index(sample_pos, block_index, cell_index);

	  if ((!block_loaded) || (current_block != block_index))
	    {
	    dens_set.get_block_data(block_index, dens_data);
	    vel_set.get_block_data(block_index, vel_data);

	    block_loaded = true;
	    current_block = block_index;
	    }

	  const double dens_value = dens_data[cell_index];
	  const double vel_value = vel_data[cell_index];

	  sample_densities[bin_index] = dens_value;
	  sample_radial_velocities[bin_index] = vel_value * pos_axis_mult;
	  }
	else
	  {
	  sample_densities[bin_index] = 0.0;
	  sample_radial_velocities[bin_index] = 0.0;
	  }

	current_radius += sample_spacing;
	}

      // Clean the caches

      dens_set.clear_cache();
      vel_set.clear_cache();

      ret_val = 0;  // OK
      }
    }

  return ret_val;
  }


int write_axis_dens_vel(const std::string & output_file_prefix,
			const QuickFlash::File::DataFile & df,
			const unsigned int axis, const bool positive_axis,
			const std::vector<double> & center_coords,
			const double axis_min_region_bound,
			const double axis_max_region_bound,
			const double sample_spacing)
  {
  using std::endl;

  int ret_val = -1;  // Error state

  std::vector<double> rel_sample_positions;
  std::vector<double> sample_densities;
  std::vector<double> sample_radial_velocities;

  ret_val = get_axis_dens_vel(df, axis, positive_axis, center_coords,
			      axis_min_region_bound, axis_max_region_bound,
			      sample_spacing, rel_sample_positions,
			      sample_densities, sample_radial_velocities);

  if (ret_val >= 0)
    {
    const unsigned int num_samples = rel_sample_positions.size();

    std::string axis_name;

    switch (axis)
      {
      case 0 :
	axis_name = "x";
	break;

      case 1 :
	axis_name = "y";
	break;

      case 2 :
	axis_name = "z";
	break;
      }

    std::string output_suffix = "axis_data_" + axis_name;

    if (positive_axis)
      output_suffix += "_pos";
    else 
      output_suffix += "_neg";

    const std::string filename = output_file_prefix + output_suffix + ".csv";

    std::ofstream f(filename.c_str());

    f << "## Axis samples" << endl;
    f << "DataFile " << df.get_filename() << endl;
    f << "Axis " << axis << endl;
    f << "AxisName " << axis_name << endl;

    f << "PositiveAxis ";

    if (positive_axis)
      f << "true";
    else
      f << "false";

    f << endl;

    f << "NumSamples " << num_samples << endl;
    f << "CenterCoords " << center_coords << endl;
    f << "SampleSpacing " << sample_spacing << endl;
    
    f << endl << "Data" << endl;

    f << endl;
    
    f << "Rel_Sample_Pos" << "\t" << "Sample_Density" << "\t" 
      << "Sample_Radial_Vel" << endl;

    for (unsigned int sample_index = 0; sample_index < num_samples;
	 sample_index++)
      {
      f << rel_sample_positions[sample_index] << "\t"
	<< sample_densities[sample_index] << "\t" 
	<< sample_radial_velocities[sample_index] << endl;
      }

    f << endl;
    }

  return ret_val;
  }


int get_cardinal_slices(const QuickFlash::File::DataFile & df,
			const std::vector<double> & center_coords,
			const unsigned int num_axis_pixels, 
			const double pixel_width, 
			const std::string & variable_name,
			QuickFlash::Slice::SliceGrid & xy_grid,
			QuickFlash::Slice::SliceGrid & yz_grid,
			QuickFlash::Slice::SliceGrid & zx_grid,
			const unsigned int axis_subsamples=0)
  {
  int ret_val = -1;  // Error state

  const QuickFlash::File::MeshInfo & meshinfo = df.get_mesh_info();

  const unsigned int dims = meshinfo.get_dims();

  if (dims > 1)
    {
    std::vector<unsigned int> subsample_dims(dims, 1);

    if (axis_subsamples > 1)
      for (unsigned int n = 0; n < 2; n++)
	subsample_dims[n] = axis_subsamples;

    const double axis_grid_width 
      = pixel_width * static_cast<double>(num_axis_pixels);

    const std::vector<double> grid_width(2, axis_grid_width);
    const std::vector<unsigned int> pixel_count(2, num_axis_pixels);

    const QuickFlash::File::Dataset & var_set = df.get_dataset(variable_name);
  
    if (dims == 2)
      {
      xy_grid.reset(center_coords, grid_width, pixel_count);
      yz_grid.reset();
      zx_grid.reset();

      QuickFlash::Slice::slice_simple(meshinfo, var_set, subsample_dims, 
				    xy_grid);

      ret_val = 0;  // OK
      }
    else if (dims == 3)
      {
      xy_grid.reset(center_coords, grid_width, pixel_count);

      yz_grid.reset(center_coords, grid_width, pixel_count, yz_azim_angle,
		    yz_polar_angle, yz_roll_angle);

      zx_grid.reset(center_coords, grid_width, pixel_count, zx_azim_angle,
		    zx_polar_angle, zx_roll_angle);

      QuickFlash::Slice::slice_simple(meshinfo, var_set, subsample_dims, 
				    xy_grid);

      QuickFlash::Slice::slice_simple(meshinfo, var_set, subsample_dims, 
				    yz_grid);

      QuickFlash::Slice::slice_simple(meshinfo, var_set, subsample_dims, 
				    zx_grid);

      ret_val = 0;  // OK
      }

    var_set.clear_cache();
    }

  return ret_val;
  }


int get_cardinal_slices(const QuickFlash::File::DataFile & df,
			const std::vector<double> & center_coords,
			const unsigned int num_axis_pixels, 
			const double pixel_width, 
			const char variable_name[],
			QuickFlash::Slice::SliceGrid & xy_grid,
			QuickFlash::Slice::SliceGrid & yz_grid,
			QuickFlash::Slice::SliceGrid & zx_grid,
			const unsigned int axis_subsamples=0)
  {
  const std::string variable_name_str = variable_name;

  return get_cardinal_slices(df, center_coords, num_axis_pixels, 
			     pixel_width, variable_name_str,
			     xy_grid, yz_grid, zx_grid, axis_subsamples);
  }


int write_cardinal_slices(const std::string & output_file_prefix,
			  const QuickFlash::File::DataFile & df,
			  const std::vector<double> & center_coords,
			  const unsigned int num_axis_pixels, 
			  const double pixel_width, 
			  const std::string & variable_name,
			  const unsigned int axis_subsamples=0)
  {
  int ret_val = -1;  // Error state

  QuickFlash::Slice::SliceGrid xy_grid;
  QuickFlash::Slice::SliceGrid yz_grid;
  QuickFlash::Slice::SliceGrid zx_grid;

  ret_val = get_cardinal_slices(df, center_coords, num_axis_pixels, 
				pixel_width, variable_name,
				xy_grid, yz_grid, zx_grid, axis_subsamples);

  if (ret_val >= 0)
    {
    const std::string xy_grid_name 
      = output_file_prefix + "_" + variable_name + "_xy.csv";

    QuickFlash::Slice::write_grid(xy_grid_name, xy_grid);

    const std::string yz_grid_name 
      = output_file_prefix + "_" + variable_name + "_yz.csv";

    QuickFlash::Slice::write_grid(yz_grid_name, yz_grid);

    const std::string zx_grid_name 
      = output_file_prefix + "_" + variable_name + "_zx.csv";

    QuickFlash::Slice::write_grid(zx_grid_name, zx_grid);
    }

  return ret_val;
  }


int get_dens_prog_velz_slices(const std::string & output_file_prefix,
			      const QuickFlash::File::DataFile & df,
			      const std::vector<double> & center_coords,
			      const unsigned int num_axis_pixels, 
			      const double pixel_width, 
			      const std::string & progress_variable_name,
			      QuickFlash::Slice::SliceGrid & xy_dens_grid,
			      QuickFlash::Slice::SliceGrid & xy_prog_grid,
			      QuickFlash::Slice::SliceGrid & xy_velz_grid,
			      const unsigned int axis_subsamples=0)
  {
  int ret_val = -1;  // Error state

  // Set up some temporary variables

  QuickFlash::Slice::SliceGrid yz_grid;
  QuickFlash::Slice::SliceGrid zx_grid;

  // Process the density grids

  ret_val = get_cardinal_slices(df, center_coords, num_axis_pixels, 
				pixel_width, "dens", xy_dens_grid, 
				yz_grid, zx_grid, axis_subsamples);

  if (ret_val >= 0)
    {
    const std::string xy_grid_name = output_file_prefix + "_dens_xy.csv";

    QuickFlash::Slice::write_grid(xy_grid_name, xy_dens_grid);

    const std::string yz_grid_name = output_file_prefix + "_dens_yz.csv";

    QuickFlash::Slice::write_grid(yz_grid_name, yz_grid);

    const std::string zx_grid_name = output_file_prefix + "_dens_zx.csv";

    QuickFlash::Slice::write_grid(zx_grid_name, zx_grid);
    }

  if (ret_val >= 0)
    {
    ret_val = get_cardinal_slices(df, center_coords, num_axis_pixels, 
				  pixel_width, progress_variable_name, 
				  xy_prog_grid, yz_grid, zx_grid,
				  axis_subsamples);

    if (ret_val >= 0)
      {
      const std::string xy_grid_name 
	= output_file_prefix + "_" + progress_variable_name + "_xy.csv";

      QuickFlash::Slice::write_grid(xy_grid_name, xy_prog_grid);

      const std::string yz_grid_name 
	= output_file_prefix + "_" + progress_variable_name + "_yz.csv";

      QuickFlash::Slice::write_grid(yz_grid_name, yz_grid);

      const std::string zx_grid_name 
	= output_file_prefix + "_" + progress_variable_name + "_zx.csv";

      QuickFlash::Slice::write_grid(zx_grid_name, zx_grid);
      }
    }

  if (ret_val >= 0)
    {
    ret_val = get_cardinal_slices(df, center_coords, num_axis_pixels, 
				  pixel_width, "velz", xy_velz_grid,
				  yz_grid, zx_grid, axis_subsamples);

    if (ret_val >= 0)
      {
      const std::string xy_grid_name = output_file_prefix + "_velz_xy.csv";

      QuickFlash::Slice::write_grid(xy_grid_name, xy_velz_grid);

      const std::string yz_grid_name = output_file_prefix + "_velz_yz.csv";

      QuickFlash::Slice::write_grid(yz_grid_name, yz_grid);

      const std::string zx_grid_name = output_file_prefix + "_velz_zx.csv";

      QuickFlash::Slice::write_grid(zx_grid_name, zx_grid);
      }
    }

  return ret_val;
  }


int main(int argc, char * argv[])
  {
  using std::cerr;
  using std::endl;

  // Parse the command line arguments

  std::vector<std::string> argument_list;
  std::list<QuickFlash::Utils::Option> option_list;

  QuickFlash::Utils::parse_command_line(argc, argv, argument_list, option_list);

  const unsigned int num_args = argument_list.size();

  const unsigned int min_args = 6;

  if (num_args < min_args)
    {
    const std::string full_exec = argv[0];

    std::string exec_path;
    std::string exec_name;

    QuickFlash::Utils::split_name_path_file(full_exec, exec_path, exec_name);

    cerr << endl << "Usage: " << exec_name << " main_output_file min_density"
	 << endl << "\t"
	 << "progress_var progress_var_min_value radial_bin_width"
	 << endl << "\t"
	 << "[ --output-prefix=OUTPUT_PREFIX ]"
	 << endl << "\t"
	 << "[ --slice-datasets=DATASET1,DATASET2,... ]"
	 << endl << "\t"
	 << "[ --max-slice-refine-level=REFINEMENT_LEVEL ]"
	 << endl << "\t"
	 << "DATA_FILE_1 [ DATA_FILE_2 ... ]"
	 << endl << endl;

    return 1;
    }

  // Read the arguments

  unsigned int argptr = 0;

  const std::string main_output_filename_input = argument_list[argptr++];

  const double min_density = std::atof(argument_list[argptr++].c_str());

  if (!(min_density > 0.0))
    {
    cerr << endl << "Error: Min density [ " << min_density 
	 << " ] must be positive" << endl << endl;
    return -1;
    }

  const std::string progress_var = argument_list[argptr++];

  const double progress_var_min_value 
    = std::atof(argument_list[argptr++].c_str());

  if ((!(progress_var_min_value > 0.0)) || (!(progress_var_min_value < 1.0)))
    {
    cerr << endl << "Error: Progress variable min value [ " 
	 << progress_var_min_value << " ] must be between zero and one" 
	 << endl << endl;
    return -1;
    }

  const double radial_bin_width = std::atof(argument_list[argptr++].c_str());

  if (!(radial_bin_width > 0.0))
    {
    cerr << endl << "Error: Radial bin width [ " << radial_bin_width 
	 << " ] must be positive" << endl << endl;
    return -1;
    }

  std::list<std::string> datafile_name_list;

  while (argptr < num_args)
    datafile_name_list.push_back(argument_list[argptr++]);

  // Read the optional arguments

  std::string base_output_prefix;

  std::vector<std::string> user_slice_var_list;

  bool use_slice_refine_level_limit = false;
  unsigned int slice_refine_level_limit = 0;

  std::list<QuickFlash::Utils::Option>::const_iterator option_iter
    = option_list.begin();

  while (option_iter != option_list.end())
    {
    const QuickFlash::Utils::Option & current_option = *option_iter;

    const std::string & keyword = current_option.get_key();

    if (keyword == "output-prefix")
      {
      if (!(current_option.value_present()))
	{
	cerr << endl << "Error: Output prefix is required with option"
	     << endl << endl;
	return -1;
	}

      base_output_prefix = current_option.get_value();

      // Add an underscore if prefix contains more than just a path

      std::string base_prefix_path;
      std::string base_prefix_name;

      QuickFlash::Utils::split_name_path_file(base_output_prefix, 
					    base_prefix_path, 
					    base_prefix_name);

      if (base_prefix_name.size() > 0)
	base_output_prefix += "_";
      }
    else if (keyword == "slice-datasets")
      {
      if (!(current_option.value_present()))
	{
	cerr << endl << "Error: Slice dataset option requires "
	     << "comma-separated list of dataset names"
	     << endl << endl;
	return -1;
	}

      current_option.get_value(user_slice_var_list);
      }
    else if (keyword == "max-slice-refine-level")
      {
      if (!(current_option.value_present()))
	{
	cerr << endl << "Error: Max slice refinenment level requires a "
	     << "positive integer" << endl << endl;
	return -1;
	}

      unsigned int level = 0;

      current_option.get_value(level);

      if (level < 1)
	{
	cerr << endl << "Error: Max slice refinenment level requires a "
	     << "positive integer" << endl << endl;
	return -1;
	}

      use_slice_refine_level_limit = true;

      slice_refine_level_limit = level;
      }
    else
      {
      cerr << endl << "Error: Unrecognized option " << keyword << endl << endl;
      return -1;
      }

    ++option_iter;
    }

  const unsigned int num_user_slice_vars = user_slice_var_list.size();

  // Set up the main output file -- if the main output file is not qualified 
  // (has no directory path), add the output prefix

  std::string main_output_file_path;
  std::string main_output_file_base_name;

  QuickFlash::Utils::split_name_path_file(main_output_filename_input,
					main_output_file_path, 
					main_output_file_base_name);

  std::string main_output_filename = main_output_filename_input;

  if (main_output_file_path.size() == 0)
    main_output_filename = base_output_prefix + main_output_file_base_name;

  std::ofstream f_main(main_output_filename.c_str(), std::ios::trunc);

  f_main << "Data_File" << "\t" << "Sim_Time" 
	 << "\t" << "Min_Slice_Cell_Width" << "\t" << "Min_Dens_Bounds" 
	 << "\t" << "Max_Dens_Bounds" << "\t" << "Fuel_Mass" 
	 << "\t" << "Fuel_CM" << "\t"  << "Fuel_CM_velz" 
	 << "\t" << "Ash_z_Momenutum_Flow_Rate" 
	 << "\t" << "Ash_z_Mass_Flow_Rate" 
	 << endl;

  // Process the files

  std::list<std::string>::const_iterator datafile_name_iter 
    = datafile_name_list.begin();

  while (datafile_name_iter != datafile_name_list.end())
    {
    const std::string & datafile_name = *datafile_name_iter;

    // Set up the output filename prefix

    std::string datafile_path;
    std::string datafile_base_name;

    QuickFlash::Utils::split_name_path_file(datafile_name, datafile_path, 
					  datafile_base_name);

    const std::string output_prefix = base_output_prefix + datafile_base_name;

    // Open the file

    cerr << "Opening file " << datafile_name << endl;

    QuickFlash::File::DataFile df(datafile_name, buffer_size, true);

    const QuickFlash::File::MeshInfo & meshinfo = df.get_mesh_info();

    const double sim_time = meshinfo.get_sim_time();

    const unsigned int dims = meshinfo.get_dims();

    if (dims != 3)
      {
      cerr << endl << "Error: Three dimensions required for file [ "
	   << datafile_name << " ]" << endl;

      return -2;
      }

    const unsigned int mesh_max_refine_level 
      = meshinfo.get_max_leaf_refine_level();

    // Prep the datasets

    const QuickFlash::File::Dataset & dens_set = df.get_dataset("dens");

    dens_set.set_cache_size(cache_size);
    dens_set.set_report_stats();

    const QuickFlash::File::Dataset & velz_set = df.get_dataset("velz");

    velz_set.set_cache_size(cache_size);
    velz_set.set_report_stats();

    const QuickFlash::File::Dataset & prog_set = df.get_dataset(progress_var);

    prog_set.set_cache_size(cache_size);
    prog_set.set_report_stats();

    for (unsigned int var_index = 0; var_index < num_user_slice_vars; 
	 var_index++)
      {
      const QuickFlash::File::Dataset & var_set 
	= df.get_dataset(user_slice_var_list[var_index]);

      var_set.set_cache_size(cache_size);
      var_set.set_report_stats();
      }

    // Determine the smallest cell size

    unsigned int slice_refine_level = 0;
    double min_slice_cell_width = 0.0;

    if (get_min_cell_width(meshinfo, slice_refine_level, min_slice_cell_width,
			   use_slice_refine_level_limit, 
			   slice_refine_level_limit) < 0)
      {
      cerr << endl << "Error: Unable to find smallest cell size for file [ "
	   << datafile_name << " ]" << endl << endl;

      return -2;
      }

    cerr << "\t" << "Min slice cell width " << min_slice_cell_width << endl;

    const double slice_pixel_width = min_slice_cell_width;
    const double slice_pixel_area = slice_pixel_width * slice_pixel_width;

    // Determine subsampling for slices

    unsigned int slice_axis_subsamples = 1;

    if (slice_refine_level < mesh_max_refine_level)
      {
      // We want 2^(difference in levels)

      const unsigned int level_diff 
        = mesh_max_refine_level - slice_refine_level;

      slice_axis_subsamples = 1 << level_diff;

      cerr << "\t" << "Slice refinement level reduction "
           << level_diff << endl;

      cerr << "\t" << "Slice axis subsamples " << slice_axis_subsamples 
	   << endl;
      }

    // Determine the maximum volume of interest (by density)

    std::vector<double> min_dens_bounds;
    std::vector<double> max_dens_bounds;

    if (find_min_var_bounds(df, "dens", min_density, min_dens_bounds, 
			    max_dens_bounds) < 0)
      {
      cerr << endl << "Error: Unable to find density boundaries for file [ "
	   << datafile_name << " ]" << endl << endl;
      return -2;
      }

    cerr << "\t" << "Dens boundary min [ " << min_dens_bounds << " ] max [ "
	 << max_dens_bounds << " ]" << endl;

    // Determine the maximum volume of interest (by progress variable

    std::vector<double> min_prog_bounds;
    std::vector<double> max_prog_bounds;

    if (find_min_var_bounds(df, progress_var, progress_var_min_value, 
			    min_prog_bounds, max_prog_bounds) < 0)
      {
      cerr << endl 
	   << "Error: Unable to find progess variable boundaries for file [ "
	   << datafile_name << " ]" << endl << endl;
      return -2;
      }

    cerr << "\t" << "Progress variable boundary min [ " << min_prog_bounds 
	 << " ] max [ " << max_prog_bounds << " ]" << endl;

    // Get the maximal boundaries for the slices

    std::vector<double> min_wd_bounds(dims, 0.0);
    std::vector<double> max_wd_bounds(dims, 0.0);

    for (unsigned int i = 0; i < dims; i++)
      {
      min_wd_bounds[i] = std::min(min_dens_bounds[i], min_prog_bounds[i]);
      max_wd_bounds[i] = std::max(max_dens_bounds[i], max_prog_bounds[i]);
      }

    cerr << "\t" << "White dwarf slice boundary min [ " << min_wd_bounds
	 << " ] max [ " << max_wd_bounds << " ]" << endl;

    // Find the center of mass of the fuel

    std::vector<double> fuel_cm;
    double fuel_cm_velz;
    double fuel_mass = 0.0;

    if (find_fuel_cm(df, progress_var, min_wd_bounds, max_wd_bounds,
		     fuel_mass, fuel_cm, fuel_cm_velz) < 0)
      {
      cerr << endl << "Error: Unable to find CM for fuel in file [ "
	   << datafile_name << " ]" << endl << endl;
      return -2;
      }

    cerr << "\t" << "Fuel mass [ " << fuel_mass << " ] CM [ " << fuel_cm
	 << " ] CM vel_z [ " << fuel_cm_velz << " ]" << endl;

    // Get the radially averaged densities (with statistics)

    QuickFlash::Stats::BinArray rad_mass_bins;
    QuickFlash::Stats::BinArray rad_vel_bins;

    double total_radial_mass = 0.0;

    if (get_radial_mass_vel_bins(df, fuel_cm, min_wd_bounds, max_wd_bounds,
				 radial_bin_width, rad_mass_bins, rad_vel_bins,
				 total_radial_mass) < 0)
      {
      cerr << endl << "Error: Unable to get radial mass/vel bins for file [ "
	   << datafile_name << " ]" << endl << endl;
      return -2;
      }

    const unsigned int num_radial_bins = rad_mass_bins.get_num_bins();

    cerr << "\t" << "Total radial mass " << total_radial_mass << endl;

    const std::string rad_mass_bins_filename 
      = output_prefix + "_rad_mass_bins.csv";

    std::ofstream f_rad(rad_mass_bins_filename.c_str());

    f_rad << "Datafile " << datafile_name << endl;
    f_rad << "SimTime " << sim_time << endl;
    f_rad << "TotalBinnedMass " << total_radial_mass << endl;
    f_rad << "RadialBinWidth " << radial_bin_width << endl;
    f_rad << "FuelCM " << fuel_cm << endl;
    f_rad << "DensMinBounds " << min_wd_bounds << endl;
    f_rad << "DensMaxBounds " << max_wd_bounds << endl;
    f_rad << "NumRadBins " << num_radial_bins << endl;

    f_rad << endl;

    f_rad << "Data" << endl;

    f_rad << endl;

    f_rad << "BinIndex" << "\t" << "BinRadius" << "\t" << "BinMass" << "\t"
	  << "BinMeanDens" << "\t" << "MinBinDens" << "\t" << "MaxBinDens" 
	  << "BinMeanRadVel" << "\t" << "MinBinRadVel" << "\t" 
	  << "MaxBinRadVel" << endl;

    for (unsigned int bin_index = 0; bin_index < num_radial_bins; bin_index++)
      {
      const double bin_center = rad_mass_bins.get_bin_center(bin_index);

      const QuickFlash::Stats::Bin & mass_bin = rad_mass_bins[bin_index];

      const double bin_mass = mass_bin.get_weight_value_sum();
      const double bin_mean_dens = mass_bin.get_mean();
      const double bin_min_dens = mass_bin.get_min_value();
      const double bin_max_dens = mass_bin.get_max_value();

      const QuickFlash::Stats::Bin & vel_bin = rad_vel_bins[bin_index];

      const double bin_mean_rad_vel = vel_bin.get_mean();
      const double bin_min_rad_vel = vel_bin.get_min_value();
      const double bin_max_rad_vel = vel_bin.get_max_value();

      f_rad << bin_index << "\t" << bin_center << "\t" << bin_mass << "\t"
	    << bin_mean_dens << "\t" << bin_min_dens << "\t" << bin_max_dens 
	    << "\t" << bin_mean_rad_vel << "\t" << bin_min_rad_vel << "\t"
	    << bin_max_rad_vel << endl;
      }

    f_rad << endl;

    // Get values for density and radial velocity along the cardinal axes

    for (unsigned int i = 0; i < dims; i++)
      {
      if (write_axis_dens_vel(output_prefix, df, i, true, fuel_cm,
			      min_wd_bounds[i], max_wd_bounds[i], 
			      radial_bin_width) < 0)
	{
	cerr << endl << "Error: Unable to get cardial axis samples for file [ "
	     << datafile_name << " ]" << endl;
	}

      if (write_axis_dens_vel(output_prefix, df, i, false, fuel_cm,
			      min_wd_bounds[i], max_wd_bounds[i],
			      radial_bin_width) < 0)
	{
	cerr << endl << "Error: Unable to get cardial axis samples for file [ "
	     << datafile_name << " ]" << endl;
	}
      }

    // Process the slices, and save the xy progress variable, density, and velz

    QuickFlash::Slice::SliceGrid xy_dens_grid;
    QuickFlash::Slice::SliceGrid xy_prog_grid;
    QuickFlash::Slice::SliceGrid xy_velz_grid;

    double max_grid_width = 0.0;  // Cover the entire density-selected region

    for (unsigned int i = 0; i < dims; i++)
      {
      const double axis_cm = fuel_cm[i];

      const double center_min_dist = std::fabs(axis_cm - min_wd_bounds[i]);
      const double center_max_dist = std::fabs(axis_cm - max_wd_bounds[i]);

      const double axis_region_width = 2.0 * std::max(center_min_dist, 
						      center_max_dist);

      if (axis_region_width > max_grid_width)
	max_grid_width = axis_region_width;
      }

    const unsigned int num_axis_pixels
      = static_cast<unsigned int>(max_grid_width / slice_pixel_width) + 1;

    cerr << "\t" << "Slice axis pixels [ " << num_axis_pixels << " ]" << endl;

    if (get_dens_prog_velz_slices(output_prefix, df, fuel_cm, num_axis_pixels,
				  slice_pixel_width, progress_var, 
				  xy_dens_grid, xy_prog_grid, 
				  xy_velz_grid, slice_axis_subsamples) < 0)
      {
      cerr << endl 
	   << "Error: Unable to get required slices for processing file [ " 
	   << datafile_name << " ]" << endl << endl;
      return -2;
      }

    // Get the user-requested slices

    for (unsigned int var_index = 0; var_index < num_user_slice_vars; 
	 var_index++)
      {
      const std::string & var_name = user_slice_var_list[var_index];

      if (write_cardinal_slices(output_prefix, df, fuel_cm, num_axis_pixels, 
				slice_pixel_width, var_name, 
				slice_axis_subsamples) < 0)
	{
	cerr << endl << "Error: Unable to get slices for variable [ " 
	     << var_name << " ] in file [ " << datafile_name << " ]" 
	     << endl << endl;
	return -2;
	}
      }

    // Compute the flux of ash through the xy plane

    const unsigned int xy_cells = xy_dens_grid.get_num_pixels();

    double xy_ash_z_momentum_flow_rate = 0.0;
    double xy_ash_z_mass_flow_rate = 0.0;

    // DO WE NEED TO EXCLUDE THE CENTER??

    for (unsigned int cell_index = 0; cell_index < xy_cells; cell_index++)
      {
      const double cell_dens = xy_dens_grid[cell_index];
      const double cell_velz = xy_velz_grid[cell_index];
      const double cell_prog = xy_prog_grid[cell_index];

      const double cell_ash_dens = cell_prog * cell_dens;

      // Use the z-velocity relative to the fuel center-of-mass

      const double rel_cell_velz = cell_velz - fuel_cm_velz;

      // Compute the ash mass fluxes with this relative velocity

      const double cell_ash_mass_flux = cell_ash_dens * rel_cell_velz;

      const double cell_ash_mom_flux = cell_ash_mass_flux * rel_cell_velz;

      xy_ash_z_momentum_flow_rate += cell_ash_mom_flux * slice_pixel_area;
      xy_ash_z_mass_flow_rate += cell_ash_mass_flux * slice_pixel_area;
      }

    // Place entry in main output file

    f_main << datafile_base_name << "\t" << sim_time 
	   << "\t" << min_slice_cell_width
	   << "\t" << min_wd_bounds << "\t" << max_wd_bounds 
	   << "\t" << fuel_mass << "\t" << fuel_cm << "\t" << fuel_cm_velz 
	   << "\t" << xy_ash_z_momentum_flow_rate 
	   << "\t" << xy_ash_z_mass_flow_rate 
	   << endl;

    ++datafile_name_iter;
    }

  return 0;
  }
