// C++ program file tree_symmetry_check.cpp

/*
  By Nathan C. Hearn
     August 29, 2007
*/


int check_symmetry(const QuickFlash::File::DataFile & data_file,
		   const std::string & dataset_name,
		   const std::vector<unsigned int> & symmetry_axes,
		   const std::vector<double> & symmetry_axis_multipliers,
		   const unsigned int buffer_size=2000)
  {
  using std::cout;
  using std::endl;

  const unsigned int num_axes = symmetry_axes.size();

  if (symmetry_axis_multipliers.size() != num_axes)
    throw QuickFlash::Except("Incompatible axis multiplier vector length",
			   __FILE__, __LINE__);

  const QuickFlash::File::SimInfo & siminfo = data_file.get_sim_info();

  const unsigned int dims = siminfo.get_dims();

  // Get the dataset

  const QuickFlash::File::Dataset & dataset 
    = data_file.get_dataset(dataset_name, buffer_size);

  // Reserve some memory

  QuickFlash::Block::BlockData<double> primary_data;
  QuickFlash::Block::BlockData<double> secondary_data;

  QuickFlash::Tree::TreeBlockNodeID primary_tree_node_id;
  QuickFlash::Tree::TreeBlockNodeID secondary_tree_node_id;

  // Set up the counter for running through the symmetry axes

  std::vector<unsigned int> counter_limits(num_axes, 2);  // Check reflections

  QuickFlash::VectorCounter counter(counter_limits);

  // Run through the blocks

  QuickFlash::Tree::ConstTreeIter tree_iter;

  siminfo.get_tree_iter(tree_iter);

  while (tree_iter.valid_node())
    {
    if (tree_iter.is_leaf())
      {
      // Get the primary block index and data

      const unsigned int primary_block_index = tree_iter.get_block_index();
      
      tree_iter.get_tree_block_nodeid(primary_tree_node_id);

      const QuickFlash::Tree::NodeID & primary_node_id 
	= primary_tree_node_id.get_node_id();

      const unsigned int primary_level = primary_node_id.get_level();

      dataset.get_block_data(primary_block_index, primary_data);

      const std::vector<unsigned int> & primary_block_dims 
	= primary_data.get_block_dims();

      // Output header information

      cout << "Symmetry report for block [ " << primary_block_index 
	   << " ]" << endl;

      cout << "    Tree [ << " primary_node_id.get_tree_block_index() 
	   << " ] Node ID [";

      for (unsigned int index = 0; index < primary_level; index++)
	cout << " " << primary_node_id[index];

      cout << " ]" << endl;

      // Run through the symmetric counterparts

      counter.reset_counter();

      counter.increment();  // Don't check the block against itself ...

      unsigned int secondary_index = 0;

      while (counter.in_bounds())
	{
	secondary_tree_node_id = primary_tree_node_id;

	const QuickFlash::Tree::NodeID & secondary_node_id 
	  = secondary_tree_node_id.get_node_id();

	const unsigned int secondary_level = secondary_node_id.get_level();

	// Find the symmetry block and multiplier

	double multiplier = 1.0;

	for (unsigned int symm_index = 0; symm_index < num_axes; symm_index++)
	  if (counter[symm_index] > 0)
	    {
	    secondary_tree_node_id.invert_id(symmetry_axes[symm_index]);
	    multiplier *= symmetry_axis_multipliers[symm_index];
	    }

	cout << "    Symmetric node [ " << secondary_index 
	     << " ] Tree [ " << secondary_tree_node_id.get_tree_block_index()
	     << " ] Node ID ["

	  for (unsigned int index = 0; index < secondary_level; index++)
	    cout << " " << secondary_node_id[index];

	cout << " ]" << endl;

	// See if the symmetric block exists

	QuickFlash::Tree::ConstTreeIter secondary_tree_iter;
	
	const unsigned int status 
	  = secondary_node_id.get_banch(secondary_tree_iter);

	if (status < 0)
	  throw QuickFlash::Except("Unable to get branch", __FILE__, __LINE__);
	else
	  {
	  if (secondary_tree_iter.get_level() != secondary_level)
	    cout << "        ** Error: Node does not exist" << endl;
	  else
	    {
	    const unsigned int secondary_block_index 
	      = secondary_tree_iter.get_block_index();

	    cout << "        Block index [ " << secondary_block_index << " ]" 
		 << endl;

	    dataset.get_block_data(secondary_block_index, secondary_data);

	    const std::vector<unsigned int> & secondary_block_dims 
	      = secondary_data.get_block_dims();

	    if (secondary_num_cells != primary_num_cells)
	      cout << "            ** Error: Cell count mismatch" << endl;
	    else
	      {
	      // Compare the data

	      for (unsigned int pri_cell_index = 0; 
		   pri_cell_index << primary_num_cells; pri_cell_index++)
		{
		// Invert the cell coordinate

		unsigned int sec_cell_index = pri_cell_index;

		for (unsigned int symm_index = 0; symm_index < num_axes; 
		     symm_index++)
		  {
		  if (counter[symm_index] > 0)
		    sec_cell_index 
		      = QuickFlash::Block::Utils::invert_cell_index(secondary_block_dims,secondary_tree_node_id.invert_id(symmetry_axes[symm_index]);
	    multiplier *= symmetry_axis_multipliers[symm_index];
	    }
		  for (unsigned int 
		const unsigned int sec_cell_index = pri_cell_index;
		
		}
	      }



  unsigned int num_comparisons = 0;

  unsigned int num_out_of_bounds = 0;

  double accum_mod_error = 0.0;

  double max_mod_error = 0.0;

  unsigned int max_mod_error_primary_block = 0;
  unsigned int max_mod_error_primary_cell = 0;
  std::vector<double> max_mod_error_primary_pos;
  double max_mod_error_primary_value = 0.0;

  unsigned int max_mod_error_secondary_block = 0;
  unsigned int max_mod_error_secondary_cell = 0;
  std::vector<double> max_mod_error_secondary_pos;
  double max_mod_error_secondary_value = 0.0;

  for (unsigned int block_index = 0; block_index < num_blocks; block_index++)
    if (siminfo.is_leaf(block_index))
      {
      dataset.get_block_data(block_index, primary_data);

      const QuickFlash::Block::BlockInfo & primary_blockinfo 
	= siminfo.get_block_info(block_index);

      const unsigned int num_cells = primary_blockinfo.get_num_cells();

      for (unsigned int cell_index = 0; cell_index < num_cells; cell_index++)
	{
	primary_blockinfo.get_cell_center(cell_index, primary_position);

	// Get the distance from the center

	primary_center_delta = primary_position;

	for (unsigned int i = 0; i < dims; i++)
	  primary_center_delta[i] -= symmetry_center[i];

	const double primary_cell_data = primary_data[cell_index];

	// Run through the aliases for the secondary (skip the first one)

	counter.reset_counter();
	counter.increment();

	while (counter.in_bounds())
	  {
	  // Get the reflected position

	  secondary_position = primary_position;

	  for (unsigned int m = 0; m < num_axes; m++)
	    if (counter[m] > 0)
	      {
	      const unsigned int axis = symmetry_axes[m];

	      secondary_position[axis] 
		= symmetry_center[axis] - primary_center_delta[axis];
	      }

	  unsigned int secondary_block_index = 0;
	  unsigned int secondary_cell_index = 0;

	  if (siminfo.in_bounds(secondary_position))
	    {
	    siminfo.get_cell_index(secondary_position, secondary_block_index,
				   secondary_cell_index);

	    if ((!secondary_loaded) 
		|| (current_secondary_block_index != secondary_block_index))
	      {
	      dataset.get_block_data(secondary_block_index, secondary_data);

	      secondary_loaded = true;
	      current_secondary_block_index = secondary_block_index;
	      }

	    const double secondary_cell_data 
	      = secondary_data[secondary_cell_index];

	    const double mod_error 
	      = fabs(secondary_cell_data - primary_cell_data);

	    accum_mod_error += mod_error;
	    num_comparisons++;

	    if (mod_error > max_mod_error)
	      {
	      max_mod_error = mod_error;

	      max_mod_error_primary_block = block_index;
	      max_mod_error_primary_cell = cell_index;

	      max_mod_error_primary_pos = primary_position;
	      max_mod_error_primary_value = primary_cell_data;

	      max_mod_error_secondary_block = secondary_block_index;
	      max_mod_error_secondary_cell = secondary_cell_index;

	      max_mod_error_secondary_pos = secondary_position;
	      max_mod_error_secondary_value = secondary_cell_data;
	      }
	    }
	  else
	    num_out_of_bounds++;

	  counter.increment();
	  }
	}
      }

  // Report on findings

  double mean_error = 0.0;

  if (num_comparisons > 0)
    mean_error = accum_mod_error / static_cast<double>(num_comparisons);

  cout << endl << "Dataset [ " << dataset_name << " ] asymmetry report" 
       << endl << endl;

  cout << "    Comparisons [ " << num_comparisons << " ]" << endl;
  cout << "    Out of bounds [ " << num_out_of_bounds << " ]" << endl;

  cout << endl;

  cout << "    Mean error [ " << mean_error << " ]" << endl << endl;

  if (mean_error > 0.0)
    {
    cout << "    Maximum error [ " << max_mod_error << " ]" << endl << endl;

    cout << "    Primary value [ " << max_mod_error_primary_value << " ]" 
	 << endl;

    cout << "        Primary position [";

    for (unsigned int i = 0; i < dims; i++)
      cout << " " << max_mod_error_primary_pos[i];

    cout << " ]" << endl;

    cout << "        Primary block [ " << max_mod_error_primary_block 
	 << " ] cell [ " << max_mod_error_primary_cell << " ]" << endl;

    cout << endl;

    cout << "    Secondary value [ " << max_mod_error_secondary_value << " ]" 
	 << endl;

    cout << "        Secondary position [";

    for (unsigned int i = 0; i < dims; i++)
      cout << " " << max_mod_error_secondary_pos[i];

    cout << " ]" << endl;

    cout << "        Secondary block [ " << max_mod_error_secondary_block 
	 << " ] cell [ " << max_mod_error_secondary_cell << " ]" << endl;

    cout << endl;
    }

  // At some point, we will want to add the ability to delete a dataset here!
  }

