// C++ program file analyze_rt.cpp

/*
  By Nathan C. Hearn
     February 17, 2008

  Analysis tool for RT simulations.
*/


#include <iostream>
#include <string>
#include <sstream>
#include <vector>
#include <list>
#include "quickflash_utils_commandline.hpp"
#include "quickflash_file_datafile.hpp"
#include "quickflash_file_meshinfo.hpp"
#include "layer_finder.hpp"
#include "mixing_layer.hpp"
#include "cubic_volume.hpp"



const unsigned int buffer_size = 1000;

const char dens_dataset_name[] = "dens";
const char light_dataset_name[] = "ligh";

const double min_envelope_bins = 5.0;
const double envelope_width_pad_bins = 3.0;


void build_datafile_name(const std::string & filename_base,
			 const unsigned int index, const unsigned int digits,
			 std::string & filename)
  {
  std::stringstream namestream;

  namestream << filename_base;

  namestream.width(digits);
  namestream.fill('0');

  namestream << index;

  filename = namestream.str();
  }


int main(const int argc, char * const argv[])
  {
  using std::cout;
  using std::cerr;
  using std::endl;

  std::vector<QuickFlash::Utils::Option> argument_list;
  std::list<QuickFlash::Utils::Option> option_list;

  QuickFlash::Utils::parse_command_line(argc, argv, argument_list, option_list);

  const unsigned int num_args = argument_list.size();

  const unsigned int req_args = 11;

  if (num_args != req_args)
    {
    cerr << endl << "Usage: " << argv[0] << " filename_base first_index "
	 << "last_index"
	 << endl << "\t"
	 << "flow_axis_index envelope_detection_dataset"
	 << endl << "\t"
	 << "envelope_min_value envelope_max_value"
	 << endl << "\t"
	 << "layer_detection_dataset"
	 << endl << "\t"
	 << "min_pos_data_value max_pos_data_value search_bin_width"
	 << endl << "\t"
	 << "[ --sample-volume-center=coord|<x,y,...> ]"
	 << endl << "\t"
	 << "[ --follow-layer-center ]"
	 << endl << "\t"
	 << "[ --sample-volume-width=width|<dx,dy,...> ]"
         << endl << "\t"
	 << "[ --use-layer-width-fraction ]"
	 << endl << "\t"
	 << "[ --use-peak-value-fraction-layer ]"
	 << endl << "\t"
	 << "[ --use-mass-weight ] [ --use-abs-value ]"
	 << endl << endl;
    return 1;
    }

  // Read the arguments

  unsigned int argptr = 0;

  const std::string filename_base = argument_list[argptr++].get_value();

  unsigned int first_file_index = 0;
  unsigned int last_file_index = 0;

  argument_list[argptr++].get_value(first_file_index);
  argument_list[argptr++].get_value(last_file_index);

  if (last_file_index < first_file_index)
    {
    cerr << endl << "Error: Last index can not be less than first index"
	 << endl << endl;
    return -1;
    }

  unsigned int flow_axis_index = 0;

  argument_list[argptr++].get_value(flow_axis_index);

  const std::string envelope_detection_dataset 
    = argument_list[argptr++].get_value();

  double envelope_min_value = 0.0;
  double envelope_max_value = 0.0;

  argument_list[argptr++].get_value(envelope_min_value);
  argument_list[argptr++].get_value(envelope_max_value);

  if (!(envelope_min_value < envelope_max_value))
    {
    cerr << endl << "Error: Envelope min value must be less than max value"
	 << endl << endl;
    return -1;
    }

  const std::string layer_detection_dataset 
    = argument_list[argptr++].get_value();

  double min_pos_data_value = 0.0;
  double max_pos_data_value = 0.0;

  argument_list[argptr++].get_value(min_pos_data_value);
  argument_list[argptr++].get_value(max_pos_data_value);

  double search_bin_width = 0.0;

  argument_list[argptr++].get_value(search_bin_width);

  if (!(search_bin_width > 0.0))
    {
    cerr << endl << "Error: Search bin width must be positive" << endl << endl;
    return -1;
    }

  // Parse the options

  bool sample_volume_center_set = false;
  std::vector<double> sample_volume_center;

  bool axis_center_only = false;  // True if only one element in volume center

  bool follow_layer_center = false;

  bool sample_volume_width_set = false;
  std::vector<double> sample_volume_width;

  bool axis_width_only = false;  // True if only one element in sample width

  bool use_layer_width_fraction = false;

  bool use_peak_value_fraction_layer = false;

  bool use_mass_weight = false;

  bool use_abs_value = false;

  std::list<QuickFlash::Utils::Option>::const_iterator option_iter 
    = option_list.begin();

  while (option_iter != option_list.end())
    {
    const QuickFlash::Utils::Option & current_option = *option_iter;

    const std::string & keyword = current_option.get_key();

    if (keyword == "sample-volume-center")
      {
      if (!current_option.value_present())
	{
	cerr << endl << "Error: Volume center requires at least one element"
	     << endl << endl;
	return -1;
	}

      current_option.get_value(sample_volume_center);

      sample_volume_center_set = true;

      if (sample_volume_center.size() == 1)
	axis_center_only = true;
      }
    else if (keyword == "follow-layer-center")
      {
      follow_layer_center = true;
      }
    else if (keyword == "sample-volume-width")
      {
      if (!current_option.value_present())
	{
	cerr << endl << "Error: Volume width requires at least one element"
	     << endl << endl;
	return -1;
	}

      current_option.get_value(sample_volume_width);

      sample_volume_width_set = true;

      if (sample_volume_width.size() == 1)
	axis_width_only = true;
      }
    else if (keyword == "use-layer-width-fraction")
      {
      use_layer_width_fraction = true;
      }
    else if (keyword == "use-peak-value-fraction-layer")
      {
      use_peak_value_fraction_layer = true;
      }
    else if (keyword == "use-mass-weight")
      {
      use_mass_weight = true;
      }
    else if (keyword == "use-abs-value")
      {
      use_abs_value = true;
      }
    else
      {
      cerr << endl << "Error: Unrecognized option " << keyword << endl << endl;
      return -1;
      }

    ++option_iter;
    }

  const double min_envelope_width = min_envelope_bins * search_bin_width;
  const double envelope_pad_width = envelope_width_pad_bins * search_bin_width;

  // Run through the files

  const unsigned int num_files = (last_file_index + 1) - first_file_index;

  for (unsigned int file_loop_index = 0; file_loop_index < num_files;
       file_loop_index++)
    {
    const unsigned int file_index = file_loop_index + first_file_index;

    // Load the data file

    std::string filename;

    build_datafile_name(filename_base, file_index, 4, filename);

    QuickFlash::File::DataFile df(filename, buffer_size, true);

    cerr << "* Opened file " << filename << endl;

    // Get some basic information about the file

    const QuickFlash::File::MeshInfo & meshinfo = df.get_mesh_info();

    const double sim_time = meshinfo.get_sim_time();

    const unsigned int dims = meshinfo.get_dims();

    const std::vector<double> & min_sim_bounds 
      = meshinfo.get_volume_minbounds();
    const std::vector<double> & max_sim_bounds 
      = meshinfo.get_volume_maxbounds();

    const double flow_axis_min_bounds = min_sim_bounds[flow_axis_index];
    const double flow_axis_max_bounds = max_sim_bounds[flow_axis_index];

    const double flow_axis_center_coord 
      = 0.5 * (flow_axis_min_bounds + flow_axis_max_bounds);

    // Perform a few checks

    if (sample_volume_center_set && (!axis_center_only))
      if (sample_volume_center.size() != dims)
	{
	cerr << endl << "Error: Volume center vector size mismatch for file "
	     << filename << endl << endl;
	return -2;
	}

    if (sample_volume_width_set && (!axis_width_only))
      if (sample_volume_width.size() != dims)
	{
	cerr << endl << "Error: Volume width vector size mismatch for file "
	     << filename << endl << endl;
	return -2;
	}

    if (use_peak_value_fraction_layer && use_mass_weight)
      {
      cerr << endl << "Error: Mass weight option not supported with layer "
	   << "peak value fraction" << endl;
      return -2;
      }

    // Determine the maximal mixing layer boundary

    double min_envelope_pos = 0.0;
    double max_envelope_pos = 0.0;

    if (get_maximal_envelope_range(df, envelope_detection_dataset, 
				   flow_axis_index,
				   envelope_min_value, envelope_max_value,
				   min_envelope_pos, max_envelope_pos) < 0)
      {
      cerr << endl << "Warning: Unable to find maximal layer envelope" 
	   << endl << endl;

      min_envelope_pos = flow_axis_center_coord;
      max_envelope_pos = flow_axis_center_coord;
      }

    // Check against a minimum width

    const double envelope_width = max_envelope_pos - min_envelope_pos;

    if (envelope_width < min_envelope_width)
      {
      const double envelope_delta = min_envelope_width - envelope_width;

      const double half_envelope_delta = 0.5 * envelope_delta;

      min_envelope_pos -= half_envelope_delta;
      max_envelope_pos += half_envelope_delta;
      }

    // Add a padding region around the envelope

    min_envelope_pos -= envelope_pad_width;
    max_envelope_pos += envelope_pad_width;

    if (min_envelope_pos < flow_axis_min_bounds)
      min_envelope_pos = flow_axis_min_bounds;
    
    if (max_envelope_pos > flow_axis_max_bounds)
      max_envelope_pos = flow_axis_max_bounds;

    // SHOULD WE CHECK AGAINST THE MINIMUM WIDTH AGAIN???

    // Get a more-refined mixing layer boundary based on bin averages

    MixingLayer mixing_layer;

    if (use_mass_weight)
      get_mixing_layer_meanValue_massWeight(df, layer_detection_dataset, 
					    flow_axis_index, min_envelope_pos,
					    max_envelope_pos, search_bin_width,
					    min_pos_data_value, 
					    max_pos_data_value, mixing_layer,
					    use_abs_value);
    else
      {
      if (use_peak_value_fraction_layer)
	get_mixing_layer_meanValue_peakFract(df, layer_detection_dataset, 
					     flow_axis_index,
					     min_envelope_pos, 
					     max_envelope_pos, 
					     search_bin_width,
					     min_pos_data_value, 
					     max_pos_data_value,
					     mixing_layer, use_abs_value);
      else
	get_mixing_layer_meanValue(df, layer_detection_dataset, 
				   flow_axis_index,
				   min_envelope_pos, max_envelope_pos, 
				   search_bin_width,
				   min_pos_data_value, max_pos_data_value,
				   mixing_layer, use_abs_value);
      }

    const double min_layer_pos = mixing_layer.get_min_pos();
    const double max_layer_pos = mixing_layer.get_max_pos();

    const double layer_center_pos = 0.5 * (min_layer_pos + max_layer_pos);
    const double layer_width = max_layer_pos - min_layer_pos;

    cerr << "  * Layer found: " << min_layer_pos << " " << max_layer_pos << endl;

    // Determine the center and width for the desired sample region ...

    std::vector<double> sample_center(dims, 0.0);
    std::vector<double> sample_width(dims, 0.0);

    // ... based on simulation volume

    for (unsigned int i = 0; i < dims; i++)
      {
      const double axis_sim_min = min_sim_bounds[i];
      const double axis_sim_max = max_sim_bounds[i];

      sample_center[i] = 0.5 * (axis_sim_min + axis_sim_max);
      sample_width[i] = axis_sim_max - axis_sim_min;
      }

    // ... based on user-specified options or found mixing layer

    sample_center[flow_axis_index] = layer_center_pos;

    if (sample_volume_center_set)
      {
      if (follow_layer_center)
	{
	if (axis_center_only)
	  sample_center[flow_axis_index] += sample_volume_center[0];
	else
	  for (unsigned int i = 0; i < dims; i++)
	    sample_center[i] += sample_volume_center[i];
	}
      else
	{
	if (axis_center_only)
	  sample_center[flow_axis_index] = sample_volume_center[0];
	else
	  sample_center = sample_volume_center;
	}
      }

    sample_width[flow_axis_index] = layer_width;

    if (sample_volume_width_set)
      {
      if (use_layer_width_fraction)
	{
	if (axis_width_only)
	  sample_width[flow_axis_index] *= sample_volume_width[0];
	else
	  for (unsigned int i = 0; i < dims; i++)
	    sample_width[i] *= sample_volume_width[i];
	}
      else
	{
	if (axis_width_only)
	  sample_width[flow_axis_index] = sample_volume_width[0];
	else
	  sample_width = sample_volume_width;
	}
      }

    std::vector<double> min_sample_bounds(dims);
    std::vector<double> max_sample_bounds(dims);

    for (unsigned int i = 0; i < dims; i++)
      {
      const double center_coord = sample_center[i];
      const double half_width = 0.5 * sample_width[i];

      min_sample_bounds[i] = center_coord - half_width;
      max_sample_bounds[i] = center_coord + half_width;
      }

    const double min_sample_pos = min_sample_bounds[flow_axis_index];
    const double max_sample_pos = max_sample_bounds[flow_axis_index];

    // Set up the datasets

    const QuickFlash::File::Dataset & dens_set = df.get_dataset("dens");

    const QuickFlash::File::Dataset & ligh_set = df.get_dataset("ligh");  // rt

    std::vector<const QuickFlash::File::Dataset *> vel_set(dims);

    if (dims > 0)
      vel_set[0] = &(df.get_dataset("velx"));

    if (dims > 1)
      vel_set[1] = &(df.get_dataset("vely"));

    if (dims > 2)
      vel_set[2] = &(df.get_dataset("velz"));

    // Parse the blocks

    double accum_mass = 0.0;

    double accum_mass_light = 0.0;
    double accum_mass_heavy = 0.0;

    std::vector<double> accum_mass_pos(dims, 0.0);
    std::vector<double> accum_mass_vel(dims, 0.0);

    std::vector<double> accum_mass_pos_light(dims, 0.0);
    std::vector<double> accum_mass_pos_heavy(dims, 0.0);

    std::vector<double> accum_mass_vel_light(dims, 0.0);
    std::vector<double> accum_mass_vel_heavy(dims, 0.0);

    double accum_kinetic_energy = 0.0;
    double accum_potential_energy = 0.0;

    double accum_kinetic_energy_light = 0.0;
    double accum_kinetic_energy_heavy = 0.0;

    double accum_potential_energy_light = 0.0;
    double accum_potential_energy_heavy = 0.0;

    QuickFlash::Tree::ConstTreeIter tree_iter = meshinfo.get_tree_iter();

    std::vector<double> cell_min_coords(dims);
    std::vector<double> cell_max_coords(dims);

    std::vector<double> cell_overlap_pos(dims);
    std::vector<double> cell_vel(dims);

    QuickFlash::Block::BlockData<double> dens;
    QuickFlash::Block::BlockData<double> ligh;
    
    std::vector< QuickFlash::Block::BlockData<double> > vel(dims);

    while (tree_iter.valid_node())
      {
      const QuickFlash::Block::BlockInfo & blockinfo 
	= tree_iter.get_block_info();

      // Check for overlap

      const std::vector<double> & block_min = blockinfo.get_min_coords();
      const std::vector<double> & block_max = blockinfo.get_max_coords();
      
      if (!(volumes_overlap(min_sample_bounds, max_sample_bounds,
			    block_min, block_max)))
	tree_iter.next_branch();
      else
	{
	if (tree_iter.is_leaf())
	  {
	  // Read the data

	  const unsigned int block_index = tree_iter.get_block_index();

	  bool block_data_loaded = false;

	  // Parse the cells

	  const unsigned int num_cells = blockinfo.get_num_cells();

	  // const double cell_volume = blockinfo.get_cell_volume();

	  for (unsigned int cell_index = 0; cell_index < num_cells; 
	       cell_index++)
	    {
	    blockinfo.get_cell_bounds(cell_index, cell_min_coords, 
				      cell_max_coords);

	    double overlap_volume = 0.0;
	    double control_fract = 0.0;
	    double cell_fract = 0.0;
	    double sqVelocity = 0.0;

	    if (volumes_overlap(min_sample_bounds, max_sample_bounds, 
				cell_min_coords, cell_max_coords, 
				overlap_volume, cell_overlap_pos, 
				control_fract, cell_fract))
	      {
	      if (!block_data_loaded)
		{
		dens_set.get_block_data(block_index, dens);

		for (unsigned int i = 0; i < dims; i++)
		  vel_set[i]->get_block_data(block_index, vel[i]);

		ligh_set.get_block_data(block_index, ligh);

		block_data_loaded = true;
		}

	      const double cell_ligh = ligh[cell_index];
	      const double cell_heav = 1.0 - cell_ligh;

	      const double cell_dens = dens[cell_index];

	      for (unsigned int i = 0; i < dims; i++)
		cell_vel[i] = vel[i][cell_index];

	      for (unsigned int i = 0; i < dims; i++)
		{
		const double vel_elem = cell_vel[i];
		sqVelocity += vel_elem * vel_elem;
		}

	      const double height = cell_overlap_pos[1];

	      const double overlap_mass = cell_dens * overlap_volume;

	      const double overlap_mass_light = cell_ligh * overlap_mass;
	      const double overlap_mass_heavy = cell_heav * overlap_mass;

	      accum_mass += overlap_mass;

	      accum_mass_light += overlap_mass_light;
	      accum_mass_heavy += overlap_mass_heavy;

	      for (unsigned int i = 0; i < dims; i++)
		accum_mass_pos[i] += cell_overlap_pos[i] * overlap_mass;

	      for (unsigned int i = 0; i < dims; i++)
		accum_mass_vel[i] += cell_vel[i] * overlap_mass;

	      for (unsigned int i = 0; i < dims; i++)
		accum_mass_pos_light[i] 
		  += cell_overlap_pos[i] * overlap_mass_light;

	      for (unsigned int i = 0; i < dims; i++)
		accum_mass_pos_heavy[i] 
		  += cell_overlap_pos[i] * overlap_mass_heavy;

	      for (unsigned int i = 0; i < dims; i++)
		accum_mass_vel_light[i] += cell_vel[i] * overlap_mass_light;

	      for (unsigned int i = 0; i < dims; i++)
		accum_mass_vel_heavy[i] += cell_vel[i] * overlap_mass_heavy;

	      const double kinetic_energy = 0.5 * sqVelocity * overlap_mass;
	      const double potential_energy = height * overlap_mass;  // g??

	      accum_kinetic_energy += kinetic_energy;
	      accum_potential_energy += potential_energy;

	      accum_kinetic_energy_light += kinetic_energy * cell_ligh;
	      accum_kinetic_energy_heavy += kinetic_energy * cell_heav;

	      accum_potential_energy_light += potential_energy * cell_ligh;
	      accum_potential_energy_heavy += potential_energy * cell_heav;
	      }
	    }
	  }
	tree_iter.next();
	}
      }

    std::vector<double> cm(dims, 0.0);

    if (accum_mass > 0.0)
      {
      const double oneOver_mass = 1.0 / accum_mass;

      for (unsigned int i = 0; i < dims; i++)
	cm[i] = accum_mass_pos[i] * oneOver_mass;
      }

    std::vector<double> cm_light(dims, 0.0);
    std::vector<double> cm_heavy(dims, 0.0);

    if (accum_mass_light > 0.0)
      {
      const double oneOver_mass_light = 1.0 / accum_mass_light;

      for (unsigned int i = 0; i < dims; i++)
	cm_light[i] = accum_mass_pos_light[i] * oneOver_mass_light;
      }

    if (accum_mass_heavy > 0.0)
      {
      const double oneOver_mass_heavy = 1.0 / accum_mass_heavy;

      for (unsigned int i = 0; i < dims; i++)
	cm_heavy[i] = accum_mass_pos_heavy[i] * oneOver_mass_heavy;
      }

    // Write out header on first pass

    if (file_loop_index == 0)
      {
      cout << "# analyze_rt output" << endl;
      cout << "#" << endl;

      cout << "# Command:";

      for (int idx = 0; idx < argc; idx++)
	cout << " " << argv[idx];

      cout << "#" << endl;
      cout << "# Filename base [ " << filename_base << " ]" << endl;
      cout << "# Start index [ " << first_file_index << " ]" << endl;
      cout << "# End index [ " << last_file_index << " ]" << endl;
      cout << "# Dims [ " << dims << " ]" << endl;

      cout << "# Sample volume center [ "<< sample_center << " ]" << endl;

      cout << "# Sample volume width [ " << sample_width << " ]" << endl;

      cout << endl;

      cout << "File_Index" << "\t" << "Time" << "\t" "Envelope_Min_Pos" << "\t"
	   << "Envelope_Max_Pos" << "\t" << "Layer_Min_Pos" << "\t"
	   << "Layer_Max_Pos" << "\t" << "Sample_Min_Pos" << "\t" 
	   << "Sample_Max_Pos" << "\t" << "Mass" << "\t" << "Mass_light" 
	   << "\t" << "Mass_heavy";

      for (unsigned int i = 0; i < dims; i++)
	cout << "\t" << "cm_" << i;

      for (unsigned int i = 0; i < dims; i++)
	cout << "\t" << "p_" << i;

      for (unsigned int i = 0; i < dims; i++)
	cout << "\t" << "cm_light_" << i;

      for (unsigned int i = 0; i < dims; i++)
	cout << "\t" << "cm_heavy_" << i;

      for (unsigned int i = 0; i < dims; i++)
	cout << "\t" << "p_light_" << i;

      for (unsigned int i = 0; i < dims; i++)
	cout << "\t" << "p_heavy_" << i;

      cout << "\t" << "kinetic_energy" << "\t" << "potential_energy";

      cout << "\t" << "kinetic_energy_light" << "\t" << "kinetic_energy_heavy";
      cout << "\t" << "potential_energy_light" 
	   << "\t" << "potential_energy_heavy";

      cout << endl;
      }

    cout << file_index << "\t" << sim_time << "\t" << min_envelope_pos << "\t"
	 << max_envelope_pos << "\t" << min_layer_pos << "\t"
	 << max_layer_pos << "\t" << min_sample_pos << "\t" << max_sample_pos
	 << "\t" << accum_mass << "\t" << accum_mass_light << "\t"
	 << accum_mass_heavy;

    for (unsigned int i = 0; i < dims; i++)
      cout << "\t" << cm[i];

    for (unsigned int i = 0; i < dims; i++)
      cout << "\t" << accum_mass_vel[i];

    for (unsigned int i = 0; i < dims; i++)
      cout << "\t" << cm_light[i];

    for (unsigned int i = 0; i < dims; i++)
      cout << "\t" << cm_heavy[i];

    for (unsigned int i = 0; i < dims; i++)
      cout << "\t" << accum_mass_vel_light[i];

    for (unsigned int i = 0; i < dims; i++)
      cout << "\t" << accum_mass_vel_heavy[i];

    cout << "\t" << accum_kinetic_energy << "\t" << accum_potential_energy;

    cout << "\t" << accum_kinetic_energy_light
	 << "\t" << accum_kinetic_energy_heavy;

    cout << "\t" << accum_potential_energy_light
	 << "\t" << accum_potential_energy_heavy;

    cout << endl;
    }

  return 0;
  }
