// C++ program file analyze_volume_fixed.cpp

/*
  By Nathan C. Hearn
     May 28, 2007

  Analysis of fixed control volume.
*/


#include <iostream>
#include <string>
#include <sstream>
#include <stdlib.h>
#include "quickflash.hpp"
#include "cubic_volume.hpp"


const unsigned int buffer_size = 1000;


void build_datafile_name(const std::string & filename_base,
			 const unsigned int index, const unsigned int digits,
			 std::string & filename)
  {
  std::stringstream namestream;

  namestream << filename_base;

  namestream.width(digits);
  namestream.fill('0');

  namestream << index;

  filename = namestream.str();
  }


int main(int argc, char * argv[])
  {
  using std::cout;
  using std::cerr;
  using std::endl;

  const int min_args = 5;

  if (argc < min_args)
    {
    cerr << "Usage: " << argv[0] << " filename_base start_index end_index" 
	 << endl;
    cerr << "dims <control_volume_center> <control_volume_width>" << endl;
    cerr << endl;

    return 1;
    }

  // Read the arguments

  int req_args = min_args;

  int argptr = 1;

  const std::string filename_base = argv[argptr++];

  const unsigned int start_index 
    = static_cast<unsigned int>(atoi(argv[argptr++]));
  
  const unsigned int end_index 
    = static_cast<unsigned int>(atoi(argv[argptr++]));

  const unsigned int dims = static_cast<unsigned int>(atoi(argv[argptr++]));

  req_args += 2 * dims;

  if (argc < req_args)
    {
    cerr << "Error: Too few arguments" << endl;
    return -1;
    }
  
  std::vector<double> control_volume_center(dims, 0.0);

  for (unsigned int i = 0; i < dims; i++)
    control_volume_center[i] = atof(argv[argptr++]);

  std::vector<double> control_volume_width(dims, 0.0);

  for (unsigned int i = 0; i < dims; i++)
    control_volume_width[i] = atof(argv[argptr++]);

  // Helpful constants

  std::vector<double> control_volume_halfwidth = control_volume_width;

  for (unsigned int i = 0; i < dims; i++)
    control_volume_halfwidth[i] *= 0.5;

  // Perform the main loop

  const unsigned int loop_count = (end_index + 1) - start_index;

  std::vector<double> control_volume_minbounds = control_volume_center;
  std::vector<double> control_volume_maxbounds = control_volume_center;

  for (unsigned int i = 0; i < dims; i++)
    {
    const double half_width = control_volume_width[i];

    control_volume_minbounds[i] -= half_width;
    control_volume_maxbounds[i] += half_width;
    }

  for (unsigned int loop_index = 0; loop_index < loop_count; loop_index++)
    {
    const unsigned int file_index = loop_index + start_index;

    // Load the data file

    std::string filename;

    build_datafile_name(filename_base, file_index, 4, filename);

    QuickFlash::File::DataFile df(filename, buffer_size, true);

    const QuickFlash::File::MeshInfo & meshinfo = df.get_mesh_info();

    const double sim_time = meshinfo.get_sim_time();

    const unsigned int dims = meshinfo.get_dims();

    for (unsigned int i = 1; i < dims; i++)
      {
      control_volume_minbounds[i] 
	= control_volume_center[i] - control_volume_halfwidth[i];

      control_volume_maxbounds[i] 
	= control_volume_center[i] + control_volume_halfwidth[i];
      }

    // Set up the datasets

    const QuickFlash::File::Dataset & dens_set = df.get_dataset("dens");

//     std::vector<const QuickFlash::File::Dataset *> vel_set(dims);

//     if (dims > 0)
//       vel_set[0] = &(df.get_dataset("velx"));

//     if (dims > 1)
//       vel_set[1] = &(df.get_dataset("vely"));

//     if (dims > 2)
//       vel_set[2] = &(df.get_dataset("velz"));

    // Parse the blocks

    double accum_mass = 0.0;
    std::vector<double> accum_mass_pos(dims, 0.0);
    std::vector<double> accum_mass_vel(dims, 0.0);

    QuickFlash::Tree::ConstTreeIter tree_iter = meshinfo.get_tree_iter();

    std::vector<double> cell_min_coords(dims);
    std::vector<double> cell_max_coords(dims);

    std::vector<double> cell_overlap_pos(dims);
    std::vector<double> cell_vel(dims);

    QuickFlash::Block::BlockData<double> dens;
    
//     std::vector< QuickFlash::Block::BlockData<double> > vel(dims);

    while (tree_iter.valid_node())
      {
      const QuickFlash::Block::BlockInfo & blockinfo 
	= tree_iter.get_block_info();

      // Check for overlap

      const std::vector<double> & block_min = blockinfo.get_min_coords();
      const std::vector<double> & block_max = blockinfo.get_max_coords();
      
      if (!(volumes_overlap(control_volume_minbounds, control_volume_maxbounds,
			    block_min, block_max)))
	tree_iter.next_branch();
      else
	{
	if (tree_iter.is_leaf())
	  {
	  // Read the data

	  const unsigned int block_index = tree_iter.get_block_index();

	  dens_set.get_block_data(block_index, dens);

// 	  for (unsigned int i = 0; i < dims; i++)
// 	    vel_set[i]->get_block_data(block_index, vel[i]);

	  // Parse the cells

	  const unsigned int num_cells = blockinfo.get_num_cells();

	  // const double cell_volume = blockinfo.get_cell_volume();

	  for (unsigned int cell_index = 0; cell_index < num_cells; 
	       cell_index++)
	    {
	    blockinfo.get_cell_bounds(cell_index, cell_min_coords, 
				      cell_max_coords);

	    double overlap_volume = 0.0;
	    double control_fract = 0.0;
	    double cell_fract = 0.0;

	    if (volumes_overlap(control_volume_minbounds, 
				control_volume_maxbounds, cell_min_coords, 
				cell_max_coords, overlap_volume, 
				cell_overlap_pos, control_fract, cell_fract))
	      {
	      const double cell_dens = dens[cell_index];

// 	      for (unsigned int i = 0; i < dims; i++)
// 		cell_vel[i] = vel[i][cell_index];

	      const double overlap_mass = cell_dens * overlap_volume;

	      accum_mass += overlap_mass;

	      for (unsigned int i = 0; i < dims; i++)
		accum_mass_pos[i] += cell_overlap_pos[i] * overlap_mass;

// 	      for (unsigned int i = 0; i < dims; i++)
// 		accum_mass_vel[i] += cell_vel[i] * overlap_mass;
	      }
	    }
	  }
	tree_iter.next();
	}
      }

    std::vector<double> cm(dims, 0.0);

    if (accum_mass > 0.0)
      {
      const double oneOver_mass = 1.0 / accum_mass;

      for (unsigned int i = 0; i < dims; i++)
	cm[i] = accum_mass_pos[i] * oneOver_mass;
      }

    // Write out header on first pass

    if (loop_index == 0)
      {
      cout << "# analyze_volume_fixed output" << endl;
      cout << "#" << endl;

      cout << "# Command:";

      for (int idx = 0; idx < argc; idx++)
	cout << " " << argv[idx];

      cout << "#" << endl;
      cout << "# Filename base [ " << filename_base << " ]" << endl;
      cout << "# Start index [ " << start_index << " ]" << endl;
      cout << "# End index [ " << end_index << " ]" << endl;
      cout << "# Dims [ " << dims << " ]" << endl;

      cout << "# Control volume center [";

      for (unsigned int i = 0; i < dims; i++)
	cout << " " << control_volume_center[i];

      cout << " ]" << endl;

      cout << "# Control volume width [";

      for (unsigned int i = 0; i < dims; i++)
	cout << " " << control_volume_width[i];

      cout << " ]" << endl;

      cout << endl;

      cout << "File_Index" << "\t" << "Time" << "\t" << "Mass";

      for (unsigned int i = 0; i < dims; i++)
	cout << "\t" << "cm_" << i;

      for (unsigned int i = 0; i < dims; i++)
	cout << "\t" << "p_" << i;

      cout << endl;
      }

    cout << file_index << "\t" << sim_time << "\t" << accum_mass;

    for (unsigned int i = 0; i < dims; i++)
      cout << "\t" << cm[i];

    for (unsigned int i = 0; i < dims; i++)
      cout << "\t" << accum_mass_vel[i];

    cout << endl;
    }

  return 0;
  }
