// C++ program file analyze_volume_fixedlayer.cpp

/*
  By Nathan C. Hearn
     May 28, 2007

  Analysis of fixed control volume.
*/


#include <iostream>
#include <string>
#include <sstream>
#include <stdlib.h>
#include "quickflash.hpp"
#include "cubic_volume.hpp"
#include "mixing_layer.hpp"
#include "layer_finder.hpp"


const unsigned int buffer_size = 1000;


void build_datafile_name(const std::string & filename_base,
			 const unsigned int index, const unsigned int digits,
			 std::string & filename)
  {
  std::stringstream namestream;

  namestream << filename_base;

  namestream.width(digits);
  namestream.fill('0');

  namestream << index;

  filename = namestream.str();
  }


int main(int argc, char * argv[])
  {
  using std::cout;
  using std::cerr;
  using std::endl;

  const int min_args = 8;

  if (argc < min_args)
    {
    cerr << "Usage: " << argv[0] << " filename_base start_index end_index" 
	 << endl
	 << "dims <control_volume_center> <control_volume_width> " << endl
	 << "min_data_value max_data_value search_bin_size" << endl
	 << " [ --width-fraction ] [ --follow-layer ]" << endl
	 << endl;

    return 1;
    }

  // Read the arguments

  int req_args = min_args;

  int argptr = 1;

  const std::string filename_base = argv[argptr++];

  const unsigned int start_index 
    = static_cast<unsigned int>(atoi(argv[argptr++]));
  
  const unsigned int end_index 
    = static_cast<unsigned int>(atoi(argv[argptr++]));

  const unsigned int dims = static_cast<unsigned int>(atoi(argv[argptr++]));

  req_args += 2 * dims;

  if (argc < req_args)
    {
    cerr << "Error: Too few arguments" << endl << endl;
    return -1;
    }
  
  std::vector<double> control_volume_center(dims, 0.0);

  for (unsigned int i = 0; i < dims; i++)
    control_volume_center[i] = atof(argv[argptr++]);

  std::vector<double> control_volume_width(dims, 0.0);

  for (unsigned int i = 0; i < dims; i++)
    control_volume_width[i] = atof(argv[argptr++]);

  const double min_data_value = atof(argv[argptr++]);
  const double max_data_value = atof(argv[argptr++]);

  const double search_bin_size = atof(argv[argptr++]);

  if (!(search_bin_size > 0.0))
    {
    cerr << "Error: Search bin size must be positive" << endl << endl;
    return -1;
    }

  bool use_width_fraction = false;
  bool follow_layer = false;

  while (argptr < argc)
    {
    const std::string keyword = argv[argptr++];

    if (keyword == "--width-fraction")
      {
      use_width_fraction = true;

      req_args++;
      }
    else if (keyword == "--follow-layer")
      {
      follow_layer = true;

      req_args++;
      }
    else
      {
      cerr << endl << "Error: Unrecognized keyword [ " << keyword << " ]"
	   << endl;
      return -1;
      }
    }

  // Helpful constants

  std::vector<double> control_volume_halfwidth = control_volume_width;

  for (unsigned int i = 0; i < dims; i++)
    control_volume_halfwidth[i] *= 0.5;

  // Perform the main loop

  const unsigned int loop_count = (end_index + 1) - start_index;

  for (unsigned int loop_index = 0; loop_index < loop_count; loop_index++)
    {
    const unsigned int file_index = loop_index + start_index;

    // Load the data file

    std::string filename;

    build_datafile_name(filename_base, file_index, 4, filename);

    QuickFlash::File::DataFile df(filename, buffer_size, true);

    const QuickFlash::File::MeshInfo & meshinfo = df.get_mesh_info();

    const double sim_time = meshinfo.get_sim_time();

    const unsigned int dims = meshinfo.get_dims();

    const std::vector<double> & min_sim_bounds 
      = meshinfo.get_volume_minbounds();
    const std::vector<double> & max_sim_bounds 
      = meshinfo.get_volume_maxbounds();

    const double min_sim_y = min_sim_bounds[1];
    const double max_sim_y = max_sim_bounds[1];

    // Set up the datasets

    const QuickFlash::File::Dataset & dens_set = df.get_dataset("dens");

    const QuickFlash::File::Dataset & ligh_set = df.get_dataset("ligh");  // rt

    std::vector<const QuickFlash::File::Dataset *> vel_set(dims);

    if (dims > 0)
      vel_set[0] = &(df.get_dataset("velx"));

    if (dims > 1)
      vel_set[1] = &(df.get_dataset("vely"));

    if (dims > 2)
      vel_set[2] = &(df.get_dataset("velz"));

    // Determine the depth of the mixing layer across simulation

    MixingLayer mixing_layer;

    // DO WE WANT TO USE THE MEAN VALUE VERSION???

    get_mixing_layer_meanValue(df, "heav", 1, min_sim_y, max_sim_y, 
			       search_bin_size, min_data_value, 
			       max_data_value, mixing_layer);

    const double layer_min_y = mixing_layer.get_min_pos();
    const double layer_max_y = mixing_layer.get_max_pos();

    const double layer_center_y = 0.5 * (layer_min_y + layer_max_y);
    const double layer_depth_y = layer_max_y - layer_min_y;

    double control_volume_depth_y = control_volume_width[1];

    if (use_width_fraction)
      control_volume_depth_y *= layer_depth_y;

    const double control_volume_half_depth_y = 0.5 * control_volume_depth_y;

    std::vector<double> control_volume_minbounds = control_volume_center;
    std::vector<double> control_volume_maxbounds = control_volume_center;

    for (unsigned int i = 0; i < dims; i++)
      {
      const double half_width = control_volume_halfwidth[i];

      control_volume_minbounds[i] -= half_width;
      control_volume_maxbounds[i] += half_width;
      }

    double sample_min_y = 0.0;
    double sample_max_y = 0.0;

    if (follow_layer)
      {
      sample_min_y = layer_center_y - control_volume_half_depth_y;
      sample_max_y = layer_center_y + control_volume_half_depth_y;

      control_volume_minbounds[1] = sample_min_y;
      control_volume_maxbounds[1] = sample_max_y;
      }
    else
      {
      sample_min_y = control_volume_minbounds[1];
      sample_max_y = control_volume_maxbounds[1];
      }

    // Parse the blocks

    double accum_mass = 0.0;

    double accum_mass_light = 0.0;
    double accum_mass_heavy = 0.0;

    std::vector<double> accum_mass_pos(dims, 0.0);
    std::vector<double> accum_mass_vel(dims, 0.0);

    std::vector<double> accum_mass_pos_light(dims, 0.0);
    std::vector<double> accum_mass_pos_heavy(dims, 0.0);

    std::vector<double> accum_mass_vel_light(dims, 0.0);
    std::vector<double> accum_mass_vel_heavy(dims, 0.0);

    double accum_kinetic_energy = 0.0;
    double accum_potential_energy = 0.0;

    double accum_kinetic_energy_light = 0.0;
    double accum_kinetic_energy_heavy = 0.0;

    double accum_potential_energy_light = 0.0;
    double accum_potential_energy_heavy = 0.0;

    QuickFlash::Tree::ConstTreeIter tree_iter = meshinfo.get_tree_iter();

    std::vector<double> cell_min_coords(dims);
    std::vector<double> cell_max_coords(dims);

    std::vector<double> cell_overlap_pos(dims);
    std::vector<double> cell_vel(dims);

    QuickFlash::Block::BlockData<double> dens;
    QuickFlash::Block::BlockData<double> ligh;
    
    std::vector< QuickFlash::Block::BlockData<double> > vel(dims);

    while (tree_iter.valid_node())
      {
      const QuickFlash::Block::BlockInfo & blockinfo 
	= tree_iter.get_block_info();

      // Check for overlap

      const std::vector<double> & block_min = blockinfo.get_min_coords();
      const std::vector<double> & block_max = blockinfo.get_max_coords();
      
      if (!(volumes_overlap(control_volume_minbounds, control_volume_maxbounds,
			    block_min, block_max)))
	tree_iter.next_branch();
      else
	{
	if (tree_iter.is_leaf())
	  {
	  // Read the data

	  const unsigned int block_index = tree_iter.get_block_index();

	  bool block_data_loaded = false;

	  // Parse the cells

	  const unsigned int num_cells = blockinfo.get_num_cells();

	  // const double cell_volume = blockinfo.get_cell_volume();

	  for (unsigned int cell_index = 0; cell_index < num_cells; 
	       cell_index++)
	    {
	    blockinfo.get_cell_bounds(cell_index, cell_min_coords, 
				      cell_max_coords);

	    double overlap_volume = 0.0;
	    double control_fract = 0.0;
	    double cell_fract = 0.0;
	    double sqVelocity = 0.0;

	    if (volumes_overlap(control_volume_minbounds, 
				control_volume_maxbounds, cell_min_coords, 
				cell_max_coords, overlap_volume, 
				cell_overlap_pos, control_fract, cell_fract))
	      {
	      if (!block_data_loaded)
		{
		dens_set.get_block_data(block_index, dens);

		for (unsigned int i = 0; i < dims; i++)
		  vel_set[i]->get_block_data(block_index, vel[i]);

		ligh_set.get_block_data(block_index, ligh);

		block_data_loaded = true;
		}

	      const double cell_ligh = ligh[cell_index];
	      const double cell_heav = 1.0 - cell_ligh;

	      const double cell_dens = dens[cell_index];

	      for (unsigned int i = 0; i < dims; i++)
		cell_vel[i] = vel[i][cell_index];

	      for (unsigned int i = 0; i < dims; i++)
		{
		const double vel_elem = cell_vel[i];
		sqVelocity += vel_elem * vel_elem;
		}

	      const double height = cell_overlap_pos[1];

	      const double overlap_mass = cell_dens * overlap_volume;

	      const double overlap_mass_light = cell_ligh * overlap_mass;
	      const double overlap_mass_heavy = cell_heav * overlap_mass;

	      accum_mass += overlap_mass;

	      accum_mass_light += overlap_mass_light;
	      accum_mass_heavy += overlap_mass_heavy;

	      for (unsigned int i = 0; i < dims; i++)
		accum_mass_pos[i] += cell_overlap_pos[i] * overlap_mass;

	      for (unsigned int i = 0; i < dims; i++)
		accum_mass_vel[i] += cell_vel[i] * overlap_mass;

	      for (unsigned int i = 0; i < dims; i++)
		accum_mass_pos_light[i] 
		  += cell_overlap_pos[i] * overlap_mass_light;

	      for (unsigned int i = 0; i < dims; i++)
		accum_mass_pos_heavy[i] 
		  += cell_overlap_pos[i] * overlap_mass_heavy;

	      for (unsigned int i = 0; i < dims; i++)
		accum_mass_vel_light[i] += cell_vel[i] * overlap_mass_light;

	      for (unsigned int i = 0; i < dims; i++)
		accum_mass_vel_heavy[i] += cell_vel[i] * overlap_mass_heavy;

	      const double kinetic_energy = 0.5 * sqVelocity * overlap_mass;
	      const double potential_energy = height * overlap_mass;  // g??

	      accum_kinetic_energy += kinetic_energy;
	      accum_potential_energy += potential_energy;

	      accum_kinetic_energy_light += kinetic_energy * cell_ligh;
	      accum_kinetic_energy_heavy += kinetic_energy * cell_heav;

	      accum_potential_energy_light += potential_energy * cell_ligh;
	      accum_potential_energy_heavy += potential_energy * cell_heav;
	      }
	    }
	  }
	tree_iter.next();
	}
      }

    std::vector<double> cm(dims, 0.0);

    if (accum_mass > 0.0)
      {
      const double oneOver_mass = 1.0 / accum_mass;

      for (unsigned int i = 0; i < dims; i++)
	cm[i] = accum_mass_pos[i] * oneOver_mass;
      }

    std::vector<double> cm_light(dims, 0.0);
    std::vector<double> cm_heavy(dims, 0.0);

    if (accum_mass_light > 0.0)
      {
      const double oneOver_mass_light = 1.0 / accum_mass_light;

      for (unsigned int i = 0; i < dims; i++)
	cm_light[i] = accum_mass_pos_light[i] * oneOver_mass_light;
      }

    if (accum_mass_heavy > 0.0)
      {
      const double oneOver_mass_heavy = 1.0 / accum_mass_heavy;

      for (unsigned int i = 0; i < dims; i++)
	cm_heavy[i] = accum_mass_pos_heavy[i] * oneOver_mass_heavy;
      }

    // Write out header on first pass

    if (loop_index == 0)
      {
      cout << "# analyze_volume_fixed output" << endl;
      cout << "#" << endl;

      cout << "# Command:";

      for (int idx = 0; idx < argc; idx++)
	cout << " " << argv[idx];

      cout << "#" << endl;
      cout << "# Filename base [ " << filename_base << " ]" << endl;
      cout << "# Start index [ " << start_index << " ]" << endl;
      cout << "# End index [ " << end_index << " ]" << endl;
      cout << "# Dims [ " << dims << " ]" << endl;

      cout << "# Control volume center [";

      for (unsigned int i = 0; i < dims; i++)
	cout << " " << control_volume_center[i];

      cout << " ]" << endl;

      cout << "# Control volume width [";

      for (unsigned int i = 0; i < dims; i++)
	cout << " " << control_volume_width[i];

      cout << " ]" << endl;

      cout << endl;

      cout << "File_Index" << "\t" << "Time" << "\t" << "Layer_Min_y" << "\t"
	   << "Layer_Max_y" << "\t" << "Sample_Min_y" << "\t" 
	   << "Sample_Max_y" << "\t" << "Mass" << "\t" << "Mass_light" << "\t"
	   << "Mass_heavy";

      for (unsigned int i = 0; i < dims; i++)
	cout << "\t" << "cm_" << i;

      for (unsigned int i = 0; i < dims; i++)
	cout << "\t" << "p_" << i;

      for (unsigned int i = 0; i < dims; i++)
	cout << "\t" << "cm_light_" << i;

      for (unsigned int i = 0; i < dims; i++)
	cout << "\t" << "cm_heavy_" << i;

      for (unsigned int i = 0; i < dims; i++)
	cout << "\t" << "p_light_" << i;

      for (unsigned int i = 0; i < dims; i++)
	cout << "\t" << "p_heavy_" << i;

      cout << "\t" << "kinetic_energy" << "\t" << "potential_energy";

      cout << "\t" << "kinetic_energy_light" << "\t" << "kinetic_energy_heavy";
      cout << "\t" << "potential_energy_light" 
	   << "\t" << "potential_energy_heavy";

      cout << endl;
      }

    cout << file_index << "\t" << sim_time << "\t" << layer_min_y << "\t"
	 << layer_max_y << "\t" << sample_min_y << "\t" << sample_max_y
	 << "\t" << accum_mass << "\t" << accum_mass_light << "\t"
	 << accum_mass_heavy;

    for (unsigned int i = 0; i < dims; i++)
      cout << "\t" << cm[i];

    for (unsigned int i = 0; i < dims; i++)
      cout << "\t" << accum_mass_vel[i];

    for (unsigned int i = 0; i < dims; i++)
      cout << "\t" << cm_light[i];

    for (unsigned int i = 0; i < dims; i++)
      cout << "\t" << cm_heavy[i];

    for (unsigned int i = 0; i < dims; i++)
      cout << "\t" << accum_mass_vel_light[i];

    for (unsigned int i = 0; i < dims; i++)
      cout << "\t" << accum_mass_vel_heavy[i];

    cout << "\t" << accum_kinetic_energy << "\t" << accum_potential_energy;

    cout << "\t" << accum_kinetic_energy_light
	 << "\t" << accum_kinetic_energy_heavy;

    cout << "\t" << accum_potential_energy_light
	 << "\t" << accum_potential_energy_heavy;

    cout << endl;
    }

  return 0;
  }
