// C++ program file sim_1d.cpp

/*
  By Nathan C. Hearn
     June 1, 2007

  Tools for extracting interface position from 1D simulations.
*/


#include "sim_1d.hpp"
#include <string>
#include "quickflash.hpp"
#include "file_list.hpp"


// Local definitions

const std::string foam_name = "foam";
const double target_foam = 0.5;


// Helper functions

void read_file(const std::string & filename, double & sim_time, 
	       double & interface_pos)
  {
  // Open the file and get basic information

  QuickFlash::File::DataFile df(filename, 1000, true);

  const QuickFlash::File::MeshInfo & meshinfo = df.get_mesh_info();

  const double time = meshinfo.get_sim_time();

  const unsigned int dims = meshinfo.get_dims();

  if (dims < 1)
    throw QuickFlash::Except("Insufficient spatial dimensions", __FILE__,
			   __LINE__);

  const unsigned int num_blocks = meshinfo.get_num_blocks();

  const double min_x = meshinfo.get_volume_minbounds()[0];

  // Load the datasets

  const QuickFlash::File::Dataset & foam_set = df.get_dataset(foam_name);

  // Parse the leaf blocks

  double iface_pos = min_x;

  std::vector<double> coords;

  QuickFlash::Block::BlockData<double> foam_data;

  for (unsigned int block_index = 0; block_index < num_blocks; block_index++)
    {   
    if (meshinfo.is_leaf(block_index))
      {
      const QuickFlash::Block::BlockInfo & blockinfo 
        = meshinfo.get_block_info(block_index);

      const unsigned int num_cells = blockinfo.get_num_cells();
      
      foam_set.get_block_data(block_index, foam_data);

      for (unsigned int cell_index = 0; cell_index < num_cells; cell_index++)
	{
	blockinfo.get_cell_center(cell_index, coords);

	const double x_coord = coords[0];

	const double foam_val = foam_data[cell_index];

	if (!(foam_val > target_foam))
	  if (x_coord > interface_pos)
	    iface_pos = x_coord;
	}
      }
    }

  sim_time = time;
  interface_pos = iface_pos;
  }


void read_sim(const std::string & filename_base, 
	      std::vector<double> & time_list, 
	      std::vector<double> & interface_pos_list)
  {
  // Build the search string

  const std::string search_str = filename_base + "[0-9][0-9][0-9][0-9]";

  // Get the filenames

  std::vector<std::string> filename_list;

  if (get_file_list(search_str, filename_list) < 0)
    throw QuickFlash::Except("Unable to find 1D simulation files", __FILE__,
			   __LINE__);

  const unsigned int num_files = filename_list.size();

  time_list.resize(num_files);
  interface_pos_list.resize(num_files);

  for (unsigned int idx = 0; idx < num_files; idx++)
    {
    double time = 0.0;
    double interface_pos = 0.0;

    read_file(filename_list[idx], time, interface_pos);

    time_list[idx] = time;
    interface_pos_list[idx] = interface_pos;
    }
  }


// Class Sim1D

void Sim1D::reset()
  {
  num_times = 0;

  min_time = 0.0;
  max_time = 0.0;

  time_list.clear();
  interface_list.clear();
  }


void Sim1D::reset(const std::string & filename_base)
  {
  read_sim(filename_base, time_list, interface_list);

  num_times = time_list.size();

  min_time = time_list[0];
  max_time = time_list[num_times - 1];
  }


void Sim1D::reset(const Sim1D & source)
  {
  if (&source != this)
    {
    num_times = source.num_times;

    min_time = source.min_time;
    max_time = source.max_time;

    time_list = source.time_list;
    interface_list = source.interface_list;
    }
  }


void Sim1D::get_interface_pos(const double time, double & interface_pos,
			      double & interface_vel)
  {
  unsigned int lower_index = 0;
  unsigned int upper_index = 0;

  double lower_weight = 0.0;
  double upper_weight = 0.0;

  if (get_interp_time(time, lower_index, upper_index, lower_weight, 
		      upper_weight) < 0)
    throw QuickFlash::Except("Unable to find interpolated time", __FILE__,
			   __LINE__);

  const double lower_pos = interface_list[lower_index];
  const double upper_pos = interface_list[upper_index];

  interface_pos = (lower_pos * lower_weight) + (upper_pos * upper_weight);

  const double delta_time = time_list[upper_index] - time_list[lower_index];

  const double delta_pos = upper_pos - lower_pos;

  interface_vel = delta_pos / delta_time;
  }


int Sim1D::get_interp_time(const double time, unsigned int & lower_index,
			   unsigned int & upper_index, double & lower_weight,
			   double & upper_weight)
  {
  int ret_val = -1;  // Error state

  if (num_times > 1)
    {
    if (!((time < min_time) || (time > max_time)))
      {
      unsigned int lo_idx = 0;
      unsigned int hi_idx = num_times - 1;

      while (lo_idx < (hi_idx - 1))
	{
	const unsigned int test_idx = (lo_idx + hi_idx) / 2;

	const double test_time = time_list[test_idx];

	if (test_time > time)
	  hi_idx = test_idx;
	else
	  lo_idx = test_idx;
	}

      // Determine weights

      const double lo_time = time_list[lo_idx];
      const double hi_time = time_list[hi_idx];

      const double interval_time = hi_time - lo_time;

      if (interval_time > 0.0)
	{
	const double delta_time_lower_time = time - lo_time;

	double hi_weight = delta_time_lower_time / interval_time;

        if (hi_weight < 0.0)
	  hi_weight = 0.0;
        else if (hi_weight > 1.0)
	  hi_weight = 1.0;

	// Set the values

	lower_index = lo_idx;
	upper_index = hi_idx;

	lower_weight = 1.0 - hi_weight;
	upper_weight = hi_weight;

	ret_val = 0;  // OK
	}
      }
    }

  return ret_val;
  }
