// C++ program file flash_stat_bins.cpp

/*
  By Nathan C. Hearn
     September 20, 2006

  Statistical binning classes.
*/


#include "stat_bins.hpp"
#include <vector>
#include "quickflash_except.hpp"



namespace Stats
{

// Class Bin

void Bin::reset()
  {
  num_entries = 0;

  sum_weight = 0.0;

  sum_weight_values = 0.0;
  sum_weight_sqValues = 0.0;

  min_value = 0.0;
  max_value = 0.0;
  }


void Bin::reset(const Bin & source)
  {
  if (&source != this)
    {
    num_entries = source.num_entries;

    sum_weight = source.sum_weight;

    sum_weight_values = source.sum_weight_values;
    sum_weight_sqValues = source.sum_weight_sqValues;
    }
  }


void Bin::add_bin(const Bin & source)
  {
  num_entries += source.num_entries;

  sum_weight += source.sum_weight;

  sum_weight_values += source.sum_weight_values;
  sum_weight_sqValues += source.sum_weight_sqValues;

  const double source_min_val = source.min_value;
  const double source_max_val = source.max_value;

  if (source_min_val < min_value)
    min_value = source_min_val;

  if (source_max_val > max_value)
    max_value = source_max_val;
  }


void Bin::add_entry(const double value, const double weight)
  {
  // Should we allow zero weights???

  sum_weight += weight;

  const double weight_value = weight * value;

  sum_weight_values += weight_value;

  sum_weight_sqValues += weight_value * value;

  if (num_entries > 0)
    {
    if (value < min_value)
      min_value = value;
    else if (value > max_value)
      max_value = value;
    }
  else
    {
    min_value = value;
    max_value = value;
    }

  num_entries++;
  }


double Bin::get_pop_sqStddev() const
  {
  // Implement sigma^2 = [ SUM(w_i * x_i^2) / SUM(w_i) ] - (N * MEAN(x_i)^2)

  // const double fl_num_entries = static_cast<double>(num_entries);

  // const double oneOver_mean_weight = fl_num_entries / sum_weight;

  double sqStddev = 0.0;

  if (sum_weight > 0.0)
    {
    const double oneOver_sum_weight = 1.0 / sum_weight;

    const double mean_val = get_mean();

    // double var = (sum_weight_sqValues * oneOver_mean_weight) 
    //   - (fl_num_entries * mean_val * mean_val);

    sqStddev = (sum_weight_sqValues * oneOver_sum_weight) 
      - (mean_val * mean_val);

    if (sqStddev < 0.0)
      sqStddev = 0.0;
    }

  return sqStddev;
  }


// Class BinArray


void BinArray::reset()
  {
  num_bins = 0;

  bin_min_bounds.clear();
  bin_max_bounds.clear();

  bin_center_coords.clear();

  bins.clear();
  }


void BinArray::reset(const unsigned int num_bins, const double min_boundary, 
		     const double max_boundary)
  {
  if (num_bins < 1)
    throw QuickFlash::Except("No bins requested", __FILE__, __LINE__);

  if (!(min_boundary < max_boundary))
    throw QuickFlash::Except("Incompatible boundaries", __FILE__, __LINE__);

  // Build the arrays of boundaries

  std::vector<double> bin_min_boundaries(num_bins);
  std::vector<double> bin_max_boundaries(num_bins);

  const double total_width = max_boundary - min_boundary;

  const double bin_width = total_width / static_cast<double>(num_bins);

  // Make sure boundaries are consistent between bins

  double current_min = min_boundary;

  for (unsigned int i = 0; i < num_bins; i++)
    {
    bin_min_boundaries[i] = current_min;

    current_min += bin_width;

    bin_max_boundaries[i] = current_min;
    }

  bin_max_boundaries[num_bins-1] = max_boundary;  // For good measure

  // Set up the bins

  reset(bin_min_boundaries, bin_max_boundaries);
}


void BinArray::reset(const std::vector<double> bin_min_boundaries, 
		     const std::vector<double> bin_max_boundaries)
  {
  reset();

  num_bins = bin_min_boundaries.size();

  // Make sure arrays are compatible

  if (bin_max_boundaries.size() != num_bins)
    throw QuickFlash::Except("Incompatible bin boundary arrays", __FILE__, 
			__LINE__);

  // Set up the structures

  bin_min_bounds = bin_min_boundaries;
  bin_max_bounds = bin_max_boundaries;

  bin_center_coords.resize(num_bins);
  bin_widths.resize(num_bins);

  bins.resize(num_bins);

  // Fill in arrays and check the boundaries ... do not allow overlaps??

  for (unsigned int bin_index = 0; bin_index < num_bins; bin_index++)
    {
    const double bin_min = bin_min_boundaries[bin_index];
    const double bin_max = bin_max_boundaries[bin_index];

    if (!(bin_min < bin_max))
      throw QuickFlash::Except("Incompatible min/max bin boundaries", __FILE__, 
			  __LINE__);

    bin_center_coords[bin_index] = (bin_min + bin_max) * 0.5;

    bin_widths[bin_index] = bin_max - bin_min;
    }
  }


void BinArray::add_data(const double coordinate, const double value, 
			const double weight)
  {
  // Run through the bins adding the weight and value to any overlapping bin

  for (unsigned int bin_index = 0; bin_index < num_bins; bin_index++)
    {
    if (in_bin(bin_index, coordinate))
      bins[bin_index].add_entry(value, weight);
    }
  }


void BinArray::add_data_spread(const double coordinate, const double value, 
			       const double spread, const double weight)
  {
  // Run through the bins, assigning the fraction of the weight to each bin
  // ... according to the fraction in which the bin covers the spread

  for (unsigned int bin_index = 0; bin_index < num_bins; bin_index++)
    {
    double bin_weight = weight;

    if (spread > 0.0)
      bin_weight *= get_overlap_fract(bin_index, coordinate, spread);
    else if (!(in_bin(bin_index, coordinate)))
      bin_weight = 0.0;

    if (bin_weight > 0.0)
      bins[bin_index].add_entry(value, bin_weight);
    }
  }


bool BinArray::in_bin(const unsigned int bin_index, const double coordinate, 
		      const double spread) const
  {
  // Determine if this position fits in the specified bin

  // CHECK VALUE FOR SPREAD??

  double coord_min = coordinate;
  double coord_max = coordinate;

  if (spread > 0.0)
    {
    const double half_spread = spread * 0.5;

    coord_min -= half_spread;
    coord_max += half_spread;
    }

  // Get the boundaries of the bin

  double bin_min = bin_min_bounds[bin_index];
  double bin_max = bin_max_bounds[bin_index];

  // See if any of the range [min_coord,max_coord) falls in [bin_min,bin_max)

  bool inside = false;

  if ((coord_min < bin_max) && (coord_max > bin_min))
    inside = true;

  return inside;
  }


double BinArray::get_overlap_fract(const unsigned int bin_index, 
				   const double coordinate,
				   const double spread) const
  {
  /*
    Returns the fraction of the range 

        [ coordinate-(spread/2) , coordinate+(spread/2) ) 

    that falls within the specified bin.

    The value returned is in the range [0.0,1.0], i.e., "none of the spread 
    is within the bin" up to "all of the spread is in the bin."
  */

  double ret_fract = 1.0;  // Spread is completely within block

  if (spread > 0.0)
    {
    const double half_spread = spread * 0.5;

    const double coord_min = coordinate - half_spread;
    const double coord_max = coordinate + half_spread;

    // Determine if the spread is not completely within the block

    const double bin_min = bin_min_bounds[bin_index];
    const double bin_max = bin_max_bounds[bin_index];

    if (!(coord_max > bin_max))
      ret_fract = 0.0;
    else if (!(coord_min < bin_max))
      ret_fract = 0.0;
    else if ((coord_min > bin_min) || (coord_max < bin_max))
      {
      // Compute partial coverage

      double overlap_width = spread;

      const double max_excess = coord_max - bin_max;

      if (max_excess > 0.0)
	overlap_width -= max_excess;

      const double min_excess = bin_min - coord_min;

      if (min_excess > 0.0)
	overlap_width -= min_excess;

      // Determine the fraction

      ret_fract = overlap_width / spread;

      if (ret_fract < 0.0)
	ret_fract = 0.0;
      }
    else
      ret_fract = 1.0;  // Full coverage
    }
  else if (!(in_bin(bin_index, coordinate)))
    ret_fract = 0.0;  // No overlap
  else
    ret_fract = 1.0;  // Full overlap

  return ret_fract;
  }


double BinArray::get_transition_pos(const double threshhold_value,
				    const bool start_lower) const
  {
  if (num_bins < 2)
    throw QuickFlash::Except("Insufficient number of bins for transition search",
			__FILE__, __LINE__);

  const unsigned int start_bin = (start_lower ? 0 : (num_bins - 1));

  double trans_pos = get_bin_center(start_bin);

  double prev_value = bins[start_bin].get_mean();

  const double init_value_delta = prev_value - threshhold_value;

  bool found_it = false;

  unsigned int loop_index = 1;

  while ((!found_it) && (loop_index < num_bins))
    {
    const unsigned int bin_index 
      = start_lower ? loop_index : (num_bins - (loop_index + 1));

    const double current_pos = get_bin_center(bin_index);

    const double current_value = bins[bin_index].get_mean();

    const double current_value_delta = current_value - threshhold_value;

    // See if we have crossed the threshhold

    if ((current_value_delta * init_value_delta) < 0.0)
      {
      const double delta_pos = current_pos - trans_pos;

      const double delta_value = current_value - prev_value;
      const double delta_threshhold = threshhold_value - current_value;

      trans_pos += (delta_threshhold / delta_value) * delta_pos;

      found_it = true;
      }
    else
      {
      trans_pos = current_pos;

      prev_value = current_value;

      loop_index++;
      }
    }

  if (!found_it)
    throw QuickFlash::Except("Unable to find threshhold crossing", __FILE__, 
			__LINE__);

  return trans_pos;
  }


double BinArray::get_transition_pos_min(const double threshhold_value,
					const bool start_lower) const
  {
  if (num_bins < 2)
    throw QuickFlash::Except("Insufficient number of bins for transition search",
			__FILE__, __LINE__);

  const unsigned int start_bin = (start_lower ? 0 : (num_bins - 1));

  double trans_pos = get_bin_center(start_bin);

  double prev_value = bins[start_bin].get_min_value();

  const double init_value_delta = prev_value - threshhold_value;

  bool found_it = false;

  unsigned int loop_index = 1;

  while ((!found_it) && (loop_index < num_bins))
    {
    const unsigned int bin_index 
      = start_lower ? loop_index : (num_bins - (loop_index + 1));

    const double current_pos = get_bin_center(bin_index);

    const double current_value = bins[bin_index].get_min_value();

    const double current_value_delta = current_value - threshhold_value;

    // See if we have crossed the threshhold

    if ((current_value_delta * init_value_delta) < 0.0)
      {
      const double delta_pos = current_pos - trans_pos;

      const double delta_value = current_value - prev_value;
      const double delta_threshhold = threshhold_value - current_value;

      trans_pos += (delta_threshhold / delta_value) * delta_pos;

      found_it = true;
      }
    else
      {
      trans_pos = current_pos;

      prev_value = current_value;

      loop_index++;
      }
    }

  if (!found_it)
    throw QuickFlash::Except("Unable to find threshhold crossing", __FILE__, 
			__LINE__);

  return trans_pos;
  }


double BinArray::get_transition_pos_max(const double threshhold_value,
					const bool start_lower) const
  {
  if (num_bins < 2)
    throw QuickFlash::Except("Insufficient number of bins for transition search",
			__FILE__, __LINE__);

  const unsigned int start_bin = (start_lower ? 0 : (num_bins - 1));

  double trans_pos = get_bin_center(start_bin);

  double prev_value = bins[start_bin].get_max_value();

  const double init_value_delta = prev_value - threshhold_value;

  bool found_it = false;

  unsigned int loop_index = 1;

  while ((!found_it) && (loop_index < num_bins))
    {
    const unsigned int bin_index 
      = start_lower ? loop_index : (num_bins - (loop_index + 1));

    const double current_pos = get_bin_center(bin_index);

    const double current_value = bins[bin_index].get_max_value();

    const double current_value_delta = current_value - threshhold_value;

    // See if we have crossed the threshhold

    if ((current_value_delta * init_value_delta) < 0.0)
      {
      const double delta_pos = current_pos - trans_pos;

      const double delta_value = current_value - prev_value;
      const double delta_threshhold = threshhold_value - current_value;

      trans_pos += (delta_threshhold / delta_value) * delta_pos;

      found_it = true;
      }
    else
      {
      trans_pos = current_pos;

      prev_value = current_value;

      loop_index++;
      }
    }

  if (!found_it)
    throw QuickFlash::Except("Unable to find threshhold crossing", __FILE__, 
			__LINE__);

  return trans_pos;
  }


}  // End namespace Stats
