#!/usr/bin/env python

# tile_left_right.py

# By Nathan C. Hearn
#    November 16, 2008
#
# Tiles pairs of images horizontally


## Image operation functions


def make_tiled_image(left_image, right_image) :

    from PIL import Image

    ## ALPHA CHANNELS???

    ## Find the final image size

    left_image_size = left_image.size
    right_image_size = right_image.size

    left_min_coords = (0, 0)
    left_max_coords = (left_min_coords[0] + left_image_size[0],
                       left_min_coords[1] + left_image_size[1])

    right_min_coords = (left_image_size[0], 0)
    right_max_coords = (right_min_coords[0] + right_image_size[0],
                        right_min_coords[1] + right_image_size[1])

    image_size = (right_max_coords[0] - left_min_coords[0],
                  right_max_coords[1] - left_min_coords[1])

    ## ERROR CHECKING???

    ## Shift the images for the proper offset (use cropping)

    tiled_image = Image.new("RGB", image_size)

    tiled_image.paste(left_image, left_min_coords)
    tiled_image.paste(right_image, right_min_coords)
    
    return tiled_image


## Main

def main() :

    from sys import exit, stderr
    from optparse import OptionParser
    
    from PIL import Image

    ## Set up the command-line parser

    usage = str("%prog [options] output_prefix image1_left image1_right "
                + "[ image2_left image2_right ... ]")
    version = str("%prog 0.1")

    argparser = OptionParser(usage, version=version)

    output_index_start_str = str("output-index-start")
    argparser.add_option("--" + output_index_start_str, type="int",
                         dest="output_index_start",
                         help="Starting index for output filenames")
    argparser.set_defaults(output_index_start=None)

    output_index_digits_str = str("output-index-digits")
    argparser.add_option("--" + output_index_digits_str, type="int",
                         dest="output_index_digits",
                         help="Number of digits in output filename index")
    argparser.set_defaults(output_index_digits=4)

    separate_left_right_lists_str = str("separate-left-right-lists")
    argparser.add_option("--" + separate_left_right_lists_str,
                         action="store_true",
                         dest="separate_left_right_lists",
                         help="All left-side files listed before right-side")
    argparser.set_defaults(separate_left_right_lists=False)

    quiet_str = str("quiet")
    argparser.add_option("--" + quiet_str, action="store_true",
                         dest="quiet_operation",
                         help="Suppress non-error output")
    argparser.set_defaults(quiet_operation=False)


    ## Process the command line input

    opts, args = argparser.parse_args()

    argc = len(args)

    min_args = 3

    if (argc < min_args) :
        stderr.write("\n")
        argparser.print_help(file=stderr)
        stderr.write("\n\n")
        exit(1)


    ## Get the optional arguments

    output_index_start = opts.output_index_start

    output_index_digits = opts.output_index_digits

    if (output_index_digits < 1) :
        stderr.write("\nError: Output index digits must be positive\n\n")
        exit(-1)

    separate_left_right_lists = opts.separate_left_right_lists

    quiet_operation = opts.quiet_operation


    ## Get the required arguments

    argptr = int(0)

    output_prefix = str(args[argptr])
    argptr += 1

    if (((argc - argptr) % 2) != 0) :
        stderr.write("\nError: Pairs of image files required\n\n")
        exit(-1)

    left_image_file_list = list()
    right_image_file_list = list()

    if (separate_left_right_lists) :

        num_pairs = (argc - argptr) / 2

        start_left_index = argptr
        start_right_index = argptr + num_pairs

        for pr_idx in range(num_pairs) :
            
            left_image_file_list.append(str(args[start_left_index + pr_idx]))
            right_image_file_list.append(str(args[start_right_index + pr_idx]))

        # Set argptr to end of file list

        argptr = start_right_index + num_pairs

    else :
        
        while (argptr < argc) :
        
            left_image_file_list.append(str(args[argptr]))
            argptr += 1

            right_image_file_list.append(str(args[argptr]))
            argptr += 1

    num_image_pairs = len(left_image_file_list)


    # Make sure there is an index if more than one image pair

    if ((output_index_start is None) and (num_image_pairs > 1)) :
        output_index_start = 0


    ## Process the files

    for file_index in range(num_image_pairs) :

        output_filename = output_prefix

        if (output_index_start is not None) :

            output_index = output_index_start + file_index

            output_filename += \
                "_" + (("%(idx)0" + str(output_index_digits) + "d")
                       % { "idx" : output_index })

        output_filename += ".png"


        left_image_filename = left_image_file_list[file_index]
        right_image_filename = right_image_file_list[file_index]

        left_image = Image.open(left_image_filename)
        right_image = Image.open(right_image_filename)

        tiled_image = make_tiled_image(left_image, right_image)

        tiled_image.save(output_filename)

        if (not quiet_operation) :
            stderr.write("Wrote image " + output_filename + "\n")

    if (not quiet_operation) :
        stderr.write("\nDone.\n\n")


if (__name__ == "__main__") :
    main()

