// C++ program file colorize_image.cpp

/*
  By Nathan C. Hearn
     May 30, 2008

  Apply a colormap to the luminosity value of the image pixels
*/


#ifndef USE_MAGICK

#error "ImageMagick support not present"

#else

#include <iostream>
#include <string>
#include <list>
#include <cmath>
#include <Magick++.h>
#include "quickflash_color_colormap.hpp"
#include "quickflash_utils_commandline.hpp"
#include "quickflash_utils_file.hpp"


const char default_image_suffix[] = "png";


int main(int argc, char * argv[])
  {
  using std::cerr;
  using std::endl;

  std::vector<std::string> argument_list;
  std::list<QuickFlash::Utils::Option> option_list;

  QuickFlash::Utils::parse_command_line(argc, argv, argument_list, 
					option_list);

  const unsigned int num_args = argument_list.size();

  const unsigned int min_args = 3;

  if (num_args < min_args)
    {
    const std::string qualified_exec_name = argv[0];

    std::string exec_path;
    std::string exec_name;

    QuickFlash::Utils::split_name_path_file(qualified_exec_name, exec_path, 
					  exec_name);

    cerr << endl << "Usage: " << exec_name << " colormap_file output_prefix"
	 << endl << "\t"
	 << "[ --output-suffix=OUTPUT_SUFFIX ]"
	 << endl << "\t"
	 << "[ --image-type=IMAGE_SUFFIX ]"
	 << endl << "\t"
	 << "[ --expand-color-range ] [ --logscale ] [ --quiet ]"
	 << endl << "\t"
	 << "image_file1 [ image_file2 ... ]"
	 << endl << endl;

    return 1;
    }

  // Read the arguments

  unsigned int argptr = 0;

  const std::string colormap_file = argument_list[argptr++];

  const std::string output_prefix = argument_list[argptr++];

  std::list<std::string> filename_list;

  while (argptr < num_args)
    filename_list.push_back(argument_list[argptr++]);

  // Parse the options

  bool use_output_suffix = false;
  std::string output_suffix;

  std::string image_suffix = default_image_suffix;

  // bool expand_color_range = false;

  bool use_logscale = false;

  bool quiet_operation = false;

  std::list<QuickFlash::Utils::Option>::const_iterator option_iter 
    = option_list.begin();

  while (option_iter != option_list.end())
    {
    const QuickFlash::Utils::Option & current_option = *option_iter;

    const std::string & keyword = current_option.get_key();

    if (keyword == "output-suffix")
      {
      if (!current_option.value_present())
	{
	cerr << endl << "Error: Output suffix is required with option"
	     << endl << endl;
	return -1;
	}

      use_output_suffix = true;

      output_suffix = current_option.get_value();
      }
    else if (keyword == "image-type")
      {
      if (!current_option.value_present())
	{
	cerr << endl 
	     << "Error: Image-type option requires an image file suffix"
	     << endl << endl;
	return -1;
	}

      image_suffix = current_option.get_value();
      }
    else if (keyword == "expand-color-range")
      {
      // expand_color_range = true;

	{
	cerr << endl 
	     << "Error: Option expand-color-range not supported at this time"
	     << endl << endl;
	return -1;
	}
      }
    else if (keyword == "logscale")
      {
      use_logscale = true;
      }
    else if (keyword == "quiet")
      {
      quiet_operation = true;
      }
    else
      {
      cerr << endl << "Error: Unrecognized option " << keyword << endl << endl;
      return -1;
      }

    ++option_iter;
    }

  // Perform some basic checks and set up preliminary objects

  QuickFlash::Color::Colormap colormap;

  QuickFlash::Color::load_colormap(colormap_file, colormap, use_logscale);

  // Run through the files

  std::list<std::string>::const_iterator file_iter = filename_list.begin();

  while (file_iter != filename_list.end())
    {
    const std::string & filename = *file_iter;

    // Read the image

    Magick::Image orig_image;

    orig_image.read(filename);

    const Magick::Geometry orig_size = orig_image.size();

    const unsigned int width = orig_size.width();
    const unsigned int height = orig_size.height();

    const unsigned int num_pixels = width * height;

    // Get the image data

    const Magick::PixelPacket * orig_pixels 
      = orig_image.getPixels(0, 0, width, height);

    // Rescale the colormap -- EXPAND RANGE???  (2-pass algorithm???)

    double min_colormap_value = 0.0;
    double max_colormap_value = 1.0;

    colormap.rescale_values(min_colormap_value, max_colormap_value, 
			    use_logscale);

    // Create the output image

    const Magick::ColorRGB magick_black(0.0, 0.0, 0.0);

    Magick::Image new_image(orig_size, magick_black);

    new_image.type(Magick::TrueColorMatteType);
    new_image.depth(8);

    Magick::Pixels pixel_cache(new_image);

    Magick::PixelPacket * new_pixels = pixel_cache.get(0, 0, width, height);

    // Process the data

    QuickFlash::Color::Color color;

    Magick::ColorRGB magick_color(0.0, 0.0, 0.0);

    for (unsigned int pixel_index = 0; pixel_index < num_pixels; pixel_index++)
      {
      Magick::ColorGray orig_color(orig_pixels[pixel_index]);

      const double orig_value = orig_color.shade();

      colormap.get_color(orig_value, color);

      magick_color.red(color.get_red());
      magick_color.green(color.get_green());
      magick_color.blue(color.get_blue());

      // ImageMagick (prior to 6.3.0) defines alpha as level of transparency
      // ... Magick++ seems to follow this definition for now ...

      const double transparency = 1.0 - color.get_alpha();

      magick_color.alpha(transparency);

      new_pixels[pixel_index] = magick_color;
      }

    new_image.syncPixels();

    // Save the image

    std::string orig_path;
    std::string orig_name;

    QuickFlash::Utils::split_name_path_file(filename, orig_path, orig_name);

    std::string output_filename = output_prefix + "_" + orig_name;

    if (use_output_suffix)
      output_filename += "_" + output_suffix;

    output_filename += "." + image_suffix;

    new_image.write(output_filename);

    ++file_iter;
    }

  return 0;
  }


#endif  // USE_MAGICK
