#!/usr/bin/env python

# colorize_image.py

# By Nathan C. Hearn
#    May 27, 2008
#
# Colorizer (color remapper) for images


try :
    import psyco
except ImportError :
    pass


from colormap import Colormap


default_image_type = "png"


def colorize_image(image, colormap) :

    from PIL import Image

    ## For now, assume that image is type L

    image_size = image.size

    i_size = image_size[0]
    j_size = image_size[1]

    ## Build the pixel array from the image

    image_pixels = image.load()

    new_image = Image.new("RGBA", image_size)

    new_pixels = new_image.load()

    for i in range(i_size) :
        for j in range(j_size) :

            value = float(image_pixels[i, j])

            pixel_color = colormap.get_color(value)

            new_pixels[i, j] = pixel_color

    return new_image


def main() :

    from sys import exit, stderr
    from os import path
    from optparse import OptionParser
    
    from PIL import Image

    ## Set up the command-line parser

    usage = str("%prog [options] colormap output_prefix image1 [ image2 ... ]")
    version = str("%prog 0.1")

    argparser = OptionParser(usage, version=version)

    output_suffix_str = str("output-suffix")
    argparser.add_option("--" + output_suffix_str, type="string",
                         dest="output_suffix", help="Output filename suffix")
    argparser.set_defaults(output_suffix=None)

    image_type_str = str("image-type")
    argparser.add_option("--" + image_type_str, type="string",
                         dest="image_type",
                         help="Image type (output filename extension)")
    argparser.set_defaults(image_type=default_image_type)

    expand_color_range_str = str("expand-color-range")
    argparser.add_option("--" + expand_color_range_str, action="store_true",
                         dest="expand_color_range",
                         help="Restrict colormap to values present in image")
    argparser.set_defaults(expand_color_range=False)

    logscale_str = str("logscale")
    argparser.add_option("--" + logscale_str, action="store_true",
                         dest="logscale",
                         help="Use logarithmic scale for color interpolation")
    argparser.set_defaults(logscale=False)

    quiet_str = str("quiet")
    argparser.add_option("--" + quiet_str, action="store_true",
                         dest="quiet_operation",
                         help="Suppress non-error output")
    argparser.set_defaults(quiet_operation=False)


    ## Process the command line input

    opts, args = argparser.parse_args()

    argc = len(args)

    min_args = 3

    if (argc < min_args) :
        stderr.write("\n")
        argparser.print_help(file=stderr)
        stderr.write("\n\n")
        exit(1)


    ## Get the required arguments

    argptr = int(0)

    colormap_file = str(args[argptr])
    argptr += 1

    output_prefix = str(args[argptr])
    argptr += 1

    image_file_list = list()

    while (argptr < argc) :
        image_file_list.append(str(args[argptr]))
        argptr += 1


    ## Get the optional arguments

    output_suffix = opts.output_suffix

    image_type = opts.image_type

    expand_color_range = opts.expand_color_range

    logscale = opts.logscale

    quiet_operation = opts.quiet_operation


    ## Load the colormap

    colormap = Colormap(colormap_file, logscale)


    ## Run through the images

    for image_file in image_file_list :

        base_filename = path.split(image_file)[1]

        output_filename = output_prefix + base_filename

        if (output_suffix is not None) :
            output_filename += "_" + output_suffix

        output_filename += "." + image_type

        orig_image = Image.open(image_file)

        gray_image = orig_image.convert("L")

        min_value = 0
        max_value = 255

        if (expand_color_range) :
            min_value, max_value = gray_image.get_extrema()

        colormap.rescale_colormap(float(min_value), float(max_value))

        new_image = colorize_image(gray_image, colormap)

        new_image.save(output_filename)

        if (not quiet_operation) :
            stderr.write("Wrote image " + output_filename + "\n")

    if (not quiet_operation) :
        stderr.write("\nDone.\n\n")


if (__name__ == "__main__") :
    main()
