// C++ program file drawimage.cpp

/*
  By Nathan C. Hearn
     Februrary 15, 2008

  C++ version of drawimage.py using ImageMagick
*/


#ifndef USE_MAGICK

#error "ImageMagick support not present"

#else

#include <iostream>
#include <string>
#include <list>
#include <cmath>
#include "quickflash_color_colormap.hpp"
#include "quickflash_block_blockdata.hpp"
#include "quickflash_image_imagegrid.hpp"
#include "quickflash_image_imagemagick.hpp"
#include "quickflash_utils_commandline.hpp"
#include "quickflash_utils_file.hpp"
#include "read_grid_datafile.hpp"


const char default_image_suffix[] = "png";
const char info_suffix[] = "info";


int main(int argc, char * argv[])
  {
  using std::cerr;
  using std::endl;

  using QuickFlash::Image::ImageMagick::write_colormap;
  using QuickFlash::Image::ImageMagick::write_colormap_info;

  std::vector<std::string> argument_list;
  std::list<QuickFlash::Utils::Option> option_list;

  QuickFlash::Utils::parse_command_line(argc, argv, argument_list, 
					option_list);

  const unsigned int num_args = argument_list.size();

  const unsigned int min_args = 1;

  if (num_args < min_args)
    {
    cerr << endl << "Usage: " << argv[0]
	 << endl << "\t"
	 << "[ --output-prefix=output_prefix ]"
	 << endl << "\t"
	 << "[ --min-value=min_value ] [ --max-value=max_value ]"
	 << endl << "\t"
	 << "[ --logscale ] [ --colormap=colormap_file ]"
	 << endl << "\t"
	 << "[ --rescale-colormap ] [ --show-colormap ]"
	 << endl << "\t"
	 << "[ --save-colormap=length,width ] [ --save-colormap-info ]"
	 << endl << "\t"
	 << "[ --transpose] [ --image-type=image_suffix ]"
	 << endl << "\t"
	 << "filename1 [ filename2 ... ]"
	 << endl << endl;

    return 1;
    }

  // Read the arguments

  std::list<std::string> filename_list;

  for (unsigned int arg_index = 0; arg_index < num_args; arg_index++)
    filename_list.push_back(argument_list[arg_index]);

  // Parse the options

  bool use_output_prefix = false;
  std::string output_prefix;

  bool min_value_set = false;
  double min_value = 0.0;

  bool max_value_set = false;
  double max_value = 0.0;

  bool use_logscale = false;

  bool use_colormap = false;
  std::string colormap_filename;
  
  bool rescale_colormap = false;

  bool show_colormap = false;

  bool save_colormap = false;
  unsigned int colormap_value_axis_length = 0;
  unsigned int colormap_width = 0;

  bool save_colormap_info = false;

  bool transpose_image = false;

  std::string image_suffix = default_image_suffix;

  std::list<QuickFlash::Utils::Option>::const_iterator option_iter 
    = option_list.begin();

  while (option_iter != option_list.end())
    {
    const QuickFlash::Utils::Option & current_option = *option_iter;

    const std::string & keyword = current_option.get_key();

    if (keyword == "output-prefix")
      {
      if (!current_option.value_present())
	{
	cerr << endl << "Error: Output prefix is required with option"
	     << endl << endl;
	return -1;
	}

      use_output_prefix = true;

      output_prefix = current_option.get_value();
      }
    else if (keyword == "min-value")
      {
      if (!current_option.value_present())
	{
	cerr << endl << "Error: Min-value option requires value"
	     << endl << endl;
	return -1;
	}

      min_value_set = true;

      current_option.get_value(min_value);
      }
    else if (keyword == "max-value")
      {
      if (!current_option.value_present())
	{
	cerr << endl << "Error: Max-value option requires value"
	     << endl << endl;
	return -1;
	}

      max_value_set = true;

      current_option.get_value(max_value);
      }
    else if (keyword == "logscale")
      {
      use_logscale = true;
      }
    else if (keyword == "colormap")
      {
      if (!current_option.value_present())
	{
	cerr << endl << "Error: Colormap option requires a colormap filename"
	     << endl << endl;
	return -1;
	}

      use_colormap = true;

      colormap_filename = current_option.get_value();
      }
    else if (keyword == "rescale-colormap")
      {
      rescale_colormap = true;
      }
    else if (keyword == "show-colormap")
      {
      show_colormap = true;
      }
    else if (keyword == "save-colormap")
      {
      if (!current_option.value_present())
	{
	cerr << endl 
	     << "Error: Save colormap option requires a length and width"
	     << endl << endl;
	return -1;
	}

      std::vector<unsigned int> length_width;

      current_option.get_value(length_width);

      if (length_width.size() != 2)
	{
	cerr << endl 
	     << "Error: Save colormap option requies a length and width"
	     << endl << endl;
	return -1;
	}

      colormap_value_axis_length = length_width[0];
      colormap_width = length_width[1];

      if (colormap_value_axis_length < 2)
	{
	cerr << endl
	     << "Error: Colormap length must be greater than one"
	     << endl << endl;
	return -1;
	}

      if (colormap_width < 1)
	{
	cerr << endl << "Error:Colormap width must be at least one"
	     << endl << endl;
	return -1;
	}

      save_colormap = true;
      }
    else if (keyword == "save-colormap-info")
      {
      save_colormap_info = true;
      }
    else if (keyword == "transpose")
      {
      transpose_image = true;
      }
    else if (keyword == "image-type")
      {
      if (!current_option.value_present())
	{
	cerr << endl 
	     << "Error: Image-type option requires an image file suffix"
	     << endl << endl;
	return -1;
	}

      image_suffix = current_option.get_value();
      }
    else
      {
      cerr << endl << "Error: Unrecognized option " << keyword << endl << endl;
      return -1;
      }

    ++option_iter;
    }

  // Perform some basic checks and set up preliminary objects

  if (min_value_set && max_value_set)
    if (!(min_value < max_value))
      {
      cerr << endl << "Error: Min value must be less than max value"
	   << endl << endl;
      return -1;
      }

  QuickFlash::Color::Colormap colormap;

  if (use_colormap)
    QuickFlash::Color::load_colormap(colormap_filename, colormap, 
				     use_logscale);

  QuickFlash::Block::BlockData<double> grid_data;
  QuickFlash::Image::ImageGrid image_grid;

  // Run through the files

  std::list<std::string>::const_iterator file_iter = filename_list.begin();

  while (file_iter != filename_list.end())
    {
    const std::string & filename = *file_iter;

    // Read the data

    read_grid_datafile(filename, grid_data);

    double grid_min_value = 0.0;
    double grid_max_value = 0.0;

    if (min_value_set && max_value_set)
      {
      grid_min_value = min_value;
      grid_max_value = max_value;
      }
    else
      {
      grid_data.get_minmax_values(grid_min_value, grid_max_value);
   
      if (min_value_set)
	grid_min_value = min_value;

      if (max_value_set)
	grid_max_value = max_value;

      if (!min_value_set)
	{
	if (grid_min_value > grid_max_value)
	  grid_min_value = grid_max_value;
	}

      if (!max_value_set)
	{
	if (grid_max_value < grid_min_value)
	  grid_max_value = grid_min_value;
	}
      }

    // Set up the colormap

    if ((!use_colormap) || rescale_colormap)
      {
      double colormap_min_value = grid_min_value;
      double colormap_max_value = grid_max_value;

      if (!(colormap_min_value < colormap_max_value))
	{
	// Make a fake maximum value that is greater than the minimum value

	colormap_max_value = (1.1 * std::fabs(colormap_min_value)) + 1.0;
	}

      if (use_colormap)
	colormap.rescale_values(colormap_min_value, colormap_max_value, 
				use_logscale);
      else
	colormap.reset(colormap_min_value, colormap_max_value, use_logscale);
      }

    // Build the image grid

    image_grid.reset(grid_data, colormap);

    // Save the image

    std::string output_filename = "";

    if (use_output_prefix)
      output_filename += output_prefix + "_";

    output_filename += filename + "." + image_suffix;

    if (show_colormap)
      QuickFlash::Image::ImageMagick::write_image_colormap(output_filename,
							 image_grid, colormap,
							 transpose_image, 
							 true);
    else
      QuickFlash::Image::ImageMagick::write_image(output_filename, image_grid,
						transpose_image);

    if (save_colormap)
      {
      std::string colormap_image_filename = "";

      if (use_output_prefix)
	colormap_image_filename += output_prefix + "_";

      colormap_image_filename += filename + ".cm." + image_suffix;

      write_colormap(colormap_image_filename, colormap, 
		     colormap_value_axis_length, colormap_width, false,
		     true, grid_min_value, grid_max_value);

      if (save_colormap_info)
	{
	const std::string colormap_info_filename 
	  = colormap_image_filename + "." + info_suffix;

	std::string colormap_image_path;
	std::string unqualified_colormap_image_name;

	QuickFlash::Utils::split_name_path_file(colormap_image_filename,
					      colormap_image_path,
					      unqualified_colormap_image_name);

	write_colormap_info(unqualified_colormap_image_name, 
			    colormap_info_filename, colormap, 
			    colormap_value_axis_length, colormap_width, false,
			    true, grid_min_value, grid_max_value);

	}
      }
    else if (save_colormap_info)
      {
      std::string colormap_info_filename = "";

      if (use_output_prefix)
	colormap_info_filename += output_prefix + "_";

      colormap_info_filename += filename + ".cm." + info_suffix;

      write_colormap_info(colormap_info_filename, colormap, 
			  colormap_value_axis_length, colormap_width, 
			  false, true, grid_min_value, grid_max_value);
      }

    ++file_iter;
    }

  return 0;
  }


#endif  // USE_MAGICK
