// C++ program file read_grid_datafile.cpp

/*
  By Nathan C. Hearn
     February 15, 2008

  Reader for ASCII grid datafiles (the same as used by drawimage.py)
*/


#include "read_grid_datafile.hpp"
#include <string>
#include <sstream>
#include <vector>
#include <list>
#include <fstream>
#include "quickflash_block_blockdata.hpp"
#include "quickflash_except.hpp"
#include "quickflash_utils_text.hpp"


const char file_comment_marker = '#';
const char file_delimiter = ' ';


int read_file_line(std::ifstream & f, std::string & line, 
		   const char comment_marker)
  {
  int ret_val = -1;  // Error state (EOF)

  if ((!(f.eof())) && (f.tellg() >= 0))
    {
    std::stringstream sstream;

    // Read the next line, and dump the newline character

    f.get(*(sstream.rdbuf()));

    if (!(f.eof()))
      f.ignore(1);

    std::string work_line = sstream.str();

    // Clean up the line (remove any comments)

    std::string::size_type num_characters = work_line.size();

    std::string::size_type start_pos = 0;

    bool found_start = false;

    while ((!found_start) && (start_pos < num_characters))
      {
      if (work_line[start_pos] != ' ')
	found_start = true;
      else
	start_pos++;
      }

    std::string::size_type end_pos = num_characters;

    if (comment_marker != 0)
      {
      bool found_comment = false;

      end_pos = start_pos;

      while ((!found_comment) && (end_pos < num_characters))
	{
	if (work_line[end_pos] == comment_marker)
	  found_comment = true;
	else
	  end_pos++;
	}
      }

    // Remove any trailing spaces
    
    bool found_end = false;

    while ((!found_end) && (end_pos > start_pos))
      {
      const char previous_character = work_line[end_pos - 1];

      if (previous_character != ' ')
	found_end = true;
      else
	end_pos--;
      }

    line = work_line.substr(start_pos, (end_pos - start_pos));

    ret_val = 0;  // Normal line
    }

  return ret_val;
  }
  

void read_grid_datafile(const std::string & filename, 
			QuickFlash::Block::BlockData<double> & grid)
  { read_grid_datafile(filename.c_str(), grid); }


void read_grid_datafile(const char filename[], 
			QuickFlash::Block::BlockData<double> & grid)
  {
  // Open the file

  std::ifstream f(filename);

  // Read the data rows

  std::list< std::vector<double> > row_list;

  std::string current_row_string;
  std::vector<double> current_row_values;

  while (read_file_line(f, current_row_string, file_comment_marker) == 0)
    {
    if (current_row_string.size() > 0)
      {
      std::vector<double> current_row_values;

      QuickFlash::Utils::read_vector_string(current_row_string, 
					  current_row_values, file_delimiter);

      row_list.push_back(current_row_values);
      }
    }
  
  // Determine the number of rows and columns

  const unsigned int num_rows = row_list.size();

  if (num_rows == 0)
    grid.reset();
  else
    {
    const unsigned int num_columns = row_list.front().size();

    if (num_columns == 0)
      grid.reset();
    else
      {
      std::vector<unsigned int> grid_dims(2);

      grid_dims[0] = num_rows;
      grid_dims[1] = num_columns;

      grid.reset(grid_dims);

      unsigned int elem_counter = 0;

      std::list< std::vector<double> >::const_iterator row_iter 
	= row_list.begin();

      while (row_iter != row_list.end())
	{
	const std::vector<double> & current_row = *row_iter;

	if (current_row.size() != num_columns)
	  throw QuickFlash::Except("Incompatible column count", __FILE__, 
				 __LINE__);

	for (unsigned int index = 0; index < num_columns; index++)
	  {
	  grid[elem_counter] = current_row[index];

	  elem_counter++;
	  }

	++row_iter;
	}

      // GET RID OF THIS CHECK AFTER VERIFICATION!!!

      if (elem_counter != grid.get_num_cells())
	throw QuickFlash::Except("Incorrect number of data points", __FILE__,
			       __LINE__);
      }
    }
  }


    
