// C++ program file star.cpp

/*
  By Nathan C. Hearn
     June 14, 2008

  Star element prototype.
*/


#include "star.hpp"
#include <cmath>
#include "quickflash_color_color.hpp"
#include "quickflash_except.hpp"


// Class Star

const double Star::deg_to_rad = (M_PI) / 180.0;

Star::Star() :
  star_x(0.0), star_y(0.0), rad_scale(1.0), oneOver_sq_rad_scale(1.0),
  max_rad(1.0), sq_max_rad(1.0), use_spikes(false), spike_angle_freq(0.0),
  spike_scale_fact(1.0), spike_phase_rad(0.0), spike_power(4.0), colormap()
  { reset(); }


Star::Star(const double center_x, const double center_y, 
	   const double radial_scale, const double max_radius,
	   const QuickFlash::Color::Color & star_color, 
	   const unsigned int num_spikes, const double spike_scale_factor,
	   const double spike_phase_deg, const double spike_steepening_power) :
  star_x(0.0), star_y(0.0), rad_scale(1.0), oneOver_sq_rad_scale(1.0),
  max_rad(1.0), sq_max_rad(1.0), use_spikes(false), spike_angle_freq(0.0),
  spike_scale_fact(1.0), spike_phase_rad(0.0), spike_power(4.0), colormap()
  {
  reset(center_x, center_y, radial_scale, max_radius, star_color, num_spikes, 
	spike_scale_factor, spike_phase_deg, spike_steepening_power);
  }


Star::Star(const Star & source) :
  star_x(0.0), star_y(0.0), rad_scale(1.0), oneOver_sq_rad_scale(1.0),
  max_rad(1.0), sq_max_rad(1.0), use_spikes(false), spike_angle_freq(0.0),
  spike_scale_fact(1.0), spike_phase_rad(0.0), spike_power(4.0), colormap()
  { reset(source); }


void Star::reset()
  {
  set_pos(0.0, 0.0);

  set_size(1.0, 1.0);

  set_spikes(0);

  QuickFlash::Color::Color star_color(1.0);  // Default is white star

  set_star_color(star_color);
  }


void Star::reset(const double center_x, const double center_y, 
		 const double radial_scale, const double max_radius,
		 const QuickFlash::Color::Color & star_color, 
		 const unsigned int num_spikes, 
		 const double spike_scale_factor,
		 const double spike_phase_deg, 
		 const double spike_steepening_power)
  {
  set_pos(center_x, center_y);

  set_size(radial_scale, max_radius);

  set_spikes(num_spikes, spike_scale_factor, spike_phase_deg, 
	     spike_steepening_power);

  set_star_color(star_color);
  }


void Star::reset(const Star & source)
  {
  if (&source != this)
    {
    star_x = source.star_x;
    star_y = source.star_y;

    rad_scale = source.rad_scale;
    oneOver_sq_rad_scale = source.oneOver_sq_rad_scale;

    max_rad = source.max_rad;
    sq_max_rad = source.sq_max_rad;

    use_spikes = source.use_spikes;

    spike_angle_freq = source.spike_angle_freq;
    spike_scale_fact = source.spike_scale_fact;
    spike_phase_rad = source.spike_phase_rad;
    spike_power = source.spike_power;

    colormap = source.colormap;
    }
  }


void Star::set_size(const double radial_scale, const double max_radius)
  {
  if (!(radial_scale > 0.0))
    throw QuickFlash::Except("Radial scale must be positive", __FILE__, 
			   __LINE__);

  if (!(max_radius > 0.0))
    throw QuickFlash::Except("Max radius must be positive", __FILE__, 
			     __LINE__);

  rad_scale = radial_scale;

  oneOver_sq_rad_scale = 1.0 / (rad_scale * rad_scale);

  max_rad = max_radius;

  sq_max_rad = max_rad * max_rad;
  }


void Star::set_spikes(const unsigned int num_spikes, 
		      const double spike_scale_factor,
		      const double spike_phase_deg,
		      const double spike_steepening_power)
  {
  if (num_spikes < 1)
    {
    use_spikes = false;

    spike_angle_freq = 0.0;
    spike_scale_fact = 1.0;
    spike_phase_rad = 0.0;
    spike_power = 4.0;
    }
  else
    {
    if (!(spike_steepening_power > 0.0))
      throw QuickFlash::Except("Spike steepening power must be positive", 
			     __FILE__, __LINE__);

    spike_angle_freq = static_cast<double>(num_spikes) * 0.5;

    spike_scale_fact = spike_scale_factor;

    spike_phase_rad = spike_phase_deg * deg_to_rad;

    spike_power = spike_steepening_power;
    }
  }


void Star::set_star_color(const QuickFlash::Color::Color & star_color)
  {
  // Colormap values go from 1.0 (center of star) to 0.0 (background)

  const QuickFlash::Color::ValueColor center_color(1.0, star_color);

  // Background is fully-transparent star color

  QuickFlash::Color::ValueColor background_color(0.0, star_color);

  background_color.set_alpha(0.0);

  // Build the colormap

  std::list<QuickFlash::Color::ValueColor> color_list;

  color_list.push_back(background_color);
  color_list.push_back(center_color);

  colormap.reset(color_list);
  }


bool Star::get_color(const double x, const double y, 
		     QuickFlash::Color::Color & pos_color) const
  {
  bool ret_val = false;  // True if within range

  const double rel_x = x - star_x;
  const double rel_y = y - star_y;

  const double value = get_value(rel_x, rel_y);

  if (value > 0.0)
    {
    colormap.get_color(value, pos_color);
    ret_val = true;
    }
  else
    {
    pos_color.reset(0.0, 0.0);  // Transparent black
    }

  return ret_val;
  }


double Star::get_spike_factor(const double rel_x, const double rel_y) const
  {
  double spike_factor = 1.0;  // Spikes are an addition to this value

  if (use_spikes)
    {
    const double pos_angle = std::atan2(rel_y, rel_x);

    const double spike_angle 
      = (pos_angle * spike_angle_freq) + spike_phase_rad;

    const double cos_value = std::cos(spike_angle);

    spike_factor += spike_scale_fact * std::pow(cos_value, spike_power);
    }

  return spike_factor;
  }


double Star::get_value(const double rel_x, const double rel_y) const
  {
  double ret_val = 0.0;

  const double sq_rad = (rel_x * rel_x) + (rel_y * rel_y);

  if (sq_rad < sq_max_rad)
    {
    const double exp_arg = sq_rad * oneOver_sq_rad_scale;

    ret_val = std::exp(-exp_arg);

    if (use_spikes)
      ret_val *= get_spike_factor(rel_x, rel_y);

    // Limit the values

    if (ret_val < 0.0)
      ret_val = 0.0;
    else if (ret_val > 1.0)
      ret_val = 1.0;
    }

  return ret_val;
  }

