// C++ program file star_field.cpp

/*
  By Nathan C. Hearn
     June 14, 2008

  Star field generator.
*/


#ifndef USE_MAGICK

#error "ImageMagick support not present"

#else

#include <cstdlib>
#include <cmath>
#include <iostream>
#include <vector>
#include <string>
#include <list>
#include "quickflash_utils_commandline.hpp"
#include "quickflash_utils_file.hpp"
#include "quickflash_block_blockinfo.hpp"
#include "quickflash_geometry.hpp"
#include "quickflash_image_imagegrid.hpp"
#include "quickflash_image_imagemagick.hpp"
#include "quickflash_color_color.hpp"
#include "star.hpp"
#include "star_generator.hpp"
#include "star_array.hpp"


const char default_image_suffix[] = "png";

const double default_min_saturation = 0.0;
const double default_max_saturation = 0.0;

const double default_min_brightness = 1.0;
const double default_max_brightness = 1.0;

const unsigned int default_num_spikes = 0;
const double default_spike_scale_factor = 4.0;
const double default_spike_phase_deg = 0.0;
const double default_spike_steepening_power = 4.0;


int main(int argc, char * argv[])
  {
  using std::cerr;
  using std::endl;

  std::srand(7);

  std::vector<std::string> argument_list;
  std::list<QuickFlash::Utils::Option> option_list;

  QuickFlash::Utils::parse_command_line(argc, argv, argument_list, 
					option_list);

  const unsigned int num_args = argument_list.size();

  const unsigned int min_args = 8;

  if (num_args < min_args)
    {
    const std::string full_exec = argv[0];

    std::string exec_path;
    std::string exec_name;

    QuickFlash::Utils::split_name_path_file(full_exec, exec_path, exec_name);

    cerr << endl << "Usage: " << exec_name 
	 << " output_file_prefix width height"
	 << endl << "\t" << "width_pixels height_pixels num_stars" 
	 << endl << "\t" << "min_radial_scale max_radial_scale "
	 << "max_radius_factor"
	 << endl << "\t" << "[ --file-type=FILE_TYPE_SUFFIX ]"
	 << endl << "\t" << "[ --saturation-range=MIN_SAT,MAX_SAT ]"
	 << endl << "\t" << "[ --brightness-range=MIN_BRIGHT,MAX_BRIGHT ]"
	 << endl << "\t" << "[ --spikes=NUM_SPIKES,SPIKE_PHASE_ANGLE ]"
	 << endl << endl;
    return 1;
    }

  // Read the requied arguments

  unsigned int argptr = 0;

  const std::string output_file_prefix = argument_list[argptr++];

  const double width = std::atof(argument_list[argptr++].c_str());

  if (!(width > 0.0))
    {
    cerr << endl << "Error: Width must be positive" << endl << endl;
    return -1;
    }

  const double height = std::atof(argument_list[argptr++].c_str());

  if (!(height > 0.0))
    {
    cerr << endl << "Error: Height must be positive" << endl << endl;
    return -1;
    }

  const int raw_width_pixels = std::atoi(argument_list[argptr++].c_str());

  if (raw_width_pixels < 1)
    {
    cerr << endl << "Error: Width pixels must be positive" << endl << endl;
    return -1;
    }

  const unsigned int width_pixels 
    = static_cast<unsigned int>(raw_width_pixels);
  
  const int raw_height_pixels = std::atoi(argument_list[argptr++].c_str());

  if (raw_height_pixels < 1)
    {
    cerr << endl << "Error: Height pixels must be positive" << endl << endl;
    return -1;
    }

  const unsigned int height_pixels 
    = static_cast<unsigned int>(raw_height_pixels);

  const int raw_num_stars = std::atoi(argument_list[argptr++].c_str());

  if (raw_num_stars < 0)
    {
    cerr << endl << "Error: Number of stars can not be negative" << endl 
	 << endl;
    return -1;
    }
  
  const unsigned int num_stars = static_cast<unsigned int>(raw_num_stars);
  
  const double min_radial_scale = std::atof(argument_list[argptr++].c_str());

  if (!(min_radial_scale > 0.0))
    {
    cerr << endl << "Error: Min radial scale must be positive" << endl << endl;
    return -1;
    }

  const double max_radial_scale = std::atof(argument_list[argptr++].c_str());

  if (max_radial_scale < min_radial_scale)
    {
    cerr << endl << "Error: Max radial scale can not be less than min scale" 
	 << endl << endl;
    return -1;
    }

  const double max_radius_factor = std::atof(argument_list[argptr++].c_str());

  if (!(max_radius_factor > 0.0))
    {
    cerr << endl << "Error: Max radius factor must be positive" << endl 
	 << endl;
    return -1;
    }


  // Process the optional arguments

  std::string file_type_suffix = default_image_suffix;

  double min_saturation = default_min_saturation;
  double max_saturation = default_max_saturation;

  double min_brightness = default_min_brightness;
  double max_brightness = default_max_brightness;

  unsigned int num_spikes = default_num_spikes;
  double spike_scale_factor = default_spike_scale_factor;
  double spike_phase_deg = default_spike_phase_deg;
  double spike_steepening_power = default_spike_steepening_power;

  std::list<QuickFlash::Utils::Option>::const_iterator option_iter
    = option_list.begin();

  while (option_iter != option_list.end())
    {
    const QuickFlash::Utils::Option & current_option = *option_iter;

    const std::string & keyword = current_option.get_key();

    if (keyword == "file-type")
      {
      if (!current_option.value_present())
	{
	cerr << endl 
	     << "Error: File type suffix is required with file type option"
	     << endl << endl;
	return -1;
	}

      file_type_suffix = current_option.get_value();
      }
    else if (keyword == "saturation-range")
      {
      if (!current_option.value_present())
	{
	cerr << endl << "Error: Saturation range requies a pair of values"
	     << endl << endl;
	return -1;
	}

      std::vector<double> sat_range_values;

      current_option.get_value(sat_range_values);

      if (sat_range_values.size() != 2)
	{
	cerr << endl << "Error: Saturation range requires min/max values"
	     << endl << endl;
	return -1;
	}

      min_saturation = sat_range_values[0];
      max_saturation = sat_range_values[1];

      if (min_saturation < 0.0)
	{
	cerr << endl << "Error: Min saturation can not be negative" << endl 
	     << endl;
	return -1;
	}

      if (max_saturation > 1.0)
	{
	cerr << endl << "Error: Max saturation can not be greater than one"
	     << endl << endl;
	return -1;
	}

      if (max_saturation < min_saturation)
	{
	cerr << endl << "Error: Min saturation can not exceed max saturation"
	     << endl;
	return -1;
	}
      }
    else if (keyword == "brightness-range")
      {
      if (!current_option.value_present())
	{
	cerr << endl << "Error: Brightness range requies a pair of values"
	     << endl << endl;
	return -1;
	}

      std::vector<double> bright_range_values;

      current_option.get_value(bright_range_values);

      if (bright_range_values.size() != 2)
	{
	cerr << endl << "Error: Brightness range requires min/max values"
	     << endl << endl;
	return -1;
	}

      min_brightness = bright_range_values[0];
      max_brightness = bright_range_values[1];

      if (min_brightness < 0.0)
	{
	cerr << endl << "Error: Min brightness can not be negative" << endl 
	     << endl;
	return -1;
	}

      if (max_brightness > 1.0)
	{
	cerr << endl << "Error: Max brightness can not be greater than one"
	     << endl << endl;
	return -1;
	}

      if (max_brightness < min_brightness)
	{
	cerr << endl << "Error: Min brightness can not exceed max brightness"
	     << endl;
	return -1;
	}
      }
    else if (keyword == "spikes")
      {
      if (!current_option.value_present())
	{
	cerr << endl << "Error:Spike option requires values" << endl << endl;
	return -1;
	}

      std::vector<std::string> spike_value_strings;

      current_option.get_value(spike_value_strings);

      if (spike_value_strings.size() != 2)
	{
	cerr << endl 
	     << "Error: Spike option requires number of spikes and spike phase"
	     << endl << endl;
	return -1;
	}

      const int raw_num_spikes = std::atoi(spike_value_strings[0].c_str());

      if (raw_num_spikes < 0)
	{
	cerr << endl << "Error: Spike count can not be negative" << endl 
	     << endl;
	return -1;
	}

      num_spikes = static_cast<unsigned int>(raw_num_spikes);

      spike_phase_deg = std::atof(spike_value_strings[1].c_str());
      }
    else
      {
      cerr << endl << "Error: Unrecognized option " << keyword << endl << endl;
      return -1;
      }

    ++option_iter;
    }


  // Build the star generator and star array

  StarGenerator star_gen(width, height, min_radial_scale, max_radial_scale,
			 max_radius_factor, min_saturation, max_saturation,
			 min_brightness, max_brightness, num_spikes,
			 spike_scale_factor, spike_phase_deg, 
			 spike_steepening_power);


  StarArray star_array;

  Star work_star;

  for (unsigned int i = 0; i < num_stars; i++)
    {
    star_gen.get_star(work_star);

    star_array.add_star_top(work_star);
    }


  // Background color

  const QuickFlash::Color::Color bg_color(0.0, 0.0);  // Transparent black

  // Build the grids

  const std::vector<double> grid_min_coords(2, 0.0);

  std::vector<double> grid_max_coords(2);

  grid_max_coords[0] = width;
  grid_max_coords[1] = height;

  std::vector<unsigned int> grid_dims(2);

  grid_dims[0] = width_pixels;
  grid_dims[1] = height_pixels;

  const QuickFlash::Block::BlockInfo grid_info(grid_min_coords, 
					       grid_max_coords, grid_dims, 
					     QuickFlash::Geometry::Cartesian);

  QuickFlash::Image::ImageGrid image_grid(grid_dims, bg_color);

  // Fill the grid

  const unsigned int num_pixels = image_grid.get_num_pixels();

  std::vector<double> pos;

  QuickFlash::Color::Color work_color;

  for (unsigned int pixel_index = 0; pixel_index < num_pixels; pixel_index++)
    {
    grid_info.get_cell_center(pixel_index, pos);

    star_array.get_color(pos[0], pos[1], work_color);

    image_grid[pixel_index].overlay_color(work_color);
    }


  // Save the image

  const std::string output_file = output_file_prefix + "." + file_type_suffix;

  QuickFlash::Image::ImageMagick::write_image(output_file, image_grid);

  return 0;
  }


#endif  // USE_MAGICK
